//
// $XConsortium$
//

//
// Copyright (c) 1993-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

//
// Base object, type, and request support.
//

//
// Protect includes so that this file can be used with a different set
// of include files.
//
#ifndef __CORBA_hh
#include <Ox/env.h>
#include <Ox/ix_rep.h>
#include <Ox/marshal.h>
#include <Ox/request.h>
#include <Ox/schema.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#endif

#define array_length(A) (sizeof(A) / sizeof(A[0]))

struct GenericSeq {
    Long _maximum;
    Long _length;
    void* _buffer;
    Boolean _release;
};

// class CORBA::Object

_Ix_ctdt_1(CORBA,5,Object,6)

_Ix_ImplementField(CORBA::Object,Object,_Ix_NullParameter)

_Ix_InterfaceDescriptor(CORBA_Object,"Object",nil,nil);

_Ix_ImplementProxy(CORBA::Object,Object,CORBA_Object)

CORBA::Object_ptr CORBA::Object::_duplicate(CORBA::Object_ptr obj) {
    return (CORBA::Object_ptr)_ox_duplicate(
	obj, obj, _Ix__tid(CORBA_Object), _Ix_CreateProxy(CORBA_Object)
    );
}

CORBA::TypeId CORBA::Object::_type_id() {
    return _Ix__tid(CORBA_Object);
}

void* CORBA::Object::_narrow_tid(CORBA::TypeId type_id) {
    if (_Ix__tid(CORBA_Object) == type_id) {
        return this;
    }
    return nil;
}

//
// Compare two type ids assuming they are not short.  The first word
// is assumed to be equal.
//

Boolean _ox_equal_long_tid(CORBA::TypeId t1, CORBA::TypeId t2) {
    ULong len = ((OxShortTypeId*)t1)->format;
    switch (len) {
    case IX_TID_MEDIUM_2:
    case IX_TID_MEDIUM_3:
    case IX_TID_MEDIUM_4:
    case IX_TID_LONG:
	len = ((OxShortTypeId*)t1)->tid;
	break;
    default:
	return false;
    }
    CORBA::TypeId i1 = t1, i2 = t2;
    for (ULong i = len; i != 0; i--) {
	i1++;
	i2++;
	if (*i1 != *i2) {
	    return false;
	}
    }
    return true;
}

//
// Put a type id into a marshal buffer.
//

void _ox_CORBA_TypeId_put(OxMarshalBuffer& b, CORBA::TypeId t) {
    b.put_unsigned_long(t[0]);
    ULong len = ((OxShortTypeId*)t)->format;
    if (len != IX_TID_SHORT) {
	if (len == IX_TID_LONG) {
	    len = ((OxShortTypeId*)t)->tid;
	}
	ULong* tp = t;
	for (ULong i = len; i > 0; i--, tp++) {
	    b.put_unsigned_long(*tp);
	}
    }
}

void* CORBA::Object::_this_ptr() { return this; }
CORBA::TypeId CORBA::Object::_tid() { return _Ix__tid(CORBA_Object); }

//
// DO NOT USE!  This is kept around for sideways compatibility.
//
void CORBA::Object::_release() {
    _exchange()->release(this);
}

Boolean CORBA::Object::_is_local() { return _exchange()->is_local(); }
Boolean CORBA::Object::_ping(_Ix_Env* _env) { return _exchange()->ping(_env); }

//
// Duplicate an object reference.
//

void* _ox_duplicate(
    CORBA::Object_ptr base, void* obj, CORBA::TypeId tid, OxStubCreator f
) {
    return base == nil ? nil : base->_exchange()->duplicate(obj, tid, f);
}

//
// Release an object reference.  This operation is just forwarded on
// to the exchange.  The this pointer is passed as a parameter so that
// the exchange can decide to delete this object.
//

void CORBA::release(CORBA::Object_ptr obj) {
    if (obj != nil) {
	obj->_exchange()->release(obj);
    }
}

//
// Return the interface definition associated with the object.
// We ask the exchange for the tid to lookup so that the remote case
// may work by having the exchange either query or store the remote tid.
//

CORBA::Type_ptr CORBA::Object::_get_interface() {
#if OX_DII
    return OxSchema::schema()->map(_exchange()->tid(this));
#else
    return nil;
#endif
}

#if OX_DII
CORBA::Request_ptr CORBA::Object::_request(const char* operation) {
    return new OxRequestImpl(this, operation);
}
#else
CORBA::Request_ptr CORBA::Object::_request(const char*) {
    return nil;
}
#endif

// class CORBA::String_field

//
// Some systems unfortunately don't have strdup.
//

static inline char* ox_strdup(const char* buf) {
    return strcpy(new char[strlen(buf) + 1], buf);
}

//
// Convenience function for handling duplicate of a string
// that might be nil.
//

static char* strdup_or_nil(const char* p) {
    return p == nil ? nil : ox_strdup(p);
}

CORBA::String_field::String_field(const char* p) {
    _str = strdup_or_nil(p);
}

CORBA::String_field::String_field(const String_field& s) {
    _str = strdup_or_nil(s._str);
}

CORBA::String_field& CORBA::String_field::operator =(const String_field& s) {
    CORBA::String_var::asg(_str, s._str);
    return *this;
}

CORBA::String_field& CORBA::String_field::operator =(char* p) {
    if (p != _str) {
	delete _str;
	_str = p;
    }
    return *this;
}

CORBA::String_field& CORBA::String_field::operator =(const char* p) {
    CORBA::String_var::asg(_str, p);
    return *this;
}

// class CORBA::String_var

void CORBA::String_var::asg(char*& s1, const char* s2) {
    if (s1 != s2) {
	delete s1;
	s1 = strdup_or_nil(s2);
    }
}

CORBA::String_var::String_var(const String_var& s) {
    _str = strdup_or_nil((const char*)s);
}

CORBA::String_var::String_var(const String_field& s) {
    _str = strdup_or_nil(s._str);
}

CORBA::String_var& CORBA::String_var::operator =(char* p) {
    if (p != _str) {
	delete _str;
	_str = p;
    }
    return *this;
}

CORBA::String_var& CORBA::String_var::operator =(const char* p) {
    asg(_str, p);
    return *this;
}

CORBA::String_var& CORBA::String_var::operator =(const String_var& s) {
    asg(_str, (const char*)s);
    return *this;
}

CORBA::String_var& CORBA::String_var::operator =(const String_field& s) {
    asg(_str, s._str);
    return *this;
}

// class CORBA::StringSeq

// Some compilers can't inline these ...
CORBA::StringSeq::StringSeq() { init(0, 0, 0, 0); }
CORBA::StringSeq::StringSeq(ULong m) { alloc(m); }
CORBA::StringSeq::StringSeq(ULong m, ULong n) { alloc(m); _length = n; }
CORBA::StringSeq::StringSeq(ULong m, ULong n, char** b, Boolean r) {
    init(m, n, b, r);
}
CORBA::StringSeq::StringSeq(const StringSeq& s) { copy(s); }
CORBA::StringSeq::~StringSeq() { free(); }

void CORBA::StringSeq::alloc(ULong m) {
    _maximum = m;
    _length = 0;
    _release = true;
    char** buf = new char*[m];
    for (ULong i = 0; i < m; i++) {
	buf[i] = 0;
    }
    _buffer = (String_field*)buf;
}

void CORBA::StringSeq::copy(const StringSeq& s) {
    _maximum = s._maximum;
    _length = s._length;
    if (_maximum == 0) {
	_buffer = nil;
	_release = false;
    } else {
	char** buf = new char*[_maximum];
	for (ULong i = 0; i < _length; i++) {
	    buf[i] = strdup_or_nil(s[i]);
	}
	for (ULong j = _length; j < _maximum; j++) {
	    buf[j] = nil;
	}
	_buffer = (String_field*)buf;
	_release = true;
    }
}

void CORBA::StringSeq::free() {
    if (_release) {
	char** buf = (char**)_buffer;
	for (ULong i = 0; i < _length; i++) {
	    delete buf[i];
	}
	delete [] buf;
    }
}

void CORBA::StringSeq::_put(OxMarshalBuffer& b, const StringSeq& s) {
    b.put_seq_hdr(&s);
    for (ULong i = 0; i < s.length(); i++) {
	b.put_string(s[i]);
    }
}

void CORBA::StringSeq::_get(OxMarshalBuffer& b, StringSeq& s) {
    b.get_seq_hdr(&s);
    if (s.maximum() == 0) {
	s._buffer = 0;
    } else {
	char** buf = new char*[s.maximum()];
	for (ULong i = 0; i < s.length(); i++) {
	    buf[i] = b.get_string();
	}
	s._buffer = (String_field*)buf;
    }
}

//
// Initialize a sequence.  Pointer sequences may need the pointers
// to be set to nil.
//

void CORBA::_init_seq(void* seq, Long m, Long n, void* b, Boolean r) {
    GenericSeq* s = (GenericSeq*)seq;
    s->_maximum = m;
    s->_length = n;
    s->_buffer = b;
    s->_release = r;
}

void CORBA::_init_ptr_seq(void* seq, Long m, Long n, void* b, Boolean r) {
    _init_seq(seq, m, n, b, r);
    CORBA::Object_ptr* buf = (CORBA::Object_ptr*)b;
    for (Long i = 0; i < m; i++) {
	buf[i] = nil;
    }
}

void CORBA::_copy_fixed_seq(const void* src, void* dst, ULong size) {
    const GenericSeq* s = (const GenericSeq*)src;
    GenericSeq* _this = (GenericSeq*)dst;
    _this->_maximum = s->_maximum;
    _this->_length = s->_length;
    if (_this->_maximum == 0) {
	_this->_buffer = nil;
	_this->_release = false;
    } else {
	_this->_buffer = new char[size * _this->_maximum];
	_this->_release = true;
	memcpy(_this->_buffer, s->_buffer, size_t(size * _this->_length));
    }
}

// class CORBA::Exception

CORBA::Exception::Exception() {
    _major_ = 0;
}

CORBA::Exception::~Exception() { }

void CORBA::Exception::_put(OxMarshalBuffer& _b) const {
    _ox_CORBA_TypeId_put(_b, _interface_);
    _b.put_long(_major_);
}

const char* CORBA::Exception::_name() const { return _name_; }

//
// Try to cast the given exception to the exception type
// identified by the given index in the given type.
// The types here are strange (const param, void* return)
// to provide the most flexibility for passing and returning
// const or non-const using a single function.
//

void* CORBA::Exception::_try_cast(
    const Exception* e, CORBA::ULong index, CORBA::TypeId t
) {
    return (
	(
	    e != nil && ULong(e->_major()) == index &&
	    _ox_equal_tid(e->_interface(), t)
	) ? (void*)e : nil
    );
}

// class CORBA::SystemException

CORBA::SystemException::SystemException() { }
CORBA::SystemException::~SystemException() { }

// class CORBA::UserException

CORBA::UserException::UserException() { }
CORBA::UserException::~UserException() { }

// class Schema

// Predefined types
// Keep 0 free to catch/avoid uninitialized ids
_Ix_DefineTypeId(CORBA_Object, 1);

#if OX_DII
_Ix_DefineType(CORBA_void, tk_void, "void")
_Ix_DefineType(CORBA_Boolean, tk_boolean, "boolean")
_Ix_DefineType(CORBA_Char, tk_char, "char")
_Ix_DefineType(CORBA_Octet, tk_octet, "octet")
_Ix_DefineType(CORBA_Short, tk_short, "short")
_Ix_DefineType(CORBA_UShort, tk_ushort, "unsigned short")
_Ix_DefineType(CORBA_Long, tk_long, "long")
_Ix_DefineType(CORBA_ULong, tk_ulong, "unsigned long")
_Ix_DefineType(CORBA_LongLong, tk_longlong, "long long")
_Ix_DefineType(CORBA_ULongLong, tk_ulonglong, "unsigned long long")
_Ix_DefineType(CORBA_Float, tk_float, "float")
_Ix_DefineType(CORBA_Double, tk_double, "double")
_Ix_DefineType(CORBA_String, tk_string, "string")
#endif

OxSchema* OxSchema::schema_;

OxSchema::OxSchema() {
    object_ = &_Ix__type(CORBA_Object);

    types_.length = 10;
    types_.used = 0;
    types_.listlist = new InterfaceList[types_.length];
    for (Long t = 0; t < array_length(types_.cache); t++) {
	InterfaceDescriptor** p = types_.cache[t];
	p[0] = nil;
	p[1] = nil;
    }

    servers_.length = 10;
    servers_.used = 0;
    servers_.listlist = new ServerList[servers_.length];
    for (Long s = 0; s < array_length(servers_.cache); s++) {
	Receiver** p = servers_.cache[s];
	p[0] = nil;
	p[1] = nil;
    }
}

OxSchema::~OxSchema() {
    delete [] types_.listlist;
    delete [] servers_.listlist;
}

OxSchema* OxSchema::schema() {
    if (schema_ == nil) {
	schema_ = new OxSchema;
    }
    return schema_;
}

//
// Add the list of type descriptors to the current schema.
//

void OxSchema::load_schema(InterfaceDescriptor** list, Long length) {
    InterfaceListList* t = &types_;
    if (t->used >= t->length) {
	Long n = t->length + 10;
	InterfaceList* new_list = new InterfaceList[n];
	for (Long i = 0; i < t->length; i++) {
	    new_list[i] = t->listlist[i];
	}
	delete [] t->listlist;
	t->listlist = new_list;
	t->length = n;
    }
    InterfaceList* e = &t->listlist[t->used];
    e->list = list;
    e->length = length;
    t->used += 1;
}

//
// Add a list of servers to the schema, which effectively registers
// the receive methods.
//

void OxSchema::load_servers(OxSchema::Receiver* list, Long length) {
    OxSchema::ServerListList* t = &servers_;
    if (t->used >= t->length) {
	Long n = t->length + 10;
	OxSchema::ServerList* new_list = new OxSchema::ServerList[n];
	for (Long i = 0; i < t->length; i++) {
	    new_list[i] = t->listlist[i];
	}
	delete [] t->listlist;
	t->listlist = new_list;
	t->length = n;
    }
    OxSchema::ServerList* e = &t->listlist[t->used];
    e->list = list;
    e->length = length;
    t->used += 1;
}

#if OX_DII
CORBA::Type_ptr OxSchema::map(CORBA::TypeId t) {
    return OxTypeImpl::make_type(descriptor(t));
}
#else
CORBA::Type_ptr OxSchema::map(CORBA::TypeId) { return nil; }
#endif

//
// Compare function for a binary search that involves comparing ids.
//

static int cmp_tids(CORBA::TypeId t1, CORBA::TypeId t2) {
    ULong w1 = *t1;
    ULong w2 = *t2;
    if (w1 > w2) {
	return 1;
    } else if (w1 < w2) {
	return -1;
    }

    Long n = ((OxShortTypeId*)t1)->format;
    if (n == IX_TID_LONG) {
	n = ((OxShortTypeId*)t1)->tid;
    }
    for (Long i = 1; i < n; i++) {
	w1 = t1[i];
	w2 = t2[i];
	if (w1 > w2) {
	    return 1;
	} else if (w1 < w2) {
	    return -1;
	}
    }
    return 0;
}

//
// Compare function for the binary search on descriptors.
// Because the array we are searching contains
// InterfaceDescriptor* elements, the compare function
// is passed InterfaceDescriptor** values.
//

static int cmp_descriptors(const void* d1, const void* d2) {
    return cmp_tids(
	(*((OxSchema::InterfaceDescriptor**)d1))->id,
	(*((OxSchema::InterfaceDescriptor**)d2))->id
    );

}

//
// Compare function for the binary search on receivers.
//

static int cmp_receivers(const void* r1, const void* r2) {
    return cmp_tids(
	((OxSchema::Receiver*)r1)->id, ((OxSchema::Receiver*)r2)->id
    );
}

//
// Look up the given type id and find the corresponding descriptor.
// Currently, we use a very small hash table with 2 entries per bucket.
// If the cache misses, then we do a binary search on each list
// using the C library bsearch function.  This function returns
// a pointer to the element that matches, and since the elements
// are themselves pointers to descriptors the result
// is InterfaceDescriptor**.
//

OxSchema::InterfaceDescriptor* OxSchema::descriptor(CORBA::TypeId t) {
    if (t[0] == object_->id[0]) {
	return object_;
    }
    ULong e = *t & (array_length(types_.cache) - 1);
    InterfaceDescriptor** dp = types_.cache[e];
    InterfaceDescriptor* d = dp[0];
    if (d != nil) {
	if (_ox_equal_tid(d->id, t)) {
	    return d;
	}
	d = dp[1];
	if (d != nil && _ox_equal_tid(d->id, t)) {
	    dp[1] = dp[0];
	    dp[0] = d;
	    return d;
	}
    }

    //
    // The bsearch function wants a pointer to a key item, items
    // are pointers to descriptors, and the "real" key type is
    // the id, so we create a dummy descriptor to pass that contains
    // the desired id.
    //
    InterfaceDescriptor dummy;
    dummy.id = t;
    InterfaceDescriptor* key = &dummy;
    InterfaceList* end = &types_.listlist[types_.used];
    for (InterfaceList* s = types_.listlist; s < end; s++) {
	void* found = (InterfaceDescriptor**)bsearch(
	    &key, s->list, int(s->length), sizeof(InterfaceDescriptor*),
	    cmp_descriptors
	);
	if (found != nil) {
	    d = *((InterfaceDescriptor**)found);
	    dp[1] = dp[0];
	    dp[0] = d;
	    return d;
	}
    }
    return nil;
}

//
// Find the receive function for a given type id.  This works
// just like the descriptor function, so much so that it might
// be better to define it as a template.
//

OxTypeObj_ReceiveFunc OxSchema::receiver(CORBA::TypeId t) {
    if (t[0] == object_->id[0]) {
	// CORBA::Object doesn't define any operations
	return nil;
    }
    ULong e = *t & (array_length(servers_.cache) - 1);
    OxSchema::Receiver** p = servers_.cache[e];
    OxSchema::Receiver* r = p[0];
    if (r != nil) {
	if (_ox_equal_tid(r->id, t)) {
	    return r->func;
	}
	r = p[1];
	if (r != nil && _ox_equal_tid(r->id, t)) {
	    p[1] = p[0];
	    p[0] = r;
	    return r->func;
	}
    }

    OxSchema::Receiver key;
    key.id = t;
    OxSchema::ServerList* end = &servers_.listlist[servers_.used];
    for (OxSchema::ServerList* s = servers_.listlist; s < end; s++) {
	OxSchema::Receiver* r = (OxSchema::Receiver*)bsearch(
	    &key, s->list, int(s->length), sizeof(OxSchema::Receiver),
	    cmp_receivers
	);
	if (r != nil) {
	    p[1] = p[0];
	    p[0] = r;
	    return r->func;
	}
    }
    return nil;
}

//
// Given the first word in a type id, get the remainder (if any)
// from the marshal buffer.  Then look up the result and return
// the corresponding descriptor.
//

OxSchema::InterfaceDescriptor* OxSchema::get_descriptor(
    ULong start, OxMarshalBuffer& b
) {
    ULong len = ((OxShortTypeId*)&start)->format;
    if (len == IX_TID_SHORT) {
	return descriptor(&start);
    }

    if (len == IX_TID_LONG) {
	len = ((OxShortTypeId*)&start)->tid;
    }
    ULong static_buf[10];
    ULong* buf = (len < 10) ? static_buf : new ULong[len];
    ULong* end = buf + len;
    buf[0] = start;
    for (ULong* t = buf; t < end; t++) {
	*t = b.get_unsigned_long();
    }
    InterfaceDescriptor* d = descriptor(buf);
    if (buf != static_buf) {
	delete buf;
    }
    return d;
}

//
// Get a type id from the marshal buffer.  Use the given buffer
// (assumed to be non-zero in length) if it will fit.  Otherwise,
// allocate space for the id on the heap and return that.
//

CORBA::TypeId OxSchema::get_type_id(
    OxMarshalBuffer& b, CORBA::TypeId buf, Long length
) {
    buf[0] = b.get_unsigned_long();
    ULong len = ((OxShortTypeId*)buf)->format;
    if (len == IX_TID_SHORT) {
	return buf;
    }

    CORBA::TypeId r = buf;
    if (len == IX_TID_LONG) {
	len = ((OxShortTypeId*)buf)->tid;
    }
    if (length >= 0 && len > ULong(length)) {
	r = new ULong[len];
    }
    for (ULong i = 1; i < len; i++) {
	r[i] = b.get_unsigned_long();
    }
    return r;
}

//
// May need to write back dynamically-added types at some point.
//

void OxSchema::save() { }

void OxSchema::deactivate() {
    save();
    delete this;
}

//
// Initialization through static objects.
//

OxInitSchema::OxInitSchema(OxSchema::InterfaceDescriptor** list, Long length) {
    OxSchema::schema()->load_schema(list, length);
}

OxInitSchema::~OxInitSchema() { }

OxInitServer::OxInitServer(OxSchema::Receiver* list, Long length) {
    OxSchema::schema()->load_servers(list, length);
}

OxInitServer::~OxInitServer() { }

#if OX_DII

// class CORBA::Type

//
// Assume that duplicate on a Type object doesn't need a type id.
//
CORBA::Type_ptr CORBA::Type::_duplicate(CORBA::Type_ptr obj) {
    return (CORBA::Type_ptr)_ox_duplicate(obj, obj, 0, 0);
}

CORBA::Type::ParamInfo::ParamInfo() {
    name._str = 0;
    type.ptr_ = 0;
}

CORBA::Type::ParamInfo::ParamInfo(const ParamInfo& s) : 
    name(s.name), mode(s.mode), type(s.type) { }

CORBA::Type::ParamInfo::~ParamInfo() {
    delete name._str;
    _Ix_release(type.ptr_);
}

CORBA::Type::ParamInfo& CORBA::Type::ParamInfo::operator =(
    const ParamInfo& s
) {
    name = s.name;
    mode = s.mode;
    type = s.type;
    return *this;
}

_Ix_CopyNormalSeq(CORBA::Type::ParamInfoSeq,ParamInfoSeq,ParamInfo)

CORBA::Type::OpInfo::OpInfo() {
    name._str = 0;
    result.ptr_ = 0;
}

CORBA::Type::OpInfo::OpInfo(const OpInfo& s) : 
    name(s.name), index(s.index), result(s.result), params(s.params) { }

CORBA::Type::OpInfo::~OpInfo() {
    delete name._str;
    _Ix_release(result.ptr_);
}

CORBA::Type::OpInfo& CORBA::Type::OpInfo::operator =(const OpInfo& s) {
    name = s.name;
    index = s.index;
    result = s.result;
    params = s.params;
    return *this;
}

_Ix_ctdt_1(CORBA,5,Type,4)

_Ix_ImplementField(CORBA::Type,Type,_Ix_NullParameter)

//
// Helper classes for implementing types
//

#define OxTypeObj_declare_hdr(Name,Desc) \
class Name : public OxTypeImpl { \
public: \
    Name(Desc*); \
    ~Name(); \
protected: \
    Desc* info_;

#define OxTypeObj_ctdt(Name,Desc) \
Name::Name(Desc* d) : OxTypeImpl(d) { \
    info_ = d; \
} \
Name::~Name() { }

// Enumerations

OxTypeObj_declare_hdr(OxEnumTypeObj,OxSchema::EnumDescriptor)
public:
    Long enum_info();
protected:
    Long max_;
};

OxTypeObj_ctdt(OxEnumTypeObj,OxSchema::EnumDescriptor)

Long OxEnumTypeObj::enum_info() {
    return info_->count;
}

// Arrays

OxTypeObj_declare_hdr(OxArrayTypeObj,OxSchema::ArrayDescriptor)
public:
    void array_info(CORBA::Type_out type, Long& size);
};

OxTypeObj_ctdt(OxArrayTypeObj,OxSchema::ArrayDescriptor)

//
// Bug -- only handles single-dimension arrays
//

void OxArrayTypeObj::array_info(CORBA::Type_out type, Long& size) {
    type = CORBA::Type::_duplicate(OxTypeImpl::make_type(info_->element));
    size = info_->subscripts[0];
}

// Structs

OxTypeObj_declare_hdr(OxStructTypeObj,OxSchema::StructDescriptor)
public:
    Long members();
    CORBA::Type_ptr member_info(Long n);
};

OxTypeObj_ctdt(OxStructTypeObj,OxSchema::StructDescriptor)

Long OxStructTypeObj::members() { return info_->count; }

CORBA::Type_ptr OxStructTypeObj::member_info(Long n) {
    if (n < 0 || n >= info_->count) {
	return nil;
    }
    return CORBA::Type::_duplicate(OxTypeImpl::make_type(info_->members[n]));
}

// Unions

OxTypeObj_declare_hdr(OxUnionTypeObj,OxSchema::UnionDescriptor)
public:
    Long members();
    CORBA::Type_ptr member_info(Long n);
};

OxTypeObj_ctdt(OxUnionTypeObj,OxSchema::UnionDescriptor)

Long OxUnionTypeObj::members() { return info_->count; }

CORBA::Type_ptr OxUnionTypeObj::member_info(Long n) {
    if (n < 0 || n >= info_->count) {
	return nil;
    }
    return CORBA::Type::_duplicate(OxTypeImpl::make_type(info_->members[n]));
}

// Sequences

OxTypeObj_declare_hdr(OxSequenceTypeObj,OxSchema::SequenceDescriptor)
public:
    void sequence_info(CORBA::Type_out type, Long& size);
};

OxTypeObj_ctdt(OxSequenceTypeObj,OxSchema::SequenceDescriptor)

void OxSequenceTypeObj::sequence_info(CORBA::Type_out type, Long& size) {
    type = CORBA::Type::_duplicate(OxTypeImpl::make_type(info_->element));
    size = 0;
}

// Typedefs

OxTypeObj_declare_hdr(OxTypedefObj,OxSchema::TypedefDescriptor)
public:
    CORBA::Type_ptr typedef_info();
};

OxTypeObj_ctdt(OxTypedefObj,OxSchema::TypedefDescriptor)

CORBA::Type_ptr OxTypedefObj::typedef_info() {
    return CORBA::Type::_duplicate(OxTypeImpl::make_type(info_->descriptor));
}

// class OxTypeImpl

OxTypeImpl::OxTypeImpl(OxSchema::TypeDescriptorPtr p) {
    OxSchema::TypeDescriptor* d = (OxSchema::TypeDescriptor*)p;
    descriptor_ = d;
    d->type = this;
    if (d->kind == CORBA::Type::tk_objref) {
	OxSchema::InterfaceDescriptor* i = (OxSchema::InterfaceDescriptor*)d;
	if (i->methods == nil) {
	    opcount_ = 0;
	} else {
	    Long m;
	    for (m = 0; i->methods[m].name != nil; m++);
	    opcount_ = m;
	}
    }
}

OxTypeImpl::~OxTypeImpl() { }

OxExchange* OxTypeImpl::_exchange() { return &exchange_; }

char* OxTypeImpl::name() { return ox_strdup(descriptor_->name); }
CORBA::Type::TCKind OxTypeImpl::kind() { return descriptor_->kind; }

Boolean OxTypeImpl::op_info(CORBA::Type::OpInfo& op, Long n) {
    if (n < 0 || n >= opcount_) {
	return false;
    }
    if (descriptor_->kind != CORBA::Type::tk_objref) {
	return false;
    }
    OxSchema::InterfaceDescriptor* d = (
	(OxSchema::InterfaceDescriptor*)descriptor_
    );
    if (d->methods == nil || d->params == nil) {
	return false;
    }
    const OxSchema::OperationData& opdata = d->methods[n];
    op.name = ox_strdup(opdata.name);
    op.index = n;
    op.result = make_type(opdata.result);
    // clear out the old stuff
    op.params.load(0, 0, nil);
    Long nparams = opdata.nparams;
    if (nparams != 0) {
	op.params.alloc(nparams);
	op.params._length = nparams;
	Long p = 0;
	for (Long i = 0; i < n; i++) {
	    p += d->methods[i].nparams;
	}
	for (Long j = 0; j < nparams; j++) {
	    const OxSchema::ParameterData& pdata = d->params[p];
	    CORBA::Type::ParamInfo& param = op.params[j];
	    param.name = ox_strdup(pdata.name);
	    param.mode = CORBA::Type::ParamMode(pdata.mode);
	    param.type = make_type(pdata.type);
	    ++p;
	}
    }
    return true;
}

Long OxTypeImpl::enum_info() { return 0; }

void OxTypeImpl::array_info(CORBA::Type_out type, Long& size) {
    type = nil;
    size = 0;
}

Long OxTypeImpl::members() { return 0; }

CORBA::Type_ptr OxTypeImpl::member_info(Long) { return nil; }

void OxTypeImpl::sequence_info(CORBA::Type_out type, Long& size) {
    type = nil;
    size = 0;
}

CORBA::Type_ptr OxTypeImpl::typedef_info() { return nil; }

OxTypeImpl* OxTypeImpl::make_type(OxSchema::TypeDescriptorPtr p) {
    if (p == nil) {
	return nil;
    }
    OxSchema::TypeDescriptor* d = (OxSchema::TypeDescriptor*)p;
    OxTypeImpl* t = d->type;
    if (t == nil) {
	switch (d->kind) {
	case CORBA::Type::tk_enum:
	    t = new OxEnumTypeObj((OxSchema::EnumDescriptor*)d);
	    break;
	case CORBA::Type::tk_struct:
	    t = new OxStructTypeObj((OxSchema::StructDescriptor*)d);
	    break;
	case CORBA::Type::tk_union:
	    t = new OxUnionTypeObj((OxSchema::UnionDescriptor*)d);
	    break;
	case CORBA::Type::tk_array:
	    t = new OxArrayTypeObj((OxSchema::ArrayDescriptor*)d);
	    break;
	case CORBA::Type::tk_sequence:
	    t = new OxSequenceTypeObj((OxSchema::SequenceDescriptor*)d);
	    break;
	case CORBA::Type::tk_TypeCode:
	    t = new OxTypedefObj((OxSchema::TypedefDescriptor*)d);
	    break;
	case CORBA::Type::tk_except:
	    // assume ExceptDescriptor is compatible with StructDescriptor
	    t = new OxStructTypeObj((OxSchema::StructDescriptor*)d);
	    break;
	default:
	    t = new OxTypeImpl(d);
	    break;
	}
	d->type = t;
    }
    // assumes ref counting
    CORBA::Type::_duplicate(t);
    return t;
}

// class CORBA::Request

//
// Assume that duplicate on a Type object doesn't need a type id.
//
CORBA::Request_ptr CORBA::Request::_duplicate(CORBA::Request_ptr obj) {
    return (CORBA::Request_ptr)_ox_duplicate(obj, obj, 0, 0);
}

_Ix_ctdt_1(CORBA,5,Request,7)

_Ix_ImplementField(CORBA::Request,Request,_Ix_NullParameter)

// class OxRequestImpl

OxRequestImpl::OxRequestImpl(CORBA::Object_ptr o, const char* operation) {
    target_ = CORBA::Object::_duplicate(o);
    init();
    operation_ = ox_strdup(operation);
}

OxRequestImpl::~OxRequestImpl() {
    _Ix_release(target_);
    free();
}

OxExchange* OxRequestImpl::_exchange() { return &exchange_; }

CORBA::Object_ptr OxRequestImpl::target() const {
    return CORBA::Object::_duplicate(target_);
}

const char* OxRequestImpl::operation() const { return ox_strdup(operation_); }
_Ix_Env* OxRequestImpl::env() { return &env_; }

CORBA::Request::CallStatus OxRequestImpl::invoke() {
    return send(true);
}

CORBA::Request::CallStatus OxRequestImpl::send_oneway() {
    return send(false);
}

CORBA::Request::CallStatus OxRequestImpl::send(Boolean wait) {
    CORBA::Request::CallStatus s;
    except_ = nil;
    if (resolved_ == 0) {
	s = CORBA::Request::unknown_operation;
    } else if (status_ == CORBA::Request::ambiguous_operation) {
	s = status_;
    } else {
	Long r = target_->_exchange()->invoke(*buffer_, !wait);
	if (r == 0) {
	    s = CORBA::Request::ok;
	} else {
	    s = CORBA::Request::invoke_failed;
	    OxSchema::InterfaceDescriptor* d = (
		OxSchema::schema()->get_descriptor(r, *buffer_)
	    );
	    Long index = get_long();
	    except_ = d->excepts[index - 1];
	}
    }
    return status(s);
}

CORBA::Request::CallStatus OxRequestImpl::op_info(CORBA::Type::OpInfo& op) {
    if (resolved_ == 0) {
	status_ = CORBA::Request::unknown_operation;
	CORBA::Type_ptr type = target_->_get_interface();
	if (find_op(type, index_)) {
	    resolved_ = target_->_exchange()->tid(target_);
	    status_ = CORBA::Request::ok;
	} else {
	    resolve(
		OxSchema::schema()->descriptor(
		    target_->_exchange()->tid(target_)
		)
	    );
	}
	if (is_ok()) {
	    target_->_exchange()->op(*buffer_, target_, resolved_, index_);
	}
    }
    if (is_ok()) {
	CORBA::Type_ptr type = OxSchema::schema()->map(resolved_);
	type->op_info(op, index_);
    }
    return status_;
}

CORBA::Type_ptr OxRequestImpl::except_info() {
    return OxTypeImpl::make_type(except_);
}

CORBA::Request::CallStatus OxRequestImpl::status(
    CORBA::Request::CallStatus s
) {
    status_ = s;
    return s;
}

void OxRequestImpl::put_char(Char value) {
    buffer_->put_char(value);
}

Char OxRequestImpl::get_char() { return buffer_->get_char(); }

void OxRequestImpl::put_boolean(Boolean value) {
    buffer_->put_boolean(value);
}

Boolean OxRequestImpl::get_boolean() { return buffer_->get_boolean(); }

void OxRequestImpl::put_octet(Octet value) {
    buffer_->put_octet(value);
}

Octet OxRequestImpl::get_octet() {
    return buffer_->get_octet();
}

void OxRequestImpl::put_short(Short value) {
    buffer_->put_short(value);
}

Short OxRequestImpl::get_short() { return buffer_->get_short(); }

void OxRequestImpl::put_unsigned_short(UShort value) {
    buffer_->put_unsigned_short(value);
}

UShort OxRequestImpl::get_unsigned_short() {
    return buffer_->get_unsigned_short();
}

void OxRequestImpl::put_long(Long value) {
    buffer_->put_long(value);
}

Long OxRequestImpl::get_long() {
    return buffer_->get_long();
}

void OxRequestImpl::put_unsigned_long(ULong value) {
    buffer_->put_unsigned_long(value);
}

ULong OxRequestImpl::get_unsigned_long() {
    return buffer_->get_unsigned_long();
}

void OxRequestImpl::put_float(Float value) {
    buffer_->put_float(value);
}

Float OxRequestImpl::get_float() {
    return buffer_->get_float();
}

void OxRequestImpl::put_double(Double value) {
    buffer_->put_double(value);
}

Double OxRequestImpl::get_double() {
    return buffer_->get_double();
}

void OxRequestImpl::put_string(const char* value) {
    buffer_->put_string(value);
}

char* OxRequestImpl::get_string() {
    return buffer_->get_string();
}

void OxRequestImpl::put_object(CORBA::Object_ptr obj) {
    buffer_->put_object(obj);
}

CORBA::Object_ptr OxRequestImpl::get_object() {
    return CORBA::Object_ptr(
	buffer_->get_object(&_Ix_CreateProxy(CORBA_Object))
    );
}

void OxRequestImpl::begin_aggregate() { ++nesting_; }

void OxRequestImpl::end_aggregate() {
    if (nesting_ > 0) {
	--nesting_;
    }
}

void OxRequestImpl::init() {
    operation_ = nil;
    index_ = -1;
    resolved_ = 0;
    status_ = CORBA::Request::initial;
    env_.exception(nil);
    buffer_ = new OxMarshalBuffer;
    buffer_->init();
    buffer_->env(&env_);
    except_ = nil;
    nesting_ = 0;
}

void OxRequestImpl::free() {
    delete operation_;
    buffer_->destroy();
}

Boolean OxRequestImpl::find_op(CORBA::Type_ptr type, Long& index) {
    Boolean b = false;
    CORBA::Type::OpInfo op;
    for (Long i = 0; type->op_info(op, i); i++) {
	if (strcmp(operation_, op.name) == 0) {
	    index = i;
	    opinfo_ = op;
	    b = true;
	    break;
	}
    }
    return b;
}

void OxRequestImpl::resolve(OxSchema::InterfaceDescriptor* d) {
    OxSchema::InterfaceDescriptor** p = d->parents;
    if (p != nil) {
	for (; *p != nil; p++) {
	    CORBA::Type_ptr type = OxTypeImpl::make_type(*p);
	    Long index;
	    if (find_op(type, index)) {
		CORBA::TypeId tid = (*p)->id;
		if (resolved_ == 0 || *resolved_ == *tid) {
		    resolved_ = tid;
		    index_ = index;
		    status_ = CORBA::Request::ok;
		    //
		    // Found first occurrence.  Now continue search in siblings
		    // for ambiguities.
		    //
		} else {
		    status_ = CORBA::Request::ambiguous_operation;
		    return;
		}
	    } else {
		resolve(*p);
	    }
	}
    }
}

#endif
