//
// $XConsortium$
//

//
// Copyright (c) 1993-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef __CORBA_hh
#define __CORBA_hh

#ifdef WIN32
#define export __declspec(dllexport)
#define import __declspec(dllimport)
#else
#define export
#define import
#endif

//- CORBA
class export CORBA {
    //. The CORBA namespace (defined as a class if namespaces are not
    //. supported) provides common definitions for the IDL C++ mapping.
public:

#if defined(__GNUC__)
typedef bool Boolean;
#else
typedef unsigned char Boolean;
#endif
typedef char Char;
typedef unsigned char UChar;
typedef unsigned char Octet;
typedef short Short;
typedef unsigned short UShort;

//
// IDL longs are 32-bits, which is a potential problem
// with respect to 32/64-bit machine ambiguities.
//

#if (_MIPS_SZLONG == 64)
typedef __int32_t Long;
typedef __uint32_t ULong;
#else
#if defined(__alpha)
typedef int Long;
typedef unsigned int ULong;
#else
typedef long Long;
typedef unsigned long ULong;
#endif
#endif

#if defined(__sgi) || defined(__GNUC__)
// Add to #if platforms that support long long
typedef long long LongLong;
typedef unsigned long long ULongLong;
#endif

typedef float Float;
typedef double Double;

typedef char* String_ptr;
_Ix_OutParam(String_out,char*)
class String_var;

class export String_field {
public:
    char* _str;

    String_field() { }
    String_field(char* p) { _str = p; }
    String_field(const char* p);
    String_field(const String_field& s);
    String_field& operator =(const String_field& s);
    String_field& operator =(const String_var& s);
    String_field& operator =(char* p);
    String_field& operator =(const char* p);
    operator char*() { return _str; }
    operator const char*() const { return _str; }
    Boolean operator ==(const char* p) const { return _str == p; }
    Boolean operator !=(const char* p) const { return _str != p; }
    const char& operator [](ULong index) const { return _str[index]; }
    char& operator [](ULong index) { return _str[index]; }
};

class export String_var {
private:
    char* _str;
public:
    String_var() { _str = 0; }
    String_var(char* p) { _str = p; }
    String_var(const String_var& s);
    String_var(const String_field& s);
    ~String_var() { delete _str; }

    String_var& operator =(char* p);
    String_var& operator =(const char* p);
    String_var& operator =(const String_var& s);
    String_var& operator =(const String_field& s);

    operator char*() { return _str; }
    operator const char*() const { return _str; }
    operator String_out() { return String_out(_str); }

    static void asg(char*& s1, const char* s2);
};

//
// Built-in sequence<string>.
//
// We cast rather than allocate String_field elements because
// some "clever" compilers try to call the new[] with a constructor
// even though the constructor is a nop.
//

class export StringSeq {
public:
    ULong _maximum;
    ULong _length;
    String_field* _buffer;
    Boolean _release;
private:
    void init(ULong m, ULong n, char** b, Boolean r) {
	_maximum = m; _length = n; _buffer = (String_field*)b; _release = r;
    }
    void copy(const StringSeq&);
    void free();
public:
    StringSeq();
    StringSeq(ULong m);
    StringSeq(ULong m, ULong n);
    StringSeq(ULong m, ULong n, char** b, Boolean r = 0);
    StringSeq(const StringSeq& s);
    ~StringSeq();

    StringSeq& operator =(const StringSeq& s) {
	free(); copy(s); return *this;
    }
    const String_field& operator[](ULong i) const { return _buffer[i]; }
    String_field& operator[](ULong i) { return _buffer[i]; }
    ULong maximum() const { return _maximum; }
    void length(ULong n) { _length = n; }
    ULong length() const { return _length; }
    char** buffer() const { return (char**)_buffer; }
    void load(ULong m, ULong n, char** e, Boolean r = 0) {
	free();
	init(m, n, e, r);
    }

    void alloc(ULong m);
    static char** allocbuf(ULong n) { return new char*[n]; }

    static void _put(OxMarshalBuffer&, const StringSeq&);
    static void _get(OxMarshalBuffer&, StringSeq&);
};

_Ix_ManagedSeq(StringSeq,String_field);

_Ix_Forward(Object)

static Boolean is_nil(Object_ptr p) { return p == 0; }
static void release(Object_ptr p);

static void _init_seq(void* seq, Long m, Long n, void* b, Boolean r);
static void _init_ptr_seq(void* seq, Long m, Long n, void* b, Boolean r);
static void _copy_fixed_seq(const void* src, void* dst, ULong size);

_Ix_Forward(Type)
_Ix_Forward(Request)

//
// A type id in memory is a pointer to the variable-length representation.
// Because we want to be able to generate code with a statically-initialized
// id, we specify is a pointer to unsigned long.  To interpret the first
// word, which contains the control information, we cast the pointer
// to the ShortTypeId struct.
//
typedef ULong* TypeId;

//- CORBA::Object*
class export Object {
    //. All objects support CORBA::Object operations, which
    //. include access to an object's type and holding or
    //. releasing a reference to the object.
public:
    virtual OxExchange* _exchange() = 0;
    virtual void* _this_ptr();
    virtual TypeId _tid();
    virtual void* _narrow_tid(TypeId);
    static TypeId _type_id();
    static Object* _narrow(Object* object) { return object; }

    // Anachronism: _release as a non-static member function
    virtual void _release();

    //
    // Convenience functions, might be better to call the exchange
    // operations directly.
    //
    Boolean _is_local();
    Boolean _isLocal() { return _is_local(); }
    Boolean _ping(_Ix_Env*);
public:
    static Object_ptr _duplicate(Object_ptr p);

    //- _get_interface
    virtual Type_ptr _get_interface();
	//. Return a reference to the object's interface type.

    //- _request
    virtual Request_ptr _request(const char* operation);
	//. Create a request for performing dynamically-chosen operations
	//. on the object.
public:
    // Object constructor and destructor are public so that one can
    // have local (stack-based) objects.
    Object();
    virtual ~Object();
private:
    Object(const Object&);
    void operator =(const Object&);
};

_Ix_OutParam(Object_out,Object_ptr)
_Ix_ManagedRef(Object,_Ix_NullParameter)
_Ix_DeclareProxy(Object)

//- CORBA::Type*
class export Type : public Object {
    //. CORBA::Type contains information about the type of an object.
public:
    static Type_ptr _duplicate(Type_ptr p);

    //- ParamMode
    enum ParamMode { param_in, param_out, param_inout };
	//. ParamMode is an enumeration for the various parameter
	//. passing modes.

    //- ParamInfo
    struct export ParamInfo {
	ParamInfo();
	ParamInfo(const ParamInfo&);
	~ParamInfo();
	ParamInfo& operator=(const ParamInfo&);

	String_field name;
	ParamMode mode;
	Type_field type;
    };
	//. This structure contains information about
	//. a particular parameter.

    //- ParamInfoSeq
    _Ix_NormalSeq(ParamInfoSeq,ParamInfo);
	//. This type defines a sequence of parameter information.

    //- OpInfo
    struct export OpInfo {
	OpInfo();
	OpInfo(const OpInfo&);
	~OpInfo();
	OpInfo& operator =(const OpInfo&);

	String_field name;
	Long index;
	Type_field result;
	ParamInfoSeq params;
    };
	//. This structure contains information about
	//. an operation or attribute.

    //- TCKind
    enum TCKind {
	tk_null, tk_void,
	tk_short, tk_long, tk_ushort, tk_ulong,
	tk_float, tk_double, tk_boolean, tk_char,
	tk_octet, tk_any,
	tk_TypeCode, tk_Principal, tk_objref,
	tk_struct, tk_union, tk_enum, tk_string,
	tk_sequence, tk_array,

	// additions
	tk_longlong, tk_ulonglong,
	tk_except, tk_module
    };
	//. TCKind is an enumeration of the different categories of
	//. type objects.

    //- name
    virtual char* name() = 0;
	//. Return the name of the type.  The name is as defined
	//. in the source, not qualified by any containing scopes.

    //- op_info
    virtual Boolean op_info(Type::OpInfo& op, Long n) = 0;
	//. Retrieve the information about the operation referred
	//. to by the given index, if any.  This operation returns
	//. true if the index is valid.

    //- kind
    virtual TCKind kind() = 0;
	//. Return the kind of this type.

    //- enum_info
    virtual Long enum_info() = 0;
	//. Return the maximum value for an enumeration constant.

    //- array_info
    virtual void array_info(Type_out type, Long& size) = 0;
	//. Return the information about an array type.

    //- members
    virtual Long members() = 0;
	//. Return the number of members in a structure or union.

    //- member_info
    virtual Type_ptr member_info(Long n) = 0;
	//. Return the type of the member identified by the given index.

    //- sequence_info
    virtual void sequence_info(Type_out type, Long& size) = 0;
	//. Return the information about a sequence type.

    //- typedef_info
    virtual Type_ptr typedef_info() = 0;
	//. Return the type to which a typedef refers.
protected:
    Type();
    ~Type();
};

_Ix_OutParam(Type_out,Type_ptr)
_Ix_ManagedRef(Type,_Ix_NullParameter)

//- CORBA::Request*
class export Request : public Object {
    //. A request object is used to perform dynamic invocation
    //. of operations on objects.  These objects are useful
    //. for script of other kinds of interpreters.
public:
    static Request_ptr _duplicate(Request_ptr p);

    //- CallStatus
    enum CallStatus {
	initial, ok,
	unknown_operation, ambiguous_operation,
	bad_argument_count, bad_argument_type,
	invoke_failed
    };
	//. This type defines the possible return status values
	//. for a call.

    //- target
    virtual Object_ptr target() const = 0;
	//. Return the object for which this request is intended.

    //- operation
    virtual const char* operation() const = 0;
	//. Return the operation for which this request is intended.

    //- env
    virtual _Ix_Env* env() = 0;

    //- invoke, send_oneway
    virtual CallStatus invoke() = 0;
    virtual CallStatus send_oneway() = 0;
	//. Call the operation on the object.  The send_oneway operation
	//. assumes that the request is defined as a oneway operation
	//. on the target and does not wait for a reply.

    //- op_info
    virtual CallStatus op_info(Type::OpInfo& op) = 0;
	//. Retrieve the information for the operation associated
	//. with the request, if any.  The return value is true
	//. if the request corresponds to a valid operation.

    //- except_info
    virtual Type_ptr except_info() = 0;
	//. If the most recent invocation on a request
	//. raised an exception, return the type information
	//. for the type of the raised exception.  Otherwise,
	//. this operation will return nil.

    //- put_char, put_boolean, put_octet, put_short, put_unsigned_short,
    //- put_long, put_unsigned_long, put_float, put_double,
    //- get_char, get_boolean, get_octet, get_short, get_unsigned_short,
    //- get_long, get_unsigned_long, get_float, get_double
    virtual void put_char(Char value) = 0;
    virtual void put_boolean(Boolean value) = 0;
    virtual void put_octet(Octet value) = 0;
    virtual void put_short(Short value) = 0;
    virtual void put_unsigned_short(UShort value) = 0;
    virtual void put_long(Long value) = 0;
    virtual void put_unsigned_long(ULong value) = 0;
    virtual void put_float(Float value) = 0;
    virtual void put_double(Double value) = 0;
    virtual Char get_char() = 0;
    virtual Boolean get_boolean() = 0;
    virtual Octet get_octet() = 0;
    virtual Short get_short() = 0;
    virtual UShort get_unsigned_short() = 0;
    virtual Long get_long() = 0;
    virtual ULong get_unsigned_long() = 0;
    virtual Float get_float() = 0;
    virtual Double get_double() = 0;
	//. Put or get a primitive type into or from the
	//. request's parameter list.

    //- put_string, get_string
    virtual void put_string(const char* value) = 0;
    virtual char* get_string() = 0;
	//. Put or get a string into or from the
	//. request's parameter list.

    //- put_object, get_object
    virtual void put_object(Object_ptr obj) = 0;
    virtual Object_ptr get_object() = 0;
	//. Put or get an object reference into or from the
	//. request's parameter list.

    //- begin_aggregate, end_aggregate
    virtual void begin_aggregate() = 0;
    virtual void end_aggregate() = 0;
	//. These operations bracket a set of values for
	//. an aggregate parameter.
protected:
    Request();
    ~Request();
};

_Ix_OutParam(Request_out,Request_ptr)
_Ix_ManagedRef(Request,_Ix_NullParameter)

//
// Base classes for exceptions
//

class export Exception {
protected:
    Exception();
public:
    virtual ~Exception();

    TypeId _interface() const { return _interface_; }
    Long _major() const { return _major_; }
    Long _hash() const { return _hash_; }

    virtual void _put(OxMarshalBuffer&) const;
    virtual const char* _name() const;
protected:
    TypeId _interface_;
    Long _major_;
    Long _hash_;
    const char* _name_;

    static void* _try_cast(const Exception* e, ULong index, TypeId t);
};

class export SystemException : public Exception {
protected:
    SystemException();
public:
    ~SystemException();

    static SystemException* _cast(Exception*);
};

class export UserException : public Exception {
protected:
    UserException();
public:
    ~UserException();

    static UserException* _cast(Exception*);
};

};

#endif
