//
// $XConsortium$
//

//
// Copyright (c) 1993-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

//
// Exchange support.
//

#include <Ox/env.h>
#include <Ox/exchange.h>
#include <Ox/ix_rep.h>
#include <Ox/marshal.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#if OX_RPC_EXCHANGE
#include <Ox/rpc_exch.h>
#include <Ox/transport.h>
#endif

#if defined(linux) || defined(__alpha)
#include <rpc/types.h>
extern "C" void exit(int);
#endif

#define array_length(A) (sizeof(A) / sizeof(A[0]))

//
// Error handling -- needs to be internationalized
//
// Most of these functions are inline because they are only called
// once and inlining should result in less code.  Error functions
// called more than once, such as overflow or underflow, should not
// be inline.
//

static inline void err_oxcall(Long r) {
    fprintf(stderr, "oxCall returns %d\n", r);
    exit(1);
}

static inline void err_narrow(Long r) {
    fprintf(stderr, "Unexpected reply %d for _narrow\n", r);
    exit(1);
}

static inline void err_release() {
    fprintf(stderr, "Release on zero refcount object\n");
    exit(1);
}

//
// Put an exchange id into a marshal buffer.
//

void _OxExchangeId_put(OxMarshalBuffer& b, OxExchangeId e) {
    b.put_unsigned_long(e);
}

void _OxExchangeId_put_nil(OxMarshalBuffer& b) {
    _Ix_extern_xid(Ox_nilref);
    b.put_unsigned_long(_Ix__xid(Ox_nilref));
}

//
// Get an exchange id from a marshal buffer and
// return the unmarshalled object.
//

extern void* OxRpcExchange_unmarshal(OxStubCreator, OxMarshalBuffer&);
extern void* OxLibraryExchange_unmarshal(OxStubCreator, OxMarshalBuffer&);

static void* (*_exch_unmarshal_[])(OxStubCreator, OxMarshalBuffer&) = {
    nil,
    nil, /* nil xid */
    &OxRpcExchange_unmarshal,
    &OxLibraryExchange_unmarshal,
    nil
};

void* _OxExchangeId_get(OxMarshalBuffer& b, OxStubCreator f) {
    _Ix_extern_xid(Ox_nilref);
    _Ix_extern_xid(Ox_maximum_known);
    OxExchangeId x = b.get_unsigned_long();
    if (x == 0 || x > _Ix__xid(Ox_maximum_known)) {
	// should set a system exception in the marshal buffer's env
	return nil;
    }
    return (x == _Ix__xid(Ox_nilref)) ? nil : (*_exch_unmarshal_[x])(f, b);
}

//
// Test if two object references denote the same implementation object.
//

Boolean _ox_is_eq(CORBA::Object_ptr /* p1 */, CORBA::Object_ptr /* p2 */) {
    // return is_not_nil(p1) && p1->_exchange()->eq(p1, p2);
    return false;
}

// class Exchange

// Predefined exchanges
// Keep 0 free to catch/avoid uninitialized ids
_Ix_DefineExchangeId(Ox_nilref, 1);
_Ix_DefineExchangeId(OxRpcExchange, 2);
_Ix_DefineExchangeId(OxLibraryExchange, 3);
_Ix_DefineExchangeId(Ox_maximum_known, 3);

_Ix_ctdt_0(OxExchange,10)

Boolean OxExchange::is_local() { return false; }
Boolean OxExchange::ping(_Ix_Env*) { return false; }

// class LibraryExchange

OxLibraryExchange::OxLibraryExchange() { refcount_ = 1; }
OxLibraryExchange::~OxLibraryExchange() {
    // depends on implementation of release
    if (refcount_ != 1) {
	fprintf(stderr, "Exchange refcount is %d in destructor\n", refcount_);
	exit(1);
    }
    // just trying to catch multiple deletes here
    refcount_ = 0;
}

//
// The LibraryExchange op and invoke operations are used
// for doing dynamic invocation within the same address space.
//

void OxLibraryExchange::op(
    OxMarshalBuffer& b, CORBA::Object_ptr obj, CORBA::TypeId t, Long method
) {
    b.put_unsigned_long((ULong)obj);
    _ox_CORBA_TypeId_put(b, t);
    b.put_unsigned_long(method);
}

Long OxLibraryExchange::invoke(OxMarshalBuffer& b, Boolean /* oneway */) {
    b.make_out_in();
    CORBA::Object_ptr obj = (CORBA::Object_ptr)b.get_unsigned_long();
    b.dispatch(obj);
    b.make_out_in();
    return b.get_unsigned_long();
}

//
// The LibraryExchange marshal/unmarshal operations just send the
// object's address through the marshal buffer.  This is pretty
// dangerous but is what we want to send DII calls through
// the marshal buffer receive code.
//

void OxLibraryExchange::marshal(OxMarshalBuffer& b, CORBA::Object_ptr o) {
    _Ix_extern_xid(OxLibraryExchange);
    b.put_unsigned_long(_Ix__xid(OxLibraryExchange));
    b.put_unsigned_long((ULong)CORBA::Object::_duplicate(o));

    
//---     CORBA::Object::_duplicate(o);
//---     b.put_unsigned_long((ULong)o->_this_ptr());
}

void* OxLibraryExchange_unmarshal(OxStubCreator, OxMarshalBuffer& b) {
    return (void*)b.get_unsigned_long();

//---     CORBA::Object_ptr obj = (CORBA::Object*)b.get_unsigned_long();
//---     return (void*)obj->_this_ptr();
}

//
// The LibraryExchange narrow operation trys to narrow and
// then increments the refcount if successful.
//

void* OxLibraryExchange::narrow(
    CORBA::Object_ptr obj, CORBA::TypeId obj_tid, CORBA::TypeId narrow_tid,
    OxStubCreator
) {
    void* v = obj->_narrow_tid(narrow_tid);
    if (v != nil) {
	refcount_ += 1;
    }
    return v;
}

//
// The LibraryExchange duplicate operation simply increments
// the refcount.
//

void* OxLibraryExchange::duplicate(void* obj, CORBA::TypeId, OxStubCreator) {
    refcount_ += 1;
    return obj;
}

//
// The LibraryExchange release operation decrements the refcount, and
// deallocates the object if the count would become zero.
//

void OxLibraryExchange::release(CORBA::Object_ptr obj) {
    if (refcount_ > 1) {
	refcount_ -= 1;
    } else if (refcount_ == 1) {
	// Releasing the last refcount will delete the object,
	// which implicitly or explicitly should delete "this" exchange.
	// So no access to either the object or the exchange
	// should occur after the delete statement.
	delete obj;
	return;
    } else {
	err_release();
    }
}

//
// LibraryExchange doesn't store the object's tid, so it assumes
// the object knows it.
//

CORBA::TypeId OxLibraryExchange::tid(CORBA::Object_ptr obj) {
    return obj->_tid();
}

// class LockingExchange

OxLockingExchange::OxLockingExchange() {
    refcount_ = 1;
}

OxLockingExchange::~OxLockingExchange() { }

//
// The LockingExchange narrow operation tries to narrow and
// then increments the refcount if successful.
//

void* OxLockingExchange::narrow(
    CORBA::Object_ptr obj, CORBA::TypeId obj_tid, CORBA::TypeId narrow_tid,
    OxStubCreator
) {
    void* v = obj->_narrow_tid(narrow_tid);
    if (v != nil) {
	lock();
	refcount_ += 1;
	unlock();
    }
    return v;
}

//
// The LockingExchange duplicate operation increments the refcount.
// The assumption of duplicate is that the given object pointer
// is already the desired type.  The type has to be passed
// for the case where we create a new serverless object, in which case
// we would then need to widen up to the requested type.
//

void* OxLockingExchange::duplicate(void* obj, CORBA::TypeId, OxStubCreator) {
    lock();
    refcount_ += 1;
    unlock();
    return obj;
}

//
// The LockingExchange release operation decrements the refcount and
// deletes the object if the count would become zero.
//

void OxLockingExchange::release(CORBA::Object_ptr obj) {
    lock();
    if (refcount_ > 1) {
	refcount_ -= 1;
    } else if (refcount_ == 1) {
	// Releasing the last refcount will delete the object,
	// which implicitly or explicitly should delete "this" exchange.
	// So no access to either the object or the exchange
	// should occur after the delete statement.
	delete obj;
	return;
    } else {
	err_release();
    }
    unlock();
}

//
// Locking not implemented yet ...
//

void OxLockingExchange::lock() { }
void OxLockingExchange::unlock() { }

// class ForeignExchange

OxForeignExchange::OxForeignExchange() {
    tid_ = 0;
}

OxForeignExchange::~OxForeignExchange() { }

//
// The ForeignExchange narrow operation uses the tid stored in the proxy
// (as opposed to the one passed from the object).  If the type
// is not known in this address space, then we need to send a message.
// If narrowing is ok, then create a new proxy using the stub creator.
//

void* OxForeignExchange::narrow(
    CORBA::Object_ptr obj, CORBA::TypeId, CORBA::TypeId narrow_tid,
    OxStubCreator f
) {
    void* v = obj->_narrow_tid(narrow_tid);
    if (v != nil) {
	lock();
	refcount_ += 1;
	unlock();
	v = (*f)(this);
    } else {
	OxMarshalBuffer buf;
	buf.init();
	op(buf, obj, narrow_tid, OxForeignExchange_narrow_op);
	Long r = invoke(buf, false);
	if (r != 0) {
	    err_narrow(r);
	}
	if (buf.get_long() != 0) {
	    lock();
	    refcount_ += 1;
	    unlock();
	    v = (*f)(this);
	}
    }
    return v;
}

//
// The ForeignExchange duplicate operation requires a new proxy
// for the desired type.
//

void* OxForeignExchange::duplicate(void*, CORBA::TypeId, OxStubCreator f) {
    lock();
    refcount_ += 1;
    unlock();
    return (*f)(this);
}

//
// The ForeignExchange release operation _always_ deallocates
// the proxy, and if the refcount becomes zero then release
// will destroy the exchange itself.
//

void OxForeignExchange::release(CORBA::Object_ptr obj) {
    lock();
    delete obj;
    if (refcount_ > 1) {
	refcount_ -= 1;
    } else if (refcount_ == 1) {
	// Releasing the last refcount will delete this exchange, so
	// no access to the exchange should occur after the delete statement.
	delete this;
	return;
    } else {
	err_release();
    }
    unlock();
}

CORBA::TypeId OxForeignExchange::tid(CORBA::Object_ptr) {
    return tid_;
}

#if OX_RPC_EXCHANGE

// class RpcExchange

OxRpcExchange::OxRpcExchange(
    char* host, ULong port, ULong impl, ULong obj
) {
    hostname_ = host;
    port_ = port;
    impl_ = impl;
    obj_ = obj;
}

OxRpcExchange::~OxRpcExchange() { }

void OxRpcExchange::op(
    OxMarshalBuffer& b, CORBA::Object_ptr, CORBA::TypeId t, Long method
) {
    b.put_unsigned_long(impl_);
    b.put_unsigned_long(obj_);
    _ox_CORBA_TypeId_put(b, t);
    b.put_unsigned_long(method);
}

Long OxRpcExchange::invoke(OxMarshalBuffer& b, Boolean /* oneway */) {
    int r = oxCall(hostname_, port_, &b);
    if (r != 0) {
	err_oxcall(r);
    }
    return b.get_long();
}

void OxRpcExchange::marshal(OxMarshalBuffer& b, CORBA::Object_ptr) {
    _Ix_extern_xid(OxRpcExchange);
    b.put_unsigned_long(_Ix__xid(OxRpcExchange));
    b.put_string(hostname_);
    b.put_unsigned_long(port_);
    b.put_unsigned_long(impl_);
    b.put_unsigned_long(obj_);
}

void* OxRpcExchange_unmarshal(OxStubCreator f, OxMarshalBuffer& b) {
    char* host = b.get_string();
    ULong port = b.get_unsigned_long();
    ULong impl = b.get_unsigned_long();
    ULong obj = b.get_unsigned_long();
    // should do lookup to avoid creating same object twice in client
    return (*f)(new OxRpcExchange(host, port, impl, obj));
}

void* OxRpcExchange::narrow(
    CORBA::Object_ptr obj, CORBA::TypeId obj_tid, CORBA::TypeId narrow_tid,
    OxStubCreator f
) {
    void* v = obj->_narrow_tid(narrow_tid);
    if (v != nil) {
	lock();
	refcount_ += 1;
	unlock();
	v = (*f)(this);
    } else {
	OxMarshalBuffer buf;
	buf.init();
	op(buf, obj, narrow_tid, OxForeignExchange_narrow_op);
	Long r = invoke(buf, false);
	if (r != 0) {
	    err_narrow(r);
	}
	if (buf.get_long() != 0) {
	    lock();
	    refcount_ += 1;
	    unlock();
	    v = (*f)(this);
	}
    }
    return v;
}

#endif
