//
// $XConsortium$
//

//
// Copyright (c) 1993-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef ox_exchange_h
#define ox_exchange_h

#include "base.h"

class OxMarshalBuffer;

class export OxExchange {
public:
    virtual void op(
	OxMarshalBuffer&, CORBA::Object_ptr, CORBA::TypeId, Long method
    ) = 0;
    virtual Long invoke(OxMarshalBuffer&, Boolean oneway) = 0;
    virtual void marshal(OxMarshalBuffer&, CORBA::Object_ptr) = 0;

    virtual void* narrow(
	CORBA::Object_ptr obj, CORBA::TypeId obj_tid, CORBA::TypeId narrow_tid,
	OxStubCreator f
    ) = 0;
    virtual void* duplicate(
	void* obj, CORBA::TypeId narrow_tid, OxStubCreator f
    ) = 0;
    virtual void release(CORBA::Object_ptr) = 0;
    virtual CORBA::TypeId tid(CORBA::Object_ptr) = 0;

    virtual Boolean is_local();
    virtual Boolean ping(_Ix_Env* _env = 0);
    // compatibility hack
    Boolean isLocal() { return is_local(); }
public:
    OxExchange();
    virtual ~OxExchange();
};

//
// LibraryExchange maintains a ref count without locking, and passes
// an address for the object reference (which only really makes sense
// for DII).
//

class export OxLibraryExchange : public OxExchange {
public:
    OxLibraryExchange();
    ~OxLibraryExchange();

    void op(OxMarshalBuffer&, CORBA::Object_ptr, CORBA::TypeId, Long method);
    Long invoke(OxMarshalBuffer&, Boolean oneway);
    void marshal(OxMarshalBuffer&, CORBA::Object_ptr);

    void* narrow(
	CORBA::Object_ptr obj, CORBA::TypeId obj_tid, CORBA::TypeId narrow_tid,
	OxStubCreator f
    );
    void* duplicate(
	void* obj, CORBA::TypeId narrow_tid, OxStubCreator f
    );
    void release(CORBA::Object_ptr);
    CORBA::TypeId tid(CORBA::Object_ptr);

    static void* try_narrow(
	CORBA::Object_ptr obj, CORBA::TypeId obj_tid, CORBA::TypeId narrow_tid
    );
protected:
    Long refcount_;
};

//
// LockingExchange maintains a ref count with mutual exclusion.
// This is typically a base class for serverless implementations.
//

class export OxLockingExchange : public OxExchange {
public:
    OxLockingExchange();
    ~OxLockingExchange();

    void* narrow(
	CORBA::Object_ptr obj, CORBA::TypeId obj_tid, CORBA::TypeId narrow_tid,
	OxStubCreator f
    );
    void* duplicate(
	void* obj, CORBA::TypeId narrow_tid, OxStubCreator f
    );
    void release(CORBA::Object_ptr);

    void lock();
    void unlock();
protected:
    Long refcount_;
};

//
// Exchange implementation for crossing address spaces.
// Just defines duplicate by creating a new proxy.
//

class export OxForeignExchange : public OxLockingExchange {
public:
    OxForeignExchange();
    ~OxForeignExchange();

    void* narrow(
	CORBA::Object_ptr obj, CORBA::TypeId obj_tid, CORBA::TypeId narrow_tid,
	OxStubCreator f
    );
    void* duplicate(
	void* obj, CORBA::TypeId narrow_tid, OxStubCreator f
    );
    void release(CORBA::Object_ptr);
    CORBA::TypeId tid(CORBA::Object_ptr);
protected:
    CORBA::TypeId tid_;
};

static const Long OxForeignExchange_narrow_op = -1;

//
// Macros for common use of exchanges in implementation classes.
//

#define Ox_implementation_decl(Name,Interface,Exchange) \
class export Name : public Interface { \
public: \
    Exchange exch_; \
public: \
    OxExchange* _exchange();

#define Ox_implementation_body(QualifiedName) \
OxExchange* QualifiedName::_exchange() { return &exch_; }

#endif
