//
// $XConsortium$
//

//
// Copyright (c) 1993-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef ox_ix_h
#define ox_ix_h

//
// Ix translator-specific definitions.
// These are macros that Ix emits so that we can control the generated
// code by modifying this header file rather than by modifying the translator.
//
// For some of these macros, we'd prefer to use templates but
// some cfront-based compilers still can't handle mixing templates and
// nested classes.
//

//
// Name concatentation
//
// T_ptr and TRef are special because they are typedefs and
// so we don't want a prefix for them.
//

#if !defined(UNIXCPP)
#define __Ix_concat(a,b) a ## b
#define _Ix_concat(a,b) __Ix_concat(a,b)
#else
#define __Ix_concat(a,b) a/**/b
#define _Ix_concat(a,b) __Ix_concat(a,b)
#endif

#define _Ix_T_ptr(A) _Ix_concat(_Ix_concat(A,_),ptr)
#define _Ix_TRef(A) _Ix_concat(A,Ref)
#define _Ix_T_var(A) _Ix_concat(A,_var)
#define _Ix_T_field(A) _Ix_concat(A,_field)
#define _Ix_T_out(A) _Ix_concat(A,_out)

//
// Some preprocessors can't handle empty parameters, e.g., f(a,).
//

#define _Ix_NullParameter

//
// Define generated type names for local environment.
//

#ifndef _Ix_MarshalBuffer
#define _Ix_MarshalBuffer OxMarshalBuffer
#endif

#ifndef _Ix_Exception
#define _Ix_Exception CORBA::Exception
#endif

#ifndef _Ix_Exception_put
#ifdef WIN32
// workaround for Visual C++ bug (2.0 and earlier)
#define _Ix_Exception_put Exception::_put
#else
#define _Ix_Exception_put CORBA::Exception::_put
#endif
#endif

#ifndef _Ix_UserException
#define _Ix_UserException CORBA::UserException
#endif

#ifndef _Ix_Env
#define _Ix_Env OxEnv
#endif

//
// Declare a C struct named N for a sequence of type T.
//

#define _Ix_CSeq(S,T) \
typedef struct { Long _maximum, _length; T* _buffer; Boolean _release; } S

//
// Declare a C++ class named S a sequence of type T, where T is
// variable-length and not an interface or string type.
//

#define _Ix_NormalSeq(S,T) \
class export S { \
public: \
    ULong _maximum; \
    ULong _length; \
    T* _buffer; \
    Boolean _release; \
private: \
    void free() { if (_release) { delete [] _buffer; } } \
    void copy(const S&); \
public: \
    S() { CORBA::_init_seq(this, 0, 0, 0, 0); } \
    S(ULong m) { CORBA::_init_seq(this, m, 0, new T[m], 1); } \
    S(ULong m, ULong n) { CORBA::_init_seq(this, m, n, new T[m], 1); } \
    S(ULong m, ULong n, T* e, Boolean r = 0) { \
	CORBA::_init_seq(this, m, n, e, r); \
    } \
    S(const S& s) { copy(s); } \
    ~S() { free(); } \
    S& operator =(const S& s) { free(); copy(s); return *this; } \
    S* operator ->() { return this; } \
    const T& operator[](ULong index) const { return _buffer[index]; } \
    T& operator[](ULong index) { return _buffer[index]; } \
    ULong maximum() const { return _maximum; } \
    void length(ULong n) { _length = n; } \
    ULong length() const { return _length; } \
    void load(ULong m, ULong n, T* e, Boolean r = 0) { \
	free(); \
	CORBA::_init_seq(this, m, n, e, r); \
    } \
    void alloc(ULong m) { CORBA::_init_seq(this, m, 0, new T[m], 1); } \
    static T* allocbuf(ULong n) { return new T[n]; } \
}

//
// Declare a C++ class for a sequence of T where T is fixed length.
//

#define _Ix_FixedSeq(S,T) \
class export S { \
public: \
    ULong _maximum; \
    ULong _length; \
    T* _buffer; \
    Boolean _release; \
private: \
    void free() { if (_release) { delete [] _buffer; } } \
    void copy(const S& s) { CORBA::_copy_fixed_seq(&s, this, sizeof(T)); } \
public: \
    S() { CORBA::_init_seq(this, 0, 0, 0, 0); } \
    S(ULong m) { CORBA::_init_seq(this, m, 0, new T[m], 1); } \
    S(ULong m, ULong n) { CORBA::_init_seq(this, m, n, new T[m], 1); } \
    S(ULong m, ULong n, T* e, Boolean r = 0) { \
	CORBA::_init_seq(this, m, n, e, r); \
    } \
    S(const S& s) { copy(s); } \
    ~S() { free(); } \
    S& operator =(const S& s) { free(); copy(s); return *this; } \
    S* operator ->() { return this; } \
    const T& operator[](ULong index) const { return _buffer[index]; } \
    T& operator[](ULong index) { return _buffer[index]; } \
    ULong maximum() const { return _maximum; } \
    void length(ULong n) { _length = n; } \
    ULong length() const { return _length; } \
    T* buffer() const { return _buffer; } \
    void load(ULong m, ULong n, T* e, Boolean r = 0) { \
	free(); \
	CORBA::_init_seq(this, m, n, e, r); \
    } \
    void alloc(ULong m) { CORBA::_init_seq(this, m, 0, new T[m], 1); } \
    static T* allocbuf(ULong n) { return new T[n]; } \
}

//
// Declare a C++ class named S for a sequence of an interface type T.
//

#define _Ix_InterfaceSeq(S,T) \
class export S { \
public: \
    ULong _maximum; \
    ULong _length; \
    _Ix_T_field(T)* _buffer; \
    Boolean _release; \
private: \
    void free(); \
    void copy(const S&); \
public: \
    S() { CORBA::_init_seq(this, 0, 0, 0, 0); } \
    S(ULong m) { CORBA::_init_ptr_seq(this, m, 0, new _Ix_T_ptr(T)[m], 1); } \
    S(ULong m, Long n) { \
	CORBA::_init_ptr_seq(this, m, n, new _Ix_T_ptr(T)[m], 1); \
    } \
    S(ULong m, ULong n, _Ix_T_ptr(T)* e, Boolean r = 0) { \
	CORBA::_init_seq(this, m, n, e, r); \
    } \
    S(const S& s) { copy(s); } \
    ~S() { free(); } \
    S& operator =(const S& s) { free(); copy(s); return *this; } \
    S* operator ->() { return this; } \
    const _Ix_T_field(T)& operator[](ULong i) const { return _buffer[i]; } \
    _Ix_T_field(T)& operator[](ULong i) { return _buffer[i]; } \
    ULong maximum() const { return _maximum; } \
    void length(Long n) { _length = n; } \
    ULong length() const { return _length; } \
    _Ix_T_ptr(T)* buffer() const { return (_Ix_T_ptr(T)*)_buffer; } \
    void load(ULong m, ULong n, _Ix_T_ptr(T)* e, Boolean r = 0) { \
	free(); \
	CORBA::_init_seq(this, m, n, e, r); \
    } \
    void alloc(ULong m) { \
	CORBA::_init_ptr_seq(this, m, 0, new _Ix_T_ptr(T)[m], 1); \
    } \
    static _Ix_T_ptr(T)* allocbuf(ULong n) { return new _Ix_T_ptr(T)[n]; } \
}

//
// Declare a C++ class named S for a sequence of strings.
//

#define _Ix_StringSeqType CORBA::StringSeq
#define _Ix_StringSeq(S) \
typedef _Ix_StringSeqType S; \
typedef _Ix_T_var(_Ix_StringSeqType) _Ix_T_var(S); \
typedef _Ix_T_out(_Ix_StringSeqType) _Ix_T_out(S)

#define _Ix_StringSeq_put(b,s) _Ix_StringSeqType::_put(b, s)
#define _Ix_StringSeq_get(b,s) _Ix_StringSeqType::_get(b, s)

//
// Declare a field type T_field for an object reference pointer type T_ptr.
//

#define _Ix_DeclareField(T) \
class export _Ix_T_field(T) { \
public: \
    _Ix_T_ptr(T) ptr_; \
\
    _Ix_T_field(T)() { } \
    _Ix_T_field(T)(const _Ix_T_field(T)& f); \
    _Ix_T_field(T)& operator =(_Ix_T_ptr(T) p); \
    _Ix_T_field(T)& operator =(const _Ix_T_field(T)& f); \
    operator _Ix_T_ptr(T)() const { return ptr_; } \
    _Ix_T_ptr(T) operator ->() const { return ptr_; } \
};

//
// Declare a type name for a string or objref out parameter.  The idea
// is to store the address of the pointer in an object and access
// the pointer indirectly.  Also, assignment to the object will release
// the old storage.
//

#define _Ix_OutParam(P,T) \
class export P { \
public: \
    T* ptr_; \
public: \
    P(T& p) { ptr_ = &p; } \
    P(const P& p) { ptr_ = p.ptr_; } \
    P& operator =(T p) { *ptr_ = p; return *this; } \
    P& operator =(const P& p) { *ptr_ = *p.ptr_; return *this; } \
    operator T() const { return *ptr_; } \
};

//
// Declare the types associated with a forward reference of
// an interface type.
//

#define _Ix_Forward(A) \
class A; \
typedef A* _Ix_T_ptr(A); \
typedef _Ix_T_ptr(A) _Ix_TRef(A); \
_Ix_DeclareField(A) \
class _Ix_T_out(A);

//
// Declare the C types for an interface.
//

#define _Ix_C_Interface(Q) \
typedef CORBA::Object_ptr Q, _Ix_T_ptr(Q), _Ix_TRef(Q);

//
// Declare the inheritance for a subclass.
// Make this virtual public if you want virtual base classes.
//

#define _Ix_public virtual public

//
// The derivation for types that do not inherit from anything.
//

#define _Ix_BaseClass(T) class export T : public CORBA::Object

//
// The common operations for an interface type.
//

#define _Ix_InterfaceOps(R) \
    R _this() { return this; } \
    virtual void* _this_ptr(); \
    virtual CORBA::TypeId _tid(); \
    virtual void* _narrow_tid(CORBA::TypeId tid); \
    static R _narrow(CORBA::Object_ptr obj); \
    static R _duplicate(R obj);

#define _Ix_ClassInterfaceOps(R) \
    virtual void* _this_ptr() { return R::_this_ptr(); } \
    virtual CORBA::TypeId _tid() { return R::_tid(); } \
    virtual void* _narrow_tid(CORBA::TypeId tid) { return R::_narrow_tid(tid); }

#define _Ix_MaintOps(T) \
protected: \
    T(); \
    ~T();

//
// Declare the types associated with the declaration of
// an interface type.
//

#define _Ix_Interface(A) \
_Ix_OutParam(_Ix_T_out(A),_Ix_T_ptr(A))

//
// Declare the extra names when defining a typedef for an interface.
//

#define _Ix_InterfaceTypedef(A,T) \
typedef _Ix_T_ptr(A) _Ix_T_ptr(T); \
typedef _Ix_T_ptr(T) _Ix_TRef(T); \
typedef _Ix_T_var(A) _Ix_T_var(T); \
typedef _Ix_T_field(A) _Ix_T_field(T); \
typedef _Ix_T_out(A) _Ix_T_out(T)

//
// Declare a proxy type for an interface.
//

#define _Ix_Proxy(T) _Ix_concat(T,Stub)

#define _Ix_DeclareProxy(T) \
class export _Ix_Proxy(T) : public T { \
public: \
    _Ix_Proxy(T)(OxExchange*); \
    ~_Ix_Proxy(T)(); \
\
    OxExchange* _exchange(); \
protected: \
    OxExchange* exch_; \
};

//
// Call release on an object reference.
//

#define _Ix_release(obj) CORBA::release(obj)

//
// The HP C++ compiler has a problem with pre-mature
// destruction of temporaries -- function arguments wrapped
// in _vars are destroyed before the end of expression.
// This temporary solution means all _vars will leak for HP.
//
#if defined(HPArchitecture) && defined(__cplusplus)
#define _Ix_tmp_release(obj) _Ix_release(obj)
#else
#define _Ix_tmp_release(obj)
#endif

//
// Declare types for managed object references.
//
// This definition assumes that the given types T and T_ptr are
// already defined (as opposed to allowing forward references).
//
// Right now, these templates avoid operator T_ptr&() because
// some compilers get confused by this definition.
//

#define _Ix_ManagedRef(T,cast) \
class export _Ix_T_var(T) { \
private: \
    _Ix_T_ptr(T) ptr_; \
public: \
    _Ix_T_var(T)() { ptr_ = 0; } \
    _Ix_T_var(T)(_Ix_T_ptr(T) p) { ptr_ = p; } \
    _Ix_T_var(T)(const _Ix_T_var(T)& r) { \
	ptr_ = T::_duplicate(r.ptr_); \
    } \
    ~_Ix_T_var(T)() { _Ix_tmp_release(cast ptr_); } \
\
    _Ix_T_var(T)& operator =(const _Ix_T_var(T)& r) { \
	if (r.ptr_ != ptr_) { \
	    _Ix_release(cast ptr_); \
	    ptr_ = T::_duplicate(r.ptr_); \
	} \
	return *this; \
    } \
\
    _Ix_T_var(T)& operator =(const _Ix_T_field(T)& f) { \
	if (f.ptr_ != ptr_) { \
	    _Ix_release(cast ptr_); \
	    ptr_ = T::_duplicate(f.ptr_); \
	} \
	return *this; \
    } \
\
    _Ix_T_var(T)& operator =(_Ix_T_ptr(T) p) { \
	if (p != ptr_) { \
	    _Ix_release(cast ptr_); \
	    ptr_ = p; \
	} \
	return *this; \
    } \
\
    _Ix_T_ptr(T) operator ->() const { return ptr_; } \
    operator _Ix_T_ptr(T)() const { return ptr_; } \
\
    _Ix_T_ptr(T)& inout() { return ptr_; } \
    _Ix_T_ptr(T)& out() { return ptr_; } \
    operator _Ix_T_out(T)() { return _Ix_T_out(T)(ptr_); } \
};

//
// Helper classes for managing pointers to aggregate types.
//

#define _Ix_CommonManaged(T) \
_Ix_OutParam(_Ix_T_out(T),T*) \
\
class export _Ix_T_var(T) { \
private: \
    T* ptr_; \
public: \
    _Ix_T_var(T)() { ptr_ = 0; } \
    _Ix_T_var(T)(T* p) { ptr_ = p; } \
    _Ix_T_var(T)(const _Ix_T_var(T)& r) { \
	ptr_ = new T(*r.ptr_); \
    } \
    ~_Ix_T_var(T)() { delete ptr_; } \
\
    _Ix_T_var(T)& operator =(const _Ix_T_var(T)& r) { \
	if (r.ptr_ != ptr_) { \
	    delete ptr_; \
	    ptr_ = new T(*r.ptr_); \
	} \
	return *this; \
    } \
\
    _Ix_T_var(T)& operator =(T* p) { \
	if (p != ptr_) { \
	    delete ptr_; \
	    ptr_ = p; \
	} \
	return *this; \
    } \
\
    T* operator ->() const { return ptr_; } \
    operator _Ix_T_out(T)() { delete ptr_; return _Ix_T_out(T)(ptr_); } \
    T*& out() { delete ptr_; return ptr_; }

#define _Ix_Managed(T) \
_Ix_CommonManaged(T) \
}

#define _Ix_ManagedSeq(T,E) \
_Ix_CommonManaged(T) \
    Long maximum() const { return ptr_->maximum(); } \
    Long length() const { return ptr_->length(); } \
    E& operator[](Long index) { return ptr_->_buffer[index]; } \
}

//
// Declare the type names for a typedef to an aggregate.
//

#define _Ix_AggrTypedef(S,T) \
typedef _Ix_T_var(S) _Ix_T_var(T); \
typedef _Ix_T_out(S) _Ix_T_out(T)

//
// Declare the cast (narrow) functions for an exception, if any.
//

#define _Ix_ExceptCast(T) \
static T* _cast(_Ix_Exception* e) { return (T*)_check_cast(e); } \
static const T* _cast(const _Ix_Exception* e) { \
    return (const T*)_check_cast(e); \
} \
static void* _check_cast(const _Ix_Exception*);

#endif
