//
// $XConsortium$
//

//
// Copyright (c) 1993-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef ox_ix_rep_h
#define ox_ix_rep_h

//
// This file contains definitions for manipulating the binary representation
// of type ids and exchange ids.

#include "base.h"
#include "schema.h"

//
// Common 32-bit type id.  Formats are defined as follows:
//
#define IX_TID_RESERVED_0 0
#define IX_TID_SHORT 1
#define IX_TID_MEDIUM_2 2
#define IX_TID_MEDIUM_3 3
#define IX_TID_MEDIUM_4 4
#define IX_TID_RESERVED_5 5
#define IX_TID_RESERVED_6 6
#define IX_TID_RESERVED_7 7
#define IX_TID_RESERVED_8 8
#define IX_TID_RESERVED_9 9
#define IX_TID_RESERVED_10 10
#define IX_TID_RESERVED_11 11
#define IX_TID_RESERVED_12 12
#define IX_TID_RESERVED_13 13
#define IX_TID_RESERVED_14 14
#define IX_TID_LONG 15

//
// Need a better way to figure this out.
//

#if defined(WIN32) || defined(i386) || defined (__alpha)
#define BigEndian 0
#else
#define BigEndian 1
#endif

struct OxShortTypeId {
#if BigEndian
    ULong format : 4;
    ULong tid : 28;
#else
    ULong tid : 28;
    ULong format : 4;
#endif
};

//
// Compare two type ids.
//
Boolean _ox_equal_long_tid(CORBA::TypeId t1, CORBA::TypeId t2);

inline Boolean _ox_equal_tid(CORBA::TypeId t1, CORBA::TypeId t2) {
    return (
	*t1 == *t2 && (
	    ((OxShortTypeId*)t1)->format == IX_TID_SHORT ||
	    _ox_equal_long_tid(t1, t2)
	)
    );
}

//
// Put a type id into a marshal buffer.
//

void _ox_CORBA_TypeId_put(OxMarshalBuffer&, CORBA::TypeId);

//
// Put an exchange id into a marshal buffer.
//

void _OxExchangeId_put(OxMarshalBuffer&, OxExchangeId);
void _OxExchangeId_put_nil(OxMarshalBuffer&);
void* _OxExchangeId_get(OxMarshalBuffer&, OxStubCreator);

#endif
