//
// $XConsortium$
//

//
// Copyright (c) 1993-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef ox_request_h
#define ox_request_h

#include "env.h"
#include "schema.h"

class OxMarshalBuffer;

class OxRequestImpl : public CORBA::Request {
public:
    OxRequestImpl(CORBA::Object_ptr, const char* operation);
    ~OxRequestImpl();

    OxExchange* _exchange();

    CORBA::Object_ptr target() const;
    const char* operation() const;
    _Ix_Env* env();
    CallStatus invoke();
    CallStatus send_oneway();
    CallStatus op_info(CORBA::Type::OpInfo& op);
    CORBA::Type_ptr except_info();
    void put_char(Char value);
    Char get_char();
    void put_boolean(Boolean value);
    Boolean get_boolean();
    void put_octet(Octet value);
    Octet get_octet();
    void put_short(Short value);
    Short get_short();
    void put_unsigned_short(UShort value);
    UShort get_unsigned_short();
    void put_long(Long value);
    Long get_long();
    void put_unsigned_long(ULong value);
    ULong get_unsigned_long();
    void put_float(Float value);
    Float get_float();
    void put_double(Double value);
    Double get_double();
    void put_string(const char* value);
    char* get_string();
    void put_object(CORBA::Object_ptr obj);
    CORBA::Object_ptr get_object();
    void begin_aggregate();
    void end_aggregate();

    CORBA::Request::CallStatus status(CORBA::Request::CallStatus);
    Boolean is_ok();
protected:
    OxLibraryExchange exchange_;
    CORBA::Object_ptr target_;
    char* operation_;
    CORBA::TypeId resolved_;
    CORBA::Type::OpInfo opinfo_;
    Long index_;
    OxMarshalBuffer* buffer_;
    OxEnv env_;
    CORBA::Request::CallStatus status_;
    OxSchema::ExceptDescriptor* except_;
    ULong argc_;
    ULong nesting_;

    void init();
    void free();
    Boolean find_op(CORBA::Type_ptr, Long& index);
    void resolve(OxSchema::InterfaceDescriptor*);
    CORBA::Request::CallStatus send(Boolean wait);
};

inline Boolean OxRequestImpl::is_ok() {
    return status_ == CORBA::Request::ok;
}

#endif
