//
// Copyright (c) 1992-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef dish_h
#define dish_h

#include "base.h"
#include "schema.h"
#include <tcl.h>

//
// Check whether this platform supports %p for printf/scanf
// on pointers.
//

#undef pformat
#if (defined(sun) && !defined(SVR4)) || defined(linux)
#define pformat 0
#else
#define pformat 1
#endif

#define Tcl_proc(name) \
static int name(ClientData, Tcl_Interp*, int argc, char* argv[])

#define Tcl_proc_body(name,data,argv) \
int name(ClientData data, Tcl_Interp* interp, int argc, char* argv[])

#define marshal_aggregate(name) \
CORBA::Boolean name( \
    CORBA::Type_ptr, char*, CORBA::Request_ptr, Tcl_Interp* \
)

#define marshal_aggregate_body(name,type,interp) \
CORBA::Boolean Dish::name( \
    CORBA::Type_ptr type, char* arg, CORBA::Request_ptr req, \
    Tcl_Interp* interp \
)

#define unmarshal_aggregate(name) \
void name( \
    CORBA::Type_ptr, CORBA::Request_ptr, Tcl_Interp* \
)

#define unmarshal_aggregate_body(name,type,interp) \
void Dish::name( \
    CORBA::Type_ptr type, CORBA::Request_ptr req, Tcl_Interp* interp \
)

class Dish {
public:
    Dish(Tcl_Interp*);
    ~Dish();
    
    virtual void init(Tcl_Interp*);
    virtual CORBA::Object_ptr resolve(const char* name);

    CORBA::Boolean interactive() { return interactive_; }
    CORBA::Boolean verbose() { return verbose_; }
    CORBA::Object_ptr context() { return context_; }

    CORBA::Boolean arg_to_object(
	Tcl_Interp*, const char* arg, CORBA::Type_ptr, CORBA::Object_ptr&
    );
    CORBA::Boolean string_to_object(const char*, CORBA::Object_ptr&);
    char* object_to_string(CORBA::Object_ptr, char* result);

    // thread support
    Tcl_Interp* clone(Tcl_Interp*);
    void save_globals(Tcl_Interp*, FILE*);
    void save_global(Tcl_Interp*, char* var, FILE*);
    void save_procs(Tcl_Interp*, FILE* f);
    void save_proc(Tcl_Interp*, char* proc, FILE*);

    char* try_expand(char* arg, Tcl_Interp*);

    static Dish* instance();
    static void cleanup();
protected:
    void restore_args(Tcl_Interp*, int& argc, char**& argv);

    CORBA::Object_ptr context_;
private:
    CORBA::Boolean dispatch(
	CORBA::Object_ptr, int argc, char* argv[], Tcl_Interp*
    );
    void set_call_status(
	CORBA::Request_ptr, CORBA::Request::CallStatus,
	CORBA::Object_ptr, int, char* op, Tcl_Interp*
    );
    CORBA::Boolean parse_args(
        int argc, char* argv[], CORBA::Request_ptr, CORBA::Type::OpInfo&,
	Tcl_Interp*
    );
    CORBA::Boolean add_arg(
	CORBA::Type_ptr, char* arg, CORBA::Request_ptr, Tcl_Interp*
    );

    CORBA::Boolean interactive_;
    CORBA::Boolean verbose_;
    static Dish* instance_;

    // marshalling arguments
    marshal_aggregate(add_interface);
    marshal_aggregate(add_array);
    marshal_aggregate(add_struct);
    marshal_aggregate(add_sequence);

    Boolean split_list(
	Tcl_Interp*, const char* arg, int&, char**&, char* type_name
    );

    // getting results
    void get_results(
        int argc, char* argv[], CORBA::Request_ptr, CORBA::Type::OpInfo&,
	Tcl_Interp*
    ); 
    void get_result(CORBA::Type_ptr, CORBA::Request_ptr, Tcl_Interp*);
    void get_concrete(CORBA::Type::TCKind, CORBA::Request_ptr, Tcl_Interp*);
    void get_string(CORBA::Request_ptr, Tcl_Interp*);
    unmarshal_aggregate(get_interface);
    unmarshal_aggregate(get_array);
    unmarshal_aggregate(get_enum);
    unmarshal_aggregate(get_struct);
    unmarshal_aggregate(get_except);
    unmarshal_aggregate(get_sequence);

    // Dish-specific Tcl commands
    Tcl_proc(verbose);
    Tcl_proc(scope);
    Tcl_proc(object);
    Tcl_proc(object_invoke);
    Tcl_proc(invoke);
    Tcl_proc(invoke_tmp);
    Tcl_proc(duplicate);
    Tcl_proc(release);
    Tcl_proc(get_current_context);
    Tcl_proc(resolve_cmd);
    Tcl_proc(isa);
    Tcl_proc(char_string);
    Tcl_proc(drawing_kit);
	
    static void unobject(ClientData);
};

#endif
