//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * commands.h 
 */

#ifndef _commands_h
#define _commands_h

#include <Fresco/defs/display.h>
#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/table.h>
#include <Fresco/widgets/dialogs.h>
#include <Fresco/figures/figures.h>

class FCommand;
class FigViewer;
class GlyphImpl;
class KeyList;
class Manipulator;
class SelectInfo;
class Selection;

class GlyphImplMap {
public:
    GlyphImplMap();
    virtual ~GlyphImplMap();
    void clear();
    void map(Glyph* key, Glyph* obj);
    Glyph* find(Glyph* key);
private:
    KeyList* keylist_;
};
    
declarePtrList(CmdList, FCommand);

//- CmdData
//+ CmdData : XfFrescoObject
class CmdData : public XfFrescoObject {
public:
    ~CmdData();
    _Ix_InterfaceOps(CmdData*)
//+
protected:
    CmdData();
};

declareTable(CmdDataTable, long, CmdData*);

//- FCommand*
//+ FCommand : ActionImpl
class FCommand : public ActionImpl {
public:
    ~FCommand();
    _Ix_InterfaceOps(FCommand*)
//+
public:
    virtual void store(long key, CmdData* data);
    virtual CmdData* recall(long key);
    virtual void log();
    virtual void execute();

    FigViewer* figviewer();
protected:
    FCommand(FigViewer* = nil);
protected:
    CmdDataTable* data_;
    FigViewer* figviewer_;
    Selection* selection_;
};

inline FigViewer* FCommand::figviewer () { return figviewer_; }

//- GroupCmd*
//+ GroupCmd : FCommand
class GroupCmd : public FCommand {
public:
    ~GroupCmd();
    _Ix_InterfaceOps(GroupCmd*)
//+
public:
    GroupCmd(FigViewer* = nil, Manipulator* = nil);
    virtual void execute();
protected:
    Manipulator* m_;
};

//- UngroupCmd*
//+ UngroupCmd : FCommand
class UngroupCmd : public FCommand {
public:
    ~UngroupCmd();
    _Ix_InterfaceOps(UngroupCmd*)
//+
public:
    UngroupCmd(FigViewer* = nil);
};

//- DeleteCmd*
//+ DeleteCmd : FCommand
class DeleteCmd : public FCommand {
public:
    ~DeleteCmd();
    _Ix_InterfaceOps(DeleteCmd*)
//+
public:
    DeleteCmd(FigViewer* = nil);
};

//- InstanceCmd*
//+ InstanceCmd : FCommand
class InstanceCmd : public FCommand {
public:
    ~InstanceCmd();
    _Ix_InterfaceOps(InstanceCmd*)
//+
public:
    InstanceCmd(FigViewer* = nil);
};

//- CopyCmd*
//+ CopyCmd : FCommand
class CopyCmd : public FCommand {
public:
    ~CopyCmd();
    _Ix_InterfaceOps(CopyCmd*)
//+
public:
    CopyCmd(FigViewer* = nil);
public:
    static GlyphImplMap glyphmap_;
};

//- NarrowCmd*
//+ NarrowCmd : FCommand
class NarrowCmd : public FCommand {
public:
    ~NarrowCmd();
    _Ix_InterfaceOps(NarrowCmd*)
//+
    NarrowCmd(FigViewer* = nil);
    NarrowCmd(FigViewer*, Long delta);

    Long delta();
protected:
    Long delta_;
};

//- WidenCmd*
//+ WidenCmd : FCommand
class WidenCmd : public FCommand {
public:
    ~WidenCmd();
    _Ix_InterfaceOps(WidenCmd*)
//+
    WidenCmd(FigViewer* = nil);
    WidenCmd(FigViewer*, Long delta);

    virtual void execute();
    virtual void unexecute();

    Long delta();
protected:
    Long delta_;
};

//- NaturalCmd*
//+ NaturalCmd : FCommand
class NaturalCmd : public FCommand {
public:
    ~NaturalCmd();
    _Ix_InterfaceOps(NaturalCmd*)
//+
    NaturalCmd(FigViewer* = nil);
};

//- SelectCmd*
//+ SelectCmd : FCommand
class SelectCmd : public FCommand {
public:
    ~SelectCmd();
    _Ix_InterfaceOps(SelectCmd*)
//+
    SelectCmd(FigViewer* = nil, Boolean selected = true);

    Boolean selected ();
protected:
    Boolean selected_;
};

inline Boolean SelectCmd::selected () { return selected_; }

//- MacroCmd*
//+ MacroCmd : FCommand
class MacroCmd : public FCommand {
public:
    ~MacroCmd();
    _Ix_InterfaceOps(MacroCmd*)
//+
    MacroCmd(FigViewer* = nil);

    virtual Boolean reversible();
    virtual void execute();
    virtual void unexecute();

    virtual void prepend(FCommand*);
    virtual void append(FCommand*);
    virtual void insert(long index, FCommand*);
    virtual void remove(long index);
    virtual void remove_all();
    virtual long count();

    FCommand* command(long);
protected:
    CmdList* cmdlist_;

};

//- SelectInfoCmd*
//+ SelectInfoCmd : FCommand
class SelectInfoCmd : public FCommand {
public:
    ~SelectInfoCmd();
    _Ix_InterfaceOps(SelectInfoCmd*)
//+
    SelectInfoCmd(FigViewer* = nil);

    void select_info(SelectInfo*);
    SelectInfo* select_info();
protected:
    SelectInfo* sinfo_;
};

inline void SelectInfoCmd::select_info(SelectInfo* s) { sinfo_ = s; }
inline SelectInfo* SelectInfoCmd::select_info() { return sinfo_; }

//- NewCmd*
//+ NewCmd : FCommand
class NewCmd : public FCommand {
public:
    ~NewCmd();
    _Ix_InterfaceOps(NewCmd*)
//+
public:
    NewCmd(FigViewer* = nil, Manipulator* = nil);
    virtual void execute();
protected:
    Manipulator* root_;
};

//- QuitCmd*
//+ QuitCmd : FCommand
class QuitCmd : public FCommand {
public:
    ~QuitCmd();
    _Ix_InterfaceOps(QuitCmd*)
//+
public:
    QuitCmd(FigViewer* = nil);
    virtual void execute();
};

//- SelectAllCmd*
//+ SelectAllCmd : FCommand
class SelectAllCmd : public FCommand {
public:
    ~SelectAllCmd();
    _Ix_InterfaceOps(SelectAllCmd*)
//+
public:
    SelectAllCmd(FigViewer* = nil);
    virtual void execute();
};

//- AboutCmd*
//+ AboutCmd : FCommand
class AboutCmd : public FCommand {
public:
    ~AboutCmd();
    _Ix_InterfaceOps(AboutCmd*)
//+
public:
    AboutCmd(FigViewer*, DialogKit_ptr d, FigureKit_ptr f);
    virtual void execute();
protected:
    DialogKit_ptr dialogs_;
    FigureKit_ptr figures_;
};

#endif
