//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * figviewer.h
 */

#ifndef _figviewer_h
#define _figviewer_h

#include <Fresco/defs/drawing.h>
#include <Fresco/figures/figures.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/thread.h>
#include <Fresco/defs_impl/viewers.h>

class EventAction;
class NarrowCmd;
class Manipulator;
class Selection;
class Tool;
class TransformImpl;
class WidenOffsetList;

declarePtrList(WidenGlyphList,Glyph)

class FigViewer : public BgViewer {
public:
    FigViewer(Boolean* editing);
    ~FigViewer();
    
    void allocations(XfGlyph::AllocationInfoSeq& a); //+ XfGlyph::allocations
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    XfTransform_ptr transformation(); //+ XfGlyph::transformation
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    void need_resize(); //+ XfGlyph::need_resize

    Transform_ptr tform();
    Tool* current_tool();
    void current_tool(Tool*);

    void narrow(Long level);
    void narrowgt(GlyphTraversal_ptr);
    WidenGlyphList* widen_list();

    Boolean press(GlyphTraversal_ptr, Event_ptr);
    Boolean drag(GlyphTraversal_ptr, Event_ptr);
    Boolean move(GlyphTraversal_ptr, Event_ptr);
    Boolean release(GlyphTraversal_ptr, Event_ptr);

    void visit_trail(GlyphTraversal_ptr t);
    void child_allocate(Glyph::AllocationInfo&);
    
    Manipulator* root();
    void root(Manipulator*);
    Selection* selection();
protected:
    Boolean* editing_;
    Tool* curtool_;
    Selection* sel_;
    Manipulator* active_;
    EventAction* action_;
    Thread_ptr thread_;
    Transform_ptr tx_;

    GlyphTraversal_var narrowgt_;
    WidenGlyphList* glyphs_;
    WidenOffsetList* offsets_;
};

#endif
