//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * main.cxx
 */

#include <Fresco/defs/fresco.h>
#include <Fresco/widgets/widgets.h>
#include "fdraw.h"

/* WinMain for Win32 */
#ifdef WIN32
#include <wtypes.h>

int main(int, char**);

int WINAPI WinMain(
    HINSTANCE  hInstance,       // handle of current instance
    HINSTANCE  hPrevInstance,   // handle of previous instance
    LPSTR  lpszCmdLine, // address of command line
    int  nCmdShow       // show state of window
) {
    int argc = 1;
    char* argv[] = {"fdraw", 0};
    return main(argc, argv);
}
#endif

static Option options[] = {
    { nil }
};

int main(int argc, char** argv) {
    FrescoContext* fc = FrescoLib::init_default_context(
	"FDraw", argc, argv, options
    );

    Viewer_var fdraw = new FDraw;

    WidgetKit_var widgets = Fresco_resolve_object(WidgetKit, "widgets_impl");
    Viewer_var viewer = widgets->viewer_group(fdraw);
    viewer->append_viewer(fdraw);
 
    Display_var display = fc->display_ptr();
    Screen_var screen = display->default_screen();
    Window_var window = screen->application(viewer);
 
    viewer->request_focus(fdraw, false);
 
    window->map();
    display->run(true);

    return 0;
}
