//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * selection.c
 */

#include "manipulators.h"
#include "selection.h"

Selection::Selection (long size) {
    mlist_ = new ManipList(size);
}

Selection::~Selection () {
    delete mlist_;
}

Selection* Selection::copy () {
    Selection* newcopy = new Selection(mlist_->count());
    for (long i = 0; i < mlist_->count(); i++) {
        newcopy->add(mlist_->item(i));
    }
    return newcopy;
}

long Selection::count () const {
    return mlist_->count();
}

Manipulator* Selection::item (long index) {
    return mlist_->item(index);
}

void Selection::add (Manipulator* m) {
    if (!selected(m)) {
        mlist_->append(m);
    }
}

void Selection::remove(Manipulator* m) {
    for (long i = 0; i < mlist_->count(); i++) {
        if (mlist_->item(i) == m) {
            mlist_->remove(i);
            break;
        }
    }
}

void Selection::remove_all () {
    mlist_->remove_all();
}

Boolean Selection::selected (Manipulator* m) {
    for (long i = 0; i < mlist_->count(); i++) {
        if (mlist_->item(i) == m) {
            return true;
        }
    }
    return false;
}
