//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * tools.h
 */

#ifndef _tools_h
#define _tools_h

#include <Fresco/defs/glyph.h>
#include <Fresco/defs/types.h>
#include <Fresco/defs/viewer.h>
#include <Fresco/defs_impl/transform.h>

class FigViewer;
class Manipulator;
class RegionImp;

class SelectInfo {
public:
    SelectInfo();
    virtual ~SelectInfo();
    SelectInfo* copy();
public:
    RegionImpl* a_; // traversal allocation for manipulator
    TransformImpl t_; // traversal transform for manipulator
    Manipulator* m_; // selected manipulator
    long level_; // depth of a selected manipulator
};

//- Tool*
//+ Tool : FrescoObject
class Tool : public FrescoObject {
public:
    ~Tool();
    _Ix_InterfaceOps(Tool*)
//+
public:
    OxExchange* _exchange();

    virtual SelectInfo* create_manipulator(
        GlyphTraversal_ptr, Event_ptr, FigViewer*
    ) = 0;
    void anchor(Coord&, Coord&);  // for multiple selections
protected:
    Tool();
protected:
    SharedFrescoObjectImpl object_;
    Coord ax_, ay_;
};

//- SelectTool*
//+ SelectTool : Tool
class SelectTool : public Tool {
public:
    ~SelectTool();
    _Ix_InterfaceOps(SelectTool*)
//+
public:
    SelectTool();

    virtual SelectInfo* create_manipulator(
        GlyphTraversal_ptr, Event_ptr, FigViewer*
    );
};

//- CreateTool*
//+ CreateTool : Tool
class CreateTool : public Tool {
public:
    ~CreateTool();
    _Ix_InterfaceOps(CreateTool*)
//+
public:
    CreateTool(Manipulator*);

    virtual SelectInfo* create_manipulator(
        GlyphTraversal_ptr, Event_ptr, FigViewer*
    );
protected:
    Manipulator* manip_;
};

//- CommonTool*
//+ CommonTool : Tool
class CommonTool : public Tool {
public:
    ~CommonTool();
    _Ix_InterfaceOps(CommonTool*)
//+
public:
    virtual SelectInfo* create_manipulator(
	GlyphTraversal_ptr, Event_ptr, FigViewer*
    );
protected:
    CommonTool();
};

//- MoveTool*
//+ MoveTool : CommonTool
class MoveTool : public CommonTool {
public:
    ~MoveTool();
    _Ix_InterfaceOps(MoveTool*)
//+
public:
    MoveTool();
};

//- ScaleTool*
//+ ScaleTool : CommonTool
class ScaleTool : public CommonTool {
public:
    ~ScaleTool();
    _Ix_InterfaceOps(ScaleTool*)
//+
public:
    ScaleTool();
};

//- RotateTool*
//+ RotateTool : CommonTool
class RotateTool : public CommonTool {
public:
    ~RotateTool();
    _Ix_InterfaceOps(RotateTool*)
//+
public:
    RotateTool();
};

//- AlterTool*
//+ AlterTool : CommonTool
class AlterTool : public CommonTool {
public:
    ~AlterTool();
    _Ix_InterfaceOps(AlterTool*)
//+
public:
    AlterTool();
};

//- ResizeTool*
//+ ResizeTool : CommonTool
class ResizeTool : public CommonTool {
public:
    ~ResizeTool();
    _Ix_InterfaceOps(ResizeTool*)
//+
public:
    ResizeTool();
};

//- NarrowTool*
//+ NarrowTool : CommonTool
class NarrowTool : public CommonTool {
public:
    ~NarrowTool();
    _Ix_InterfaceOps(NarrowTool*)
//+
public:
    NarrowTool();
    virtual SelectInfo* create_manipulator(
	GlyphTraversal_ptr, Event_ptr, FigViewer*
    );
    Long delta();
protected:
    Long delta_;
};

#endif
