//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * visitors.h
 */

#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/list.h>

class FCommand;
class ManipList;

declarePtrList(AllocationList, Region);
declarePtrList(TransformList, Transform);

class GlyphVisitor {
public:
    GlyphVisitor();
    virtual ~GlyphVisitor();

    virtual Boolean visit(Glyph_ptr g, GlyphOffset_ptr offset);
    virtual void visit_children(Glyph_ptr g);
    virtual void visit_parents(Glyph_ptr g);
};

class Appender : public GlyphVisitor {
public:
    Appender(Glyph_ptr, Boolean resize = true);
    ~Appender();
    Boolean visit(Glyph_ptr, GlyphOffset_ptr);
protected:
    Glyph_ptr glyph_;
    Boolean resize_;
};

class Remover : public GlyphVisitor {
public:
    Remover();
    ~Remover();
    Boolean visit(Glyph_ptr, GlyphOffset_ptr);
protected:
    GlyphOffsetList list_;
};

class Counter : public GlyphVisitor {
public:
    Counter();
    Boolean visit(Glyph_ptr, GlyphOffset_ptr);
    long count();
protected:
    long count_;
};

inline long Counter::count () { return count_; }

class CmdVisitor : public GlyphVisitor {
public:
    CmdVisitor(FCommand* cmd, Boolean execute = true);
    ~CmdVisitor();

    Boolean visit(Glyph_ptr, GlyphOffset_ptr);
protected:
    FCommand* cmd_;
    Boolean execute_;
};

class ManipCopier : public GlyphVisitor {
public:
    ManipCopier(Boolean shallow = true);
    ~ManipCopier();

    Boolean visit(Glyph_ptr, GlyphOffset_ptr);
    ManipList* manipulators();
private:
    ManipList* maniplist_;
    Boolean shallow_;
};

inline ManipList* ManipCopier::manipulators () { return maniplist_; }

class TAManipCopier : public ManipCopier {
public:
    TAManipCopier(Region_ptr a, Boolean shallow = true);
    ~TAManipCopier();

    Boolean visit(Glyph_ptr, GlyphOffset_ptr);

    AllocationList* allocations();
    TransformList* transforms();
private:
    AllocationList* alist_;
    TransformList* tlist_;
    Region_ptr a_;
};

inline AllocationList* TAManipCopier::allocations () { return alist_; }
inline TransformList* TAManipCopier::transforms () { return tlist_; }

class OffsetVisitor : public GlyphVisitor {
public:
    OffsetVisitor();
    ~OffsetVisitor();

    Boolean visit(Glyph_ptr, GlyphOffset_ptr);
    GlyphOffset_ptr offset(long);
    long offset_count();
protected:
    GlyphOffsetList* glist_;
};

