/*
 * Copyright (c) 1992-1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_defs_fcontext_h
#define Fresco_defs_fcontext_h

#include <Fresco/defs/display.h>
#include <Fresco/defs/types.h>
#include <Fresco/defs_impl/naming.h>

//- FrescoContext
class export FrescoContext : public Naming::NamingContext {
    //. A FrescoContext is a special naming context that resolves
    //. Fresco objects and exports information specified when
    //. the context is created, such as a class name and argument list.
protected:
    FrescoContext();
    ~FrescoContext();
public:
    //- name
    virtual const char* name() = 0;
	//. Return the class name associated with the context.

    //- argc
    virtual int argc() = 0;
	//. Return the number of arguments associated with the context.

    //- argv
    virtual char** argv() = 0;
	//. Return the argument strings associated with the context.

    //- style
    virtual StyleContext_ptr style() = 0;
	//. Return the style associated with the context.

    //- display_context
    virtual DisplayContext_ptr display_context() = 0;
	//. Return the display context accessible from this context.
	//. This operation is equivalent to doing a resolve
	//. on the name "fdisplay" and narrowing the result
	//. to a DisplayContext.

    //- display_ptr
    virtual Display_ptr display_ptr() = 0;
	//. Return the open display associated with the creation
	//. of this context, or nil if none was specified.

    //- default_drawing_kit
    virtual void default_drawing_kit(DrawingKit_ptr) = 0;
    virtual DrawingKit_ptr default_drawing_kit() = 0;
	//. Set or get the default drawing kit for use
	//. by other objects.  Initially, the drawing kit will
	//. be defined by the display opened when this context
	//. is created.
};

#define Fresco_resolve_object_impl(T,name) \
    T::_narrow( \
	FrescoLib::tmp_resolve_string(FrescoLib::current_context(), name) \
    )

//- Fresco_resolve_object
#define Fresco_resolve_object(T,name) Fresco_resolve_object_impl(T,name)
    //. This macro is short-hand for looking up a name using the
    //. current context and narrowing the result to a particular type.
    //. The return value is an object reference for interface type T
    //. that results from narrowing the result of resolving the name
    //. on the current Fresco naming context.
//-

#endif
