/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/Ox/schema.h>

#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/memory.h>
#include <Fresco/defs_impl/rasters.h>
#include <Fresco/defs_impl/table.h>
#include <Fresco/defs_impl/transform.h>

#include <Fresco/defs_Mac/MacDisplay.h>
#include <Fresco/defs_Mac/MacDrawing.h>

// class MacDrawingKit

MacDrawingKit::MacDrawingKit(MacDisplay* d) {
    display_ = Display::_duplicate(d);
}

MacDrawingKit::~MacDrawingKit() {
    CORBA::release(display_);
}

//+ MacDrawingKit(XfDrawingKit::find_font)
XfFont_ptr MacDrawingKit::find_font(XfCharString_ptr fullname) {
    Font_ptr f = nil;
    if (!fonts_->find(f, fullname)) {
        CharStringBuffer cs(fullname);
        const char* lfd = cs.string();

        short font;
        short face;
        short size = 12;
        short mode;

        f = new MacFont(this, font, face, size, mode);
        fonts_->insert(fullname, f);
    }
    return Font::_duplicate(f);
}

//+ MacDrawingKit(XfDrawingKit::find_font_match)
XfFont_ptr MacDrawingKit::find_font_match(XfCharString_ptr font_family, XfCharString_ptr font_style, XfCoord ptsize) {
    return nil;
}

//+ MacDrawingKit(XfDrawingKit::bitmap_from_char)
XfRaster_ptr MacDrawingKit::bitmap_from_char(XfFont_ptr f, XfCharCode c) {
    MacFont* fi = MacFont::_narrow(f);
    return fi->bitmap(c);
}

Display_ptr MacDrawingKit::display_ptr() {
    return Display::_duplicate(display_);
}

// class MacFont

//+ MacFont(MacFont)
void* MacFont::_this_ptr() { return this; }
_Ix_extern__tid(MacFont);
void* MacFont::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(MacFont) == t) return this;
    void* p;
    if (p = XfFont::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(MacFont);
_Ix_ImplementGetTid(MacFont,_Ix__tid(MacFont))
_Ix_ImplementNarrow(MacFont*,MacFont,_Ix_NullParameter,MacFont,0)
_Ix_ImplementDuplicate(MacFont*,MacFont,_Ix_NullParameter,_Ix__tid(MacFont),0)
//+

MacFont::MacFont(
    MacDrawingKit* drawing_kit,
    short font, short face, short size, short mode
) {
    drawing_kit_ = drawing_kit;
    font_ = font;
    face_ = face;
    size_ = size;
    mode_ = mode;
    
	TextFont(font_);
	TextFace(face_);
	TextSize(size_);
	TextMode(mode_);	
	GetFontInfo(&info_);
	widths_ = new int[255];
    for (int c = 0; c < 255; ++c) {
        widths_[c] = CharWidth(c);
    }   
}

MacFont::~MacFont() {
    delete widths_;
}

OxExchange* MacFont::_exchange() { return &object_; } 

//+ MacFont(XfFrescoObject::=object_.)
XfTag MacFont::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void MacFont::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void MacFont::disconnect() {
    object_.disconnect();
}
void MacFont::notify_observers() {
    object_.notify_observers();
}
void MacFont::update() {
    object_.update();
}
//+

//+ MacFont(XfFont::hash)
ULong MacFont::hash() {
    return (ULong) font_;
}

//+ MacFont(XfFont::equal)
Boolean MacFont::equal(XfFont_ptr f) {
    return f == this;
}

//+ MacFont(XfFont::fullname)
XfCharString_ptr MacFont::fullname() {
    // return CharString::_duplicate(name_);
    return nil;
}

//+ MacFont(XfFont::encoding)
XfCharString_ptr MacFont::encoding() {
    // return CharString::_duplicate(encoding_);
    return nil;
}

//+ MacFont(XfFont::point_size)
XfCoord MacFont::point_size() {
    return size_;
}

//+ MacFont(XfFont::font_info)
void MacFont::font_info(XfFont::Info& i) {
    i.left_bearing = 0;
    i.right_bearing = 0;
    i.width = info_.widMax;
    i.ascent = info_.ascent;
    i.descent = info_.descent;
    i.font_ascent = i.ascent;
    i.font_descent = i.descent;
    i.height = i.font_ascent + i.font_descent;
}

//+ MacFont(XfFont::char_info)
void MacFont::char_info(XfCharCode c, XfFont::Info& i) {
    i.left_bearing = 0;
    i.right_bearing = 0;
    i.width = widths_[c];

    i.ascent = info_.ascent;
    i.descent = info_.descent;
    i.font_ascent = i.ascent;
    i.font_descent = i.descent;
    i.height = i.font_ascent + i.font_descent;
}

//+ MacFont(XfFont::string_info)
void MacFont::string_info(XfCharString_ptr s, XfFont::Info& i) {
    i.left_bearing = 0;
    i.right_bearing = 0;
    i.width = 0;
    CharString::Data data(s->count(), s->count());
    s->get_data(data);
    for (ULong j = 0; j < data.length(); ++j) {
	    i.width += widths_[data[j]];
    }
    
    i.ascent = info_.ascent;
    i.descent = info_.descent;
    i.font_ascent = i.ascent;
    i.font_descent = i.descent;
    i.height = i.font_ascent + i.font_descent;
}

Raster_ptr MacFont::bitmap(CharCode c) {
    // !!! unimplemented
    return nil;
}
