/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef MacDrawing_h
#define MacDrawing_h

#include <Fresco/defs_impl/drawings.h>

class MacDisplay;

class MacDrawingKit : public DrawingKitBase {
public:
    MacDrawingKit(MacDisplay*);
    ~MacDrawingKit();
    
    XfFont_ptr find_font(XfCharString_ptr fullname);
    XfFont_ptr find_font_match(XfCharString_ptr font_family, XfCharString_ptr font_style, XfCoord ptsize);
    XfRaster_ptr bitmap_from_char(XfFont_ptr f, XfCharCode c);

    Display_ptr display_ptr();

private:
    Display_ptr display_;
};

class MacFont : public XfFont {
public:
    MacFont(
        MacDrawingKit* drawing_kit,
        short font, short face, short size, short mode
    );
    ~MacFont();
    
    _Ix_InterfaceOps(MacFont*)
    OxExchange* _exchange();

    //+ XfFont::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfFont */
    Boolean equal(XfFont_ptr f);
    ULong hash();
    XfCharString_ptr fullname();
    XfCharString_ptr encoding();
    XfCoord point_size();
    void font_info(XfFont::Info& i);
    void char_info(XfCharCode c, XfFont::Info& i);
    void string_info(XfCharString_ptr s, XfFont::Info& i);
    //+

    Raster_ptr bitmap(CharCode);

    short font() { return font_; }
    short face() { return face_; }
    short size() { return size_; }
    short mode() { return mode_; }
    
protected:
    SharedFrescoObjectImpl object_;
    MacDrawingKit* drawing_kit_;
    short font_;
    short face_;
    short size_;
    short mode_;
    FontInfo info_;
    int* widths_;
};

#endif
