/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/Ox/schema.h>

#include <Fresco/defs_Mac/MacDisplay.h>
#include <Fresco/defs_Mac/MacEvent.h>
#include <Fresco/defs_Mac/MacWindow.h>

// class MacEvent

//+ MacEvent(MacEvent)
void* MacEvent::_this_ptr() { return this; }
_Ix_extern__tid(MacEvent);
void* MacEvent::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(MacEvent) == t) return this;
    void* p;
    if (p = XfEvent::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(MacEvent);
_Ix_ImplementGetTid(MacEvent,_Ix__tid(MacEvent))
_Ix_ImplementNarrow(MacEvent*,MacEvent,_Ix_NullParameter,MacEvent,0)
_Ix_ImplementDuplicate(MacEvent*,MacEvent,_Ix_NullParameter,_Ix__tid(MacEvent),0)
//+

MacEvent::MacEvent(MacWindow* window, const EventRecord& event) {
    Window::_duplicate(window);
    window_ = window;
    event_ = event;
}

MacEvent::~MacEvent() {
    CORBA::release(window_);
}

OxExchange* MacEvent::_exchange() { return &object_; }

//+ MacEvent(XfFrescoObject::=object_.)
XfTag MacEvent::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void MacEvent::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void MacEvent::disconnect() {
    object_.disconnect();
}
void MacEvent::notify_observers() {
    object_.notify_observers();
}
void MacEvent::update() {
    object_.update();
}
//+

//+ MacEvent(XfEvent::hits)
Boolean MacEvent::hits(XfViewer_ptr v) {
    Boolean b = false;
    Window* w = window_;
    if (is_not_nil(w) && is_eq(v, w->main_viewer())) {
	b = true;
    }
    //
    // This implementation is incomplete.  We should now check
    // if v's main viewer is the same as the event's window, and if so
    // do a pick and check the result to see if v is hit.
    //
    return b;
}

//+ MacEvent(XfEvent::type)
XfEvent::TypeId MacEvent::type() {
    switch (event_.what) {
        case osEvt:
            if (((event_.modifiers && osEvtMessageMask) >> 24) == mouseMovedMessage) {
                return Event::motion;
            } else {
                break;
            }

	    case mouseDown:
            return Event::down;

	    case mouseUp: 
            return Event::up;

        case activateEvt:
            return event_.modifiers & activeFlag ? Event::focus_in : Event::focus_out;

        case keyDown:
            return Event::key_press;

        case keyUp:
            return Event::key_release;
    }

    return Event::other;
}

//+ MacEvent(XfEvent::time)
XfEvent::TimeStamp MacEvent::time() {
    return (Event::TimeStamp) event_.when;
}

//+ MacEvent(XfEvent::double_click)
Boolean MacEvent::double_click(XfEvent::TimeStamp previous) {
    Screen_var screen = window_->screen_ptr();
    Display_var display = screen->display_ptr();
    return (time() - previous) < (display->pointer_double_click_threshold() * 60);
}

//+ MacEvent(XfEvent::positional)
Boolean MacEvent::positional() {
    switch (type()) {
        case Event::motion:
        case Event::enter:
        case Event::leave:
        case Event::down:
        case Event::up:
            return true;
    }
    return false;
}

//+ MacEvent(XfEvent::pointer_x)
XfCoord MacEvent::pointer_x() {
    PixelCoord x = event_.where.h - window_->left();
    Screen_var screen = window_->screen_ptr();
    return screen->to_coord(x, X_axis);
}

//+ MacEvent(XfEvent::pointer_y)
XfCoord MacEvent::pointer_y() {
    PixelCoord y = window_->bottom() - event_.where.v;
    Screen_var screen = window_->screen_ptr();
    return screen->to_coord(y, Y_axis);
}

//+ MacEvent(XfEvent::pointer_button)
XfEvent::ButtonIndex MacEvent::pointer_button() {
	return 1;
}

//+ MacEvent(XfEvent::button_is_down)
Boolean MacEvent::button_is_down(XfEvent::ButtonIndex b) {
    switch (b) {
        case 1:
            return ((event_.what == mouseUp) || (! (event_.modifiers & btnState))) ? true : false;
     }
    return false;
}

//+ MacEvent(XfEvent::modifier_is_down)
Boolean MacEvent::modifier_is_down(XfEvent::Modifier m) {
    switch (m) {
	case Event::control:
            return event_.modifiers & controlKey ? true : false;
	case Event::shift:
            return event_.modifiers & shiftKey ? true : false;
	case Event::capslock:
            return event_.modifiers & alphaLock ? true : false;
    case Event::meta: // there is no meta key, so we make meta be command
            return event_.modifiers & cmdKey ? true : false;
	case Event::alt:
	    return event_.modifiers & optionKey ? true : false;
    }
    return false;
}

//+ MacEvent(XfEvent::key)
XfEvent::KeySym MacEvent::key() {
    return (event_.message & keyCodeMask) >> 8;
}

//+ MacEvent(XfEvent::character)
XfCharCode MacEvent::character() {
    return event_.message & charCodeMask;
}

//+ MacEvent(XfEvent::unread)
void MacEvent::unread() {
    switch (event_.what) {
	    case mouseDown:
	    case mouseUp: 
        case keyDown:
        case keyUp:
            EvQEl* qel;
            PPostEvent(event_.what, event_.message, &qel);
            qel->evtQWhen = event_.when;
            qel->evtQWhere = event_.where;
            qel->evtQModifiers = event_.modifiers;
    }
}
