/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/Ox/schema.h>
	  
#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/memory.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/transform.h>

#include <Fresco/defs_Mac/MacDrawing.h>
#include <Fresco/defs_Mac/MacLib.h>
#include <Fresco/defs_Mac/MacPainter.h>
#include <Fresco/defs_Mac/MacScreen.h>

#include <QDOffscreen.h>
	  
/*
 * Two useful array macros: determine the number of elements and
 * return the address of the end of the array (after last element).
 * We assume the compiler will fold any constant arithmetic.
 */

#define array_nelements(array) (sizeof(array) / sizeof(array[0]))
#define array_end(array) (&array[array_nelements(array)])

declareList(MacClippingStack,RgnHandle)
implementList(MacClippingStack,RgnHandle)

// class MacTDB

class MacTDB {
public:
    MacTDB(XfRaster_ptr raster, XfTransform_ptr transform, unsigned long age);
    ~MacTDB();

    XfTransform_ptr transform() { return &transform_; }

    unsigned long age() { return age_; }
    void age(unsigned long age) { age_ = age; }

    PixelCoord raster_origin_x() { return raster_origin_x_; }
    PixelCoord raster_origin_y() { return raster_origin_y_; }

    PixelCoord width() { return width_; }
    PixelCoord height() { return height_; }
    PixelCoord origin_x() { return origin_x_; }
    PixelCoord origin_y() { return origin_y_; }

    unsigned long size() { return width_ * height_ * 3; }

    void create();

/*    
    void* color_bits() { return color_bits_; }
    BITMAPINFO* color_info() { return &color_info_; }

    HBITMAP mask() { return mask_; }
*/

private:
    XfRaster_ptr raster_;
    TransformImpl transform_;
    unsigned long age_;

    PixelCoord raster_origin_x_;
    PixelCoord raster_origin_y_;

    PixelCoord width_;
    PixelCoord height_;
    PixelCoord origin_x_;
    PixelCoord origin_y_;

    TransformImpl* it_;

/*
    void* color_bits_;
    BITMAPINFO color_info_;

    HBITMAP mask_;
*/
};

inline int pad(int c, int s) {
    return (s - (c & (s - 1))) & (s - 1);
}

MacTDB::MacTDB(XfRaster_ptr raster, XfTransform_ptr transform, unsigned long age) {
    raster_ = XfRaster::_duplicate(raster);
    transform_.copy(transform);
    age_ = age;
/*
    color_bits_ = nil;
    mask_ = nil;
*/

    Raster::Info ri;
    raster->raster_info(ri);
    raster_origin_x_ = ri.origin_x;
    raster_origin_y_ = ri.origin_y;

    XfVertex v1;
    v1.x = Coord(0);
    v1.y = Coord(0);
    v1.z = Coord(0);
    transform_.transform_vertex(v1);
    XfVertex v2;
    v2.x = Coord(0);
    v2.y = Coord(ri.rows - 1);
    v2.z = Coord(0);
    transform_.transform_vertex(v2);
    XfVertex v3;
    v3.x = Coord(ri.columns - 1);
    v3.y = Coord(ri.rows - 1);
    v3.z = Coord(0);
    transform_.transform_vertex(v3);
    XfVertex v4;
    v4.x = Coord(ri.columns - 1);
    v4.y = Coord(0);
    v4.z = Coord(0);
    transform_.transform_vertex(v4);

    Coord bbl = Math::min(v1.x, v2.x, v3.x, v4.x);
    Coord bbb = Math::min(v1.y, v2.y, v3.y, v4.y);
    Coord bbr = Math::max(v1.x, v2.x, v3.x, v4.x);
    Coord bbt = Math::max(v1.y, v2.y, v3.y, v4.y);

    width_ = PixelCoord(bbr - bbl) + 1;
    height_ = PixelCoord(bbt - bbb) + 1;

    origin_x_ = PixelCoord(-bbl);
    origin_y_ = PixelCoord(-bbb);

    it_ = new TransformImpl;
    it_->copy(&transform_);
    Vertex vlb;
    vlb.x = -bbl;
    vlb.y = -bbb;
    vlb.z = Coord(0);
    it_->translate(vlb);
    it_->invert();
}

void MacTDB::create() {
/*
    BITMAPINFOHEADER& ch = color_info_.bmiHeader;
    ch.biSize		= sizeof(BITMAPINFOHEADER);
    ch.biWidth		= width_;
    ch.biHeight		= height_;
    ch.biPlanes		= 1;
    ch.biBitCount	= 24;
    ch.biCompression	= BI_RGB;
    ch.biSizeImage	= 0;
    ch.biXPelsPerMeter	= 0; // not used?
    ch.biYPelsPerMeter	= 0; // not used?
    ch.biClrUsed	= 0; // not used?
    ch.biClrImportant	= 0; // not used?

    int color_pad = pad(ch.biWidth * 3, sizeof(LONG));
    long color_size = (ch.biWidth * 3 + color_pad) * ch.biHeight;
    color_bits_ = new unsigned char[color_size];

    BITMAP bm;
    bm.bmType		= 0;
    bm.bmWidth		= width_;
    bm.bmHeight		= height_;
    bm.bmWidthBytes	= ((bm.bmWidth + 15) / 16) * 2;
    bm.bmPlanes		= 1;
    bm.bmBitsPixel	= 1;
    long bm_size = bm.bmWidthBytes * bm.bmHeight;
    bm.bmBits = new unsigned char[bm_size];

    Boolean mask_pad = ((bm.bmWidth + 7) / 8) % 2;

    unsigned char* cp = (unsigned char*) color_bits_;
    unsigned char* mp = (unsigned char*) bm.bmBits;
    XfVertex v;
    XfColor* color;
    XfColor::Intensity r = XfColor::Intensity(0);
    XfColor::Intensity g = XfColor::Intensity(0);
    XfColor::Intensity b = XfColor::Intensity(0);
    Float blend;
    unsigned char mb;
    int index;
    for (XfRaster::Index row = 0; row < height_; ++row) {
        mb = 0xff;
	index = 7;
        for (XfRaster::Index col = 0; col < width_; ++col) {
	    v.x = Coord(col);
            v.y = Coord(row);
            v.z = Coord(0);
	    it_->transform_vertex(v);
	    PixelCoord raster_col = PixelCoord(v.x);
	    PixelCoord raster_row = PixelCoord(v.y);
	    if (
	        (0 <= raster_col) && (raster_col < width_) &&
	        (0 <= raster_row) && (raster_row < height_)
	    ) {
	        raster_->get_color_blend(raster_row, raster_col, color, blend);
	        color->rgb(r, g, b);
		MacAssert(color != nil);
	        CORBA::release(color);
	    } else {
	        blend = 0.0f;
	    }

	    if (blend > 0.99f) {
	        mb &= ~(1 << index);
	    } else {
	        r = g = b = XfColor::Intensity(1);
	    }
	    if (index > 0) {
	        --index;
	    } else {
	        *mp++ = mb;
	        mb = 0xff;
	        index = 7;
	    }

	    *cp++ = (unsigned char) (b * 255.001);
	    *cp++ = (unsigned char) (g * 255.001);
	    *cp++ = (unsigned char) (r * 255.001);
	}
	for (int i = 0; i < color_pad; ++i) {
	    *cp++ = 0;
	}

	if (index != 7) {
	    *mp++ = mb;
	}
	if (mask_pad) {
	    *mp++ = 0xff;
	}
    }

    CORBA::release(it_);
    it_ = nil;

    CORBA::release(raster_);
    raster_ = nil;

    mask_ = CreateBitmapIndirect(&bm);
    MacAssert(mask_);
    delete bm.bmBits;
*/
}

MacTDB::~MacTDB() {
    if (! is_nil(it_)) {
        CORBA::release(it_);
    }
}

// class MacDIB

class MacDIB {
public:
    MacDIB(XfRaster_ptr raster, ULong hash);
    MacDIB(XfFont_ptr font, XfCharCode character, ULong hash);
    ~MacDIB();

    XfFont_ptr font() { return font_; }
    XfCharCode character() { return character_; }

    XfRaster_ptr raster() { return raster_; }
    ULong hash() { return hash_; }

    MacTDB* find(XfTransform_ptr transform, unsigned long age);
    MacTDB* create(XfTransform_ptr transform, unsigned long age);
    void remove(unsigned long i);

    unsigned long count() { return count_; }
    MacTDB* tdb(unsigned long i) { return buffer_[i]; }

private:
    XfFont_ptr font_;
    XfCharCode character_;

    XfRaster_ptr raster_;
    ULong hash_;

    MacTDB** buffer_;
    unsigned long size_;
    unsigned long count_;
};

MacDIB::MacDIB(XfRaster_ptr raster, ULong hash) {
    font_ = nil;
    character_ = 0;

    raster_ = XfRaster::_duplicate(raster);
    hash_ = hash;

    size_ = 8;
    buffer_ = new MacTDB*[size_];
    count_ = 0;
}

MacDIB::MacDIB(XfFont_ptr font, XfCharCode character, ULong hash) {
    font_ = XfFont::_duplicate(font);
    character_ = character;

    // !!! bitmap should be member function of XfFont.
    MacFont* fi = MacFont::_narrow(font);
    raster_ = fi->bitmap(character);
    CORBA::release(fi);
    hash_ = hash;

    size_ = 8;
    buffer_ = new MacTDB*[size_];
    count_ = 0;
}

MacDIB::~MacDIB() {
    CORBA::release(font_);

    CORBA::release(raster_);

    for (unsigned long i = 0; i < count_; ++i) {
        delete buffer_[i];
    }
    delete buffer_;
}

MacTDB* MacDIB::find(XfTransform_ptr transform, unsigned long age) {
    for (unsigned long i = 0; i < count_; ++i) {
        MacTDB* tdb = buffer_[i];
	if (transform->equal(tdb->transform())) {
	    tdb->age(age);
	    return tdb;
	}
    }
    return nil;
}

MacTDB* MacDIB::create(XfTransform_ptr transform, unsigned long age) {
    if (count_ == size_) {
        MacTDB** old_buffer = buffer_;
        size_ = size_ + size_;
        buffer_ = new MacTDB*[size_];
        Memory::copy(buffer_, old_buffer, sizeof(MacTDB*) * count_);
    }

    MacTDB* tdb = new MacTDB(raster_, transform, age);
    buffer_[count_] = tdb;
    ++count_;

    return tdb;
}

void MacDIB::remove(unsigned long i) {
    delete buffer_[i];
    for (unsigned long j = i; j < (count_ - 1); ++j) {
        buffer_[j] = buffer_[j + 1];
    }
    buffer_[count_ - 1] = nil;
    --count_;
}

// class MacDIBCache

class MacDIBCache {
public:
    MacDIBCache();
    ~MacDIBCache();

    MacTDB* find(XfRaster_ptr raster, XfTransform_ptr transform);
    MacTDB* find(XfFont_ptr font, XfCharCode character, XfTransform_ptr transform);

private:
    unsigned long slot(XfRaster_ptr raster, ULong hash);
    unsigned long slot(XfFont_ptr font, XfCharCode character, ULong hash);
    unsigned long slot(MacDIB* dib, ULong hash);
    void rehash();
    MacTDB* find(MacDIB* dib, XfTransform_ptr transform);
    void remove(unsigned long i);

    MacDIB** buffer_;
    unsigned long size_;
    unsigned long count_;
    unsigned long age_;
    unsigned long memory_size_;
    unsigned long memory_size_limit_;
};

MacDIBCache::MacDIBCache() {
    size_ = 64;
    buffer_ = new MacDIB*[size_];
    Memory::zero(buffer_, size_ * sizeof(MacDIB*));
    count_ = 0;
    age_ = 0;
    memory_size_ = 0;
    memory_size_limit_ = 4000000; // approx. 4 megabyte limit
}

MacDIBCache::~MacDIBCache() {
    for (unsigned long i = 0; i < size_; ++i) {
        MacDIB* dib = buffer_[i];
        if (dib != nil) {
            delete dib;
        }
    }
    delete buffer_;
}

inline unsigned long MacDIBCache::slot(XfRaster_ptr raster, ULong hash) {
    unsigned long index = hash % size_;
    while (
        (buffer_[index] != nil) &&
        !is_eq(raster, buffer_[index]->raster())
    ) {
        index = (index + 1) % size_;
    }
    return index;
}

inline unsigned long MacDIBCache::slot(XfFont_ptr font, XfCharCode character, ULong hash) {
    unsigned long index = hash % size_;
    while (
        (buffer_[index] != nil) &&
        !(is_eq(font, buffer_[index]->font()) && (buffer_[index]->character() == character))
    ) {
        index = (index + 1) % size_;
    }
    return index;
}

inline unsigned long MacDIBCache::slot(MacDIB* dib, ULong hash) {
    unsigned long index = hash % size_;
    while (
        (buffer_[index] != nil) &&
        (dib != buffer_[index])
    ) {
        index = (index + 1) % size_;
    }
    return index;
}

void MacDIBCache::rehash() {
    MacDIB** old_buffer = buffer_;
    unsigned long old_size = size_;
    size_ = count_ * 4;
    buffer_ = new MacDIB*[size_];
    Memory::zero(buffer_, size_ * sizeof(MacDIB*));
    for (unsigned long i = 0; i < old_size; ++i) {
        MacDIB* dib = old_buffer[i];
        if (dib != nil) {
            buffer_[slot(dib, dib->hash())] = dib;
        }
    }
    delete old_buffer;
}

MacTDB* MacDIBCache::find(MacDIB* dib, XfTransform_ptr transform) {
    // Remove the translation component for transformed object lookup.
    // We should also limit the scale and stretch blit...
    XfVertex v;
    v.x = v.y = v.z = Coord(0);
    transform->transform_vertex(v);
    v.x = -v.x;
    v.y = -v.y;
    v.z = -v.z;
    TransformImpl nt;
    nt.copy(transform);
    nt.translate(v);

    ++age_;
    MacTDB* tdb = dib->find(&nt, age_);
    if (! tdb) {
        tdb = dib->create(&nt, age_);

		// Reduce the cache to stay within the limit.
		while ((memory_size_ > 0) && ((memory_size_ + tdb->size()) > memory_size_limit_)) {
		    MacDIB* candidate_dib = nil;
		    unsigned long candidate_i = 0;
		    MacTDB* candidate_tdb = nil;
		    unsigned long candidate_j = 0;
		    for (unsigned long i = 0; i < size_; ++i) {
		        MacDIB* dib = buffer_[i];
				for (unsigned long j = 0; j < dib->count(); ++j) {
				    MacTDB* tdb = dib->tdb(j);
				    if ((candidate_tdb == nil) || (tdb->age() < candidate_tdb->age())) {
		                candidate_i = i;
				        candidate_dib = dib;
		                candidate_j = j;
				        candidate_tdb = tdb;
				    }
				}
		    }
		    if ((candidate_tdb != nil) && (candidate_tdb != tdb)) {
	            memory_size_ -= candidate_tdb->size();
	            candidate_dib->remove(candidate_j);
	            if ((candidate_dib != dib) && (candidate_dib->count() == 0)) {
	                remove(candidate_j);
	            }
		    }
		}
	
		memory_size_ += tdb->size();
		tdb->create();
    }
    return tdb;
}

MacTDB* MacDIBCache::find(XfRaster_ptr raster, XfTransform_ptr transform) {
    ULong hash = raster->hash();
    unsigned long index = slot(raster, hash);

    MacDIB* dib = buffer_[index];
    if (dib == nil) {
        dib = new MacDIB(raster, hash);
		buffer_[index] = dib;
		count_++;
		if ((count_ * 3) > size_) {
            rehash();
        }
    }

    return find(dib, transform);
}

MacTDB* MacDIBCache::find(XfFont_ptr font, XfCharCode character, XfTransform_ptr transform) {
    ULong hash = font->hash() ^ (character * 5);
    unsigned long index = slot(font, character, hash);

    MacDIB* dib = buffer_[index];
    if (dib == nil) {
        dib = new MacDIB(font, character, hash);
		buffer_[index] = dib;
		count_++;
		if ((count_ * 3) > size_) {
            rehash();
        }
    }

    return find(dib, transform);
}

void MacDIBCache::remove(unsigned long index) {
    delete buffer_[index];
    buffer_[index] = nil;
    --count_;

    // Rehash if there might have been an element that collided with
    // this one.  We should be smarter and just rehash the section of
    // the table that could have had collisions	with this element.
    if (buffer_[(index + 1) % size_] != nil) {
        rehash();
    }
}

static MacDIBCache cache;

static TransformImpl identity_transform;

// class MacPainter

MacPainter::MacPainter(MacScreen* screen) {
    screen_ = screen;

    clippers_ = new MacClippingStack;

    init_items();
}

MacPainter::~MacPainter() {            
    delete clippers_;
}

void MacPainter::begin_paint() {
    PainterBase::begin_paint();

    // ---- initialize clipping to cover entire area ----
	CGrafPtr cg;
	GDHandle gd;
	GetGWorld(&cg, &gd);
    ClipRect(&cg->portRect);
}

void MacPainter::end_paint() {
    flush_text();

    PainterBase::end_paint();
}

//+ MacPainter(XfPainter::to_coord)
XfCoord MacPainter::to_coord(XfPixelCoord p, XfAxis a) {
    return screen_->to_coord(p, a);
}

//+ MacPainter(XfPainter::to_pixels)
XfPixelCoord MacPainter::to_pixels(XfCoord c, XfAxis a) {
    return screen_->to_pixels(c, a);
}

//+ MacPainter(XfPainter::to_pixels_coord)
XfCoord MacPainter::to_pixels_coord(XfCoord c, XfAxis a) {
    return screen_->to_pixels_coord(c, a);
}

//+ MacPainter(XfPainter::current_brush=b)
void MacPainter::current_brush(XfBrush_ptr b) {
    if (brush_ == b) {
        return;
    }

    PainterBase::current_brush(b);

    // !!! Need to handle scaling the brush when not in advanced graphics mode,
    // !!! but then we need to handle arbitrary transformations (like when the
    // !!! x and y scales are different).

    int width = brush_ == nil ? 0 : screen_->to_pixels(brush_->width(), X_axis);
    PenSize(width, width);
}

//+ MacPainter(XfPainter::current_color=c)
void MacPainter::current_color(XfColor_ptr c) {
    if (color_ == c) {
        return;
    }

    flush_text();

    PainterBase::current_color(c);

    Color::Intensity r, g, b;
    color_->rgb(r, g, b);
    RGBColor rgb;
    rgb.red = r * 65535;
    rgb.green = g * 65535;
    rgb.blue = b * 65535;
    RGBForeColor(&rgb);
}

//+ MacPainter(XfPainter::current_font=f)
void MacPainter::current_font(XfFont_ptr f) {
    if (font_ == f) {
        return;
    }

    flush_text();

    PainterBase::current_font(f);

    MacFont* font = MacFont::_narrow(f);
    if (! is_nil(font)) {
		TextFont(font->font());
		TextFace(font->face());
		TextSize(font->size());
		TextMode(font->mode());
    } else {
        // We should find a Mac font that matches the font spec...
        // !!! unimplemented
    }
    CORBA::release(font);
}

RgnHandle MacPainter::select_path() {
    flush_text();

    Transform::Matrix m;
    matrix_->store_matrix(m);
    path_.transform(m);

    RgnHandle region = NewRgn();
    OpenRgn();
    for (int i = 0; i < path_.count(); ++i) {
        PathImpl::Point& p = path_.point(i);
        switch (p.type) {
	        case PathImpl::MoveTo: {
	            MoveTo(p.px, p.py);
	        } break;
	        case PathImpl::LineTo: {
	            LineTo(p.px, p.py);
	        } break;
	        case PathImpl::CurveTo: {
                PathImpl::Point& p1 = path_.point(i + 1);
                PathImpl::Point& p2 = path_.point(i + 2);
/*
	            POINT pts[3];
			    pts[2].x = p.px;
			    pts[2].y = p.py;
			    pts[0].x = p1.px;
			    pts[0].y = p1.py;
			    pts[1].x = p2.px;
			    pts[1].y = p2.py;
	            PolyBezierTo(drawable_, pts, 3);
*/
                i += 2;
	        } break;
	    }
    }
        
    if (path_.is_closed()) {
        PathImpl::Point& p = path_.point(0);
        LineTo(p.px, p.py);
    }
    
    CloseRgn(region);
    
    return region;
}

//+ MacPainter(XfPainter::stroke)
void MacPainter::stroke() {
    RgnHandle region = select_path();
    PaintRgn(region);
    DisposeRgn(region);
}

//+ MacPainter(XfPainter::fill)
void MacPainter::fill() {
    RgnHandle region = select_path();
    FrameRgn(region);
    DisposeRgn(region);
}

// The text rendering is buffered so that a larger request is made over
// to the graphics subsystem.

void MacPainter::init_items() {
    char_ = chars_;
    width_ = widths_;
    item_ = items_;
    item_->c = char_;
    item_->n = 0;
    item_->x = 0;
    item_->y = 0;
}

void MacPainter::flush_text() {
    if (char_ == chars_) {
        return;
    }
    int n = item_ - items_ + 1;
    for (int i = 0; i < n; ++i) {
        TextItem& item = items_[i];
        MoveTo(item.x, item.y);
        DrawText(item.c, 0, item.n);
    }
    init_items();
}

void MacPainter::add_item() {
    TextItem* i = item_;
    if (i->n > 0) {
        ++i;
        if (i >= array_end(items_)) {
            flush_text();
            i = item_;
        } else {
            item_ = i;
        }
        i->c = char_;
        i->n = 0;
        i->x = 0;
        i->y = 0;
    }
}

/*
 * This function tests whether a matrix is only a translation.
 * Something like this probably should be provided directly
 * by the Transform interface.
 */
static Boolean is_translation(Transform_ptr t) {
    Transform::Matrix m;
    t->store_matrix(m);
    return m[0][0] == 1 && m[0][1] == 0 && m[1][0] == 0 && m[1][1] == 1;
}

//+ MacPainter(XfPainter::character)
void MacPainter::character(XfCharCode ch, XfCoord width, XfCoord x, XfCoord y) {
    if (is_translation(matrix_)) {
        add_item();
        *width_ = 0; // coord_to_twips(width);
        width_++;
        XfVertex v;
        v.x = x;
        v.y = y;
        v.z = Coord(0);
        matrix_->transform_vertex(v);
        item_->x = screen_->to_coord(v.x, X_axis);
        item_->y = screen_->to_coord(v.y, Y_axis);
        item_->n = 1;
        *char_ = ch;
        char_++;
    } else {
        mask_blt(x, y, cache.find(font_, ch, matrix_));
    }
}

void MacPainter::mask_blt(Coord x, Coord y, MacTDB* tdb) {
    // !!! unimplemented
}

//+ MacPainter(XfPainter::image)
void MacPainter::image(XfRaster_ptr r, XfCoord x, XfCoord y) {
    flush_text();

    // !!! unimplemented
}

//+ MacPainter(XfPainter::stencil)
void MacPainter::stencil(XfRaster_ptr r, XfCoord x, XfCoord y) {
    mask_blt(x, y, cache.find(r, matrix_));
}

//+ MacPainter(XfPainter::current_matrix)
void MacPainter::current_matrix(Transform_ptr t) {
    flush_text();

    PainterBase::current_matrix(t);
}

//+ MacPainter(XfPainter::pop_matrix)
void MacPainter::pop_matrix() {
    flush_text();

    PainterBase::pop_matrix();
}

//+ MacPainter(XfPainter::premultiply)
void MacPainter::premultiply(XfTransform_ptr t) {
    flush_text();

    PainterBase::premultiply(t);
}

//+ MacPainter(XfPainter::clip)
void MacPainter::clip() {
    flush_text();

    RgnHandle region = select_path();
    RgnHandle clip = NewRgn();
    GetClip(clip);
    SectRgn(clip, region, region);
    DisposeRgn(clip);
    SetClip(region);
    DisposeRgn(region);
}

//+ MacPainter(XfPainter::push_clipping)
void MacPainter::push_clipping() {
    RgnHandle clip = NewRgn();
    GetClip(clip);
    clippers_->append(clip);
}

//+ MacPainter(XfPainter::pop_clipping)
void MacPainter::pop_clipping() {
    flush_text();

    int n = clippers_->count();
    MacAssert(n > 0);

    RgnHandle clip = clippers_->item(n - 1);
    clippers_->remove(n - 1);
    SetClip(clip);
    DisposeRgn(clip);
}

//+ MacPainter(XfPainter::is_visible)
Boolean MacPainter::is_visible(XfRegion_ptr r) {
    RegionImpl region;
    region.copy(r);
    region.apply_transform(matrix_);
    Rect tr;
    tr.left = screen_->to_pixels(region.lower_.x, X_axis);
    tr.top = screen_->to_pixels(region.upper_.y, Y_axis);
    tr.right = screen_->to_pixels(region.upper_.x, X_axis);
    tr.bottom = screen_->to_pixels(region.lower_.y, Y_axis);

    RgnHandle clip = NewRgn();
    GetClip(clip);
    Boolean result = RectInRgn(&tr, clip);
    DisposeRgn(clip);
    
    return result;
}

//+ MacPainter(XfPainter::visible)
XfRegion_ptr MacPainter::visible() {
    RgnHandle clip = NewRgn();
    GetClip(clip);
    Rect tr = (*clip)->rgnBBox;
    DisposeRgn(clip);

    RegionImpl* r = new RegionImpl;
    r->lower_.x = screen_->to_coord(tr.left, X_axis);
    r->lower_.y = screen_->to_coord(tr.bottom, Y_axis);
    r->upper_.x = screen_->to_coord(tr.right, X_axis);
    r->upper_.y = screen_->to_coord(tr.top, Y_axis);
    r->defined_ = true;
    TransformImpl it;
	it.copy(matrix_);
	it.invert();
	r->apply_transform(&it);

    return r;
}
