/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/Ox/schema.h>

#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/table.h>

#include <Fresco/defs_Mac/MacDisplay.h>
#include <Fresco/defs_Mac/MacScreen.h>
#include <Fresco/defs_Mac/MacWindow.h>

// class MacScreen

/*
 * To avoid a circularity, a screen does not hold a reference
 * to its display.  We could be safer by observing the display,
 * but for now we just assume that the display will live
 * at least as long as the screen.
 */

//+ MacScreen(MacScreen)
void* MacScreen::_this_ptr() { return this; }
_Ix_extern__tid(MacScreen);
void* MacScreen::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(MacScreen) == t) return this;
    void* p;
    if (p = XfScreen::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(MacScreen);
_Ix_ImplementGetTid(MacScreen,_Ix__tid(MacScreen))
_Ix_ImplementNarrow(MacScreen*,MacScreen,_Ix_NullParameter,MacScreen,0)
_Ix_ImplementDuplicate(MacScreen*,MacScreen,_Ix_NullParameter,_Ix__tid(MacScreen),0)
//+

MacScreen::MacScreen(MacDisplay* d, long n) {
    display_ = d;
    number_ = n;

    x_point_ = 1;
    y_point_ = 1;
    x_pixel_ = 1.0f / x_point_;
    y_pixel_ = 1.0f / y_point_;

    Rect rect = (*GetGrayRgn())->rgnBBox;
    width_ = to_coord(rect.right, X_axis);
    height_ = to_coord(rect.top, Y_axis);
}

MacScreen::~MacScreen() {
}

OxExchange* MacScreen::_exchange() { return &object_; }

//+ MacScreen(XfFrescoObject::=object_.)
XfTag MacScreen::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void MacScreen::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void MacScreen::disconnect() {
    object_.disconnect();
}
void MacScreen::notify_observers() {
    object_.notify_observers();
}
void MacScreen::update() {
    object_.update();
}
//+

//+ MacScreen(XfScreen::dpi=c)
void MacScreen::dpi(XfCoord c) {
}

//+ MacScreen(XfScreen::dpi?)
XfCoord MacScreen::dpi() {
    // probably should add axis argument
    return Coord(72.0) / x_pixel_;
}

//+ MacScreen(XfScreen::display_ptr)
XfDisplay_ptr MacScreen::display_ptr() {
    return Display::_duplicate(display_);
}

//+ MacScreen(XfScreen::width)
XfCoord MacScreen::width() {
    return width_;
}

//+ MacScreen(XfScreen::height)
XfCoord MacScreen::height() {
    return height_;
}

//+ MacScreen(XfScreen::to_pixels)
XfPixelCoord MacScreen::to_pixels(XfCoord c, XfAxis a) {
    switch(a) {
        case X_axis:
            return PixelCoord(c * x_point_ + ((c > 0) ? 0.5 : -0.5));
        case Y_axis:
            return PixelCoord(c * y_point_ + ((c > 0) ? 0.5 : -0.5));
    }
    return PixelCoord(0);
}

//+ MacScreen(XfScreen::to_coord)
XfCoord MacScreen::to_coord(XfPixelCoord p, XfAxis a) {
    switch(a) {
        case X_axis:
            return Coord(p) * x_pixel_;
        case Y_axis:
            return Coord(p) * y_pixel_;
    }
    return Coord(0);
}

//+ MacScreen(XfScreen::to_pixels_coord)
XfCoord MacScreen::to_pixels_coord(XfCoord c, XfAxis a) {
    return to_coord(to_pixels(c, a), a);
}

//+ MacScreen(XfScreen::move_pointer)
void MacScreen::move_pointer(XfCoord x, XfCoord y) {
    // !!! unimplemented
}

//+ MacScreen(XfScreen::application)
XfWindow_ptr MacScreen::application(XfViewer_ptr v) {
    return new MacApplicationWindow(display_, this, v);
}

//+ MacScreen(XfScreen::top_level)
XfWindow_ptr MacScreen::top_level(XfViewer_ptr v, XfWindow_ptr group_leader) {
    return new MacTopLevelWindow(display_, this, v, group_leader);
}

//+ MacScreen(XfScreen::transient)
XfWindow_ptr MacScreen::transient(XfViewer_ptr v, XfWindow_ptr transient_for) {
    return new MacTransientWindow(display_, this, v, transient_for);
}

//+ MacScreen(XfScreen::popup)
XfWindow_ptr MacScreen::popup(XfViewer_ptr v) {
    return new MacPopupWindow(display_, this, v);
}

//+ MacScreen(XfScreen::icon)
XfWindow_ptr MacScreen::icon(XfViewer_ptr v) {
    return new MacIconWindow(display_, this, v);
}
