/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef MacWindow_h
#define MacWindow_h

#include <Fresco/defs_impl/windows.h>

#include <QDOffscreen.h>

class MacDisplay;
class MacPainter;
class MacScreen;

class export MacWindow : public WindowBase {
public:
    MacWindow(MacDisplay*, MacScreen*, Viewer_ptr);
    ~MacWindow();

    void decor_adjust(int& x, int& y, int& w, int& h);

    void do_update();
    void do_size(int cx, int cy);
    void do_move(int x, int y);

    XfScreen_ptr screen_ptr();
    void configure(const XfWindow::Placement& p);
    void map();
    void unmap();
    Boolean is_mapped();
    void iconify();
    void deiconify();
    void raise();
    void lower();
    void repair();
    void grab_pointer(XfCursor_ptr c);
    void ungrab_pointer();

    PixelCoord left() { return left_; }
    PixelCoord bottom() { return top_ + height_; }

protected:
    virtual void create_window() = 0;
    
    MacDisplay* display_;
    MacScreen* screen_;
    WindowPtr window_;

private:
    GWorldPtr buffer_bitmap_;
    MacPainter* painter_;
};

class MacManagedWindow : public MacWindow {
public:
    MacManagedWindow(MacDisplay*, MacScreen*, Viewer_ptr);
    virtual ~MacManagedWindow();
    
protected:
    void new_window(void* proc);
};

class MacApplicationWindow : public MacManagedWindow {
public:
    MacApplicationWindow(MacDisplay*, MacScreen*, Viewer_ptr);
    virtual ~MacApplicationWindow();
    
protected:
    virtual void create_window();
};

class MacTopLevelWindow : public MacManagedWindow {
public:
    MacTopLevelWindow(
        MacDisplay*, MacScreen*, Viewer_ptr, Window_ptr group_leader
    );
    virtual ~MacTopLevelWindow();

protected:
    virtual void create_window();

private:
    Window_ptr group_leader_;
};

class MacTransientWindow : public MacManagedWindow {
public:
    MacTransientWindow(
        MacDisplay*, MacScreen*, Viewer_ptr, Window_ptr transient_for
    );
    virtual ~MacTransientWindow();

protected:
    virtual void create_window();

private:
    Window_ptr transient_for_;
};

class MacPopupWindow : public MacManagedWindow {
public:
    MacPopupWindow(MacDisplay*, MacScreen*, Viewer_ptr);
    virtual ~MacPopupWindow();
    
protected:
    virtual void create_window();
};

class MacIconWindow : public MacManagedWindow {
public:
    MacIconWindow(MacDisplay*, MacScreen*, Viewer_ptr);
    virtual ~MacIconWindow();
    
protected:
    virtual void create_window();
};

#endif
