/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/Ox/schema.h>

#include <Fresco/defs_Win32/Wcursor.h>

// class Win32Cursor

//+ XfWin32Cursor(XfWin32Cursor)
void* XfWin32Cursor::_this_ptr() { return this; }
_Ix_extern__tid(XfWin32Cursor);
void* XfWin32Cursor::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(XfWin32Cursor) == t) return this;
    void* p;
    if (p = XfCursor::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(XfWin32Cursor);
_Ix_ImplementGetTid(XfWin32Cursor,_Ix__tid(XfWin32Cursor))
_Ix_ImplementNarrow(XfWin32Cursor*,XfWin32Cursor,_Ix_NullParameter,XfWin32Cursor,0)
_Ix_ImplementDuplicate(XfWin32Cursor*,XfWin32Cursor,_Ix_NullParameter,_Ix__tid(XfWin32Cursor),0)
//+

Win32Cursor::Win32Cursor(
    short x, short y, long pattern [16], long mask [16]
) {
    // ---- put data into MS-Windows format ----
    unsigned char p[32];
    unsigned char m[32];
    unsigned char* p_ptr = p;
    unsigned char* m_ptr = m;
    for (int i = 0; i < 16; i++) {
	*(p_ptr++) = (pattern[i] >> 8) & 0xFF;
	*(m_ptr++) = (mask[i] >> 8) & 0xFF;
	*(p_ptr++) = pattern[i] & 0xFF;
	*(m_ptr++) = mask[i] & 0xFF;
    }

    // ---- create it ----
    HINSTANCE inst = GetModuleHandle(NULL);
    cursor_ = CreateCursor(inst, x, y, 16, 16, m, p);
}

Win32Cursor::Win32Cursor(Raster_ptr /* bitmap */, Raster_ptr /* mask */) {
    /* not implemented */
    cursor_ = nil;
}

Win32Cursor::Win32Cursor(Font_ptr /* f */, long /* pattern */, long /* mask */) {
    /* not implemented */
    cursor_ = nil;
}

Win32Cursor::Win32Cursor(long index) {
    // Assume that index is a resource identifier.
    cursor_ = LoadCursor(GetModuleHandle(NULL), (LPCTSTR) index);
}

Win32Cursor::~Win32Cursor() { 
    if (cursor_) {
	DestroyCursor(cursor_);
    }
}

OxExchange* Win32Cursor::_exchange() { return &object_; }

//+ Win32Cursor(XfFrescoObject::=object_.)
XfTag Win32Cursor::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void Win32Cursor::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void Win32Cursor::disconnect() {
    object_.disconnect();
}
void Win32Cursor::notify_observers() {
    object_.notify_observers();
}
void Win32Cursor::update() {
    object_.update();
}
//+
