/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/dload.h>
#include <Fresco/defs_impl/table.h>

#include <Fresco/Ox/schema.h>

#include <Fresco/defs_Win32/Wcursor.h>
#include <Fresco/defs_Win32/Wdisplay.h>
#include <Fresco/defs_Win32/Wdrawing.h>
#include <Fresco/defs_Win32/Wevent.h>
#include <Fresco/defs_Win32/Wscreen.h>
#include <Fresco/defs_Win32/Wwindow.h>

// class Win32DisplayContext

class Win32DisplayContext : public DisplayContext {
public:
    Win32DisplayContext() {}
    ~Win32DisplayContext() {}

    OxExchange* _exchange() { return &exchange_; }

    //+ XfDisplayContext::*
    /* XfDisplayContext */
    XfDisplay_ptr open_string(const char* n);
    XfDisplay_ptr open_default();
    //+

    static Win32Display* display_;

private:
    OxLibraryExchange exchange_;
};

dload(Win32DisplayContext)

Win32Display* Win32DisplayContext::display_ = nil;

//+ Win32DisplayContext(XfDisplayContext::open_string)
XfDisplay_ptr Win32DisplayContext::open_string(const char* n) {
    return open_default();
}

//+ Win32DisplayContext(XfDisplayContext::open_default)
XfDisplay_ptr Win32DisplayContext::open_default() {
    if (display_ == nil) {
        display_ = new Win32Display;
    }
    return display_;
}


// class Win32Display

declareTable(Win32DisplayWindowTable,HWND,Win32Window*)
implementTable(Win32DisplayWindowTable,HWND,Win32Window*)

//+ XfWin32Display(XfWin32Display)
void* XfWin32Display::_this_ptr() { return this; }
_Ix_extern__tid(XfWin32Display);
void* XfWin32Display::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(XfWin32Display) == t) return this;
    void* p;
    if (p = XfDisplay::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(XfWin32Display);
_Ix_ImplementGetTid(XfWin32Display,_Ix__tid(XfWin32Display))
_Ix_ImplementNarrow(XfWin32Display*,XfWin32Display,_Ix_NullParameter,XfWin32Display,0)
_Ix_ImplementDuplicate(XfWin32Display*,XfWin32Display,_Ix_NullParameter,_Ix__tid(XfWin32Display),0)
//+

Win32Display::Win32Display() {
    nscreens_ = 1;
    screen_ = new Win32Screen*[nscreens_];
    for (long i = 0; i < nscreens_; i++) {
        screen_[i] = new Win32Screen(this, i);
    }

    wtable_ = new Win32DisplayWindowTable(256);
}

Win32Display::~Win32Display() {
    for (long i = 0; i < nscreens_; i++) {
        CORBA::release(screen_[i]);
    }
    delete screen_;

    delete wtable_;
}

//+ Win32Display(XfDisplay::drawing_kit)
XfDrawingKit_ptr Win32Display::drawing_kit() {
    return new Win32DrawingKit(this);
}

//+ Win32Display(XfDisplay::number_of_screens)
XfDisplay::ScreenNumber Win32Display::number_of_screens() {
    return nscreens_;
}

//+ Win32Display(XfDisplay::default_screen)
XfScreen_ptr Win32Display::default_screen() {
    return Screen::_duplicate(screen_[0]);
}

//+ Win32Display(XfDisplay::display_screen)
XfScreen_ptr Win32Display::display_screen(XfDisplay::ScreenNumber n) {
    MWassert(n < nscreens_);
    return Screen::_duplicate(screen_[n]);
}

//+ Win32Display(XfDisplay::cursor_from_data)
XfCursor_ptr Win32Display::cursor_from_data(Short x, Short y, CursorPattern pat, CursorPattern mask) {
    return new Win32Cursor(x, y, pat, mask);
}

//+ Win32Display(XfDisplay::cursor_from_bitmap)
XfCursor_ptr Win32Display::cursor_from_bitmap(XfRaster_ptr b, XfRaster_ptr mask) {
    return new Win32Cursor(b, mask);
}

//+ Win32Display(XfDisplay::cursor_from_font)
XfCursor_ptr Win32Display::cursor_from_font(XfFont_ptr f, Long pat, Long mask) {
    return new Win32Cursor(f, pat, mask);
}

//+ Win32Display(XfDisplay::cursor_from_index)
XfCursor_ptr Win32Display::cursor_from_index(Long n) {
    return new Win32Cursor(n);
}

Boolean Win32Display::run_one_event_bool() {
    Boolean prepared = prepare_to_read();
    if (prepared) {
        MSG msg;
        Boolean result = GetMessage(&msg, NULL, 0, 0);
        after_read();
        if (result) {
            DispatchMessage(&msg);
        } else {
            // If it's a quit message, then we shut this display down.
            close();
	}
    }
    return prepared;
}

Boolean Win32Display::peek() {
    MSG msg;
    return PeekMessage(&msg,NULL,0,0,PM_NOREMOVE) == FALSE; 
}

// Convert between keysyms and strings.

static char* keysym_names[256] = {
    "", "lbutton", "rbutton", "cancel", "mbutton", "", "", "",
    "back", "tab", "", "", "clear", "return", "", "",
    "shift", "control", "menu", "pause", "capital", "", "", "",
    "", "", "", "escape", "", "", "", "",
    "space", "prior", "next", "end", "home", "left", "up", "right",
    "down", "select", "print", "execute", "snapshot", "insert", "delete", "help",
    "0", "1", "2", "3", "4", "5", "6", "7",
    "8", "9", "", "", "", "", "", "",
    "", "A", "B", "C", "D", "E", "F", "G",
    "H", "I", "J", "K", "L", "M", "N", "O",
    "P", "Q", "R", "S", "T", "U", "V", "W",
    "X", "Y", "Z", "lwin", "rwin", "apps", "", "",
    "numpad0", "numpad1", "numpad2", "numpad3", "numpad4", "numpad5", "numpad6", "numpad7",
    "numpad8", "numpad9", "multiply", "add", "separator", "subtract", "decimal", "divide",
    "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8",
    "f9", "f10", "f11", "f12", "f13", "f14", "f15", "f16",
    "f17", "f18", "f19", "f20", "f21", "f22", "f23", "f24",
    "", "", "", "", "", "", "", "",
    "numlock", "scroll", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "lshift", "rshift", "lcontrol", "rcontrol", "lmenu", "rmenu", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "attn", "crsel",
    "exsel", "ereof", "play", "zoom", "noname", "pa1", "oem_clear", ""
};

//+ Win32Display(XfDisplay::keysym_to_string)
char* Win32Display::keysym_to_string(XfEvent::KeySym k) {
    return (0 <= k) && (k <= 256) ? keysym_names[k] : "";
}

declareTable(Win32StringToKeysymTable,const char*,XfEvent::KeySym)
implementTable(Win32StringToKeysymTable,const char*,XfEvent::KeySym)

//+ Win32Display(XfDisplay::string_to_keysym)
XfEvent::KeySym Win32Display::string_to_keysym(const char* s) {
    static Win32StringToKeysymTable* string_to_keysym_table = nil;
    if (string_to_keysym_table	== nil) {
        string_to_keysym_table = new Win32StringToKeysymTable(256);
        for (Event::KeySym i = 0; i < 256; ++i) {
            if (*keysym_names[i] != '\0') {
                string_to_keysym_table->insert(keysym_names[i], i);
            }
        }
    }
    Event::KeySym keysym;
    if (string_to_keysym_table->find(keysym, s)) {
        return keysym;
    } else {
        return 0; // !!! what to return in this case? -denis
    }
}

//+ Win32Display(XfDisplay::flush)
void Win32Display::flush() {
    GdiFlush();
}

//+ Win32Display(XfDisplay::flush_and_wait)
void Win32Display::flush_and_wait() {
    GdiFlush();
}

//+ Win32Display(XfDisplay::ring_bell)
void Win32Display::ring_bell(Float pct_loudness) {
    // The windows approch is to have an enumerated set of possible
    // beep types which may be sent through the sound system, but
    // beep on the computer speaker is only one volume :-(
    //
    // Maybe we should use the audio services instead.

    if (pct_loudness >= 0) {
        MessageBeep(0xFFFFFFFF);
    }
}

//+ Win32Display(XfDisplay::auto_repeat=b)
void Win32Display::auto_repeat(Boolean b) {
    // !!! unimplemented
    DisplayBase::auto_repeat(b);
}

//+ Win32Display(XfDisplay::auto_repeat?)
Boolean Win32Display::auto_repeat() {
    // !!! unimplemented
    return DisplayBase::auto_repeat();
}

//+ Win32Display(XfDisplay::key_click_volume=v)
void Win32Display::key_click_volume(Float v) {
    // !!! unimplemented
    DisplayBase::key_click_volume(v);
}

//+ Win32Display(XfDisplay::key_click_volume?)
Float Win32Display::key_click_volume() {
    // !!! unimplemented
    return DisplayBase::key_click_volume();
}

//+ Win32Display(XfDisplay::pointer_acceleration=a)
void Win32Display::pointer_acceleration(Float a) {
    // !!! unimplemented
    DisplayBase::pointer_acceleration(a);
}

//+ Win32Display(XfDisplay::pointer_acceleration?)
Float Win32Display::pointer_acceleration() {
    // !!! unimplemented
    return DisplayBase::pointer_acceleration();
}

//+ Win32Display(XfDisplay::pointer_threshold=t)
void Win32Display::pointer_threshold(Long t) {
    // !!! unimplemented
    DisplayBase::pointer_threshold(t);
}

//+ Win32Display(XfDisplay::pointer_threshold?)
Long Win32Display::pointer_threshold() {
    // !!! unimplemented
    return DisplayBase::pointer_threshold();
}

//+ Win32Display(XfDisplay::pointer_double_click_threshold=t)
void Win32Display::pointer_double_click_threshold(Float t) {
    DisplayBase::pointer_double_click_threshold(t);
}

//+ Win32Display(XfDisplay::pointer_double_click_threshold?)
Float Win32Display::pointer_double_click_threshold() {
    return DisplayBase::pointer_double_click_threshold();
}

HINSTANCE Win32Display::module() {
    return GetModuleHandle(NULL);
}

LRESULT CALLBACK Win32Display::WndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam) {
    return Win32DisplayContext::display_->dispatch(hwnd, msg, wparam, lparam);
}

//
// Associate the given MS window id with the given window object.
//
void Win32Display::bind(HWND id, Win32Window* w) {
    lock_->acquire();
    wtable_->insert(id, w);
    lock_->release();
}

//
// Disassociate the given MS window id with the given window object.
//
void Win32Display::unbind(HWND id) {
    lock_->acquire();
    wtable_->remove(id);
    lock_->release();
}

//
// Dispatch the incoming MS-Windows message to the appropriate place.
// We get a lot of calls across this function, so we want to be 
// reasonably quick.  We also want to make as few as possible calls
// back to the user32 subsystem so we try to save information where
// we can in our process.
//
LRESULT Win32Display::dispatch(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam) {
    LRESULT result = 0;
    Win32Window* window = nil;
    if (wtable_->find(window, hwnd)) {
        switch (msg) {
            case WM_GETMINMAXINFO:
	        window->on_get_min_max_info((LPMINMAXINFO) lparam);
	    break;
	    case WM_CLOSE:
	        PostQuitMessage(0);
	    break;
	    case WM_PAINT:
		window->on_paint();
	        notify_redisplay();
	    break;
	    case WM_SIZE:
		window->on_size((UINT) wparam, LOWORD(lparam), HIWORD(lparam));
	        notify_redisplay();
	    break;
	    case WM_MOVE:
		window->on_move(LOWORD(lparam), HIWORD(lparam));
	        notify_redisplay();
	    break;
            case WM_PALETTECHANGED:
                window->on_palette_changed((HWND) wparam);
            break;
            case WM_QUERYNEWPALETTE:
		result = window->on_query_new_palette();
	    break;
	    case WM_SETCURSOR:
		if (window->on_set_cursor((HWND) wparam, LOWORD(lparam), HIWORD(lparam))) {
		    result = DefWindowProc(hwnd, msg, wparam, lparam);
		}
	    break;
	    case WM_MOUSEMOVE:
	    case WM_LBUTTONDOWN:
	    case WM_MBUTTONDOWN:
	    case WM_RBUTTONDOWN:
	    case WM_LBUTTONUP: 
	    case WM_MBUTTONUP:
	    case WM_RBUTTONUP:
	    case WM_SETFOCUS:
	    case WM_KILLFOCUS:
	    case WM_CHAR:
	    case WM_KEYDOWN:
	    case WM_SYSCHAR:
	    case WM_SYSKEYDOWN:
	    case WM_KEYUP:
	    case WM_SYSKEYUP:
	        window->window_proc(msg, wparam, lparam);
	    break;
	    default:
		result = DefWindowProc(hwnd, msg, wparam, lparam);
	    break;
	}
    } else {
	// don't know this window, do default behavior for lack of anything better.
	result = DefWindowProc(hwnd, msg, wparam, lparam);
    }

    return result;
}
