/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef Fresco_Impls_Wdisplay_h
#define Fresco_Impls_Wdisplay_h

#include <Fresco/defs_impl/displays.h>

#include <Fresco/defs_Win32/Wlib.h>

class Win32DisplayWindowTable;
class Win32Screen;
class Win32Window;

class export Win32Display : public DisplayBase {
public:
    Win32Display();
    ~Win32Display();

    _Ix_InterfaceOps(Win32Display*)

    XfDrawingKit_ptr drawing_kit();
    XfScreen_ptr display_screen(XfDisplay::ScreenNumber n);
    XfDisplay::ScreenNumber number_of_screens();
    XfScreen_ptr default_screen();
    XfCursor_ptr cursor_from_data(Short x, Short y, CursorPattern pat, CursorPattern mask);
    XfCursor_ptr cursor_from_bitmap(XfRaster_ptr b, XfRaster_ptr mask);
    XfCursor_ptr cursor_from_font(XfFont_ptr f, Long pat, Long mask);
    XfCursor_ptr cursor_from_index(Long n);
    char* keysym_to_string(XfEvent::KeySym k);
    XfEvent::KeySym string_to_keysym(const char* s);
    void flush();
    void flush_and_wait();
    void ring_bell(Float pct_loudness);
    Boolean auto_repeat();
    void auto_repeat(Boolean _p);
    Float key_click_volume();
    void key_click_volume(Float _p);
    Float pointer_acceleration();
    void pointer_acceleration(Float _p);
    Long pointer_threshold();
    void pointer_threshold(Long _p);
    Float pointer_double_click_threshold();
    void pointer_double_click_threshold(Float _p);

protected:
    virtual Boolean run_one_event_bool();
    virtual Boolean peek();


public:
    // return a handle associated with the collection of windows under
    // the care of this module.
    HINSTANCE module();

    // This is the callback function registered with Fresco windows.
    static LRESULT CALLBACK WndProc(HWND, UINT msg, WPARAM, LPARAM);

    void bind(HWND, Win32Window*);
    void unbind(HWND);

private:
    LRESULT dispatch(HWND, UINT, WPARAM, LPARAM);

    Long nscreens_;
    Win32Screen** screen_;

    Win32DisplayWindowTable* wtable_;
};

#endif
