/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/memory.h>
#include <Fresco/defs_impl/rasters.h>
#include <Fresco/defs_impl/table.h>
#include <Fresco/defs_impl/transform.h>

#include <Fresco/Ox/schema.h>

#include <Fresco/defs_Win32/Wdisplay.h>
#include <Fresco/defs_Win32/Wdrawing.h>

// class Win32DrawingKit

Win32DrawingKit::Win32DrawingKit(Win32Display* d) {
    display_ = Display::_duplicate(d);
}

Win32DrawingKit::~Win32DrawingKit() {
    CORBA::release(display_);
}

//+ Win32DrawingKit(XfDrawingKit::find_font)
XfFont_ptr Win32DrawingKit::find_font(XfCharString_ptr fullname) {
    Font_ptr f = DrawingKitBase::find_font(fullname);
    if (is_nil(f)) {
        CharStringBuffer cs(fullname);
        const char* lfd = cs.string();

        //
        // REAL IMPLEMENTATION (ie name mapping) IS TBD  !!!!
        //
        LOGFONT spec;
        spec.lfHeight = coord_to_twips(Coord(12));
        spec.lfWidth = 0;
        spec.lfEscapement = 0;
        spec.lfOrientation = 0;
        spec.lfWeight = 0;
        spec.lfItalic = 0;
        spec.lfUnderline = 0;
        spec.lfStrikeOut = 0;
        spec.lfCharSet = ANSI_CHARSET;
        spec.lfOutPrecision = OUT_DEFAULT_PRECIS;
        spec.lfClipPrecision = CLIP_DEFAULT_PRECIS;
        spec.lfQuality = DEFAULT_QUALITY;
        spec.lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;
	spec.lfFaceName[0] = '\0';

        f = new Win32Font(this, spec);
        insert_font(fullname, f);
    }
    return Font::_duplicate(f);
}

//+ Win32DrawingKit(XfDrawingKit::find_font_match)
XfFont_ptr Win32DrawingKit::find_font_match(XfCharString_ptr font_family, XfCharString_ptr font_style, XfCoord ptsize) {
    // !!! unimplemented
    return nil;
}

//+ Win32DrawingKit(XfDrawingKit::bitmap_from_char)
XfRaster_ptr Win32DrawingKit::bitmap_from_char(XfFont_ptr f, XfCharCode c) {
    Win32Font* fi = Win32Font::_narrow(f);
    return fi->bitmap(c);
}

Display_ptr Win32DrawingKit::display_ptr() {
    return Display::_duplicate(display_);
}

// class Win32Font

//+ XfWin32Font(XfWin32Font)
void* XfWin32Font::_this_ptr() { return this; }
_Ix_extern__tid(XfWin32Font);
void* XfWin32Font::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(XfWin32Font) == t) return this;
    void* p;
    if (p = XfFont::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(XfWin32Font);
_Ix_ImplementGetTid(XfWin32Font,_Ix__tid(XfWin32Font))
_Ix_ImplementNarrow(XfWin32Font*,XfWin32Font,_Ix_NullParameter,XfWin32Font,0)
_Ix_ImplementDuplicate(XfWin32Font*,XfWin32Font,_Ix_NullParameter,_Ix__tid(XfWin32Font),0)
//+

Win32Font::Win32Font(Win32DrawingKit* drawing_kit, LOGFONT& spec) {
    drawing_kit_ = drawing_kit;

    font_ = CreateFontIndirect(&spec);

    HWND hwnd = GetDesktopWindow();
    HDC hdc = GetDC(hwnd);
    HFONT old = (HFONT) SelectObject(hdc, font_);
    int result = SetMapMode(hdc, MM_TWIPS);
    MWassert(result);

    result = GetTextMetrics(hdc, &metrics_);
    MWassert(result);

    character_metrics_ = new ABCFLOAT[metrics_.tmLastChar - metrics_.tmFirstChar + 1];
    result = GetCharABCWidthsFloat(
        hdc, metrics_.tmFirstChar, metrics_.tmLastChar, character_metrics_
    );
    if (! result) {
        int count = metrics_.tmLastChar - metrics_.tmFirstChar + 1;
        INT* width = new INT[count];
        result = GetCharWidth(
            hdc, metrics_.tmFirstChar, metrics_.tmLastChar, width
        );
        MWassert(result);
        for (int i = 0; i < count; ++i) {
            character_metrics_[i].abcfA = FLOAT(0);
            character_metrics_[i].abcfB = FLOAT(width[i]);
            character_metrics_[i].abcfC = FLOAT(0);
        }
        delete width;
    }    

    SelectObject(hdc, old);
    ReleaseDC(hwnd, hdc); 
}

Win32Font::~Win32Font() {
    delete character_metrics_;

    int result = DeleteObject(font_);
    MWassert(result);
}

OxExchange* Win32Font::_exchange() { return &object_; } 

//+ Win32Font(XfFrescoObject::=object_.)
XfTag Win32Font::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void Win32Font::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void Win32Font::disconnect() {
    object_.disconnect();
}
void Win32Font::notify_observers() {
    object_.notify_observers();
}
void Win32Font::update() {
    object_.update();
}
//+

//+ Win32Font(XfFont::hash)
ULong Win32Font::hash() {
    return (ULong) font_;
}

//+ Win32Font(XfFont::equal)
Boolean Win32Font::equal(XfFont_ptr f) {
    return f == this;
}

//+ Win32Font(XfFont::fullname)
XfCharString_ptr Win32Font::fullname() {
    // return CharString::_duplicate(name_);
    return nil;
}

//+ Win32Font(XfFont::encoding)
XfCharString_ptr Win32Font::encoding() {
    // return CharString::_duplicate(encoding_);
    return nil;
}

//+ Win32Font(XfFont::point_size)
XfCoord Win32Font::point_size() {
    return twips_to_coord(metrics_.tmAscent + metrics_.tmDescent);
}

//+ Win32Font(XfFont::font_info)
void Win32Font::font_info(XfFont::Info& i) {
    i.left_bearing = twips_to_coord(metrics_.tmOverhang);
    i.right_bearing = twips_to_coord(metrics_.tmMaxCharWidth + metrics_.tmOverhang);
    i.width = twips_to_coord(metrics_.tmMaxCharWidth);
    i.ascent = twips_to_coord(metrics_.tmAscent);
    i.descent = twips_to_coord(metrics_.tmDescent);
    i.font_ascent = i.ascent;
    i.font_descent = i.descent;
    i.height = i.font_ascent + i.font_descent;
}

inline ABCFLOAT Win32Font::character_metrics(XfCharCode c) {
    ABCFLOAT cm;
    if ((metrics_.tmFirstChar <= c) && (c <= metrics_.tmLastChar)) {
        cm = character_metrics_[c - metrics_.tmFirstChar];
    } else {
        cm.abcfA = cm.abcfB = cm.abcfC = 0.0f;
    }
    return cm;
}

//+ Win32Font(XfFont::char_info)
void Win32Font::char_info(XfCharCode c, XfFont::Info& i) {
    ABCFLOAT cm = character_metrics(c);

    i.left_bearing = twips_to_coord(cm.abcfA);
    i.right_bearing = twips_to_coord(cm.abcfA + cm.abcfB);
    i.width = twips_to_coord(cm.abcfA + cm.abcfB + cm.abcfC);

    i.font_ascent = twips_to_coord(metrics_.tmAscent);
    i.font_descent = twips_to_coord(metrics_.tmDescent);
    i.ascent = i.font_ascent;
    i.descent = i.font_descent;
    i.height = i.ascent + i.descent;
}

//+ Win32Font(XfFont::string_info)
void Win32Font::string_info(XfCharString_ptr s, XfFont::Info& i) {
    CharString::Data data(s->count(), s->count());
    s->get_data(data);
    
    ABCFLOAT cm;
    cm.abcfA = cm.abcfB = cm.abcfC = 0.0f;
    FLOAT left_bearing = 0.0f;
    FLOAT width = 0.0f;
    for (ULong j = 0; j < data.length(); ++j) {
        cm = character_metrics(data[j]);
	if (j == 0) {
	    left_bearing = cm.abcfA;
	}
	width += cm.abcfA + cm.abcfB + cm.abcfC;
    }

    i.left_bearing = twips_to_coord(left_bearing);
    i.right_bearing = twips_to_coord(width - cm.abcfC);
    i.width = twips_to_coord(width);

    i.font_ascent = twips_to_coord(metrics_.tmAscent);
    i.font_descent = twips_to_coord(metrics_.tmDescent);
    i.ascent = i.font_ascent;
    i.descent = i.font_descent;
    i.height = i.ascent + i.descent;
}

Raster_ptr Win32Font::bitmap(CharCode c) {
    // setup
    HWND hwnd = GetDesktopWindow();
    HDC win_hdc = GetDC(hwnd);

    // Get the character metrics in pixels.
    Display_var display = drawing_kit_->display_ptr();
    Screen_var screen = display->default_screen();
    ABCFLOAT cm = character_metrics(c);
    XfRaster::Info info;
    info.rows = screen->to_pixels(twips_to_coord(metrics_.tmAscent + metrics_.tmDescent), Y_axis);
    info.columns = screen->to_pixels(twips_to_coord(cm.abcfA + cm.abcfB + cm.abcfC), X_axis);
    info.origin_x = 0;
    info.origin_y = screen->to_pixels(twips_to_coord(metrics_.tmDescent), Y_axis);

    // create bitmap
//    HBITMAP hbitmap = CreateCompatibleBitmap(win_hdc, info.columns, info.rows);
    HBITMAP hbitmap = CreateBitmap(info.columns, info.rows, 1, 1, NULL);
    MWassert(hbitmap);
    HDC hdc = CreateCompatibleDC(win_hdc);
    MWassert(hdc);
    void* old = SelectObject(hdc, hbitmap);
    MWassert(old);
    int result = SetMapMode(hdc, MM_TWIPS);
    MWassert(result);
    HFONT hdc_font = (HFONT) SelectObject(hdc, font_);
    SetTextAlign(hdc, TA_LEFT | TA_BOTTOM | TA_NOUPDATECP);
    SetTextColor(hdc, RGB(255, 255, 255));

    // draw character into bitmap
    char ch = c;
    result = TextOut(hdc, 0, 0, &ch, 1);
    MWassert(result);

    // create raster from bitmap
    LONG row_bytes = ((info.columns + 15) / 16) * 2;
    LONG size = info.rows * row_bytes;
    unsigned char* image = new unsigned char[size];
// !!! none of this stuff seems to work, why?
/*
    LONG amount = GetBitmapBits(hbitmap, size, image);
    MWassert(amount);
*/
/*
    BITMAPINFO bi;
    BITMAPINFOHEADER& bh = bi.bmiHeader;
    bh.biSize = sizeof(BITMAPINFO);
    bh.biBitCount = 0;
    GetDIBits(hdc, hbitmap, 0, 0, nil, &bi, DIB_RGB_COLORS);
    bh.biSize = sizeof(BITMAPINFO);
    bh.biWidth = info.columns;
    bh.biHeight = info.rows;
    bh.biPlanes = 1;
    bh.biBitCount = 1;
    bh.biCompression = BI_RGB;
    int lines = GetDIBits(hdc, hbitmap, 0, info.rows, image, &bi, DIB_RGB_COLORS);
    MWassert(lines == info.rows);
*/
    XfRaster_ptr raster = new RasterBitmapImpl(info, image);
    delete image;
// !!! so we use this really slow method...
    for (int col = 0; col < info.columns; ++col) {
        for (int row = 0; row < info.rows; ++row) {
	    raster->set_mask(row, col, GetPixel(hdc, col, row) != 0);
	}
    }

    // cleanup from create bitmap
    SelectObject(hdc, hdc_font);
    DeleteDC(hdc);

    // cleanup
    ReleaseDC(hwnd, win_hdc); 

    return raster;
}

HFONT Win32Font::mfont() {
    return font_;
}
