/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/Ox/schema.h>

#include <Fresco/defs_Win32/Wdisplay.h>
#include <Fresco/defs_Win32/Wevent.h>
#include <Fresco/defs_Win32/Wwindow.h>

// class Win32Event

//+ XfWin32Event(XfWin32Event)
void* XfWin32Event::_this_ptr() { return this; }
_Ix_extern__tid(XfWin32Event);
void* XfWin32Event::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(XfWin32Event) == t) return this;
    void* p;
    if (p = XfEvent::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(XfWin32Event);
_Ix_ImplementGetTid(XfWin32Event,_Ix__tid(XfWin32Event))
_Ix_ImplementNarrow(XfWin32Event*,XfWin32Event,_Ix_NullParameter,XfWin32Event,0)
_Ix_ImplementDuplicate(XfWin32Event*,XfWin32Event,_Ix_NullParameter,_Ix__tid(XfWin32Event),0)
//+

Win32Event::Win32Event(
    Window_ptr window, HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam
) {
    window_ = Window::_duplicate(window);
    hWnd_ = hWnd;
    message_ = message;
    wParam_ = wParam;
    lParam_ = lParam;

    // GetMessageTime and GetKeyboardState are relative to
    // the last message removed from the message queue.
    time_ = GetMessageTime();
    switch (message_) {
        case WM_MOUSEMOVE:
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_LBUTTONUP: 
	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
	case WM_KEYUP:
	case WM_SYSKEYUP:
	    BOOL result = GetKeyboardState(key_state_);
	    MWassert(result);
    }
}

Win32Event::~Win32Event() {
    CORBA::release(window_);
}

OxExchange* Win32Event::_exchange() { return &object_; }

//+ Win32Event(XfFrescoObject::=object_.)
XfTag Win32Event::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void Win32Event::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void Win32Event::disconnect() {
    object_.disconnect();
}
void Win32Event::notify_observers() {
    object_.notify_observers();
}
void Win32Event::update() {
    object_.update();
}
//+

//+ Win32Event(XfEvent::hits)
Boolean Win32Event::hits(XfViewer_ptr v) {
    Boolean b = false;
    Window* w = window_;
    if (is_not_nil(w) && is_eq(v, w->main_viewer())) {
	b = true;
    }
    //
    // This implementation is incomplete.  We should now check
    // if v's main viewer is the same as the event's window, and if so
    // do a pick and check the result to see if v is hit.
    //
    return b;
}

//+ Win32Event(XfEvent::type)
XfEvent::TypeId Win32Event::type() {
    switch (message_) {
        case WM_MOUSEMOVE:
            return Event::motion;

	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	    return Event::down;

	case WM_LBUTTONUP: 
	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	    return Event::up;

	case WM_SETFOCUS:
	    return Event::focus_in;

	case WM_KILLFOCUS:
	    return Event::focus_out;

	case WM_CHAR:
	case WM_KEYDOWN:
	case WM_SYSCHAR:
	case WM_SYSKEYDOWN:
	    return Event::key_press;

	case WM_KEYUP:
	case WM_SYSKEYUP:
	    return Event::key_release;
    }

    return Event::other;
}

//+ Win32Event(XfEvent::time)
XfEvent::TimeStamp Win32Event::time() {
    return (Event::TimeStamp) time_;
}

//+ Win32Event(XfEvent::double_click)
Boolean Win32Event::double_click(XfEvent::TimeStamp previous) {
    Screen_var screen = window_->screen_ptr();
    Display_var display = screen->display_ptr();
    return (time() - previous) < (display->pointer_double_click_threshold() * 1000);
}

//+ Win32Event(XfEvent::positional)
Boolean Win32Event::positional() {
    switch (type()) {
        case Event::motion:
        case Event::enter:
        case Event::leave:
        case Event::down:
        case Event::up:
            return true;
    }
    return false;
}

//+ Win32Event(XfEvent::pointer_x)
XfCoord Win32Event::pointer_x() {
    PixelCoord x = LOWORD(lParam_);
    Screen_var screen = window_->screen_ptr();
    return screen->to_coord(x, X_axis);
}

//+ Win32Event(XfEvent::pointer_y)
XfCoord Win32Event::pointer_y() {
    // The coordinates given as arguments to the MS-Windows message
    // are in terms of the client area, which is of course different
    // from the size of the window in most cases.  The adjustment is
    // therefore relative to the client_pheight() rather than pheight().

    PixelCoord y = HIWORD(lParam_);
    RECT rect;
    BOOL result = GetClientRect(hWnd_, &rect);
    MWassert(result);
    Screen_var screen = window_->screen_ptr();
    return screen->to_coord(rect.bottom - y, Y_axis);
}

//+ Win32Event(XfEvent::pointer_button)
XfEvent::ButtonIndex Win32Event::pointer_button() {
    switch (message_) {
	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP:
	    return 1;
	case WM_MBUTTONDOWN:
	case WM_MBUTTONUP:
	    return 2;
	case WM_RBUTTONDOWN:
	case WM_RBUTTONUP:
	    return 3;
    }
    return 0;
}

//+ Win32Event(XfEvent::button_is_down)
Boolean Win32Event::button_is_down(XfEvent::ButtonIndex b) {
    switch (b) {
        case 1:
            return ((message_ == WM_LBUTTONUP) || (key_state_[VK_LBUTTON] & 0x80)) ? true : false;
        case 2:
            return ((message_ == WM_MBUTTONUP) || (key_state_[VK_MBUTTON] & 0x80)) ? true : false;
        case 3:
            return ((message_ == WM_RBUTTONUP) || (key_state_[VK_RBUTTON] & 0x80)) ? true : false;
    }
    return false;
}

//+ Win32Event(XfEvent::modifier_is_down)
Boolean Win32Event::modifier_is_down(XfEvent::Modifier m) {
    switch (m) {
	case Event::control:
            return key_state_[VK_CONTROL] & 0x80 ? true : false;
	case Event::shift:
            return key_state_[VK_SHIFT] & 0x80 ? true : false;
	case Event::capslock:
            return key_state_[VK_CAPITAL] & 0x80 ? true : false;
        case Event::meta: // there is no meta key, so we make meta be alt
            return key_state_[VK_MENU] & 0x80 ? true : false;
	case Event::alt:
	    return false;
    }
    return false;
}

//+ Win32Event(XfEvent::key)
XfEvent::KeySym Win32Event::key() {
    Event::KeySym keysym = 0;

    switch (message_) {
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
	case WM_KEYUP:
	case WM_SYSKEYUP: {
	    keysym = wParam_;
        } break;

	case WM_CHAR:
	case WM_SYSCHAR: {
	    SHORT result = VkKeyScan(wParam_);
	    if ((LOWORD(result) != -1) || (HIWORD(result) != -1)) {
	        keysym = result & 0xff;
	    }
	} break;
    }

    return keysym;
}

//+ Win32Event(XfEvent::character)
XfCharCode Win32Event::character() {
    CharCode code = 0;

    switch (message_) {
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
	case WM_KEYUP:
	case WM_SYSKEYUP: {
	    unsigned short lpwTransKey[2];
	    int result = ToAscii(wParam_, lParam_, key_state_, lpwTransKey, 0u);
	    MWassert(result >= 0);
	    if (result == 1) {
	        code = lpwTransKey[0] & 0xff;
	    }
        } break;

	case WM_CHAR:
	case WM_SYSCHAR: {
	    code = wParam_;
	} break;
    }

    return code;
}

//+ Win32Event(XfEvent::unread)
void Win32Event::unread() {
    PostMessage(hWnd_, message_, wParam_, lParam_);
}
