/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_Impls_Wpainter_h
#define Fresco_Impls_Wpainter_h

#include <Fresco/defs/drawing.h>

#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/painters.h>
	  
#include <Fresco/defs_Win32/Wlib.h>

class Win32ClippingStack;
class Win32Screen;
class Win32TDB;

class Win32Painter : public PainterBase {
public:
    Win32Painter(Win32Screen*);
    ~Win32Painter();

    _Ix_InterfaceOps(Win32Painter*)
    OxExchange* _exchange();

    void begin_paint(HDC dc, POINT& size);
    void end_paint();

    XfCoord to_coord(XfPixelCoord p, XfAxis a);
    XfPixelCoord to_pixels(XfCoord c, XfAxis a);
    XfCoord to_pixels_coord(XfCoord c, XfAxis a);
    void current_brush(XfBrush_ptr b);
    void current_color(XfColor_ptr c);
    void current_font(XfFont_ptr f);
    void stroke();
    void fill();
    void character(XfCharCode ch, XfCoord width, XfCoord x, XfCoord y);
    void image(XfRaster_ptr r, XfCoord x, XfCoord y);
    void stencil(XfRaster_ptr r, XfCoord x, XfCoord y);
    void current_matrix(Transform_ptr t);
    void pop_matrix();
    void premultiply(XfTransform_ptr t);
    void clip();
    void push_clipping();
    void pop_clipping();
    Boolean is_visible(XfRegion_ptr r);
    XfRegion_ptr visible();

private:
    void end_path();
    void select_path();

    void init_items();
    void flush_text();
    void add_item();

    HDC prep_blt(
        Coord x, Coord y, Win32TDB* tdb,
        int& x0, int& y0, int& width, int& height
    );
    void post_blt(HDC hdc);
    void mask_blt(Coord x, Coord y, Win32TDB* tdb);

    void update_world_transform();

    Win32Screen* screen_;

    HRGN clipping_;
    Win32ClippingStack* clippers_;

    LOGBRUSH brush_stats_;

    int mode_;
    
    TCHAR* char_;
    TCHAR chars_[100];
    int* width_;
    int widths_[100];
    POLYTEXT* item_;
    POLYTEXT items_[100];

    HDC drawable_;

    HPEN orig_pen_;
    HBRUSH orig_brush_;
    HFONT orig_font_;
};

#endif
