/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_Impls_Wscreen_h
#define Fresco_Impls_Wscreen_h

#include <Fresco/defs/display.h>

#include <Fresco/defs_impl/fobjects.h>

#include <Fresco/defs_Win32/Wlib.h>

class Win32Display;

typedef int XCoord;

class export Win32Screen : public XfScreen {
public:
    Win32Screen(Win32Display*, long n);
    ~Win32Screen();

    _Ix_InterfaceOps(Win32Screen*)
    OxExchange* _exchange();

    //+ XfScreen::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfScreen */
    XfCoord dpi();
    void dpi(XfCoord _p);
    XfDisplay_ptr display_ptr();
    XfCoord width();
    XfCoord height();
    XfPixelCoord to_pixels(XfCoord c, XfAxis a);
    XfCoord to_coord(XfPixelCoord p, XfAxis a);
    XfCoord to_pixels_coord(XfCoord c, XfAxis a);
    void move_pointer(XfCoord x, XfCoord y);
    XfWindow_ptr application(XfViewer_ptr v);
    XfWindow_ptr top_level(XfViewer_ptr v, XfWindow_ptr group_leader);
    XfWindow_ptr transient(XfViewer_ptr v, XfWindow_ptr transient_for);
    XfWindow_ptr popup(XfViewer_ptr v);
    XfWindow_ptr icon(XfViewer_ptr v);
    //+

    HPALETTE find_colormap(Window*);
	// fetch the colormap for the given window.  Initially all of
	// the windows share the same colormap.  If the given window
	// doesn't need colormaps, then the value returned will be
	// null.
	//

    void find_color(Window*, COLORREF&);
	// find the best reasonable color that can match the given
	// color and return it in place of the given color.  For
	// hi-color and true-color systems the return value will be
	// the same.  For displays using colormaps, the return will
	// be altered in some way.
	//

private:
    SharedFrescoObjectImpl object_;
    Win32Display* display_;
    long number_;
    int x_pixels_;
    int y_pixels_;
    float x_pixel_;
    float y_pixel_;
    float x_point_;
    float y_point_;
};

#endif
