/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/Ox/schema.h>

#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/traversal.h>
#include <Fresco/defs_impl/viewers.h>

#include <Fresco/defs_Win32/Wcursor.h>
#include <Fresco/defs_Win32/Wdisplay.h>
#include <Fresco/defs_Win32/Wevent.h>
#include <Fresco/defs_Win32/Wpainter.h>
#include <Fresco/defs_Win32/Wscreen.h>
#include <Fresco/defs_Win32/Wwindow.h>

// class Win32Window

//+ XfWin32Window(XfWin32Window)
void* XfWin32Window::_this_ptr() { return this; }
_Ix_extern__tid(XfWin32Window);
void* XfWin32Window::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(XfWin32Window) == t) return this;
    void* p;
    if (p = XfWindow::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(XfWin32Window);
_Ix_ImplementGetTid(XfWin32Window,_Ix__tid(XfWin32Window))
_Ix_ImplementNarrow(XfWin32Window*,XfWin32Window,_Ix_NullParameter,XfWin32Window,0)
_Ix_ImplementDuplicate(XfWin32Window*,XfWin32Window,_Ix_NullParameter,_Ix__tid(XfWin32Window),0)
//+

/*
 * A window doesn't hold a reference to the display or screen,
 * as the display's style could indirectly hold a reference to
 * the main viewer.  We could attach to the display to guarantee
 * safety, but for now we assume the display will live as long
 * if not longer than a window.
 */

Win32Window::Win32Window(
    Win32Display* d, Win32Screen* s, Viewer_ptr v
) : WindowBase(d, v) {
    display_ = d;
    screen_ = s;

    buffer_bitmap_ = nil;
    painter_ = new Win32Painter(s);
    draw_->current_painter(painter_);

    entered_ = false;
    grabbed_ = false;

    left_ = CW_USEDEFAULT;
    top_ = CW_USEDEFAULT;
    width_ = CW_USEDEFAULT;
    height_ = CW_USEDEFAULT;
}

Win32Window::~Win32Window() {
    BOOL result = DeleteObject(buffer_bitmap_);
    MWassert(result);
    CORBA::release(painter_);
}

void Win32Window::get_client_geometry(
    PixelCoord& x, PixelCoord& y, PixelCoord& w, PixelCoord& h
) {
    WindowBase::get_client_geometry(x, y, w, h);
    if (! configured_) {
        x = CW_USEDEFAULT;
        y = CW_USEDEFAULT;
    }
}

void Win32Window::decor_adjust(
    DWORD dwStyle, PixelCoord& x, PixelCoord& y, PixelCoord& w, PixelCoord& h
) {
    if (dwStyle & WS_THICKFRAME) {
        int xa = GetSystemMetrics(SM_CXBORDER) + GetSystemMetrics(SM_CXFRAME);
        if (x != CW_USEDEFAULT)	{
            x -= xa;
	}
        w += 2 * xa;
        int ya = GetSystemMetrics(SM_CYBORDER) + GetSystemMetrics(SM_CYFRAME);
        if (y != CW_USEDEFAULT)	{
            y -= ya;
	}
        h += 2 * ya;
    }
    if (dwStyle & WS_CAPTION) {
        int c = GetSystemMetrics(SM_CYCAPTION);
        if (y != CW_USEDEFAULT)	{
            y -= c;
	}
        h += c;
    }
}

LRESULT Win32Window::window_proc(UINT message, WPARAM wParam, LPARAM lParam) {
    // !!! how can we know if we should pass the event for further processing? -denis
    switch (message) {
        case WM_MOUSEMOVE:
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_LBUTTONUP: 
	case WM_MBUTTONUP:
	case WM_RBUTTONUP: /* {
	    // Win32 does not seem to have enter/leave mouse events, so we have
	    // to capture the mouse input so that viewers can tell when	the
	    // mouse leaves the window.
            PixelCoord x = LOWORD(lParam);
            PixelCoord y = HIWORD(lParam);
            RECT rect;
            GetClientRect(wnd(), &rect);
            if ((rect.left <= x) && (x <= rect.right) && (rect.top <= y) && (y <= rect.bottom)) {
                if (! entered_) {
	            entered_ = true;
	            if (! grabbed_) {
	                SetCapture(wnd());
	            }
	        }
            } else {
                if (entered_) {
                    entered_ = false;
                    if (! grabbed_) {
                        BOOL result = ReleaseCapture();
                        MWassert(result);
                    }
                }
            } 
	}
*/
	case WM_SETFOCUS:
	case WM_KILLFOCUS:
	case WM_CHAR:
	case WM_KEYDOWN:
	case WM_SYSCHAR:
	case WM_SYSKEYDOWN:
	case WM_KEYUP:
	case WM_SYSKEYUP: {
            Win32Event* event = new Win32Event(this, wnd(), message, wParam, lParam);
            if (! display_->filtered(event)) {
                handle_event(event);
	    }
            CORBA::release(event);
	}
	return true;
    }

    return false;
}

void Win32Window::on_get_min_max_info(LPMINMAXINFO mmi) {
    if (is_not_nil(Viewer_ptr(viewer_))) {
        PixelCoord x_smallest = 2;
        PixelCoord y_smallest = 2;
        PixelCoord x_largest = screen_->to_pixels(screen_->width(), X_axis);
        PixelCoord y_largest = screen_->to_pixels(screen_->height(), Y_axis);

        Glyph::Requisition req;
        GlyphImpl::init_requisition(req);
        viewer_->request(req);
        PixelCoord x = left_;
        PixelCoord y = top_;
        PixelCoord w = screen_->to_pixels(req.x.minimum, X_axis);
        PixelCoord h = screen_->to_pixels(req.y.minimum, Y_axis);
        decor_adjust(GetWindowLong(wnd(), GWL_STYLE), x, y, w, h);

        // ---- minimum size ----
        mmi->ptMinTrackSize.x = Math::min(x_largest, Math::max(x_smallest, w));
        mmi->ptMinTrackSize.y = Math::min(y_largest, Math::max(y_smallest, h));

        // ---- maximum size ----
        w = screen_->to_pixels(req.x.maximum, X_axis);
        h = screen_->to_pixels(req.y.maximum, Y_axis);
        decor_adjust(GetWindowLong(wnd(), GWL_STYLE), x, y, w, h);
        mmi->ptMaxTrackSize.x = Math::min(x_largest, Math::max(x_smallest, w));
        mmi->ptMaxTrackSize.y = Math::min(y_largest, Math::max(y_smallest, h));
        mmi->ptMaxSize.x = mmi->ptMaxTrackSize.x;
        mmi->ptMaxSize.y = mmi->ptMaxTrackSize.y;
    } else {
        mmi->ptMinTrackSize.x = 0;
        mmi->ptMinTrackSize.y = 0;
        mmi->ptMaxTrackSize.x = screen_->to_pixels(screen_->width(), X_axis);
        mmi->ptMaxTrackSize.y = screen_->to_pixels(screen_->height(), Y_axis);
        mmi->ptMaxSize.x = mmi->ptMaxTrackSize.x;
        mmi->ptMaxSize.y = mmi->ptMaxTrackSize.y;
    }
}

void Win32Window::on_paint() {
    PAINTSTRUCT ps;
    BeginPaint(wnd(), &ps);

    Screen_var s = screen_ptr();
    RECT& a = ps.rcPaint;
    Coord left = s->to_coord(a.left - 1, X_axis);
    Coord bottom = s->to_coord(height_ - a.bottom - 1, Y_axis);
    Coord width = s->to_coord(a.right - a.left + 2, X_axis);
    Coord height = s->to_coord(a.bottom - a.top + 2, Y_axis);
    redraw(left, bottom, width, height);

    EndPaint(wnd(), &ps);
}

void Win32Window::on_size(UINT /* type */, int cx, int cy) {
    lock_->acquire();

    // release old double-buffer resources
    if (buffer_bitmap_) {
        BOOL result = DeleteObject(buffer_bitmap_);
        MWassert(result);
        buffer_bitmap_ = nil;
    }

    // remember the window size
    width_ = cx;
    height_ = cy;

    // fill in the viewer allocation
    viewer_->allocate(
        screen_->to_coord(width_, X_axis), screen_->to_coord(height_, Y_axis), Coord(0)
    );

    // schedule entire area to be re-rendered
    incur();

    lock_->release();
}

void Win32Window::on_move(int x, int y) {
    left_ = x;
    top_ = y;
}

void Win32Window::on_palette_changed(HWND pFocusWnd) {
    if (pFocusWnd != wnd()) {
        on_query_new_palette();
    }
}

BOOL Win32Window::on_query_new_palette() {
    HPALETTE p = screen_->find_colormap(this);
    if (p) {
        HDC hdc = GetDC(wnd());
        SelectPalette(hdc, p, FALSE);
        int i = RealizePalette(hdc);
        if (i) {
            InvalidateRgn(wnd(), NULL, FALSE);
	}
        ReleaseDC(wnd(), hdc);
        return i;
    }
    return FALSE;
}

BOOL Win32Window::on_set_cursor(HWND pWnd, UINT nHitTest, UINT message) {
    Win32Cursor* cimpl = nil;
    if (cursor_) {
        cimpl = Win32Cursor::_narrow(cursor_);
    }
    if (cimpl) {
        SetCursor(cimpl->handle());
    }
    return cimpl == nil;
}

//+ Win32Window(XfWindow::screen_ptr)
XfScreen_ptr Win32Window::screen_ptr() {
    return Screen::_duplicate(screen_);
}

//+ Win32Window(XfWindow::configure)
void Win32Window::configure(const XfWindow::Placement& p) {
    lock_->acquire();

    WindowBase::configure(p);

    if (wnd() != nil) {
        PixelCoord x = left_;
        PixelCoord y = top_;
        PixelCoord w = width_;
        PixelCoord h = height_;
        decor_adjust(GetWindowLong(wnd(), GWL_STYLE), x, y, w, h);
        MoveWindow(wnd(), x, y, w, h, TRUE);
    }

    lock_->release();
}

//+ Win32Window(XfWindow::map)
void Win32Window::map() {
    begin_update();

    lock_->acquire();
    if (wnd() == nil) {
        create_window();
    }
    MWassert(wnd()!= nil);
    ShowWindow(wnd(), SW_SHOW);
    lock_->release();

    end_update();
}

//+ Win32Window(XfWindow::unmap)
void Win32Window::unmap() {
    lock_->acquire();
    ShowWindow(wnd(), SW_HIDE);
    lock_->release();
}

//+ Win32Window(XfWindow::is_mapped)
Boolean Win32Window::is_mapped() {
    return IsWindowVisible(wnd());
}

//+ Win32Window(XfWindow::iconify)
void Win32Window::iconify() {
    lock_->acquire();
    ShowWindow(wnd(), SW_MINIMIZE);
    lock_->release();
}

//+ Win32Window(XfWindow::deiconify)
void Win32Window::deiconify() {
    lock_->acquire();
    ShowWindow(wnd(), SW_RESTORE);
    lock_->release();
}

//+ Win32Window(XfWindow::raise)
void Win32Window::raise() {
    lock_->acquire();
    SetWindowPos(wnd(), HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
    lock_->release();
}

//+ Win32Window(XfWindow::lower)
void Win32Window::lower() {
    lock_->acquire();
    SetWindowPos(wnd(), HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
    lock_->release();
}

/*
 * Initiate a draw traversal on the window's viewer.
 * Note that we keep the window locked during the traversal.
 * Not sure this is the best approach, but it is simple and
 * right not there doesn't appear to be any need to access
 * the window while drawing (except for damage, which is protected
 * by a different lock).
 */

//+ Win32Window(XfWindow::repair)
void Win32Window::repair() {
    lock_->acquire();

    // block other threads from changing the damage while we are 
    // determining what area we will update.  Stash the coordinates 
    // because we will let other threads change the damage before 
    // we are done updating.
    //
    damage_lock_->acquire();
    Coord dx0 = damage_->lower_.x;
    Coord dy0 = damage_->lower_.y;
    Coord dx1 = damage_->upper_.x;
    Coord dy1 = damage_->upper_.y;

    //
    // Prep the painter... this may be a memory-based device context
    // if double-buffered, or it might be the actual window if not
    // double-buffered.
    //
    HDC hdc;
    HPALETTE palette = screen_->find_colormap(this);
    // ---- prepare the device contexts ----
    HDC win_hdc = GetDC(wnd());
    MWassert(win_hdc);
    if (palette) {
        SelectPalette(win_hdc, palette, TRUE);
        RealizePalette(win_hdc);
    }
    HDC buffer_hdc = nil;
    if (double_buffer_) {
        if (buffer_bitmap_ == nil) {
    	    buffer_bitmap_ = CreateCompatibleBitmap(win_hdc, width_, height_ );
        }
        buffer_hdc = CreateCompatibleDC(win_hdc);
        SelectObject(buffer_hdc, buffer_bitmap_);
        if (palette) {
            SelectPalette(buffer_hdc, palette, TRUE);
            RealizePalette(buffer_hdc);
        }
        hdc = buffer_hdc;
    } else {
	hdc = win_hdc;
    }

    // --- prepare the painter with the appropriate device context ---
    POINT size;
    size.x = width_;
    size.y = height_;
    painter_->begin_paint(hdc, size);
    painter_->push_clipping();
    painter_->clip_rect(dx0, dy0, dx1, dy1);
    // 
    // The damage is marked undefined, and the lock is released to allow
    // other threads access again.  We have stashed the damage coordinates
    // in case they are needed for the double-buffering.
    //
    damage_->defined_ = false;
    damage_lock_->release();
    //
    render();
    //
    // ---- finish with painter ----
    painter_->pop_clipping();
    painter_->end_paint();

    // finalize things... if we are double-buffering than transfer the
    // updated area to the window.  Free up any resources that were used.
    if (buffer_hdc) {
        POINT xfer[2];
        xfer[0].x = coord_to_twips(dx0);
        xfer[0].y = coord_to_twips(dy0);
        xfer[1].x = coord_to_twips(dx1);
        xfer[1].y = coord_to_twips(dy1);
        LPtoDP(buffer_hdc, xfer, 2);
        SetMapMode(buffer_hdc, MM_TEXT);
	POINT point;
	SetWindowOrgEx(buffer_hdc, 0, 0, &point);

        BOOL result = BitBlt(
            win_hdc, xfer[0].x, xfer[1].y, 
            xfer[1].x - xfer[0].x, xfer[0].y - xfer[1].y,
            buffer_hdc, xfer[0].x, xfer[1].y, SRCCOPY
        );
	MWassert(result);        
    }
    // ---- free up resources ----
    if (win_hdc) {
        ReleaseDC(wnd(), win_hdc);
    }
    if (buffer_hdc) {
        DeleteDC(buffer_hdc);
    }

    lock_->release();
}

//+ Win32Window(XfWindow::grab_pointer)
void Win32Window::grab_pointer(XfCursor_ptr c) {
    grabbed_ = true;
    if (! entered_) {
        SetCapture(wnd());
    }
}

//+ Win32Window(XfWindow::ungrab_pointer)
void Win32Window::ungrab_pointer() {
    grabbed_ = false;
    if (! entered_) {
        BOOL result = ReleaseCapture();
        MWassert(result);
    }
}

// class Win32ManagedWindow

Win32ManagedWindow::Win32ManagedWindow(
    Win32Display* display, Win32Screen* screen, Viewer_ptr viewer
) : Win32Window(display, screen, viewer) {
    wnd_ = nil;
}

Win32ManagedWindow::~Win32ManagedWindow() {
    if (wnd_ != nil) {
        display_->unbind(wnd_);
        BOOL result = DestroyWindow(wnd_);
	MWassert(result);
    }
}

HWND Win32ManagedWindow::wnd() {
    return wnd_;
}

void Win32ManagedWindow::new_window(
    char* wclass, UINT wc_style, DWORD ex_wstyle, DWORD	wstyle
) {
    HINSTANCE hinst = display_->module();
    HINSTANCE fdisplay = GetModuleHandle("fdisplay");
    MWassert(fdisplay != NULL);
    WNDCLASS  wc;
    if (GetClassInfo(hinst, wclass, &wc) == FALSE) {
        // ---- class not yet registered ----
        wc.style         = wc_style;
        wc.lpfnWndProc   = Win32Display::WndProc;
        wc.cbClsExtra    = 0;
        wc.cbWndExtra    = 0;
        wc.hInstance     = hinst;
        wc.hIcon         = LoadIcon(fdisplay, MAKEINTRESOURCE(101));
	MWassert(wc.hIcon);
        wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
        wc.lpszMenuName  = NULL;
        wc.lpszClassName = wclass;

        // Register the window class
        ATOM result = RegisterClass(&wc);
	MWassert(result != 0);
    }

    PixelCoord x, y, w, h;
    get_client_geometry(x, y, w, h);
    // Fill in the viewer allocation that we expect to get.
    viewer_->allocate(
        screen_->to_coord(w, X_axis), screen_->to_coord(h, Y_axis), Coord(0)
    );
    decor_adjust(wstyle, x, y, w, h);

    CharString_var s = style_->context_name();
    if (is_nil(s)) {
	s = StyleContext_var(display_->style())->context_name();
    }
    CharStringBuffer cs(s);

    wnd_ = CreateWindowEx(
        ex_wstyle, wclass, cs.string(), wstyle, x, y, w, h,
        nil, (HMENU) nil, hinst, (LPSTR) nil            
    );
    MWassert(wnd_);

    // CreateWindowEx doesn't seem to do this. -denis
    BOOL result = SetWindowText(wnd_, cs.string());
    MWassert(result);

    // arrange for MS-Windows messages to be forwarded to this window
    display_->bind(wnd_, this);
}

// class Win32ApplicationWindow

Win32ApplicationWindow::Win32ApplicationWindow(
    Win32Display* display, Win32Screen* screen, Viewer_ptr viewer
) : Win32ManagedWindow(display, screen, viewer) {
}

Win32ApplicationWindow::~Win32ApplicationWindow() {
}

void Win32ApplicationWindow::create_window() {
    new_window("FrescoApplicationWindow", 0, 0, WS_OVERLAPPEDWINDOW);
}

// class Win32TopLevelWindow

Win32TopLevelWindow::Win32TopLevelWindow(
    Win32Display* display, Win32Screen* screen, Viewer_ptr viewer, Window_ptr group_leader
) : Win32ManagedWindow(display, screen, viewer) {
    group_leader_ = Window::_duplicate(group_leader);
}

Win32TopLevelWindow::~Win32TopLevelWindow() {
    CORBA::release(group_leader_);
}

void Win32TopLevelWindow::create_window() {
    new_window("FrescoTopLevelWindow", 0, 0, WS_OVERLAPPEDWINDOW);
}

// class Win32TransientWindow

Win32TransientWindow::Win32TransientWindow(
    Win32Display* display, Win32Screen* screen, Viewer_ptr viewer, Window_ptr transient_for
) : Win32ManagedWindow(display, screen, viewer) {
    transient_for_ = Window::_duplicate(transient_for);
}

Win32TransientWindow::~Win32TransientWindow() {
    CORBA::release(transient_for_);
}

void Win32TransientWindow::create_window() {
    new_window("FrescoTransientWindow", 0, 0, WS_OVERLAPPEDWINDOW);
}

// class Win32PopupWindow

Win32PopupWindow::Win32PopupWindow(
    Win32Display* display, Win32Screen* screen, Viewer_ptr viewer
) : Win32ManagedWindow(display, screen, viewer) {
}

Win32PopupWindow::~Win32PopupWindow() {
}

//+ Win32PopupWindow(XfWindow::map)
void Win32PopupWindow::map() {
    begin_update();

    lock_->acquire();
    if (wnd() == nil) {
        create_window();
    }
    MWassert(wnd()!= nil);
    ShowWindow(wnd(), SW_SHOWNA);
    lock_->release();

    end_update();
}

void Win32PopupWindow::create_window() {
    new_window("FrescoPopupWindow", CS_SAVEBITS, WS_EX_TOPMOST, WS_POPUP);
}

// class Win32IconWindow

Win32IconWindow::Win32IconWindow(
    Win32Display* display, Win32Screen* screen, Viewer_ptr viewer
) : Win32ManagedWindow(display, screen, viewer) {
}

Win32IconWindow::~Win32IconWindow() {
}

void Win32IconWindow::create_window() {
    new_window("FrescoIconWindow", 0, 0, WS_OVERLAPPEDWINDOW);
}
