/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_Impls_Wwindow_h
#define Fresco_Impls_Wwindow_h

#include <Fresco/defs_impl/windows.h>

#include <Fresco/defs_Win32/Wlib.h>

class Win32Display;
class Win32Painter;
class Win32Screen;

class export Win32Window : public WindowBase {
public:
    Win32Window(Win32Display*, Win32Screen*, Viewer_ptr);
    ~Win32Window();

    // The x and y values are set to CW_USEDEFAULT if the window has
    // not been configured.
    void get_client_geometry(
        PixelCoord& x, PixelCoord& y, PixelCoord& w, PixelCoord& h
    );
    void decor_adjust(
        DWORD dwStyle, PixelCoord& x, PixelCoord& y, PixelCoord& w, PixelCoord& h
    );

    // CWnd overrides
    LRESULT window_proc(UINT message, WPARAM wParam, LPARAM lParam);

    // CWnd message handlers
    void on_get_min_max_info(LPMINMAXINFO mmi);
    void on_paint();
    void on_size(UINT type, int cx, int cy);
    void on_move(int x, int y);
    void on_palette_changed(HWND pFocusWnd);
    BOOL on_query_new_palette();
    BOOL on_set_cursor(HWND pWnd, UINT nHitTest, UINT message);
    
public:
    _Ix_InterfaceOps(Win32Window*)

    XfScreen_ptr screen_ptr();
    void configure(const XfWindow::Placement& p);
    void map();
    void unmap();
    Boolean is_mapped();
    void iconify();
    void deiconify();
    void raise();
    void lower();
    void repair();
    void grab_pointer(XfCursor_ptr c);
    void ungrab_pointer();

protected:
    virtual HWND wnd() = 0;
    virtual void create_window() = 0;

    Win32Display* display_;
    Win32Screen* screen_;

private:
    HBITMAP buffer_bitmap_;
    Win32Painter* painter_;

    Boolean entered_;
    Boolean grabbed_;
};

class Win32ManagedWindow : public Win32Window {
public:
    Win32ManagedWindow(Win32Display*, Win32Screen*, Viewer_ptr);
    virtual ~Win32ManagedWindow();

protected:
    virtual HWND wnd();

protected:
    void new_window(
        char* wclass, UINT wc_style, DWORD ex_wstyle, DWORD	wstyle
    );

private:
    HWND wnd_;
};

class Win32ApplicationWindow : public Win32ManagedWindow {
public:
    Win32ApplicationWindow(Win32Display*, Win32Screen*, Viewer_ptr);
    virtual ~Win32ApplicationWindow();

protected:
    virtual void create_window();
};

class Win32TopLevelWindow : public Win32ManagedWindow {
public:
    Win32TopLevelWindow(
	Win32Display*, Win32Screen*, Viewer_ptr, Window_ptr group_leader
    );
    virtual ~Win32TopLevelWindow();

protected:
    virtual void create_window();

private:
    Window_ptr group_leader_;
};

class Win32TransientWindow : public Win32ManagedWindow {
public:
    Win32TransientWindow(
	Win32Display*, Win32Screen*, Viewer_ptr, Window_ptr transient_for
    );
    virtual ~Win32TransientWindow();

protected:
    virtual void create_window();

private:
    Window_ptr transient_for_;
};

class Win32PopupWindow : public Win32ManagedWindow {
public:
    Win32PopupWindow(Win32Display*, Win32Screen*, Viewer_ptr);
    virtual ~Win32PopupWindow();

    virtual void map();

protected:
    virtual void create_window();
};

class Win32IconWindow : public Win32ManagedWindow {
public:
    Win32IconWindow(Win32Display*, Win32Screen*, Viewer_ptr);
    virtual ~Win32IconWindow();

protected:
    virtual void create_window();
};

#endif
