//
// $XConsortium$
//

//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-93 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_X11_Xdisplay_h
#define Fresco_defs_X11_Xdisplay_h

#include <Fresco/defs/display.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/styles.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/defs_impl/thread.h>
#include <Fresco/defs_X11/Xlib.h>

class DisplayImplDamage;
class DisplayImplFilters;
class DisplayImplRemoveFilter;
class DisplayImplWindowTable;
class EventImpl;
class ScreenImpl;

//+ DisplayImpl : XfDisplay
class DisplayImpl : public XfDisplay {
public:
    ~DisplayImpl();
    _Ix_InterfaceOps(DisplayImpl*)
//+
public:
    DisplayImpl(XDisplay*);

    OxExchange* _exchange();

    //+ XfDisplay::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfDisplay */
    XfStyleContext_ptr style();
    XfDrawingKit_ptr drawing_kit();
    XfScreen_ptr display_screen(XfDisplay::ScreenNumber n);
    XfDisplay::ScreenNumber number_of_screens();
    XfScreen_ptr default_screen();
    XfCursor_ptr cursor_from_data(Short x, Short y, CursorPattern pat, CursorPattern mask);
    XfCursor_ptr cursor_from_bitmap(XfRaster_ptr b, XfRaster_ptr mask);
    XfCursor_ptr cursor_from_font(XfFont_ptr f, Long pat, Long mask);
    XfCursor_ptr cursor_from_index(Long n);
    void run(Boolean b);
    Boolean running();
    void run_one_event();
    XfTag add_filter(XfEventFilter_ptr f);
    XfTag add_general_filter(XfEventFilter_ptr f);
    void remove_filter(XfTag add_tag);
    char* keysym_to_string(XfEvent::KeySym k);
    XfEvent::KeySym string_to_keysym(const char* s);
    void need_repair(XfWindow_ptr w);
    void repair();
    void flush();
    void flush_and_wait();
    void ring_bell(Float pct_loudness);
    void close();
    Boolean auto_repeat();
    void auto_repeat(Boolean _p);
    Float key_click_volume();
    void key_click_volume(Float _p);
    Float pointer_acceleration();
    void pointer_acceleration(Float _p);
    Long pointer_threshold();
    void pointer_threshold(Long _p);
    Float pointer_double_click_threshold();
    void pointer_double_click_threshold(Float _p);
    //+

    struct FilterInfo {
	long tag;
	Boolean general;
	EventFilter_ptr filter;
    };

    XDisplay* xdisplay();
    ULong double_click_threshold_msec() { return double_click_msec_; }
    Atom wm_protocols_atom();
    Atom wm_delete_atom();
    unsigned int meta_mask() { return meta_mask_; }
    unsigned int alt_mask() { return alt_mask_; }
    void bind(XWindow, Window_ptr);
    void unbind(XWindow);

    Boolean prepare_to_read();
    void read_event(EventImpl*);
    void after_read();
    void prepare_to_request();
    void after_requesting();

    void get_modifier_masks();
protected:
    friend class DisplayImplRemoveFilter;

    SharedFrescoObjectImpl object_;
    Mutex_var lock_;
    Boolean running_;
    XDisplay* xdisplay_;
    XWindow xwindow_;
    StyleImpl* style_;
    Float double_click_;
    ULong double_click_msec_;
    Long nscreens_;
    ScreenImpl** screen_;
    Boolean trace_events_;
    Boolean reading_;
    Long waiting_to_read_;
    Boolean requesting_;
    Long waiting_to_request_;
    ConditionVariable_var wait_to_read_;
    ConditionVariable_var wait_to_request_;
    Thread_var redisplay_;
    ConditionVariable_var damage_pending_;
    DisplayImplDamage* damaged_;
    Mutex_var filter_lock_;
    DisplayImplFilters* filters_;
    long filter_tag_;
    Mutex_var table_lock_;
    DisplayImplWindowTable* wtable_;
    Atom wm_protocols_;
    Atom wm_delete_;
    unsigned int meta_mask_;
    unsigned int alt_mask_;

    void init_style();
    void load_path(const char* head, const char* tail, long priority);
    void load_path(
	const char* head, const char* middle, const char* tail, long priority
    );
    const char* home();

    void run_events(Boolean loop);
    void dispatch(EventImpl*);
    Tag add_filter_info(EventFilter_ptr, Boolean general);
    Boolean filtered(EventImpl*);
    void expose(Window_ptr, const XExposeEvent&);
    void configure(Window_ptr, const XConfigureEvent&);

    void redisplay_thread();
    void notify_redisplay();
    void do_repairs();
};

#endif
