//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_X11_Xdrawing_h
#define Fresco_defs_X11_Xdrawing_h

#include <Fresco/defs/drawing.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_X11/Xlib.h>

class DisplayImpl;
class DrawingKitColorTable;
class DrawingKitFontTable;

class DrawingKitImpl : public DrawingKit {
public:
    DrawingKitImpl(DisplayImpl*);
    ~DrawingKitImpl();

    OxExchange* _exchange();

    //+ XfDrawingKit::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfDrawingKit */
    XfStyleContext_ptr style();
    XfBrush_ptr simple_brush(XfCoord width);
    XfBrush_ptr dither_brush(XfCoord width, Long p);
    XfBrush_ptr patterned_brush(XfCoord width, const XfDrawingKit::Data32& p);
    XfPattern_ptr solid_pattern();
    XfPattern_ptr halftone_pattern();
    XfPattern_ptr stipple(const XfDrawingKit::Data32& d);
    XfPattern_ptr stipple_4x4(ULong p);
    XfColor_ptr color_rgb(XfColor::Intensity r, XfColor::Intensity g, XfColor::Intensity b);
    XfColor_ptr find_color(XfCharString_ptr n);
    XfColor_ptr resolve_color(XfStyleContext_ptr s, XfCharString_ptr n);
    XfColor_ptr foreground(XfStyleContext_ptr s);
    XfColor_ptr background(XfStyleContext_ptr s);
    XfFont_ptr find_font(XfCharString_ptr fullname);
    XfFont_ptr find_font_match(XfCharString_ptr font_family, XfCharString_ptr font_style, XfCoord ptsize);
    XfFont_ptr resolve_font(XfStyleContext_ptr s, XfCharString_ptr n);
    XfFont_ptr default_font(XfStyleContext_ptr s);
    XfRaster_ptr bitmap_from_data(const XfDrawingKit::Data8& d, XfRaster::Index rows, XfRaster::Index columns, XfRaster::Index origin_row, XfRaster::Index origin_column);
    XfRaster_ptr bitmap_from_char(XfFont_ptr f, XfCharCode c);
    XfRaster_ptr raster_open(XfCharString_ptr n);
    XfTransform_ptr transform_matrix(XfTransform::Matrix m);
    XfTransform_ptr identity_matrix();
    XfRegion_ptr bounding_box();
    XfPainter_ptr printer(XfCharString_ptr output);
    //+
protected:
    SharedFrescoObjectImpl object_;
    DisplayImpl* display_;
    DrawingKitColorTable* colors_;
    DrawingKitFontTable* fonts_;
    CharString_var foreground_str_;
    CharString_var Foreground_str_;
    CharString_var background_str_;
    CharString_var Background_str_;
    CharString_var font_str_;
    CharString_var Font_str_;

    Color_ptr resolve_colors(
	StyleContext_ptr s, CharString_ptr name1, CharString_ptr name2
    );
};

//+ BrushImpl : XfBrush
class BrushImpl : public XfBrush {
public:
    ~BrushImpl();
    _Ix_InterfaceOps(BrushImpl*)
//+
public:
    BrushImpl(Coord width);
    BrushImpl(const Long* p, Long c, Coord w);
    BrushImpl(Long pat, Coord w);

    OxExchange* _exchange();

    //+ XfBrush::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfBrush */
    XfCoord width();
    Boolean equal(XfBrush_ptr b);
    ULong hash();
    //+

    void info(Coord& width, char*& dash_list, Long& dash_count);
protected:
    SharedFrescoObjectImpl object_;
    Coord width_;
    char* dash_list_;
    long dash_count_;

    void init(const Long* p, Long c, Coord w);
    void calc_dashes(Long pat, Long* dash, Long& count);
};

//+ ColorImpl : XfColor
class ColorImpl : public XfColor {
public:
    ~ColorImpl();
    _Ix_InterfaceOps(ColorImpl*)
//+
public:
    ColorImpl(Color::Intensity r, Color::Intensity g, Color::Intensity b);

    OxExchange* _exchange();

    //+ XfColor::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfColor */
    void rgb(XfColor::Intensity& r, XfColor::Intensity& g, XfColor::Intensity& b);
    Boolean equal(XfColor_ptr c);
    ULong hash();
    //+

    static unsigned short to_short(Color::Intensity);
    static Color::Intensity to_intensity(unsigned short);
protected:
    SharedFrescoObjectImpl object_;
    Color::Intensity red_;
    Color::Intensity green_;
    Color::Intensity blue_;
};

/*
 * The assumption here is that intensities are positive,
 * so the roundoff doesn't need to do a sign check.
 */

inline unsigned short ColorImpl::to_short(Color::Intensity i) {
    return (unsigned short)(i * float(0xffff) + 0.5);
}

inline Color::Intensity ColorImpl::to_intensity(unsigned short s) {
    return float(s) / float(0xffff);
}

Ox_implementation_decl(FontImpl,Font,SharedFrescoObjectImpl)
    FontImpl(DisplayImpl*, const char* name);
    ~FontImpl();

    _Ix_InterfaceOps(FontImpl*)

    //+ XfFont::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfFont */
    Boolean equal(XfFont_ptr f);
    ULong hash();
    XfCharString_ptr fullname();
    XfCharString_ptr encoding();
    XfCoord point_size();
    void font_info(XfFont::Info& i);
    void char_info(XfCharCode c, XfFont::Info& i);
    void string_info(XfCharString_ptr s, XfFont::Info& i);
    //+

    void load();
    XDisplay* xdisplay();
    XFontStruct* xfont();
    PixelCoord char_width(CharCode);
    Raster_ptr bitmap(CharCode);
    Coord to_coord(PixelCoord, Axis);
    Coord average_width();
protected:
    CharString_ptr name_;
    CharString_ptr encoding_;
    Coord point_size_;
    Coord scale_;
    XDisplay* xdisplay_;
    XFontStruct* xfont_;
    Raster_ptr* rasters_;
};

inline XDisplay* FontImpl::xdisplay() { return xdisplay_; }
inline XFontStruct* FontImpl::xfont() { return xfont_; }

//+ PatternImpl : XfPattern
class PatternImpl : public XfPattern {
public:
    ~PatternImpl();
    _Ix_InterfaceOps(PatternImpl*)
//+
public:
    PatternImpl(ULong);
    PatternImpl(const DrawingKit::Data32&);

    OxExchange* _exchange();

    //+ XfPattern::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfPattern */
    Boolean equal(XfPattern_ptr p);
    ULong hash();
    void stipple(_Ix_T_out(XfDrawingKit::Data32) d);
    //+
protected:
    SharedFrescoObjectImpl object_;
public:
    DrawingKit::Data32 stipple_;
};

#endif
