//
// $XConsortium$
//

//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_X11_Xscreen_h
#define Fresco_defs_X11_Xscreen_h

#include <Fresco/defs/display.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_X11/Xlib.h>
#include <Fresco/defs_X11/Xutil.h>

class DisplayImpl;
class RGBTable;
class ScreenVisualList;

typedef int XCoord;

//+ ScreenImpl : XfScreen
class ScreenImpl : public XfScreen {
public:
    ~ScreenImpl();
    _Ix_InterfaceOps(ScreenImpl*)
//+
public:
    ScreenImpl(DisplayImpl*, long n);

    OxExchange* _exchange();

    //+ XfScreen::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfScreen */
    XfCoord dpi();
    void dpi(XfCoord _p);
    XfDisplay_ptr display_ptr();
    XfCoord width();
    XfCoord height();
    XfPixelCoord to_pixels(XfCoord c, XfAxis a);
    XfCoord to_coord(XfPixelCoord p, XfAxis a);
    XfCoord to_pixels_coord(XfCoord c, XfAxis a);
    void move_pointer(XfCoord x, XfCoord y);
    XfWindow_ptr application(XfViewer_ptr v);
    XfWindow_ptr top_level(XfViewer_ptr v, XfWindow_ptr group_leader);
    XfWindow_ptr transient(XfViewer_ptr v, XfWindow_ptr transient_for);
    XfWindow_ptr popup(XfViewer_ptr v);
    XfWindow_ptr icon(XfViewer_ptr v);
    //+

    struct OverlayInfo {
	VisualID id;
	long type;
	long transparent;
	long layer;
    };

    struct VisualInfo {
	XDisplay* display;
	long screen;
	long depth;
	Visual* visual;
	XColormap cmap;
	OverlayInfo overlay;
	RGBTable* rgb_table;
	XColor* local_map;
	long local_map_size;
	unsigned long red, green, blue;
	unsigned long red_shift, green_shift, blue_shift;
	unsigned long white;
	unsigned long xor;
    };

    struct RGBTableEntry {
	unsigned short red, green, blue;
    };

    XWindow root_window();
    ScreenImpl::VisualInfo* default_visual();
    ScreenImpl::VisualInfo* find_visual(StyleContext_ptr);
    void find_color(ScreenImpl::VisualInfo&, XColor&);
protected:
    SharedFrescoObjectImpl object_;
    DisplayImpl* display_;
    long number_;
    XWindow root_;
    ScreenVisualList* visuals_;
    ScreenImpl::VisualInfo* default_visual_;
    Coord width_;
    Coord height_;
    XCoord pwidth_;
    XCoord pheight_;
    Coord pixels_;
    Coord points_;

    void set_dpi();
    ScreenImpl::VisualInfo* lookup_overlay(StyleContext_ptr);
    Boolean find_layer(StyleValue_ptr, Long& layer);
    Boolean find_overlay(Long layer, XVisualInfo&, ScreenImpl::OverlayInfo&);
    Boolean lookup_visual_id(StyleContext_ptr, XVisualInfo&);
    Boolean lookup_visual(StyleContext_ptr, XVisualInfo&);
    Boolean find_visual_by_class_name(CharString_ptr name, XVisualInfo& xinfo);
    void find_overlay(Long layer, ScreenImpl::VisualInfo&);
    Boolean find_visual_by_info(XVisualInfo&, Long mask);
    void init_visual(ScreenImpl::VisualInfo&);
    void init_color_tables(ScreenImpl::VisualInfo&);
    void set_shift(unsigned long mask, unsigned long& v, unsigned long& shift);
    unsigned int MSB(unsigned long);
    unsigned long xor(const ScreenImpl::VisualInfo&, StyleContext_ptr);
    void load_map(ScreenImpl::VisualInfo&);
    void search_map(
	ScreenImpl::VisualInfo&, ScreenImpl::RGBTableEntry& rgb, XColor& xc
    );
    double distance(
	unsigned short r, unsigned short g, unsigned short b, const XColor&
    );
    unsigned long rescale(
	unsigned long value, unsigned long in_scale, unsigned long out_scale
    );
};

#endif
