//
// $XConsortium$
//

//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_X11_Xwindow_h
#define Fresco_defs_X11_Xwindow_h

#include <Fresco/defs/glyph.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/styles.h>
#include <Fresco/defs_impl/thread.h>
#include <Fresco/defs_X11/Xdisplay.h>
#include <Fresco/defs_X11/Xscreen.h>

class DamageRegion;
class GlyphTraversalImpl;
class MainViewer;
class RegionImpl;
class XPainterImpl;

//+ WindowImpl : XfWindow, XfDamage
class WindowImpl : public XfWindow, public XfDamage {
public:
    ~WindowImpl();
    _Ix_InterfaceOps(WindowImpl*)
//+
public:
    WindowImpl(DisplayImpl*, ScreenImpl*, Viewer_ptr);

    OxExchange* _exchange();

    //+ XfWindow::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfWindow */
    Boolean double_buffered();
    void double_buffered(Boolean _p);
    XfCursor_ptr default_cursor();
    void default_cursor(XfCursor_ptr _p);
    XfColor_ptr cursor_foreground();
    void cursor_foreground(XfColor_ptr _p);
    XfColor_ptr cursor_background();
    void cursor_background(XfColor_ptr _p);
    XfCharString_ptr geometry();
    void geometry(XfCharString_ptr _p);
    XfWindow_ptr icon();
    void icon(XfWindow_ptr _p);
    Boolean iconic();
    void iconic(Boolean _p);
    XfCharString_ptr title();
    void title(XfCharString_ptr _p);
    Long xor_pixel();
    void xor_pixel(Long _p);
    XfScreen_ptr screen_ptr();
    XfStyleContext_ptr style();
    XfViewer_ptr main_viewer();
    void configure(const XfWindow::Placement& p);
    void get_configuration(Boolean position, XfWindow::Placement& p);
    void configure_notify(XfCoord width, XfCoord height);
    void move_notify(XfCoord left, XfCoord bottom);
    void map();
    void unmap();
    Boolean is_mapped();
    void map_notify();
    void unmap_notify();
    void iconify();
    void deiconify();
    void raise();
    void lower();
    void redraw(XfCoord left, XfCoord bottom, XfCoord width, XfCoord height);
    void repair();
    Boolean hits(XfEvent_ptr e);
    void handle_event(XfEvent_ptr e);
    void grab_pointer(XfCursor_ptr c);
    void ungrab_pointer();
    //+

    //+ XfDamage::=
    void incur();
    void extend(XfRegion_ptr r);
    XfRegion_ptr current();
    void begin_update();
    void end_update();
    XfWindow_ptr window_ptr();
    //+

    virtual void default_geometry();
    virtual void compute_geometry();
    virtual void bind();
    virtual Boolean bound();
    virtual void set_attributes();
    virtual void set_props();
    virtual void do_map();
    virtual void unbind();

    virtual void need_repair();

    DisplayImpl* display();
    ScreenImpl* screen_impl();
    ScreenImpl::VisualInfo* visual();
    XDisplay* xdisplay();
    XWindow xwindow();
    PixelCoord pwidth() { return pwidth_; }
    PixelCoord pheight() { return pheight_; }
protected:
    SharedFrescoObjectImpl object_;
    Mutex_var lock_;
    DisplayImpl* display_;
    ScreenImpl* screen_;
    StyleImpl* style_;
    Cursor_ptr cursor_;
    Color_ptr cursor_foreground_;
    Color_ptr cursor_background_;
    Window_ptr icon_;
    ScreenImpl::VisualInfo* visual_;
    MainViewer* viewer_;
    XPainterImpl* painter_;
    Mutex_var damage_lock_;
    RegionImpl* damage_;
    Long update_nesting_;
    GlyphTraversalImpl* draw_;
    GlyphTraversalImpl* pick_;
    Window::Placement placement_;
    XWindow xwindow_;
    XSetWindowAttributes xattrs_;
    unsigned long xattrmask_;
    unsigned int xclass_;
    PixelCoord xleft_, xtop_;
    PixelCoord pwidth_, pheight_;

    Boolean placed_ : 1;
    Boolean needs_resize_ : 1;
    Boolean resized_ : 1;
    Boolean moved_ : 1;
    Boolean wm_mapped_ : 1;
    Boolean map_pending_ : 1;

    static Boolean xbound(XWindow);
    void set_attribute(const char* name, CharString_ptr value);
    CharString_ptr get_attribute(const char* name);
    void clear_mapping_info();
    void check_position();
    void compute_x_position();
    void damage(Coord l, Coord b, Coord r, Coord t);
    void do_request(Glyph::Requisition&);
    void do_draw_traversal();
    void moved(PixelCoord x, PixelCoord y);
    void resize();
};

inline DisplayImpl* WindowImpl::display() { return display_; }
inline ScreenImpl* WindowImpl::screen_impl() { return screen_; }
inline ScreenImpl::VisualInfo* WindowImpl::visual() { return visual_; }
inline XDisplay* WindowImpl::xdisplay() { return display_->xdisplay(); }
inline XWindow WindowImpl::xwindow() { return xwindow_; }

class ManagedWindow : public WindowImpl {
public:
    struct HintInfo {
	XWMHints* hints;
	XDisplay* xdisplay;
	XWindow xwindow;
    };

    typedef Boolean (ManagedWindow::*HintFunction)(ManagedWindow::HintInfo&);

    ManagedWindow(DisplayImpl*, ScreenImpl*, Viewer_ptr);
    virtual ~ManagedWindow();

    void iconify(); //+ XfWindow::iconify
    void deiconify(); //+ XfWindow::deiconify

    virtual void compute_geometry();
    virtual void set_props();
protected:
    Window_ptr group_leader_;
    Window_ptr transient_for_;
private:
    void do_set(ManagedWindow::HintFunction);

    Boolean set_name(ManagedWindow::HintInfo&);
    Boolean set_geometry(ManagedWindow::HintInfo&);
    Boolean set_group_leader(ManagedWindow::HintInfo&);
    Boolean set_transient_for(ManagedWindow::HintInfo&);
    Boolean set_icon_name(ManagedWindow::HintInfo&);
    Boolean set_icon_geometry(ManagedWindow::HintInfo&);
    Boolean set_icon(ManagedWindow::HintInfo&);
    Boolean set_icon_bitmap(ManagedWindow::HintInfo&);
    Boolean set_icon_mask(ManagedWindow::HintInfo&);
    Boolean set_all(ManagedWindow::HintInfo&);

    void wm_normal_hints();
    void wm_name();
    void wm_class();
    void wm_protocols();
    void wm_colormap_windows();
    void wm_hints();
};

class ApplicationWindow : public ManagedWindow {
public:
    ApplicationWindow(DisplayImpl*, ScreenImpl*, Viewer_ptr);
    virtual ~ApplicationWindow();

    virtual void compute_geometry();
    virtual void set_props();
};

class TopLevelWindow : public ManagedWindow {
public:
    TopLevelWindow(
	DisplayImpl*, ScreenImpl*, Viewer_ptr, Window_ptr group_leader
    );
    virtual ~TopLevelWindow();
};

class TransientWindow : public ManagedWindow {
public:
    TransientWindow(
	DisplayImpl*, ScreenImpl*, Viewer_ptr, Window_ptr transient_for
    );
    virtual ~TransientWindow();

    virtual void set_attributes();
};

class PopupWindow : public WindowImpl {
public:
    PopupWindow(DisplayImpl*, ScreenImpl*, Viewer_ptr);
    virtual ~PopupWindow();

    virtual void set_attributes();
};

class IconWindow : public ManagedWindow {
public:
    IconWindow(DisplayImpl*, ScreenImpl*, Viewer_ptr);
    virtual ~IconWindow();

    virtual void do_map();
};

#endif
