//
// $XConsortium$
//

//
// Copyright (c) 1992-1993 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the name of
// Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_impl_action_h
#define Fresco_defs_impl_action_h

#include <Fresco/defs_impl/fobjects.h>

//- ActionImpl*
class export ActionImpl : public Action {
    //. ActionImpl provides a common implementation for actions.
public:
    ActionImpl();
    ~ActionImpl();

    OxExchange* _exchange();

    //+ XfAction::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfAction */
    void execute();
    Boolean reversible();
    void unexecute();
    //+
protected:
    SharedFrescoObjectImpl object_;
};

/*
 * Action denoted by an object and member function to call on the object.
 */

#if !defined(UNIXCPP)
#define __ActionCallback(T) T##_ActionCallback
#define ActionCallback(T) __ActionCallback(T)
#define __ActionMemberFunction(T) T##_ActionMemberFunction
#define ActionMemberFunction(T) __ActionMemberFunction(T)
#else
#define __ActionCallback(T) T/**/_ActionCallback
#define ActionCallback(T) __ActionCallback(T)
#define __ActionMemberFunction(T) T/**/_ActionMemberFunction
#define ActionMemberFunction(T) __ActionMemberFunction(T)
#endif

#define declareActionCallback(T) \
typedef void (T::*ActionMemberFunction(T))(); \
class ActionCallback(T) : public ActionImpl { \
public: \
    ActionCallback(T)(T*, ActionMemberFunction(T)); \
    virtual ~ActionCallback(T)(); \
\
    void execute(); \
private: \
    T* obj_; \
    ActionMemberFunction(T) func_; \
};

#define implementActionCallback(T) \
ActionCallback(T)::ActionCallback(T)(T* obj, ActionMemberFunction(T) func) { \
    obj_ = obj; \
    func_ = func; \
} \
ActionCallback(T)::~ActionCallback(T)() { } \
\
void ActionCallback(T)::execute() { (obj_->*func_)(); }

#endif
