//
// $XConsortium$
//

//
// Copyright (c) 1993-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_impl_charstr_h
#define Fresco_defs_impl_charstr_h

#include <Fresco/defs/types.h>
#include <Fresco/defs_impl/fobjects.h>

class export CharStringImpl : public CharString {
public:
    CharStringImpl();
    CharStringImpl(const char*);
    CharStringImpl(const char*, Long length);
    CharStringImpl(CharString_ptr);
    CharStringImpl(const CharStringImpl&);
    ~CharStringImpl();

    OxExchange* _exchange();

    //+ XfCharString::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfCharString */
    XfCharString::HashValue hash();
    ULong count();
    void copy(XfCharString_ptr s);
    Boolean equal(XfCharString_ptr s);
    Boolean case_insensitive_equal(XfCharString_ptr s);
    void get_data(XfCharString::Data& d);
    void get_char_data(XfCharString::CharData& d);
    void put_data(const XfCharString::Data& d);
    void put_char_data(const XfCharString::CharData& d);
    char* string_ptr();
    //+

    static CharStringImpl* create_static(const char* s);
    static const char* index(const char* s, Long length, char c);
protected:
    SharedFrescoObjectImpl object_;
    char* data_;
    ULong length_;
    Boolean dynamic_;

    void set(const char*, Long);
    void free();
};

class export CharStringBuffer {
public:
    CharStringBuffer(const CharString_var& s) { init(s); }
    CharStringBuffer(CharString_ptr s) { init(s); }
    ~CharStringBuffer();

    char* string() { return buf_._buffer; }
    Long length() { return buf_._length; }

    Boolean operator ==(const char*) const ;
    Boolean operator ==(const CharStringBuffer&) const ;
    Boolean operator !=(const char*) const ;
    Boolean operator !=(const CharStringBuffer&) const ;
protected:
    char fixed_[100];
    CharString::CharData buf_;

    void init(CharString_ptr);
};

static inline unsigned long key_to_hash(CharString_ptr s) { return s->hash(); }
static inline key_equal(CharString_ptr s1, CharString_ptr s2) {
    return s1->equal(s2);
}

#endif
