/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-94 Silicon Graphics, Inc.
 * Copyright (c) 1993-94 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef Fresco_defs_impl_displays_h
#define Fresco_defs_impl_displays_h

#include <Fresco/defs/display.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/styles.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/defs_impl/thread.h>

class DisplayFilters;
class DisplayDamage;

class export DisplayBase : public XfDisplay {
public:
    DisplayBase();
    ~DisplayBase();

    OxExchange* _exchange();

    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfDisplay */
    XfStyleContext_ptr style();
    XfDrawingKit_ptr drawing_kit() = 0;
    XfScreen_ptr display_screen(XfDisplay::ScreenNumber n) = 0;
    XfDisplay::ScreenNumber number_of_screens() = 0;
    XfScreen_ptr default_screen() = 0;
    XfCursor_ptr cursor_from_data(Short x, Short y, CursorPattern pat, CursorPattern mask) = 0;
    XfCursor_ptr cursor_from_bitmap(XfRaster_ptr b, XfRaster_ptr mask) = 0;
    XfCursor_ptr cursor_from_font(XfFont_ptr f, Long pat, Long mask) = 0;
    XfCursor_ptr cursor_from_index(Long n) = 0;
    void run(Boolean b);
    Boolean running();
    void run_one_event();
    XfTag add_filter(XfEventFilter_ptr f);
    XfTag add_general_filter(XfEventFilter_ptr f);
    void remove_filter(XfTag add_tag);
    char* keysym_to_string(XfEvent::KeySym k) = 0;
    XfEvent::KeySym string_to_keysym(const char* s) = 0;
    void need_repair(XfWindow_ptr w);
    void repair();
    void flush() = 0;
    void flush_and_wait() = 0;
    void ring_bell(Float pct_loudness) = 0;
    void close();
    Boolean auto_repeat() = 0;
    void auto_repeat(Boolean _p) = 0;
    Float key_click_volume() = 0;
    void key_click_volume(Float _p) = 0;
    Float pointer_acceleration() = 0;
    void pointer_acceleration(Float _p) = 0;
    Long pointer_threshold() = 0;
    void pointer_threshold(Long _p) = 0;
    Float pointer_double_click_threshold() = 0;
    void pointer_double_click_threshold(Float _p) = 0;

    // Should this be part of the Display interface? -denis
    Boolean filtered(Event_ptr event);

protected:
    virtual Boolean run_one_event_bool() = 0;
    virtual Boolean peek() = 0;

protected:
    Boolean prepare_to_read();
    void after_read();

    Tag add_filter_info(EventFilter_ptr, Boolean general);

    void do_repairs();
    void redisplay_thread();
    void notify_redisplay();

    SharedFrescoObjectImpl object_;

    Mutex_var lock_;

    Boolean reading_;
    Long waiting_to_read_;
    ConditionVariable_var wait_to_read_;

    Boolean requesting_;
    Long waiting_to_request_;
    ConditionVariable_var wait_to_request_;

    Boolean running_;

    StyleImpl* style_;

    Float double_click_;

    Mutex_var filter_lock_;
    DisplayFilters* filters_;
    long filter_tag_;

    Thread_var redisplay_;
    ConditionVariable_var damage_pending_;
    DisplayDamage* damaged_;
};

#endif
