/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <stdio.h>

#include <Fresco/defs/display.h>
#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/drawings.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/memory.h>
#include <Fresco/defs_impl/rasters.h>
#include <Fresco/defs_impl/transform.h>

#include <Fresco/Ox/schema.h>

// class DrawingKitBase

implementTable(DrawingKitBaseColorTable,CharString_ptr,Color_ptr)

implementTable(DrawingKitBaseFontTable,CharString_ptr,Font_ptr)

DrawingKitBase::DrawingKitBase() {
    colors_ = new DrawingKitBaseColorTable(20);
    fonts_ = new DrawingKitBaseFontTable(20);
    foreground_str_ = FrescoLib::string_ref("foreground");
    Foreground_str_ = FrescoLib::string_ref("Foreground");
    background_str_ = FrescoLib::string_ref("background");
    Background_str_ = FrescoLib::string_ref("Background");
    font_str_ = FrescoLib::string_ref("font");
    Font_str_ = FrescoLib::string_ref("Font");
}

DrawingKitBase::~DrawingKitBase() {
    // !!! Need to release the entries in these tables. -denis
    delete colors_;
    delete fonts_;
}

OxExchange* DrawingKitBase::_exchange() { return &object_; } 

//+ DrawingKitBase(XfFrescoObject::=object_.)
XfTag DrawingKitBase::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void DrawingKitBase::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void DrawingKitBase::disconnect() {
    object_.disconnect();
}
void DrawingKitBase::notify_observers() {
    object_.notify_observers();
}
void DrawingKitBase::update() {
    object_.update();
}
//+

//+ DrawingKitBase(XfDrawingKit::style)
XfStyleContext_ptr DrawingKitBase::style() {
    return Display_var(display_ptr())->style();
}

//+ DrawingKitBase(XfDrawingKit::simple_brush)
XfBrush_ptr DrawingKitBase::simple_brush(XfCoord width) {
    return new BrushBase(width);
}

//+ DrawingKitBase(XfDrawingKit::dither_brush)
XfBrush_ptr DrawingKitBase::dither_brush(XfCoord width, Long p) {
    return new BrushBase(p, width);
}

//+ DrawingKitBase(XfDrawingKit::patterned_brush)
XfBrush_ptr DrawingKitBase::patterned_brush(XfCoord width, const XfDrawingKit::Data32& p) {
    return new BrushBase(p._buffer, p._length, width);
}

//+ DrawingKitBase(XfDrawingKit::solid_pattern)
XfPattern_ptr DrawingKitBase::solid_pattern() {
    return new PatternBase(0xffff);
}

//+ DrawingKitBase(XfDrawingKit::halftone_pattern)
XfPattern_ptr DrawingKitBase::halftone_pattern() {
    return new PatternBase(0xa5a5);
}

//+ DrawingKitBase(XfDrawingKit::stipple)
XfPattern_ptr DrawingKitBase::stipple(const XfDrawingKit::Data32& d) {
    return new PatternBase(d);
}

//+ DrawingKitBase(XfDrawingKit::stipple_4x4)
XfPattern_ptr DrawingKitBase::stipple_4x4(ULong p) {
    return new PatternBase(p);
}

//+ DrawingKitBase(XfDrawingKit::color_rgb)
XfColor_ptr DrawingKitBase::color_rgb(XfColor::Intensity r, XfColor::Intensity g, XfColor::Intensity b) {
    return new ColorBase(r, g, b);
}

//+ DrawingKitBase(XfDrawingKit::find_color)
XfColor_ptr DrawingKitBase::find_color(XfCharString_ptr n) {
    Color_ptr c = nil;
    if (! colors_->find(c, n)) {
        CharStringBuffer cs(n);
        const char* nm = cs.string();
        if (nm[0] == '#') {
            int rx;
            int gx;
            int bx;
            sscanf(&(nm[1]), "%2x", &rx);
            sscanf(&(nm[3]), "%2x", &gx);
            sscanf(&(nm[5]), "%2x", &bx);

            Color::Intensity range = Color::Intensity(255);
            Color::Intensity r = Color::Intensity(rx) / range;
            Color::Intensity g = Color::Intensity(gx) / range;
            Color::Intensity b = Color::Intensity(bx) / range;
            c = new ColorBase(r, g, b);
            colors_->insert(n, c);
        }
    }
    return Color::_duplicate(c);
}

//+ DrawingKitBase(XfDrawingKit::resolve_color)
XfColor_ptr DrawingKitBase::resolve_color(XfStyleContext_ptr s, XfCharString_ptr n) {
    Color_ptr c = nil;
    StyleValue_var a = s->resolve(n);
    if (is_not_nil(a)) {
        CharString_var v;
        if (a->read_string(v.out())) {
            c = find_color(v);
        }
    }
    return c;
}

Color_ptr DrawingKitBase::resolve_colors(
    StyleContext_ptr s, CharString_ptr name1, CharString_ptr name2
) {
    Color_ptr c = resolve_color(s, name1);
    if (is_nil(c)) {
        c = resolve_color(s, name2);
    }
    return c;
}

//+ DrawingKitBase(XfDrawingKit::foreground)
XfColor_ptr DrawingKitBase::foreground(XfStyleContext_ptr s) {
    return resolve_colors(s, foreground_str_, Foreground_str_);
}

//+ DrawingKitBase(XfDrawingKit::background)
XfColor_ptr DrawingKitBase::background(XfStyleContext_ptr s) {
    return resolve_colors(s, background_str_, Background_str_);
}

//+ DrawingKitBase(XfDrawingKit::find_font)
XfFont_ptr DrawingKitBase::find_font(XfCharString_ptr fullname) {
    Font_ptr f = nil;
    fonts_->find(f, fullname);
    return Font::_duplicate(f);
}

//+ DrawingKitBase(XfDrawingKit::find_font_match)
XfFont_ptr DrawingKitBase::find_font_match(XfCharString_ptr font_family, XfCharString_ptr font_style, XfCoord ptsize) {
    return nil;
}

//+ DrawingKitBase(XfDrawingKit::resolve_font)
XfFont_ptr DrawingKitBase::resolve_font(XfStyleContext_ptr s, XfCharString_ptr n) {
    Font_ptr f = nil;
    StyleValue_var a = s->resolve(n);
    if (is_not_nil(a)) {
        CharString_var v;
        if (a->read_string(v.out())) {
            f = find_font(v);
        }
    }
    return f;
}

//+ DrawingKitBase(XfDrawingKit::default_font)
XfFont_ptr DrawingKitBase::default_font(XfStyleContext_ptr s) {
    Font_ptr f = resolve_font(s, font_str_);
    if (is_nil(f)) {
        f = resolve_font(s, Font_str_);
    }
    return f;
}

//+ DrawingKitBase(XfDrawingKit::bitmap_from_data)
XfRaster_ptr DrawingKitBase::bitmap_from_data(const XfDrawingKit::Data8& d, XfRaster::Index rows, XfRaster::Index columns, XfRaster::Index origin_row, XfRaster::Index origin_column) {
    return new RasterBitmapImpl(d, rows, columns, origin_row, origin_column);
}

//+ DrawingKitBase(XfDrawingKit::bitmap_from_char)
XfRaster_ptr DrawingKitBase::bitmap_from_char(XfFont_ptr f, XfCharCode c) {
// !!! Should this call be a member of the Font interface? -denis
//    return f->bitmap(c);
    return nil;
}

//+ DrawingKitBase(XfDrawingKit::raster_open)
XfRaster_ptr DrawingKitBase::raster_open(XfCharString_ptr n) {
    // !!! What formats should we read? -denis
    return nil;
}

//+ DrawingKitBase(XfDrawingKit::transform_matrix)
XfTransform_ptr DrawingKitBase::transform_matrix(XfTransform::Matrix m) {
    return new TransformImpl(m);
}

//+ DrawingKitBase(XfDrawingKit::identity_matrix)
XfTransform_ptr DrawingKitBase::identity_matrix() {
    return new TransformImpl;
}

//+ DrawingKitBase(XfDrawingKit::printer)
XfPainter_ptr DrawingKitBase::printer(XfCharString_ptr output) {
    return nil;
}

void DrawingKitBase::insert_color(XfCharString_ptr fullname, XfColor_ptr color) {
    colors_->insert(fullname, XfColor::_duplicate(color));
}

void DrawingKitBase::insert_font(XfCharString_ptr fullname, XfFont_ptr font) {
    fonts_->insert(fullname, XfFont::_duplicate(font));
}

// class BrushBase

//+ BrushBase(BrushBase)
void* BrushBase::_this_ptr() { return this; }
_Ix_extern__tid(BrushBase);
void* BrushBase::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(BrushBase) == t) return this;
    if (CORBA::Object::_type_id() == t) return (CORBA::Object*) this;
    return nil;
};
_Ix_extern__tid(BrushBase);
_Ix_ImplementGetTid(BrushBase,_Ix__tid(BrushBase))
_Ix_ImplementNarrow(BrushBase*,BrushBase,_Ix_NullParameter,BrushBase,0)
_Ix_ImplementDuplicate(BrushBase*,BrushBase,_Ix_NullParameter,_Ix__tid(BrushBase),0)
//+

BrushBase::BrushBase(Coord w) {
    init(nil, 0, w);
}

BrushBase::BrushBase(const Long* p, Long c, Coord w) {
    init(p, c, w);
}

BrushBase::BrushBase(Long pat, Coord w) {
    Long dash[16];
    Long count;

    calc_dashes(pat, dash, count);
    init(dash, count, w);
}

BrushBase::~BrushBase() {
    delete dash_list_;
}

OxExchange* BrushBase::_exchange() { return &object_; } 

//+ BrushBase(XfFrescoObject::=object_.)
XfTag BrushBase::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void BrushBase::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void BrushBase::disconnect() {
    object_.disconnect();
}
void BrushBase::notify_observers() {
    object_.notify_observers();
}
void BrushBase::update() {
    object_.update();
}
//+

//+ BrushBase(XfBrush::width)
XfCoord BrushBase::width() {
    return width_;
}

//+ BrushBase(XfBrush::hash)
ULong BrushBase::hash() {
    return 0;
}

//+ BrushBase(XfBrush::equal)
Boolean BrushBase::equal(XfBrush_ptr b) {
    if (is_nil(b) || hash() != b->hash()) {
        return false;
    }
    /* should compare values */
    return this == b;
}

void BrushBase::info(Coord& w, long*& dash_list, Long& count) {
    w = width_;
    dash_list = dash_list_;
    count = dash_count_;
}

void BrushBase::init(const long* pattern, long count, Coord w) {
    width_ = w;
    dash_count_ = count;
    if (count > 0) {
        dash_list_ = new long[count];
        for (Long i = 0; i < count; i++) {
            dash_list_[i] = long(pattern[i]);
        }
    } else {
        dash_list_ = nil;
    }
}

void BrushBase::calc_dashes(long pat, long* dash, long& count) {
    unsigned long p = pat & 0xffff;

    if (p == 0 || p == 0xffff) {
        count = 0;
    } else {
        const unsigned long MSB = 1 << 15;
        while ((p & MSB) == 0) {
            p <<= 1;
        }

        if (p == 0x5555 || p == 0xaaaa) {
            dash[0] = 1;
            dash[1] = 3;
            count = 2;
        } else if (p == 0xaaaa) {
            dash[0] = 1;
            dash[1] = 1;
            count = 2;
        } else if (p == 0xcccc) {
            dash[0] = 2;
            dash[1] = 2;
            count = 2;
        } else if (p == 0xeeee) {
            dash[0] = 3;
            dash[1] = 1;
            count = 2;
        } else {
           unsigned long m = MSB;
           long index = 0;
           while (m != 0) {
                /* count the consecutive one bits */
                int length = 0;
                while (m != 0 && (p & m) != 0) {
                    ++length;
                    m >>= 1;
                }
                dash[index++] = length;

                /* count the consecutive zero bits */
                length = 0;
                while (m != 0 && (p & m) == 0) {
                    ++length;
                    m >>= 1;
                }
                if (length > 0) {
                    dash[index++] = length;
                }
            }
            count = index;
        }
    }
}

// class ColorBase

ColorBase::ColorBase(
    Color::Intensity r, Color::Intensity g, Color::Intensity b
) {
    red_ = r;
    green_ = g;
    blue_ = b;
}

ColorBase::~ColorBase() { }

OxExchange* ColorBase::_exchange() { return &object_; } 

//+ ColorBase(XfFrescoObject::=object_.)
XfTag ColorBase::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void ColorBase::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void ColorBase::disconnect() {
    object_.disconnect();
}
void ColorBase::notify_observers() {
    object_.notify_observers();
}
void ColorBase::update() {
    object_.update();
}
//+

//+ ColorBase(XfColor::rgb)
void ColorBase::rgb(XfColor::Intensity& r, XfColor::Intensity& g, XfColor::Intensity& b) {
    r = red_;
    g = green_;
    b = blue_;
}

//+ ColorBase(XfColor::equal)
Boolean ColorBase::equal(XfColor_ptr c) {
    if (is_nil(c)) {
        return false;
    }
    Color::Intensity r, g, b;
    c->rgb(r, g, b);
    float tol = float(1e-2);
    /* convoluted testing to workaround compiler inline problem */
    Boolean result = Math::equal(red_, r, tol);
    if (result) {
        result = Math::equal(green_, g, tol);
        if (result) {
            result = Math::equal(blue_, b, tol);
        }
    }
    return result;
}

/*
 * Hash function generates 18 distinct bits, 6 each from rgb.
 * The idea is to "promote" collisions a little bit, so that colors
 * that are indistinguishable will match.
 */

//+ ColorBase(XfColor::hash)
ULong ColorBase::hash() {
    Long r = Long(64 * red_) & 0x3f;
    Long g = Long(64 * green_) & 0x3f;
    Long b = Long(64 * blue_) & 0x3f;
    return (r << 12) | (g << 6) | b;
}


// class XfPatternBase

PatternBase::PatternBase(ULong p) {
    stipple_._maximum = 4;
    stipple_._length = 4;
    stipple_._buffer = new Long[4];
    stipple_._buffer[0] = (p & 0xf000) >> 12;
    stipple_._buffer[1] = (p & 0x0f00) >> 8;
    stipple_._buffer[2] = (p & 0x00f0) >> 4;
    stipple_._buffer[3] = (p & 0x000f);
}

PatternBase::PatternBase(const DrawingKit::Data32& d) {
    stipple_._maximum = d._length;
    stipple_._length = d._length;
    stipple_._buffer = new Long[d._length];
    Memory::copy(d._buffer, stipple_._buffer, sizeof(Long) * d._length);
}

PatternBase::~PatternBase() {
    /* avoid implicit deallocation */
    stipple_._buffer = nil;
}

OxExchange* PatternBase::_exchange() { return &object_; } 

//+ PatternBase(XfFrescoObject::=object_.)
XfTag PatternBase::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void PatternBase::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void PatternBase::disconnect() {
    object_.disconnect();
}
void PatternBase::notify_observers() {
    object_.notify_observers();
}
void PatternBase::update() {
    object_.update();
}
//+

//+ PatternBase(XfPattern::equal)
Boolean PatternBase::equal(XfPattern_ptr p) {
    return is_eq(this, p);
}

//+ PatternBase(XfPattern::hash)
ULong PatternBase::hash() {
    return (ULong) this;
}

//+ PatternBase(XfPattern::stipple)
void PatternBase::stipple(_Ix_T_out(XfDrawingKit::Data32) d) {
    d = new DrawingKit::Data32(stipple_);
}
