/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_defs_impls_drawings_h
#define Fresco_defs_impls_drawings_h

#include <Fresco/defs/display.h>
#include <Fresco/defs/drawing.h>

#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/table.h>

declareTable(DrawingKitBaseColorTable,CharString_ptr,Color_ptr)
declareTable(DrawingKitBaseFontTable,CharString_ptr,Font_ptr)

class export DrawingKitBase : public DrawingKit {
public:
    DrawingKitBase();
    ~DrawingKitBase();

    OxExchange* _exchange();

    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfDrawingKit */
    XfStyleContext_ptr style();
    XfBrush_ptr simple_brush(XfCoord width);
    XfBrush_ptr dither_brush(XfCoord width, Long p);
    XfBrush_ptr patterned_brush(XfCoord width, const XfDrawingKit::Data32& p);
    XfPattern_ptr solid_pattern();
    XfPattern_ptr halftone_pattern();
    XfPattern_ptr stipple(const XfDrawingKit::Data32& d);
    XfPattern_ptr stipple_4x4(ULong p);
    XfColor_ptr color_rgb(XfColor::Intensity r, XfColor::Intensity g, XfColor::Intensity b);
    XfColor_ptr find_color(XfCharString_ptr n);
    XfColor_ptr resolve_color(XfStyleContext_ptr s, XfCharString_ptr n);
    XfColor_ptr foreground(XfStyleContext_ptr s);
    XfColor_ptr background(XfStyleContext_ptr s);
    XfFont_ptr find_font(XfCharString_ptr fullname) = 0;
    XfFont_ptr find_font_match(XfCharString_ptr font_family, XfCharString_ptr font_style, XfCoord ptsize) = 0;
    XfFont_ptr resolve_font(XfStyleContext_ptr s, XfCharString_ptr n);
    XfFont_ptr default_font(XfStyleContext_ptr s);
    XfRaster_ptr bitmap_from_data(const XfDrawingKit::Data8& d, XfRaster::Index rows, XfRaster::Index columns, XfRaster::Index origin_row, XfRaster::Index origin_column);
    XfRaster_ptr bitmap_from_char(XfFont_ptr f, XfCharCode c) = 0;
    XfRaster_ptr raster_open(XfCharString_ptr n);
    XfTransform_ptr transform_matrix(XfTransform::Matrix m);
    XfTransform_ptr identity_matrix();
    XfPainter_ptr printer(XfCharString_ptr output);

    // !!! Should this be part of the DrawingKit interface? -denis
    virtual Display_ptr display_ptr() = 0;

protected:
    void insert_color(XfCharString_ptr fullname, XfColor_ptr color);
    void insert_font(XfCharString_ptr fullname, XfFont_ptr font);

private:
    Color_ptr resolve_colors(
	StyleContext_ptr s, CharString_ptr name1, CharString_ptr name2
    );

    SharedFrescoObjectImpl object_;
    DrawingKitBaseColorTable* colors_;
    DrawingKitBaseFontTable* fonts_;
    CharString_var foreground_str_;
    CharString_var Foreground_str_;
    CharString_var background_str_;
    CharString_var Background_str_;
    CharString_var font_str_;
    CharString_var Font_str_;
};

// class BrushBase

class export BrushBase : public XfBrush {
public:
    BrushBase(Coord width);
    BrushBase(const long* p, long c, Coord w);
    BrushBase(long pat, Coord w);
    ~BrushBase();

    _Ix_InterfaceOps(BrushBase*)
    OxExchange* _exchange();

    //+ XfBrush::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfBrush */
    XfCoord width();
    Boolean equal(XfBrush_ptr b);
    ULong hash();
    //+

    // !!! Something like this should be part of the Brush interface. -denis
    void info(Coord& width, long*& dash_list, long& dash_count);

protected:
    void init(const long* p, long c, Coord w);
    void calc_dashes(long pat, long* dash, long& count);

    SharedFrescoObjectImpl object_;
    Coord width_;
    long* dash_list_;
    long dash_count_;
};

// class ColorBase

class export ColorBase : public XfColor {
public:
    ColorBase(Color::Intensity r, Color::Intensity g, Color::Intensity b);
    ~ColorBase();

    OxExchange* _exchange();

    //+ XfColor::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfColor */
    void rgb(XfColor::Intensity& r, XfColor::Intensity& g, XfColor::Intensity& b);
    Boolean equal(XfColor_ptr c);
    ULong hash();
    //+

protected:
    SharedFrescoObjectImpl object_;
    Color::Intensity red_;
    Color::Intensity green_;
    Color::Intensity blue_;
};

// class PatternBase

class export PatternBase : public XfPattern {
public:
    PatternBase(ULong);
    PatternBase(const DrawingKit::Data32&);
    ~PatternBase();

    OxExchange* _exchange();

    //+ XfPattern::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfPattern */
    Boolean equal(XfPattern_ptr p);
    ULong hash();
    void stipple(_Ix_T_out(XfDrawingKit::Data32) d);
    //+

protected:
    SharedFrescoObjectImpl object_;
    DrawingKit::Data32 stipple_;
};

#endif
