//
// $XConsortium$
//

//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

//
// Common glyph implementations
//

#include <Fresco/defs/display.h>
#include <Fresco/defs/drawing.h>
#include <Fresco/defs/viewer.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/polyglyph.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/defs_impl/traversal.h>
#include <stdio.h>

implementPtrList(GlyphOffsetList,GlyphOffset)

GlyphImpl::GlyphImpl() { }
GlyphImpl::~GlyphImpl() { }

OxExchange* GlyphImpl::_exchange() { return &object_; }

//+ GlyphImpl(XfFrescoObject::=object_.)
XfTag GlyphImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void GlyphImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void GlyphImpl::disconnect() {
    object_.disconnect();
}
void GlyphImpl::notify_observers() {
    object_.notify_observers();
}
void GlyphImpl::update() {
    object_.update();
}
//+

//+ GlyphImpl(XfGlyph::style=)
void GlyphImpl::style(XfStyleContext_ptr) { }

//+ GlyphImpl(XfGlyph::style?)
XfStyleContext_ptr GlyphImpl::style() { return nil; }

//+ GlyphImpl(XfGlyph::transformation)
XfTransform_ptr GlyphImpl::transformation() { return nil; }

//+ GlyphImpl(XfGlyph::request)
void GlyphImpl::request(XfGlyph::Requisition&) { }

//+ GlyphImpl(XfGlyph::extension)
void GlyphImpl::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    default_extension(a, r);
}

//+ GlyphImpl(XfGlyph::shape)
void GlyphImpl::shape(XfRegion_ptr) { }

//+ GlyphImpl(XfGlyph::allocations)
void GlyphImpl::allocations(XfGlyph::AllocationInfoSeq& a) {
    for (ListItr(GlyphOffsetList) i(parents_); i.more(); i.next()) {
	i.cur()->allocations(a);
    }
}

//+ GlyphImpl(XfGlyph::traverse)
void GlyphImpl::traverse(XfGlyphTraversal_ptr t) {
    switch (t->op()) {
    case GlyphTraversal::draw:
	draw(t);
	break;
    case GlyphTraversal::pick_top:
    case GlyphTraversal::pick_any:
    case GlyphTraversal::pick_all:
	pick(t);
	break;
    case GlyphTraversal::other:
	t->visit();
	break;
    }
}

//+ GlyphImpl(XfGlyph::draw)
void GlyphImpl::draw(XfGlyphTraversal_ptr) { }

//+ GlyphImpl(XfGlyph::pick)
void GlyphImpl::pick(XfGlyphTraversal_ptr) { }

//+ GlyphImpl(XfGlyph::body=)
void GlyphImpl::body(XfGlyph_ptr) { }

//+ GlyphImpl(XfGlyph::body?)
XfGlyph_ptr GlyphImpl::body() { return nil; }

//+ GlyphImpl(XfGlyph::append)
void GlyphImpl::append(XfGlyph_ptr) { }

//+ GlyphImpl(XfGlyph::prepend)
void GlyphImpl::prepend(XfGlyph_ptr) { }

//+ GlyphImpl(XfGlyph::add_parent)
XfTag GlyphImpl::add_parent(XfGlyphOffset_ptr parent_offset) {
    Long n = parents_.count();
    parents_.append(parent_offset);
    return n;
}

//+ GlyphImpl(XfGlyph::remove_parent)
void GlyphImpl::remove_parent(XfTag add_tag) {
    Long n = Long(add_tag);
    if (n >= 0 && n < parents_.count()) {
	parents_.remove(n);
    }
}

//+ GlyphImpl(XfGlyph::first_child_offset)
XfGlyphOffset_ptr GlyphImpl::first_child_offset() {
    return nil;
}

//+ GlyphImpl(XfGlyph::last_child_offset)
XfGlyphOffset_ptr GlyphImpl::last_child_offset() {
    return nil;
}

//+ GlyphImpl(XfGlyph::parent_offsets)
void GlyphImpl::parent_offsets(XfGlyph::OffsetSeq& parents) {
    get_parent_offsets(parents, parents_);
}

void GlyphImpl::get_parent_offsets(
    Glyph::OffsetSeq& offsets, const GlyphOffsetList& parents
) {
    ULong p = ULong(parents.count());
    if (p > offsets.maximum()) {
	offsets.load(p, 0, new GlyphOffset_ptr[p], true);
    }
    offsets.length(p);
    Long n = 0;
    for (ListItr(GlyphOffsetList) i(parents); i.more(); i.next(), n++) {
	offsets[n] = GlyphOffset::_duplicate(i.cur());
    }
}

//+ GlyphImpl(XfGlyph::need_redraw)
void GlyphImpl::need_redraw() {
    Glyph::AllocationInfoSeq alist;
    allocations(alist);
    for (ULong i = 0; i < alist->length(); i++) {
	Glyph::AllocationInfo& a = alist[i];
	if (is_not_nil(a.damaged)) {
	    RegionImpl r;
	    extension(a, &r);
	    if (r.defined_) {
		a.damaged->extend(&r);
	    }
	}
    }
}

//+ GlyphImpl(XfGlyph::need_redraw_region)
void GlyphImpl::need_redraw_region(XfRegion_ptr r) {
    Glyph::AllocationInfoSeq alist;
    allocations(alist);
    RegionImpl dr;
    for (ULong i = 0; i < alist->length(); i++) {
	Glyph::AllocationInfo& a = alist[i];
	if (is_not_nil(a.damaged)) {
	    dr.copy(r);
	    dr.apply_transform(a.transformation);
	    a.damaged->extend(&dr);
	}
    }
}

//+ GlyphImpl(XfGlyph::need_resize)
void GlyphImpl::need_resize() {
    for (ListItr(GlyphOffsetList) i(parents_); i.more(); i.next()) {
	i.cur()->notify();
    }
}

//+ GlyphImpl(XfGlyph::restore_trail) !t
Boolean GlyphImpl::restore_trail(XfGlyphTraversal_ptr) {
    return false;
}

//+ GlyphImpl(XfGlyph::clone_glyph)
XfGlyph_ptr GlyphImpl::clone_glyph() {
    return nil;
}

void GlyphImpl::init_requisition(Glyph::Requisition& r) {
    r.x.defined = false;
    r.y.defined = false;
    r.z.defined = false;
    r.preserve_aspect = false;
}

void GlyphImpl::default_requisition(Glyph::Requisition& r) {
    Coord zero = Coord(0.0);
    require(r.x, zero, zero, zero, zero);
    require(r.y, zero, zero, zero, zero);
    require(r.z, zero, zero, zero, zero);
    r.preserve_aspect = false;
}

void GlyphImpl::require(
    Glyph::Requirement& r,
    Coord natural, Coord stretch, Coord shrink, Coord alignment
) {
    r.defined = true;
    r.natural = natural;
    r.maximum = natural + stretch;
    r.minimum = natural - shrink;
    r.align = alignment;
}

void GlyphImpl::require_lead_trail(
    Glyph::Requirement& r,
    Coord natural_lead, Coord max_lead, Coord min_lead,
    Coord natural_trail, Coord max_trail, Coord min_trail
) {
    Coord zero = Coord(0);
    r.defined = true;
    natural_lead = Math::max(min_lead, Math::min(max_lead, natural_lead));
    max_lead = Math::max(max_lead, natural_lead);
    min_lead = Math::min(min_lead, natural_lead);
    natural_trail = Math::max(min_trail, Math::min(max_trail, natural_trail));
    max_trail = Math::max(max_trail, natural_trail);
    min_trail = Math::min(min_trail, natural_trail);
    r.natural = natural_lead + natural_trail;
    if (natural_lead == zero) {
	r.minimum = min_trail;
	r.maximum = max_trail;
	r.align = Coord(0);
    } else if (natural_trail == zero) {
	r.minimum = min_lead;
	r.maximum = max_lead;
	r.align = Coord(1);
    } else {
        r.minimum = r.natural * Math::max(
	    min_lead / natural_lead, min_trail / natural_trail
	);
	r.maximum = r.natural * Math::min(
            max_lead / natural_lead, max_trail / natural_trail
        );
        if (r.natural == zero) {
            r.align = zero;
        } else {
            r.align = natural_lead / r.natural;
        }
    }
}

Glyph::Requirement* GlyphImpl::requirement(Glyph::Requisition& r, Axis a) {
    Glyph::Requirement* req;
    switch (a) {
    case X_axis:
	req = &r.x;
	break;
    case Y_axis:
	req = &r.y;
	break;
    case Z_axis:
	req = &r.z;
	break;
    default:
	req = nil;
	break;
    }
    return req;
}

void GlyphImpl::default_extension(
    const Glyph::AllocationInfo& a, Region_ptr r
) {
    if (is_not_nil(a.allocation)) {
	if (is_nil(a.transformation)) {
	    r->merge_union(a.allocation);
	} else {
	    RegionImpl tmp(a.allocation, a.transformation, nil);
	    r->merge_union(&tmp);
	}
    }
}

// class MonoGlyph

MonoGlyph::MonoGlyph() : offset_(this) { offset_.child_ = nil; }

MonoGlyph::~MonoGlyph() {
    if (is_not_nil(offset_.child_)) {
	offset_.child_->remove_parent(offset_.remove_tag_);
	CORBA::release(offset_.child_);
    }
}

//+ MonoGlyph(XfGlyph::request)
void MonoGlyph::request(XfGlyph::Requisition& r) {
    if (is_not_nil(offset_.child_)) {
	offset_.child_->request(r);
    }
}

//+ MonoGlyph(XfGlyph::extension)
void MonoGlyph::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    if (is_not_nil(offset_.child_)) {
	Glyph::AllocationInfo i;
	if (is_not_nil(a.allocation)) {
	    i.allocation = new RegionImpl;
	    i.allocation->copy(a.allocation);
	}
	if (is_not_nil(a.transformation)) {
	    i.transformation = new TransformImpl;
	    i.transformation->copy(a.transformation);
	}
	i.damaged = a.damaged;
	child_allocate(i);
	offset_.child_->extension(i, r);
    }
}

//+ MonoGlyph(XfGlyph::shape)
void MonoGlyph::shape(XfRegion_ptr r) {
    if (is_not_nil(offset_.child_)) {
	offset_.child_->shape(r);
    }
}

//+ MonoGlyph(XfGlyph::traverse)
void MonoGlyph::traverse(XfGlyphTraversal_ptr t) {
    if (is_not_nil(offset_.child_)) {
	t->traverse_child(&offset_, Region_var(t->allocation()));
    }
}

//+ MonoGlyph(XfGlyph::body=g)
void MonoGlyph::body(XfGlyph_ptr g) {
    if (is_not_nil(offset_.child_)) {
	offset_.child_->remove_parent(offset_.remove_tag_);
	CORBA::release(offset_.child_);
    }
    offset_.child_ = Glyph::_duplicate(g);
    offset_.remove_tag_ = is_not_nil(g) ? g->add_parent(&offset_) : 0;
}

//+ MonoGlyph(XfGlyph::body?)
XfGlyph_ptr MonoGlyph::body() {
    return Glyph::_duplicate(offset_.child_);
}

//+ MonoGlyph(XfGlyph::append)
void MonoGlyph::append(XfGlyph_ptr g) {
    offset_.child_->append(g);
}

//+ MonoGlyph(XfGlyph::prepend)
void MonoGlyph::prepend(XfGlyph_ptr g) {
    offset_.child_->prepend(g);
}

//+ MonoGlyph(XfGlyph::first_child_offset)
XfGlyphOffset_ptr MonoGlyph::first_child_offset() {
    return offset_.child_->first_child_offset();
}

//+ MonoGlyph(XfGlyph::last_child_offset)
XfGlyphOffset_ptr MonoGlyph::last_child_offset() {
    return offset_.child_->last_child_offset();
}

void MonoGlyph::visit_trail(GlyphTraversal_ptr t) {
    if (is_not_nil(offset_.child_)) {
	Glyph::AllocationInfo a;
	a.allocation = t->allocation();
	a.transformation = Painter_var(t->current_painter())->current_matrix();
	child_allocate(a);
    }
}

void MonoGlyph::child_allocate(Glyph::AllocationInfo&) { }

// class MonoGlyphOffset

MonoGlyphOffset::MonoGlyphOffset(MonoGlyph* g) { parent_ = g; }
MonoGlyphOffset::~MonoGlyphOffset() { }

OxExchange* MonoGlyphOffset::_exchange() { return parent_->_exchange(); }

//+ MonoGlyphOffset(XfFrescoObject::*parent_->)
/* XfFrescoObject */
XfTag MonoGlyphOffset::attach(XfFrescoObject_ptr observer) {
    return parent_->attach(observer);
}
void MonoGlyphOffset::detach(XfTag attach_tag) {
    parent_->detach(attach_tag);
}
void MonoGlyphOffset::disconnect() {
    parent_->disconnect();
}
void MonoGlyphOffset::notify_observers() {
    parent_->notify_observers();
}
void MonoGlyphOffset::update() {
    parent_->update();
}
//+

//+ MonoGlyphOffset(XfGlyphOffset::parent)
XfGlyph_ptr MonoGlyphOffset::parent() {
    return Glyph::_duplicate(parent_);
}

//+ MonoGlyphOffset(XfGlyphOffset::child)
XfGlyph_ptr MonoGlyphOffset::child() {
    return Glyph::_duplicate(child_);
}

//+ MonoGlyphOffset(XfGlyphOffset::next_child)
XfGlyphOffset_ptr MonoGlyphOffset::next_child() {
    return nil;
}

//+ MonoGlyphOffset(XfGlyphOffset::prev_child)
XfGlyphOffset_ptr MonoGlyphOffset::prev_child() {
    return nil;
}

//+ MonoGlyphOffset(XfGlyphOffset::allocations)
void MonoGlyphOffset::allocations(XfGlyph::AllocationInfoSeq& a) {
    ULong start = a.length();
    parent_->allocations(a);
    for (ULong i = start; i < a.length(); i++) {
	parent_->child_allocate(a[i]);
    }
}

//
// The aggregate operations (insert, replace, and remove) are nops for now.
//

//+ MonoGlyphOffset(XfGlyphOffset::insert)
void MonoGlyphOffset::insert(XfGlyph_ptr) { }

//+ MonoGlyphOffset(XfGlyphOffset::replace)
void MonoGlyphOffset::replace(XfGlyph_ptr) { }

//+ MonoGlyphOffset(XfGlyphOffset::remove)
void MonoGlyphOffset::remove() { }

//+ MonoGlyphOffset(XfGlyphOffset::notify)
void MonoGlyphOffset::notify() {
    parent_->need_resize();
}

//+ MonoGlyphOffset(XfGlyphOffset::traverse)
void MonoGlyphOffset::traverse(XfGlyphTraversal_ptr t) {
    child_->traverse(t);
}

//+ MonoGlyphOffset(XfGlyphOffset::visit_trail)
void MonoGlyphOffset::visit_trail(XfGlyphTraversal_ptr t) {
    parent_->visit_trail(t);
}

//+ MonoGlyphOffset(XfGlyphOffset::child_allocate)
void MonoGlyphOffset::child_allocate(XfGlyph::AllocationInfo& a) {
    parent_->child_allocate(a);
}

// class DebugGlyph

DebugGlyph::DebugGlyph(Glyph_ptr g, const char* msg, DebugGlyph::Flags f) {
    body(g);
    msg_ = msg;
    flags_ = f;
}

DebugGlyph::~DebugGlyph() { }

//+ DebugGlyph(XfGlyph::request)
void DebugGlyph::request(XfGlyph::Requisition& r) {
    MonoGlyph::request(r);
    if ((flags_ & DebugGlyph::trace_request) != 0) {
	heading("request ");
	print_requirement(r.x);
	printf(", ");
	print_requirement(r.y);
	printf(", ");
	print_requirement(r.z);
	printf("\n");
	fflush(stdout);
    }
}

//+ DebugGlyph(XfGlyph::traverse)
void DebugGlyph::traverse(XfGlyphTraversal_ptr t) {
    if ((flags_ & DebugGlyph::trace_traverse) != 0) {
	const char* op;
	DebugGlyph::Flags f;
	switch (t->op()) {
	case GlyphTraversal::draw:
	    f = DebugGlyph::trace_draw;
	    op = "draw";
	    break;
	case GlyphTraversal::pick_top:
	case GlyphTraversal::pick_any:
	case GlyphTraversal::pick_all:
	    f = DebugGlyph::trace_pick;
	    op = "pick";
	    break;
	case GlyphTraversal::other:
	    f = DebugGlyph::trace_other;
	    op = "other";
	    break;
	}
	if ((flags_ & f) != 0) {
	    heading(op);
	    if ((f & DebugGlyph::trace_draw_pick) != 0) {
		print_region(Region_var(t->allocation()));
	    }
	    printf("\n");
	    fflush(stdout);
	}
    }
    MonoGlyph::traverse(t);
}

void DebugGlyph::heading(const char* s) {
    Glyph_var g = body();
    printf("%s(0x%x) %s ", msg_, Glyph_ptr(g), s);
}

void DebugGlyph::print_requirement(Glyph::Requirement& r) {
    if (!r.defined) {
	printf("undef");
    } else {
	float tol = float(1e-2);
	if (Math::equal(r.natural, r.minimum, tol)) {
	    if (Math::equal(r.natural, r.maximum, tol)) {
		printf("%.2f", r.natural);
	    } else {
		printf("(%.2f,%.2f)", r.natural, r.maximum);
	    }
	} else if (Math::equal(r.natural, r.maximum, tol)) {
	    printf("(%.2f,%.2f)", r.minimum, r.natural);
	} else {
	    printf("(%.2f,%.2f,%.2f)", r.minimum, r.natural, r.maximum);
	}
	if (!Math::equal(r.align, float(0), tol)) {
	    printf(" @ %.1f", r.align);
	}
    }
}

void DebugGlyph::print_region(Region_ptr r) {
    Region::BoundingSpan s;
    printf("X(");
    r->span(X_axis, s);
    print_span(s);
    printf("), Y(");
    r->span(Y_axis, s);
    print_span(s);
    printf("), Z(");
    r->span(Z_axis, s);
    print_span(s);
    printf(")");
    fflush(stdout);
}

void DebugGlyph::print_span(const Region::BoundingSpan& s) {
    printf("%.2f,%.2f", s.begin, s.end);
    if (!Math::equal(s.align, float(0), float(1e-2))) {
	printf(" @ %.1f", s.align);
    }
}

// class Allocator

Allocator::Allocator() {
    requested_ = false;
}

Allocator::~Allocator() { }

//+ Allocator(XfGlyph::request)
void Allocator::request(XfGlyph::Requisition& r) {
    update_requisition();
    r = req_;
}

//+ Allocator(XfGlyph::allocations)
void Allocator::allocations(XfGlyph::AllocationInfoSeq& a) {
    update_requisition();
    MonoGlyph::allocations(a);
}

void Allocator::child_allocate(Glyph::AllocationInfo& i) {
    update_requisition();
    if (is_nil(i.allocation)) {
	i.allocation = new RegionImpl;
    }
    i.allocation->copy(&nat_);
}

//+ Allocator(XfGlyph::traverse)
void Allocator::traverse(XfGlyphTraversal_ptr t) {
    update_requisition();
    t->traverse_child(&offset_, &nat_);
}

//+ Allocator(XfGlyph::need_resize)
void Allocator::need_resize() {
    Glyph::AllocationInfoSeq alist;
    allocations(alist);
    RegionImpl region;
    region.copy(&ext_);
    requested_ = false;
    update_requisition();
    region.merge_union(&ext_);
    need_damage(&region, alist);
    MonoGlyph::need_resize();
}

void Allocator::update_requisition() {
    if (!requested_) {
	Glyph::Requisition r;
	r.x.defined = false;
	r.y.defined = false;
	r.z.defined = false;
	MonoGlyph::request(r);
	req_ = r;
	if (r.x.defined) {
	    nat_.xalign_ = r.x.align;
	    nat_.lower_.x = -r.x.align * r.x.natural;
	    nat_.upper_.x = nat_.lower_.x + r.x.natural;
	    nat_.defined_ = true;
	}
	if (r.y.defined) {
	    nat_.yalign_ = r.y.align;
	    nat_.lower_.y = -r.y.align * r.y.natural;
	    nat_.upper_.y = nat_.lower_.y + r.y.natural;
	    nat_.defined_ = true;
	}
	if (r.z.defined) {
	    nat_.lower_.z = -r.z.align * r.z.natural;
	    nat_.upper_.z = nat_.lower_.z + r.z.natural;
	    nat_.zalign_ = r.z.align;
	    nat_.defined_ = true;
	}
	requested_ = true;
	ext_.defined_ = false;
	Glyph::AllocationInfo a;
	MonoGlyph::extension(a, &ext_);
    }
}

void Allocator::need_damage(
    RegionImpl* ext, const Glyph::AllocationInfoSeq& a
) {
    RegionImpl r;
    float dot = float(1.0);
    for (ULong i = 0; i < a.length(); i++) {
	const Glyph::AllocationInfo& info = a[i];
	if (is_not_nil(info.damaged) && is_nil(info.allocation)) {
	    r.copy(ext);
	    if (is_not_nil(info.transformation)) {
		r.apply_transform(info.transformation);
	    }
	    info.damaged->extend(&r);
	}
    }
}

// class TransformAllocator

TransformAllocator::TransformAllocator(
    Alignment x_parent, Alignment y_parent, Alignment z_parent,
    Alignment x_child, Alignment y_child, Alignment z_child
) {
    x_parent_ = x_parent;
    y_parent_ = y_parent;
    z_parent_ = z_parent;
    x_child_ = x_child;
    y_child_ = y_child;
    z_child_ = z_child;
}

TransformAllocator::~TransformAllocator() { }

//+ TransformAllocator(XfGlyph::request)
void TransformAllocator::request(XfGlyph::Requisition& r) {
    Allocator::request(r);
    Coord fil = Coord(1000000.0);
    Coord zero = Coord(0.0);
    r.x.maximum = fil;
    r.x.minimum = zero;
    r.y.maximum = fil;
    r.y.minimum = zero;
    r.z.maximum = fil;
    r.z.minimum = zero;
    req_.x.maximum = fil;
    req_.x.minimum = zero;
    req_.y.maximum = fil;
    req_.y.minimum = zero;
    req_.z.maximum = fil;
    req_.z.minimum = zero;
}

void TransformAllocator::child_allocate(Glyph::AllocationInfo& i) {
    Vertex lower, upper, delta;
    TransformImpl tx;

    Allocator::child_allocate(i);
    if (is_nil(i.transformation)) {
	i.transformation = new TransformImpl;
    }
    i.allocation->bounds(lower, upper);
    compute_delta(lower, upper, delta);
    tx.translate(delta);
    i.transformation->premultiply(&tx);
    i.allocation->copy(&nat_);
}

//+ TransformAllocator(XfGlyph::traverse)
void TransformAllocator::traverse(XfGlyphTraversal_ptr t) {
    TransformImpl tx;
    update_requisition();
    Painter_var p = t->current_painter();
    Vertex lower, upper, v;
    t->bounds(lower, upper, v);
    compute_delta(lower, upper, v);
    tx.translate(v);
    p->push_matrix();
    p->premultiply(&tx);
    t->traverse_child(&offset_, &nat_);
    p->pop_matrix();
}

void TransformAllocator::compute_delta(
    const Vertex& lower, const Vertex& upper, Vertex& delta
) {
    delta.x = (
	lower.x - nat_.lower_.x +
	x_parent_ * (upper.x - lower.x) -
	x_child_ * (nat_.upper_.x - nat_.lower_.x)
    );
    delta.y = (
	lower.y - nat_.lower_.y +
	y_parent_ * (upper.y - lower.y) -
	y_child_ * (nat_.upper_.y - nat_.lower_.y)
    );
    delta.z = (
	lower.z - nat_.lower_.z +
	z_parent_ * (upper.z - lower.z) -
	z_child_ * (nat_.upper_.z - nat_.lower_.z)
    );
}

// class PolyGlyph

implementPtrList(PolyGlyphOffsetList,PolyGlyphOffset)

PolyGlyph::PolyGlyph() { }

PolyGlyph::~PolyGlyph() {
    for (ListItr(PolyGlyphOffsetList) g(children_); g.more(); g.next()) {
	CORBA::release(g.cur());
    }
}

//+ PolyGlyph(XfGlyph::append)
void PolyGlyph::append(XfGlyph_ptr g) {
    PolyGlyphOffset* offset = new_offset(children_.count(), g);
    children_.append(offset);
    modified();
}

//+ PolyGlyph(XfGlyph::prepend)
void PolyGlyph::prepend(XfGlyph_ptr g) {
    fixup(0, +1);
    PolyGlyphOffset* offset = new_offset(0, g);
    children_.prepend(offset);
    modified();
}

//+ PolyGlyph(XfGlyph::first_child_offset)
XfGlyphOffset_ptr PolyGlyph::first_child_offset() {
    Long n = children_.count();
    return n > 0 ? GlyphOffset::_duplicate(children_.item(0)) : nil;
}

//+ PolyGlyph(XfGlyph::last_child_offset)
XfGlyphOffset_ptr PolyGlyph::last_child_offset() {
    Long n = children_.count();
    return n > 0 ? GlyphOffset::_duplicate(children_.item(n - 1)) : nil;
}

PolyGlyphOffset* PolyGlyph::new_offset(Long index, Glyph_ptr child) {
    return new PolyGlyphOffset(this, index, child);
}

Glyph::Requisition* PolyGlyph::children_requests(
    Glyph::Requisition* req, Long n
) {
    Long count = children_.count();
    Glyph::Requisition* child_reqs = (
	count <= n ? req : new Glyph::Requisition[count]
    );
    Glyph::Requisition* r = child_reqs;
    for (ListItr(PolyGlyphOffsetList) i(children_); i.more(); i.next()) {
	Glyph_ptr g = i.cur()->child_;
	GlyphImpl::init_requisition(*r);
	if (is_not_nil(g)) {
	    g->request(*r);
	}
	++r;
    }
    return child_reqs;
}

void PolyGlyph::visit_trail(Long, GlyphTraversal_ptr) { }

void PolyGlyph::child_allocate(Long, Glyph::AllocationInfo&) { }

void PolyGlyph::fixup(Long start, Long delta) {
    Long n = children_.count();
    for (Long i = start; i < n; i++) {
	PolyGlyphOffset* p = children_.item(i);
	p->index_ += delta;
    }
}

void PolyGlyph::change(Long) { modified(); }

//
// Default behavior when modified is to notify any observers
// of this parent.
//

void PolyGlyph::modified() {
    notify_observers();
}

// class PolyGlyphOffset

Ox_implementation_body(PolyGlyphOffset)

PolyGlyphOffset::PolyGlyphOffset(PolyGlyph* p, Long index, Glyph_ptr child) {
    parent_ = p;
    index_ = index;
    child_ = Glyph::_duplicate(child);
    remove_tag_ = child->add_parent(this);
}

PolyGlyphOffset::~PolyGlyphOffset() {
    CORBA::release(child_);
}

//+ PolyGlyphOffset(XfFrescoObject::=exch_.)
XfTag PolyGlyphOffset::attach(XfFrescoObject_ptr observer) {
    return exch_.attach(observer);
}
void PolyGlyphOffset::detach(XfTag attach_tag) {
    exch_.detach(attach_tag);
}
void PolyGlyphOffset::disconnect() {
    exch_.disconnect();
}
void PolyGlyphOffset::notify_observers() {
    exch_.notify_observers();
}
void PolyGlyphOffset::update() {
    exch_.update();
}
//+

//+ PolyGlyphOffset(XfGlyphOffset::parent)
XfGlyph_ptr PolyGlyphOffset::parent() {
    return Glyph::_duplicate(parent_);
}

//+ PolyGlyphOffset(XfGlyphOffset::child)
XfGlyph_ptr PolyGlyphOffset::child() {
    return Glyph::_duplicate(child_);
}

GlyphOffset_ptr PolyGlyphOffset::offset(Long index) {
    GlyphOffset_ptr g = nil;
    Long n = parent_->children_.count();
    if (index >= 0 && index < n) {
	g = GlyphOffset::_duplicate(parent_->children_.item(index));
    }
    return g;
}

//+ PolyGlyphOffset(XfGlyphOffset::next_child)
XfGlyphOffset_ptr PolyGlyphOffset::next_child() {
    return offset(index_ + 1);
}

//+ PolyGlyphOffset(XfGlyphOffset::prev_child)
XfGlyphOffset_ptr PolyGlyphOffset::prev_child() {
    return offset(index_ - 1);
}

//+ PolyGlyphOffset(XfGlyphOffset::allocations)
void PolyGlyphOffset::allocations(XfGlyph::AllocationInfoSeq& a) {
    ULong start = a.length();
    parent_->allocations(a);
    for (ULong i = start; i < a.length(); i++) {
	parent_->child_allocate(index_, a[i]);
    }
}

//+ PolyGlyphOffset(XfGlyphOffset::insert)
void PolyGlyphOffset::insert(XfGlyph_ptr g) {
    PolyGlyphOffset* p = parent_->new_offset(index_, g);
    parent_->children_.insert(index_, p);
    parent_->fixup(index_ + 1, +1);
    parent_->modified();
    parent_->need_resize();
}

//+ PolyGlyphOffset(XfGlyphOffset::replace)
void PolyGlyphOffset::replace(XfGlyph_ptr g) {
    if (index_ < parent_->children_.count()) {
	child_->remove_parent(remove_tag_);
	CORBA::release(child_);
	child_ = Glyph::_duplicate(g);
	remove_tag_ = child_->add_parent(this);
	notify();
    }
}

//+ PolyGlyphOffset(XfGlyphOffset::remove)
void PolyGlyphOffset::remove() {
    PolyGlyphOffsetList& list = parent_->children_;
    if (index_ < list.count()) {
	list.remove(index_);
	parent_->fixup(index_, -1);
	parent_->modified();
	child_->remove_parent(remove_tag_);
	CORBA::release(this);
    }
}

//+ PolyGlyphOffset(XfGlyphOffset::notify)
void PolyGlyphOffset::notify() {
    parent_->change(index_);
    parent_->need_resize();
}

//+ PolyGlyphOffset(XfGlyphOffset::traverse)
void PolyGlyphOffset::traverse(XfGlyphTraversal_ptr t) {
    child_->traverse(t);
}

//+ PolyGlyphOffset(XfGlyphOffset::visit_trail)
void PolyGlyphOffset::visit_trail(XfGlyphTraversal_ptr t) {
    parent_->visit_trail(index_, t);
}

//+ PolyGlyphOffset(XfGlyphOffset::child_allocate)
void PolyGlyphOffset::child_allocate(XfGlyph::AllocationInfo& a) {
    parent_->child_allocate(index_, a);
}

// class GlyphTraversalImpl

declareList(GTStack,GlyphTraversalImpl::Info)
implementList(GTStack,GlyphTraversalImpl::Info)

class GTFilter : public EventFilter {
public:
    GTFilter(GlyphTraversal_ptr t);
    ~GTFilter();

    OxExchange* _exchange();

    //+ XfEventFilter::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfEventFilter */
    Boolean handle(XfEvent_ptr e);
    //+
protected:
    SharedFrescoObjectImpl object_;
    GlyphTraversal_ptr traversal_;
    TransformImpl inverse_;
};

GTFilter::GTFilter(GlyphTraversal_ptr t) {
    inverse_.copy(Transform_var(t->current_transform()));
    inverse_.invert();
    traversal_ = GlyphTraversal::_duplicate(t);
}

GTFilter::~GTFilter() {
    CORBA::release(traversal_);
}

OxExchange* GTFilter::_exchange() { return &object_; }

//+ GTFilter(XfFrescoObject::=object_.)
XfTag GTFilter::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void GTFilter::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void GTFilter::disconnect() {
    object_.disconnect();
}
void GTFilter::notify_observers() {
    object_.notify_observers();
}
void GTFilter::update() {
    object_.update();
}
//+

//+ GTFilter(XfEventFilter::handle)
Boolean GTFilter::handle(XfEvent_ptr e) {
    Boolean b = false;
    GlyphTraversal_ptr t = traversal_;
    Viewer_var v = t->current_viewer();
    if (is_not_nil(v)) {
	Painter_var p = t->current_painter();
	p->push_clipping();
	Vertex ev;
	ev.x = e->pointer_x();
	ev.y = e->pointer_y();
	ev.z = Coord(0);
	inverse_.transform_vertex(ev);
	p->clip_rect(ev.x, ev.y, ev.x, ev.y);
	b = v->handle(t, e);
	p->pop_clipping();
    }
    return b;
}

//
// We assume the creator of the traversal will take care of removing
// the painter, window, and damage in conjunction with releasing
// the traversal object.  Therefore, we do not hold or release a reference
// to these objects.
//

GlyphTraversalImpl::GlyphTraversalImpl(
    GlyphTraversal::Operation op, Window_ptr w, Damage_ptr damage
) {
    op_ = op;
    stack_ = new GTStack(20);
    index_ = 0;
    painter_ = nil;
    window_ = w;
    damage_ = damage;
    picked_ = nil;
    hit_info_ = 0;
}

GlyphTraversalImpl::GlyphTraversalImpl(const GlyphTraversalImpl& t) {
    stack_ = new GTStack(20);
    index_ = 0;
    op_ = t.op_;
    for (ListItr(GTStack) sp(*t.stack_); sp.more(); sp.next()) {
	GlyphTraversalImpl::Info& info = sp.cur_ref();
	GlyphTraversalImpl::Info i;
	i.viewer = Viewer::_duplicate(info.viewer);
	i.glyph = Glyph::_duplicate(info.glyph);
	i.offset = GlyphOffset::_duplicate(info.offset);
	i.allocation = Region::_duplicate(info.allocation);
	i.transformation = Transform::_duplicate(info.transformation);
	stack_->append(i);
    }
    index_ = t.index_;
    painter_ = t.painter_;
    window_ = t.window_;
    damage_ = t.damage_;
    picked_ = t.picked_;
    hit_info_ = t.hit_info_;
}

GlyphTraversalImpl::~GlyphTraversalImpl() {
    for (ListItr(GTStack) sp(*stack_); sp.more(); sp.next()) {
	GlyphTraversalImpl::Info& i = sp.cur_ref();
	CORBA::release(i.viewer);
	CORBA::release(i.glyph);
	CORBA::release(i.offset);
	CORBA::release(i.allocation);
	CORBA::release(i.transformation);
    }
    delete stack_;
    CORBA::release(picked_);
}

OxExchange* GlyphTraversalImpl::_exchange() { return &object_; }

//+ GlyphTraversalImpl(XfFrescoObject::=object_.)
XfTag GlyphTraversalImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void GlyphTraversalImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void GlyphTraversalImpl::disconnect() {
    object_.disconnect();
}
void GlyphTraversalImpl::notify_observers() {
    object_.notify_observers();
}
void GlyphTraversalImpl::update() {
    object_.update();
}
//+

//+ GlyphTraversalImpl(XfGlyphTraversal::op)
XfGlyphTraversal::Operation GlyphTraversalImpl::op() {
    return op_;
}

//+ GlyphTraversalImpl(XfGlyphTraversal::swap_op)
XfGlyphTraversal::Operation GlyphTraversalImpl::swap_op(XfGlyphTraversal::Operation op) {
    GlyphTraversal::Operation old_op = op_;
    op_ = op;
    return old_op;
}

//+ GlyphTraversalImpl(XfGlyphTraversal::is_pick_op)
Boolean GlyphTraversalImpl::is_pick_op() {
    return (
	op_ == GlyphTraversal::pick_top ||
	op_ == GlyphTraversal::pick_any ||
	op_ == GlyphTraversal::pick_all
    );
}

//
// Begin a new list of glyphs for the given viewer.
// The viewer will appear as the last entry of the previous
// viewer's list (since the parent will call traverse_child
// on the viewer before the viewer calls begin_train), so
// we copy that entry's allocation at the beginning of the glyph list.
// This approach allows us "inherit" the current allocation
// without looking past the current viewer information.
// We copy nil instead of the entry's glyph and offset fields
// to enable an object to distinguish picking a viewer from
// picking a glyph within the viewer.  If the pick glyph is nil,
// then the pick viewer is the hit object.
//

//+ GlyphTraversalImpl(XfGlyphTraversal::begin_viewer)
void GlyphTraversalImpl::begin_viewer(XfViewer_ptr v) {
    GlyphTraversalImpl::Info* i = top();
    push(v, nil, nil, i->allocation, nil);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::end_viewer)
void GlyphTraversalImpl::end_viewer() {
    pop();
}

//+ GlyphTraversalImpl(XfGlyphTraversal::event_filter)
XfEventFilter_ptr GlyphTraversalImpl::event_filter() {
    return new GTFilter(this);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::traverse_child)
void GlyphTraversalImpl::traverse_child(XfGlyphOffset_ptr o, XfRegion_ptr allocation) {
    push(nil, nil, o, allocation, nil);
    o->traverse(this);
    pop();
}

//+ GlyphTraversalImpl(XfGlyphTraversal::visit)
void GlyphTraversalImpl::visit() { }

//+ GlyphTraversalImpl(XfGlyphTraversal::trail)
XfGlyphTraversal_ptr GlyphTraversalImpl::trail() {
    return new GlyphTraversalImpl(*this);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::current_glyph)
XfGlyph_ptr GlyphTraversalImpl::current_glyph() {
    GlyphTraversalImpl::Info* i = cur();
    return (i == nil) ? nil : Glyph::_duplicate(i->glyph);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::current_viewer)
XfViewer_ptr GlyphTraversalImpl::current_viewer() {
    GlyphTraversalImpl::Info* i = cur();
    return (i == nil) ? nil : Viewer::_duplicate(i->viewer);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::current_offset)
XfGlyphOffset_ptr GlyphTraversalImpl::current_offset() {
    GlyphTraversalImpl::Info* i = cur();
    return (i == nil) ? nil : GlyphOffset::_duplicate(i->offset);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::forward)
Boolean GlyphTraversalImpl::forward() {
    if (index_ < stack_->count()) {
	++index_;
	return true;
    }
    return false;
}

//+ GlyphTraversalImpl(XfGlyphTraversal::backward)
Boolean GlyphTraversalImpl::backward() {
    if (index_ > 0) {
	--index_;
	return true;
    }
    return false;
}

//+ GlyphTraversalImpl(XfGlyphTraversal::current_painter=p)
void GlyphTraversalImpl::current_painter(XfPainter_ptr p) {
    CORBA::release(painter_);
    painter_ = Painter::_duplicate(p);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::current_display)
XfDisplay_ptr GlyphTraversalImpl::current_display() {
    return Screen_var(window_->screen_ptr())->display_ptr();
}

//+ GlyphTraversalImpl(XfGlyphTraversal::current_screen)
XfScreen_ptr GlyphTraversalImpl::current_screen() {
    return window_->screen_ptr();
}

//+ GlyphTraversalImpl(XfGlyphTraversal::current_window)
XfWindow_ptr GlyphTraversalImpl::current_window() {
    return Window::_duplicate(window_);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::current_painter?)
XfPainter_ptr GlyphTraversalImpl::current_painter() {
    return Painter::_duplicate(painter_);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::allocation)
XfRegion_ptr GlyphTraversalImpl::allocation() {
    GlyphTraversalImpl::Info* i = cur();
    return (i == nil) ? nil : Region::_duplicate(i->allocation);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::bounds)
Boolean GlyphTraversalImpl::bounds(XfVertex& lower, XfVertex& upper, XfVertex& origin) {
    Boolean b = false;
    GlyphTraversalImpl::Info* i = cur();
    if (i != nil) {
	Region_ptr r = i->allocation;
	if (is_not_nil(r)) {
	    r->bounds(lower, upper);
	    r->origin(origin);
	    b = true;
	}
    }
    return b;
}

//+ GlyphTraversalImpl(XfGlyphTraversal::allocation_is_visible)
Boolean GlyphTraversalImpl::allocation_is_visible() {
    Boolean b = false;
    GlyphTraversalImpl::Info* i = cur();
    if (i != nil) {
	Region_ptr r = i->allocation;
	if (is_not_nil(r)) {
	    b = painter_->is_visible(r);
	}
    }
    return b;
}

//+ GlyphTraversalImpl(XfGlyphTraversal::current_transform)
XfTransform_ptr GlyphTraversalImpl::current_transform() {
    GlyphTraversalImpl::Info* i = cur();
    return i == nil ? nil : Transform::_duplicate(i->transformation);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::damaged)
XfDamage_ptr GlyphTraversalImpl::damaged() {
    return Damage::_duplicate(damage_);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::hit)
void GlyphTraversalImpl::hit() {
    picked_ = new GlyphTraversalImpl(*this);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::hit_info=i)
void GlyphTraversalImpl::hit_info(Long i) { hit_info_ = i; }

//+ GlyphTraversalImpl(XfGlyphTraversal::hit_info?)
Long GlyphTraversalImpl::hit_info() { return hit_info_; }

//+ GlyphTraversalImpl(XfGlyphTraversal::picked)
XfGlyphTraversal_ptr GlyphTraversalImpl::picked() {
    return GlyphTraversal::_duplicate(picked_);
}

//+ GlyphTraversalImpl(XfGlyphTraversal::clear)
void GlyphTraversalImpl::clear() {
    CORBA::release(picked_);
    picked_ = nil;
}

//
// Add a new item at the end of the current trail.  Each trail item
// can contain a viewer, glyph, glyph offset, and allocation.  If the
// given glyph is nil, then we use the offset's child.  If the
// given allocation is nil, then use the allocation of the previous item.
// The latter case arises when a glyph wants to pass its body
// (or a child) the same allocation as it was passed.
//

void GlyphTraversalImpl::push(
    Viewer_ptr v, Glyph_ptr g, GlyphOffset_ptr o,
    Region_ptr a, Transform_ptr tx
) {
    GlyphTraversalImpl::Info i;
    i.viewer = Viewer::_duplicate(v);
    i.glyph = Glyph::_duplicate(g);
    i.offset = GlyphOffset::_duplicate(o);
    i.allocation = Region::_duplicate(a);
    i.transformation = Transform::_duplicate(tx);
    if (is_nil(g) && is_not_nil(o)) {
	i.glyph = o->child();
    }
    GlyphTraversalImpl::Info* t = top();
    if (t != nil && is_nil(v)) {
	i.viewer = Viewer::_duplicate(t->viewer);
    }
    stack_->append(i);
    ++index_;
}

void GlyphTraversalImpl::pop() {
    Long n = stack_->count() - 1;
    if (n >= 0) {
	GlyphTraversalImpl::Info& i = stack_->item_ref(n);
	CORBA::release(i.viewer);
	CORBA::release(i.glyph);
	CORBA::release(i.offset);
	CORBA::release(i.allocation);
	CORBA::release(i.transformation);
	stack_->remove(n);
	--index_;
    }
    // else raise exception
}

GlyphTraversalImpl::Info* GlyphTraversalImpl::top() {
    Long n = stack_->count() - 1;
    return n >= 0 ? &stack_->item_ref(n) : nil;
}

GlyphTraversalImpl::Info* GlyphTraversalImpl::cur() {
    Long n = index_ - 1;
    return n >= 0 ? &stack_->item_ref(n) : nil;
}

// class PickPainter

class ClippingStack;
class MatrixStack;

class PickPainter : public Painter {
public:
    PickPainter();
    ~PickPainter();

    OxExchange* _exchange();

    //+ XfPainter::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfPainter */
    XfCoord to_coord(XfPixelCoord p, XfAxis a);
    XfPixelCoord to_pixels(XfCoord c, XfAxis a);
    XfCoord to_pixels_coord(XfCoord c, XfAxis a);
    void begin_path();
    void move_to(XfCoord x, XfCoord y);
    void line_to(XfCoord x, XfCoord y);
    void curve_to(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2);
    void close_path();
    XfBrush_ptr current_brush();
    void current_brush(XfBrush_ptr _p);
    XfColor_ptr current_color();
    void current_color(XfColor_ptr _p);
    XfFont_ptr current_font();
    void current_font(XfFont_ptr _p);
    XfPattern_ptr current_pattern();
    void current_pattern(XfPattern_ptr _p);
    void stroke();
    void fill();
    void line(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1);
    void rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1);
    void fill_rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1);
    void character(XfCharCode ch, XfCoord width, XfCoord x, XfCoord y);
    void image(XfRaster_ptr r, XfCoord x, XfCoord y);
    void stencil(XfRaster_ptr r, XfCoord x, XfCoord y);
    XfTransform_ptr current_matrix();
    void current_matrix(XfTransform_ptr _p);
    void push_matrix();
    void pop_matrix();
    void premultiply(XfTransform_ptr t);
    void clip();
    void clip_rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1);
    void push_clipping();
    void pop_clipping();
    Boolean is_visible(XfRegion_ptr r);
    XfRegion_ptr visible();
    void comment(XfCharString_ptr s);
    void page_number(XfCharString_ptr s);
    //+

private:
    RegionImpl* new_clip();
    void free_clip(RegionImpl*);
    Long clip_index(Long);

    SharedFrescoObjectImpl object_;
    Brush_ptr brush_;
    Color_ptr color_;
    Font_ptr font_;
    Pattern_ptr pattern_;
    TransformImpl* matrix_;
    Tag matrix_tag_;
    MatrixStack* transforms_;
    RegionImpl* clipping_;
    ClippingStack* clippers_;
    RegionImpl* free_clippers_[10];
    Long free_clipper_head_;
    Long free_clipper_tail_;
};

//
// We only need this silly MatrixStackElement because the compiler
// (language) can't deal with functions that return arrays.
//

struct MatrixStackElement {
    Transform::Matrix matrix;
};

declareList(MatrixStack,MatrixStackElement)
implementList(MatrixStack,MatrixStackElement)

declarePtrList(ClippingStack,RegionImpl)
implementPtrList(ClippingStack,RegionImpl)

inline Long PickPainter::clip_index(Long i) {
    return i % (sizeof(free_clippers_) / sizeof(free_clippers_[0]));
}

PickPainter::PickPainter() {
    brush_ = nil;
    color_ = nil;
    font_ = nil;
    pattern_ = nil;
    matrix_ = new TransformImpl;
    matrix_tag_ = matrix_->attach(this);
    transforms_ = new MatrixStack;
    clipping_ = nil;
    clippers_ = new ClippingStack;
    free_clipper_head_ = 0;
    free_clipper_tail_ = 0;
}

PickPainter::~PickPainter() {
    CORBA::release(brush_);
    CORBA::release(color_);
    CORBA::release(font_);
    CORBA::release(pattern_);
    matrix_->detach(matrix_tag_);
    CORBA::release(matrix_);
    delete transforms_;
    CORBA::release(clipping_);
    for (ListItr(ClippingStack) j(*clippers_); j.more(); j.next()) {
	CORBA::release(j.cur());
    }
    delete clippers_;
    for (
	Long i = free_clipper_head_;
	i != free_clipper_tail_;
	i = clip_index(i + 1)
    ) {
	CORBA::release(free_clippers_[i]);
    }
}

OxExchange* PickPainter::_exchange() { return &object_; }

//+ PickPainter(XfFrescoObject::=object_.)
XfTag PickPainter::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void PickPainter::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void PickPainter::disconnect() {
    object_.disconnect();
}
void PickPainter::notify_observers() {
    object_.notify_observers();
}
void PickPainter::update() {
    object_.update();
}
//+

//+ PickPainter(XfPainter::to_coord) !a
XfCoord PickPainter::to_coord(XfPixelCoord p, XfAxis) { return Coord(p); }

//+ PickPainter(XfPainter::to_pixels) !a
XfPixelCoord PickPainter::to_pixels(XfCoord c, XfAxis) { return PixelCoord(c); }

//+ PickPainter(XfPainter::to_pixels_coord) !a
XfCoord PickPainter::to_pixels_coord(XfCoord c, XfAxis) { return c; }

//+ PickPainter(XfPainter::begin_path)
void PickPainter::begin_path() { }

//+ PickPainter(XfPainter::move_to)
void PickPainter::move_to(XfCoord, XfCoord) { }

//+ PickPainter(XfPainter::line_to)
void PickPainter::line_to(XfCoord, XfCoord) { }

//+ PickPainter(XfPainter::curve_to)
void PickPainter::curve_to(XfCoord, XfCoord, XfCoord, XfCoord, XfCoord, XfCoord) { }

//+ PickPainter(XfPainter::close_path)
void PickPainter::close_path() { }

//+ PickPainter(XfPainter::current_brush=b)
void PickPainter::current_brush(XfBrush_ptr b) {
    if (brush_ != b) {
	CORBA::release(brush_);
	brush_ = Brush::_duplicate(b);
    }
}

//+ PickPainter(XfPainter::current_brush?)
XfBrush_ptr PickPainter::current_brush() {
    return Brush::_duplicate(brush_);
}

//+ PickPainter(XfPainter::current_color=c)
void PickPainter::current_color(XfColor_ptr c) {
    if (color_ != c) {
	CORBA::release(color_);
	color_ = Color::_duplicate(c);
    }
}

//+ PickPainter(XfPainter::current_color?)
XfColor_ptr PickPainter::current_color() {
    return Color::_duplicate(color_);
}

//+ PickPainter(XfPainter::current_font=f)
void PickPainter::current_font(XfFont_ptr f) {
    if (font_ != f) {
	CORBA::release(font_);
	font_ = Font::_duplicate(f);
    }
}

//+ PickPainter(XfPainter::current_font?)
XfFont_ptr PickPainter::current_font() {
    return Font::_duplicate(font_);
}

//+ PickPainter(XfPainter::current_pattern=p)
void PickPainter::current_pattern(XfPattern_ptr p) {
    if (pattern_ != p) {
	CORBA::release(pattern_);
	pattern_ = Pattern::_duplicate(p);
    }
}

//+ PickPainter(XfPainter::current_pattern?)
XfPattern_ptr PickPainter::current_pattern() {
    return Pattern::_duplicate(pattern_);
}

//+ PickPainter(XfPainter::stroke)
void PickPainter::stroke() { }

//+ PickPainter(XfPainter::fill)
void PickPainter::fill() { }

//+ PickPainter(XfPainter::line)
void PickPainter::line(XfCoord, XfCoord, XfCoord, XfCoord) { }

//+ PickPainter(XfPainter::rect)
void PickPainter::rect(XfCoord, XfCoord, XfCoord, XfCoord) { }

//+ PickPainter(XfPainter::fill_rect)
void PickPainter::fill_rect(XfCoord, XfCoord, XfCoord, XfCoord) { }

//+ PickPainter(XfPainter::character)
void PickPainter::character(XfCharCode, XfCoord, XfCoord, XfCoord) { }

//+ PickPainter(XfPainter::image)
void PickPainter::image(XfRaster_ptr, XfCoord, XfCoord) { }

//+ PickPainter(XfPainter::stencil)
void PickPainter::stencil(XfRaster_ptr, XfCoord, XfCoord) { }

//+ PickPainter(XfPainter::current_matrix=t)
void PickPainter::current_matrix(XfTransform_ptr t) {
    matrix_->copy(t);
}

//+ PickPainter(XfPainter::current_matrix?)
XfTransform_ptr PickPainter::current_matrix() {
    return Transform::_duplicate(matrix_);
}

//+ PickPainter(XfPainter::push_matrix)
void PickPainter::push_matrix() {
    transforms_->prepend(*((MatrixStackElement*)(matrix_->matrix())));
}

//+ PickPainter(XfPainter::pop_matrix)
void PickPainter::pop_matrix() {
    MatrixStack* s = transforms_;
    if (s->count() == 0) {
	//
	// Underflow -- should raise an exception.
	//
	return;
    }
    matrix_->load_matrix(s->item_ref(0).matrix);
    s->remove(0);
}

//+ PickPainter(XfPainter::premultiply)
void PickPainter::premultiply(XfTransform_ptr t) {
    matrix_->premultiply(t);
}

//+ PickPainter(XfPainter::clip)
void PickPainter::clip() {
    // not implemented
}

//+ PickPainter(XfPainter::clip_rect)
void PickPainter::clip_rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1) {
    RegionImpl region;
    RegionImpl* r = (clipping_ == nil) ? new_clip() : &region;
    r->defined_ = true;
    r->lower_.x = x0; r->lower_.y = y0; r->lower_.z = Coord(0);
    r->upper_.x = x1; r->upper_.y = y1; r->upper_.z = Coord(0);
    if (!matrix_->identity()) {
	r->apply_transform(matrix_);
    }
    if (clipping_ == nil) {
	clipping_ = r;
    } else {
	clipping_->merge_intersect(r);
    }
}

//+ PickPainter(XfPainter::push_clipping)
void PickPainter::push_clipping() {
    clippers_->prepend(clipping_);
    if (is_not_nil(clipping_)) {
	RegionImpl* r = new_clip();
	r->defined_ = true;
	r->lower_ = clipping_->lower_;
	r->upper_ = clipping_->upper_;
	clipping_ = r;
    }
}

RegionImpl* PickPainter::new_clip() {
    RegionImpl* r;
    if (free_clipper_head_ != free_clipper_tail_) {
	r = free_clippers_[free_clipper_head_];
	free_clipper_head_ = clip_index(free_clipper_head_ + 1);
    } else {
	r = new RegionImpl;
    }
    return r;
}

//+ PickPainter(XfPainter::pop_clipping)
void PickPainter::pop_clipping() {
    ClippingStack* s = clippers_;
    if (s->count() == 0) {
	//
	// Underflow -- should raise an exception.
	//
	return;
    }
    free_clip(clipping_);
    clipping_ = s->item(0);
    s->remove(0);
}

void PickPainter::free_clip(RegionImpl* r) {
    Long new_tail = clip_index(free_clipper_tail_ + 1);
    if (new_tail != free_clipper_head_) {
	free_clippers_[free_clipper_tail_] = r;
	free_clipper_tail_ = new_tail;
    } else {
	CORBA::release(r);
    }
}

//
// Determine if the given region is (at least partially) visible
// given the current clipping region.
//

//+ PickPainter(XfPainter::is_visible)
Boolean PickPainter::is_visible(XfRegion_ptr r) {
    Boolean b = is_nil(clipping_);
    if (!b) {
	Region_ptr rr;
	RegionImpl bounds;
	if (matrix_->identity()) {
	    rr = r;
	} else {
	    rr = &bounds;
	    rr->copy(r);
	    rr->apply_transform(matrix_);
	}
	b = clipping_->intersects(rr);
    }
    return b;
}

//+ PickPainter(XfPainter::visible)
XfRegion_ptr PickPainter::visible() {
    return Region::_duplicate(clipping_);
}

//
// The comment and page_number operations are typically
// only relevent for a printer.
//

//+ PickPainter(XfPainter::comment)
void PickPainter::comment(XfCharString_ptr) { }

//+ PickPainter(XfPainter::page_number)
void PickPainter::page_number(XfCharString_ptr) { }

// class PickTraversal

PickTraversal::PickTraversal(
    GlyphTraversal::Operation op, Window_ptr w, Damage_ptr d
) : GlyphTraversalImpl(op, w, d) {
    painter_ = new PickPainter;
}

PickTraversal::PickTraversal(
    const PickTraversal& p
) : GlyphTraversalImpl(p) {
    painter_ = Painter::_duplicate(p.painter_);
}

PickTraversal::~PickTraversal() {
    CORBA::release(painter_);
}

//+ PickTraversal(XfGlyphTraversal::begin_viewer)
void PickTraversal::begin_viewer(XfViewer_ptr v) {
    truncate_stack();
    GlyphTraversalImpl::Info* i = top();
    TransformImpl* t = new TransformImpl;
    t->copy(Transform_var(painter_->current_matrix()));
    push(v, nil, nil, i->allocation, t);
    CORBA::release(t);
}

//+ PickTraversal(XfGlyphTraversal::traverse_child)
void PickTraversal::traverse_child(XfGlyphOffset_ptr o, XfRegion_ptr allocation) {
    truncate_stack();
    Transform_var t = new TransformImpl;
    t->copy(Transform_var(painter_->current_matrix()));
    push(nil, nil, o, allocation, t);
    Glyph_var(o->child())->traverse(this);
    pop();
}

//+ PickTraversal(XfGlyphTraversal::hit)
void PickTraversal::hit() {
    picked_ = new PickTraversal(*this);
}

//
// Restore the traversal stack to match the current position.
// This is necessary when starting a new pick traversal on
// an old traversal object.
//

void PickTraversal::truncate_stack() {
    Long save_index = index_;
    for (Long i = top() - cur(); i > 0; i--) {
	pop();
    }
    index_ = save_index;
}
