//
// $XConsortium$
//

//
// Copyright (c) 1992-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the name of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <Fresco/defs_impl/names.h>
#include <Fresco/Ox/stub.h>
#include <string.h>

// class NamingLib

Naming::NamingContext_ptr NamingLib::context_;

void NamingLib::current_context(Naming::NamingContext_ptr nc) {
    context_ = nc;
}

Naming::NamingContext_ptr NamingLib::current_context() {
    return context_;
}

Naming::Name* NamingLib::to_name(const char* str) {
    Naming::Name* n;
    if (str == nil) {
	n = nil;
    } else {
	ULong length = 1;
	for (const char* p = str; *p != '\0'; p++) {
	    if (*p == '\\') {
		++p;
		if (*p == '\0') {
		    break;
		}
	    } else if (*p == '/') {
		++length;
	    }
	}
	n = new Naming::Name(length, length);
	ULong i = 0;
	const char* q = str;
	for (p = str; *p != '\0'; p++) {
	    if (*p == '\\') {
		++p;
		if (*p == '\0') {
		    break;
		}
	    } else if (*p == '/') {
		set_component((*n)[i], q, p - q);
		q = p + 1;
		++i;
	    }
	}
	if (*q != '\0' || i == 0) {
	    set_component((*n)[i], q, p - q);
	}
    }
    return n;
}

char* NamingLib::to_string(const Naming::Name& n) {
    ULong len = 0;
    ULong count = n.length();
    for (ULong i = 0; i < count; i++) {
	const Naming::NameComponent& comp = n[i];
	len += strlen(comp.id) + 1;
	if (comp.kind != (char*)nil) {
	    len += strlen(comp.kind) + 1;
	}
    }
    char* s = new char[len];
    char* p = s;
    for (i = 0; i < count; i++) {
	const Naming::NameComponent& comp = n[i];
	strcpy(p, comp.id);
	p += strlen(comp.id);
	if (comp.kind != (char*)nil) {
	    strcpy(p, ".");
	    ++p;
	    strcpy(p, comp.kind);
	    p += strlen(comp.kind);
	}
	*p = '/';
	++p;
    }
    s[len - 1] = '\0';
    return s;
}

void NamingLib::set_component(
    Naming::NameComponent& comp, const char* str, unsigned int length
) {
    char* s = new char[length + 1];
    strncpy(s, str, length);
    s[length] = '\0';
    comp.id = s;
}

//+ Naming
Naming::NameComponent::NameComponent() {
    id._str = 0;
    kind._str = 0;
}
Naming::NameComponent::NameComponent(const NameComponent& __s) :
    id(__s.id),
    kind(__s.kind)
{
}
Naming::NameComponent::~NameComponent() {
    delete id._str;
    delete kind._str;
}
Naming::NameComponent& Naming::NameComponent::operator =(const NameComponent& __s) {
    id = __s.id;
    kind = __s.kind;
    return *this;
}
export void _Naming_NameComponent_put(_Ix_MarshalBuffer& _b, const Naming::NameComponent& _this) {
    _b.put_string(_this.id);
    _b.put_string(_this.kind);
}
export void _Naming_NameComponent_get(_Ix_MarshalBuffer& _b, Naming::NameComponent& _this) {
    _this.id = _b.get_string();
    _this.kind = _b.get_string();
}

_Ix_CopyNormalSeq(Naming::Name,Name,NameComponent)
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
export void _Naming_Name_put(_Ix_MarshalBuffer& _b, const Naming::Name& _this) {
    _Ix_BeginSequencePut(Naming::NameComponent,_this)
        _Naming_NameComponent_put(_b, _this._buffer[i]);
    _Ix_EndSequencePut()
}
export void _Naming_Name_get(_Ix_MarshalBuffer& _b, Naming::Name& _this) {
    _Ix_BeginSequenceGet(Naming::NameComponent,_this)
        _Naming_NameComponent_get(_b, _this._buffer[i]);
    _Ix_EndSequenceGet()
}

_Ix_CopyNormalSeq(Naming::NameList,NameList,Name)
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
extern import void _Naming_Name_put(
    _Ix_MarshalBuffer&, const Naming::Name&
);
extern import void _Naming_Name_get(
    _Ix_MarshalBuffer&, Naming::Name&
);
export void _Naming_NameList_put(_Ix_MarshalBuffer& _b, const Naming::NameList& _this) {
    _Ix_BeginSequencePut(Naming::Name,_this)
        _Naming_Name_put(_b, _this._buffer[i]);
    _Ix_EndSequencePut()
}
export void _Naming_NameList_get(_Ix_MarshalBuffer& _b, Naming::NameList& _this) {
    _Ix_BeginSequenceGet(Naming::Name,_this)
        _Naming_Name_get(_b, _this._buffer[i]);
    _Ix_EndSequenceGet()
}

_Ix_CopyInterfaceSeq(Naming::ObjectList,ObjectList,CORBA::Object)
_Ix_FreeInterfaceSeq(Naming::ObjectList,ObjectList,CORBA::Object)
_Ix_ExternProxy(CORBA_Object);
export void _Naming_ObjectList_put(_Ix_MarshalBuffer& _b, const Naming::ObjectList& _this) {
    _Ix_BeginSequencePut(CORBA::Object_ptr,_this)
        _b.put_object(_this._buffer[i]);
    _Ix_EndSequencePut()
}
export void _Naming_ObjectList_get(_Ix_MarshalBuffer& _b, Naming::ObjectList& _this) {
    _Ix_GetInterfaceSeq(CORBA::Object,CORBA_Object,_this)
}

Naming::Binding::Binding() {
}
Naming::Binding::Binding(const Binding& __s) :
    binding_name(__s.binding_name),
    binding_type(__s.binding_type)
{
}
Naming::Binding::~Binding() {
}
Naming::Binding& Naming::Binding::operator =(const Binding& __s) {
    binding_name = __s.binding_name;
    binding_type = __s.binding_type;
    return *this;
}
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
export void _Naming_Binding_put(_Ix_MarshalBuffer& _b, const Naming::Binding& _this) {
    _Naming_NameComponent_put(_b, _this.binding_name);
    _b.put_long(_this.binding_type);
}
export void _Naming_Binding_get(_Ix_MarshalBuffer& _b, Naming::Binding& _this) {
    _Naming_NameComponent_get(_b, _this.binding_name);
    _this.binding_type = Naming::BindingType(_b.get_long());
}

_Ix_CopyNormalSeq(Naming::BindingList,BindingList,Binding)
extern import void _Naming_Binding_put(
    _Ix_MarshalBuffer&, const Naming::Binding&
);
extern import void _Naming_Binding_get(
    _Ix_MarshalBuffer&, Naming::Binding&
);
export void _Naming_BindingList_put(_Ix_MarshalBuffer& _b, const Naming::BindingList& _this) {
    _Ix_BeginSequencePut(Naming::Binding,_this)
        _Naming_Binding_put(_b, _this._buffer[i]);
    _Ix_EndSequencePut()
}
export void _Naming_BindingList_get(_Ix_MarshalBuffer& _b, Naming::BindingList& _this) {
    _Ix_BeginSequenceGet(Naming::Binding,_this)
        _Naming_Binding_get(_b, _this._buffer[i]);
    _Ix_EndSequenceGet()
}

_Ix_extern__tid(Naming);
static const char* _Naming_NotFound_name = "Naming::NotFound";
Naming::NotFound::NotFound() {
    _name_ = _Naming_NotFound_name;
    _major_ = _index;
    _hash_ = _code;
    _interface_ = _Ix__tid(Naming);
}
Naming::NotFound::NotFound(Naming::NotFoundReason _why, Naming::Name _rest_of_name) : 
    why(_why), rest_of_name(_rest_of_name)
{
    _name_ = _Naming_NotFound_name;
    _major_ = _index;
    _hash_ = _code;
    _interface_ = _Ix__tid(Naming);
}
Naming::NotFound::~NotFound() {
}
Naming::NotFound::NotFound(const NotFound& __s) :
    why(__s.why),
    rest_of_name(__s.rest_of_name)
{
}
Naming::NotFound& Naming::NotFound::operator =(const NotFound& __s) {
    why = __s.why;
    rest_of_name = __s.rest_of_name;
    return *this;
}
void* Naming::NotFound::_check_cast(const _Ix_Exception* e) {
    return _try_cast(e, _index, _Ix__tid(Naming));
}
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
extern import void _Naming_Name_put(
    _Ix_MarshalBuffer&, const Naming::Name&
);
extern import void _Naming_Name_get(
    _Ix_MarshalBuffer&, Naming::Name&
);
extern import OxSchema::EnumDescriptor _Ix__type(Naming_NotFoundReason);
extern import OxSchema::SequenceDescriptor _Ix__type(Naming_Name);
_Ix_BeginMemberList(Naming_NotFound)
    &_Ix__type(Naming_NotFoundReason),
    &_Ix__type(Naming_Name)
_Ix_EndMemberList()
_Ix_ExceptDescriptorDII(Naming_NotFound,"Naming::NotFound",Naming::NotFound,2);
void Naming::NotFound::_put(_Ix_MarshalBuffer& _b) const {
    _Ix_Exception_put(_b);
    const Naming::NotFound& _this = *this;
    _b.put_long(_this.why);
    _Naming_Name_put(_b, _this.rest_of_name);
}
_Ix_Exception* Naming::NotFound::_get(_Ix_MarshalBuffer& _b) {
    Naming::NotFound* _e = new NotFound;
    if (_e != 0) {
        Naming::NotFound& _this = *_e;
        _this.why = Naming::NotFoundReason(_b.get_long());
        _Naming_Name_get(_b, _this.rest_of_name);
    }
    return _e;
}

_Ix_extern__tid(Naming);
static const char* _Naming_CannotProceed_name = "Naming::CannotProceed";
Naming::CannotProceed::CannotProceed() {
    _name_ = _Naming_CannotProceed_name;
    _major_ = _index;
    _hash_ = _code;
    _interface_ = _Ix__tid(Naming);
    cxt.ptr_ = 0;
}
Naming::CannotProceed::CannotProceed(_Ix_T_field(NamingContext)  _cxt, Naming::Name _rest_of_name) : 
    cxt(_cxt), rest_of_name(_rest_of_name)
{
    _name_ = _Naming_CannotProceed_name;
    _major_ = _index;
    _hash_ = _code;
    _interface_ = _Ix__tid(Naming);
}
Naming::CannotProceed::~CannotProceed() {
    _Ix_release(cxt.ptr_);
}
Naming::CannotProceed::CannotProceed(const CannotProceed& __s) :
    cxt(__s.cxt),
    rest_of_name(__s.rest_of_name)
{
}
Naming::CannotProceed& Naming::CannotProceed::operator =(const CannotProceed& __s) {
    cxt = __s.cxt;
    rest_of_name = __s.rest_of_name;
    return *this;
}
void* Naming::CannotProceed::_check_cast(const _Ix_Exception* e) {
    return _try_cast(e, _index, _Ix__tid(Naming));
}
_Ix_ExternProxy(Naming_NamingContext);
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
extern import void _Naming_Name_put(
    _Ix_MarshalBuffer&, const Naming::Name&
);
extern import void _Naming_Name_get(
    _Ix_MarshalBuffer&, Naming::Name&
);
extern import OxSchema::InterfaceDescriptor _Ix__type(Naming_NamingContext);
_Ix_BeginMemberList(Naming_CannotProceed)
    &_Ix__type(Naming_NamingContext),
    &_Ix__type(Naming_Name)
_Ix_EndMemberList()
_Ix_ExceptDescriptorDII(Naming_CannotProceed,"Naming::CannotProceed",Naming::CannotProceed,2);
void Naming::CannotProceed::_put(_Ix_MarshalBuffer& _b) const {
    _Ix_Exception_put(_b);
    const Naming::CannotProceed& _this = *this;
    _b.put_object(_this.cxt);
    _Naming_Name_put(_b, _this.rest_of_name);
}
_Ix_Exception* Naming::CannotProceed::_get(_Ix_MarshalBuffer& _b) {
    Naming::CannotProceed* _e = new CannotProceed;
    if (_e != 0) {
        Naming::CannotProceed& _this = *_e;
        _this.cxt = (Naming::NamingContext_ptr)_b.get_object(&_Ix_CreateProxy(Naming_NamingContext));
        _Naming_Name_get(_b, _this.rest_of_name);
    }
    return _e;
}

_Ix_extern__tid(Naming);
static const char* _Naming_InvalidName_name = "Naming::InvalidName";
Naming::InvalidName::InvalidName() {
    _name_ = _Naming_InvalidName_name;
    _major_ = _index;
    _hash_ = _code;
    _interface_ = _Ix__tid(Naming);
}
Naming::InvalidName::~InvalidName() { }
void* Naming::InvalidName::_check_cast(const _Ix_Exception* e) {
    return _try_cast(e, _index, _Ix__tid(Naming));
}
_Ix_EmptyExceptDescriptorDII(Naming_InvalidName,"Naming::InvalidName",Naming::InvalidName);
void Naming::InvalidName::_put(_Ix_MarshalBuffer& _b) const {
    _Ix_Exception_put(_b);
}
_Ix_Exception* Naming::InvalidName::_get(_Ix_MarshalBuffer&) { return new InvalidName; }

_Ix_extern__tid(Naming);
static const char* _Naming_AlreadyBound_name = "Naming::AlreadyBound";
Naming::AlreadyBound::AlreadyBound() {
    _name_ = _Naming_AlreadyBound_name;
    _major_ = _index;
    _hash_ = _code;
    _interface_ = _Ix__tid(Naming);
}
Naming::AlreadyBound::~AlreadyBound() { }
void* Naming::AlreadyBound::_check_cast(const _Ix_Exception* e) {
    return _try_cast(e, _index, _Ix__tid(Naming));
}
_Ix_EmptyExceptDescriptorDII(Naming_AlreadyBound,"Naming::AlreadyBound",Naming::AlreadyBound);
void Naming::AlreadyBound::_put(_Ix_MarshalBuffer& _b) const {
    _Ix_Exception_put(_b);
}
_Ix_Exception* Naming::AlreadyBound::_get(_Ix_MarshalBuffer&) { return new AlreadyBound; }

_Ix_extern__tid(Naming);
static const char* _Naming_NotEmpty_name = "Naming::NotEmpty";
Naming::NotEmpty::NotEmpty() {
    _name_ = _Naming_NotEmpty_name;
    _major_ = _index;
    _hash_ = _code;
    _interface_ = _Ix__tid(Naming);
}
Naming::NotEmpty::~NotEmpty() { }
void* Naming::NotEmpty::_check_cast(const _Ix_Exception* e) {
    return _try_cast(e, _index, _Ix__tid(Naming));
}
_Ix_EmptyExceptDescriptorDII(Naming_NotEmpty,"Naming::NotEmpty",Naming::NotEmpty);
void Naming::NotEmpty::_put(_Ix_MarshalBuffer& _b) const {
    _Ix_Exception_put(_b);
}
_Ix_Exception* Naming::NotEmpty::_get(_Ix_MarshalBuffer&) { return new NotEmpty; }

_Ix_BeginExceptionTable(Naming)
    _Ix_ExceptionEntry(Naming_NotFound),
    _Ix_ExceptionEntry(Naming_CannotProceed),
    _Ix_ExceptionEntry(Naming_InvalidName),
    _Ix_ExceptionEntry(Naming_AlreadyBound),
    _Ix_ExceptionEntry(Naming_NotEmpty)
_Ix_EndExceptionTable()
//+

//+ Naming::NamingContext::%init,type-funcs,client,server
Naming::NamingContext::NamingContext() { }
Naming::NamingContext::~NamingContext() { }
void* Naming::NamingContext::_this_ptr() { return this; }
_Ix_extern__tid(Naming_NamingContext);
void* Naming::NamingContext::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(Naming_NamingContext) == t) return this;
    if (CORBA::Object::_type_id() == t) return (CORBA::Object*) this;
    return nil;
};
_Ix_ImplementField(Naming::NamingContext,NamingContext,_Ix_NullParameter)

_Ix_extern__tid(Naming_NamingContext);
_Ix_ExternProxy(Naming_NamingContext);
_Ix_ImplementGetTid(Naming::NamingContext,_Ix__tid(Naming_NamingContext))
_Ix_ImplementNarrow(Naming::NamingContext_ptr,Naming::NamingContext,_Ix_NullParameter,Naming_NamingContext,&_Ix_CreateProxy(Naming_NamingContext))
_Ix_ImplementDuplicate(Naming::NamingContext_ptr,Naming::NamingContext,_Ix_NullParameter,_Ix__tid(Naming_NamingContext),&_Ix_CreateProxy(Naming_NamingContext))
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
extern import void _Naming_Name_put(
    _Ix_MarshalBuffer&, const Naming::Name&
);
extern import void _Naming_Name_get(
    _Ix_MarshalBuffer&, Naming::Name&
);
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
extern import void _Naming_Name_put(
    _Ix_MarshalBuffer&, const Naming::Name&
);
extern import void _Naming_Name_get(
    _Ix_MarshalBuffer&, Naming::Name&
);
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
extern import void _Naming_Name_put(
    _Ix_MarshalBuffer&, const Naming::Name&
);
extern import void _Naming_Name_get(
    _Ix_MarshalBuffer&, Naming::Name&
);
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
extern import void _Naming_Name_put(
    _Ix_MarshalBuffer&, const Naming::Name&
);
extern import void _Naming_Name_get(
    _Ix_MarshalBuffer&, Naming::Name&
);
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
extern import void _Naming_Name_put(
    _Ix_MarshalBuffer&, const Naming::Name&
);
extern import void _Naming_Name_get(
    _Ix_MarshalBuffer&, Naming::Name&
);
extern import void _Naming_NameComponent_put(
    _Ix_MarshalBuffer&, const Naming::NameComponent&
);
extern import void _Naming_NameComponent_get(
    _Ix_MarshalBuffer&, Naming::NameComponent&
);
extern import void _Naming_Name_put(
    _Ix_MarshalBuffer&, const Naming::Name&
);
extern import void _Naming_Name_get(
    _Ix_MarshalBuffer&, Naming::Name&
);
extern import void _Naming_Binding_put(
    _Ix_MarshalBuffer&, const Naming::Binding&
);
extern import void _Naming_Binding_get(
    _Ix_MarshalBuffer&, Naming::Binding&
);
extern import void _Naming_BindingList_put(
    _Ix_MarshalBuffer&, const Naming::BindingList&
);
extern import void _Naming_BindingList_get(
    _Ix_MarshalBuffer&, Naming::BindingList&
);
extern import void _Naming_Binding_put(
    _Ix_MarshalBuffer&, const Naming::Binding&
);
extern import void _Naming_Binding_get(
    _Ix_MarshalBuffer&, Naming::Binding&
);
extern import void _Naming_BindingList_put(
    _Ix_MarshalBuffer&, const Naming::BindingList&
);
extern import void _Naming_BindingList_get(
    _Ix_MarshalBuffer&, Naming::BindingList&
);

_Ix_ImplementProxy(Naming::NamingContext,NamingContext,Naming_NamingContext)
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_bind_pdesc[4] = { 3, 4, 1, 61 };
_Ix_MarshalBuffer::ArgMarshal _Naming_NamingContext_bind_pfunc[] = {
    &_Naming_Name_put, &_Naming_Name_get,
    &_Ix_CreateProxy(CORBA_Object),

};
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_bind_pinfo = {
    _Naming_NamingContext_tid, 0, _Naming_NamingContext_bind_pdesc, _Naming_NamingContext_bind_pfunc
};
void Naming::NamingContext::bind(const Naming::Name& n, CORBA::Object_ptr obj) {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[3];
    _arg[1].u_addr = &n;
    _arg[2].u_objref = obj;
    _b.invoke(this, _Naming_NamingContext_bind_pinfo, _arg);
}
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_bind_context_pdesc[4] = { 3, 4, 1, 61 };
_Ix_MarshalBuffer::ArgMarshal _Naming_NamingContext_bind_context_pfunc[] = {
    &_Naming_Name_put, &_Naming_Name_get,
    &_Ix_CreateProxy(Naming_NamingContext),

};
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_bind_context_pinfo = {
    _Naming_NamingContext_tid, 1, _Naming_NamingContext_bind_context_pdesc, _Naming_NamingContext_bind_context_pfunc
};
void Naming::NamingContext::bind_context(const Naming::Name& n, Naming::NamingContext_ptr nc) {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[3];
    _arg[1].u_addr = &n;
    _arg[2].u_objref = nc;
    _b.invoke(this, _Naming_NamingContext_bind_context_pinfo, _arg);
}
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_rebind_pdesc[4] = { 3, 4, 1, 61 };
_Ix_MarshalBuffer::ArgMarshal _Naming_NamingContext_rebind_pfunc[] = {
    &_Naming_Name_put, &_Naming_Name_get,
    &_Ix_CreateProxy(CORBA_Object),

};
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_rebind_pinfo = {
    _Naming_NamingContext_tid, 2, _Naming_NamingContext_rebind_pdesc, _Naming_NamingContext_rebind_pfunc
};
void Naming::NamingContext::rebind(const Naming::Name& n, CORBA::Object_ptr obj) {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[3];
    _arg[1].u_addr = &n;
    _arg[2].u_objref = obj;
    _b.invoke(this, _Naming_NamingContext_rebind_pinfo, _arg);
}
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_unbind_pdesc[3] = { 2, 4, 1 };
_Ix_MarshalBuffer::ArgMarshal _Naming_NamingContext_unbind_pfunc[] = {
    &_Naming_Name_put, &_Naming_Name_get,

};
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_unbind_pinfo = {
    _Naming_NamingContext_tid, 3, _Naming_NamingContext_unbind_pdesc, _Naming_NamingContext_unbind_pfunc
};
void Naming::NamingContext::unbind(const Naming::Name& n) {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[2];
    _arg[1].u_addr = &n;
    _b.invoke(this, _Naming_NamingContext_unbind_pinfo, _arg);
}
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_resolve_pdesc[3] = { 2, 60, 1 };
_Ix_MarshalBuffer::ArgMarshal _Naming_NamingContext_resolve_pfunc[] = {
    &_Naming_Name_put, &_Naming_Name_get,
    &_Ix_CreateProxy(CORBA_Object)
};
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_resolve_pinfo = {
    _Naming_NamingContext_tid, 4, _Naming_NamingContext_resolve_pdesc, _Naming_NamingContext_resolve_pfunc
};
CORBA::Object_ptr Naming::NamingContext::resolve(const Naming::Name& n) {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[2];
    _arg[1].u_addr = &n;
    _b.invoke(this, _Naming_NamingContext_resolve_pinfo, _arg);
    return (CORBA::Object_ptr)_arg[0].u_vaddr;
}
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_new_context_pdesc[2] = { 1, 60 };
_Ix_MarshalBuffer::ArgMarshal _Naming_NamingContext_new_context_pfunc[] = {
    &_Ix_CreateProxy(Naming_NamingContext)
};
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_new_context_pinfo = {
    _Naming_NamingContext_tid, 5, _Naming_NamingContext_new_context_pdesc, _Naming_NamingContext_new_context_pfunc
};
Naming::NamingContext_ptr Naming::NamingContext::new_context() {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[1];
    _b.invoke(this, _Naming_NamingContext_new_context_pinfo, _arg);
    return (Naming::NamingContext_ptr)_arg[0].u_vaddr;
}
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_bind_new_context_pdesc[3] = { 2, 60, 1 };
_Ix_MarshalBuffer::ArgMarshal _Naming_NamingContext_bind_new_context_pfunc[] = {
    &_Naming_Name_put, &_Naming_Name_get,
    &_Ix_CreateProxy(Naming_NamingContext)
};
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_bind_new_context_pinfo = {
    _Naming_NamingContext_tid, 6, _Naming_NamingContext_bind_new_context_pdesc, _Naming_NamingContext_bind_new_context_pfunc
};
Naming::NamingContext_ptr Naming::NamingContext::bind_new_context(const Naming::Name& n) {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[2];
    _arg[1].u_addr = &n;
    _b.invoke(this, _Naming_NamingContext_bind_new_context_pinfo, _arg);
    return (Naming::NamingContext_ptr)_arg[0].u_vaddr;
}
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_list_names_pdesc[4] = { 3, 0, 37, 37 };
_Ix_MarshalBuffer::ArgMarshal _Naming_NamingContext_list_names_pfunc[] = {
    &_Naming_BindingList_put, &_Naming_BindingList_get
};
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_list_names_pinfo = {
    _Naming_NamingContext_tid, 7, _Naming_NamingContext_list_names_pdesc, _Naming_NamingContext_list_names_pfunc
};
Naming::BindingList* Naming::NamingContext::list_names(ULong start, ULong how_many) {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[3];
    Naming::BindingList* _result;
    _arg[0].u_addr = &_result;
    _arg[1].u_unsigned_long = start;
    _arg[2].u_unsigned_long = how_many;
    _b.invoke(this, _Naming_NamingContext_list_names_pinfo, _arg);
    return _result;
}
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_list_all_pdesc[2] = { 1, 0 };
_Ix_MarshalBuffer::ArgMarshal _Naming_NamingContext_list_all_pfunc[] = {
    &_Naming_BindingList_put, &_Naming_BindingList_get
};
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_list_all_pinfo = {
    _Naming_NamingContext_tid, 8, _Naming_NamingContext_list_all_pdesc, _Naming_NamingContext_list_all_pfunc
};
Naming::BindingList* Naming::NamingContext::list_all() {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[1];
    Naming::BindingList* _result;
    _arg[0].u_addr = &_result;
    _b.invoke(this, _Naming_NamingContext_list_all_pinfo, _arg);
    return _result;
}
_Ix_MarshalBuffer::ArgDesc _Naming_NamingContext_destroy_pdesc[2] = { 1, 4 };
_Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_destroy_pinfo = {
    _Naming_NamingContext_tid, 9, _Naming_NamingContext_destroy_pdesc, 0
};
void Naming::NamingContext::destroy() {
    _Ix_MarshalBuffer _b;
    _Ix_MarshalBuffer::ArgValue _arg[1];
    _b.invoke(this, _Naming_NamingContext_destroy_pinfo, _arg);
}
_Ix_ReceiveHdr(Naming_NamingContext,_b_receive,_b_reply) {
    _Ix_CastReceiver(Naming::NamingContext_ptr,Naming_NamingContext);
    _Ix_MarshalBuffer::ArgValue _arg[3];
    _Ix_MarshalBuffer::ArgInfo* _arginfo;
    switch (_m) {
        case /* bind */ 0: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_bind_pinfo;
            _arginfo = &_Naming_NamingContext_bind_pinfo;
            Naming::Name n;
            _arg[1].u_addr = &n;
            CORBA::Object_ptr obj;
            _arg[2].u_addr = &obj;
            if (_b_receive.receive(_Naming_NamingContext_bind_pinfo, _arg)) {
                obj = CORBA::Object::_narrow((CORBA::Object_ptr)(void*)obj);
                _this->bind(n, obj);
            }
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        case /* bind_context */ 1: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_bind_context_pinfo;
            _arginfo = &_Naming_NamingContext_bind_context_pinfo;
            Naming::Name n;
            _arg[1].u_addr = &n;
            Naming::NamingContext_ptr nc;
            _arg[2].u_addr = &nc;
            if (_b_receive.receive(_Naming_NamingContext_bind_context_pinfo, _arg)) {
                nc = Naming::NamingContext::_narrow((CORBA::Object_ptr)(void*)nc);
                _this->bind_context(n, nc);
            }
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        case /* rebind */ 2: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_rebind_pinfo;
            _arginfo = &_Naming_NamingContext_rebind_pinfo;
            Naming::Name n;
            _arg[1].u_addr = &n;
            CORBA::Object_ptr obj;
            _arg[2].u_addr = &obj;
            if (_b_receive.receive(_Naming_NamingContext_rebind_pinfo, _arg)) {
                obj = CORBA::Object::_narrow((CORBA::Object_ptr)(void*)obj);
                _this->rebind(n, obj);
            }
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        case /* unbind */ 3: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_unbind_pinfo;
            _arginfo = &_Naming_NamingContext_unbind_pinfo;
            Naming::Name n;
            _arg[1].u_addr = &n;
            if (_b_receive.receive(_Naming_NamingContext_unbind_pinfo, _arg)) {
                _this->unbind(n);
            }
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        case /* resolve */ 4: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_resolve_pinfo;
            _arginfo = &_Naming_NamingContext_resolve_pinfo;
            Naming::Name n;
            _arg[1].u_addr = &n;
            if (_b_receive.receive(_Naming_NamingContext_resolve_pinfo, _arg)) {
                _arg[0].u_objref = _this->resolve(n);
            }
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        case /* new_context */ 5: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_new_context_pinfo;
            _arginfo = &_Naming_NamingContext_new_context_pinfo;
            _arg[0].u_objref = _this->new_context();
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        case /* bind_new_context */ 6: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_bind_new_context_pinfo;
            _arginfo = &_Naming_NamingContext_bind_new_context_pinfo;
            Naming::Name n;
            _arg[1].u_addr = &n;
            if (_b_receive.receive(_Naming_NamingContext_bind_new_context_pinfo, _arg)) {
                _arg[0].u_objref = _this->bind_new_context(n);
            }
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        case /* list_names */ 7: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_list_names_pinfo;
            _arginfo = &_Naming_NamingContext_list_names_pinfo;
            Naming::BindingList* _result;
            ULong start;
            _arg[1].u_addr = &start;
            ULong how_many;
            _arg[2].u_addr = &how_many;
            if (_b_receive.receive(_Naming_NamingContext_list_names_pinfo, _arg)) {
                _result = _this->list_names(start, how_many);
                _arg[0].u_addr = &_result;
            }
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        case /* list_all */ 8: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_list_all_pinfo;
            _arginfo = &_Naming_NamingContext_list_all_pinfo;
            Naming::BindingList* _result;
            _result = _this->list_all();
            _arg[0].u_addr = &_result;
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        case /* destroy */ 9: {
            extern _Ix_MarshalBuffer::ArgInfo _Naming_NamingContext_destroy_pinfo;
            _arginfo = &_Naming_NamingContext_destroy_pinfo;
            _this->destroy();
            _b_reply.reply(*_arginfo, _arg);
            break;
        }
        default:
            _Ix_BadMethod(_m);
    }
}
//+
