/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_defs_impls_painters_h
#define Fresco_defs_impls_painters_h

#include <Fresco/defs/drawing.h>

#include <Fresco/defs_impl/fobjects.h>
	  
// class PathImpl

class export PathImpl {
public:
    PathImpl();
    ~PathImpl();

    void begin_path();
    void move_to(XfCoord x, XfCoord y);
    void line_to(XfCoord x, XfCoord y);
    void curve_to(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2);
    void close_path();

    enum Type {MoveTo, LineTo, CurveTo};
    struct Point {
	Type type;
        XfCoord x;
	XfCoord y;
	XfPixelCoord px;
	XfPixelCoord py;
    };

    void transform(Transform::Matrix m);

    int count();
    PathImpl::Point& point(int i);
    Boolean is_closed();

private:
    void expand();

    Point* buffer_;
    int size_;
    int count_;
    Boolean is_closed_;
};

class PainterMatrixStack;
class TransformImpl;

class export PainterBase : public Painter {
public:
    PainterBase();
    ~PainterBase();

    OxExchange* _exchange();

    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfPainter */
    XfCoord to_coord(XfPixelCoord p, XfAxis a) = 0;
    XfPixelCoord to_pixels(XfCoord c, XfAxis a) = 0;
    XfCoord to_pixels_coord(XfCoord c, XfAxis a) = 0;
    void begin_path();
    void move_to(XfCoord x, XfCoord y);
    void line_to(XfCoord x, XfCoord y);
    void curve_to(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2);
    void close_path();
    XfBrush_ptr current_brush();
    void current_brush(XfBrush_ptr _p);
    XfColor_ptr current_color();
    void current_color(XfColor_ptr _p);
    XfFont_ptr current_font();
    void current_font(XfFont_ptr _p);
    XfPattern_ptr current_pattern();
    void current_pattern(XfPattern_ptr _p);
    void stroke() = 0;
    void fill() = 0;
    void line(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1);
    void rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1);
    void fill_rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1);
    void character(XfCharCode ch, XfCoord width, XfCoord x, XfCoord y) = 0;
    void image(XfRaster_ptr r, XfCoord x, XfCoord y) = 0;
    void stencil(XfRaster_ptr r, XfCoord x, XfCoord y) = 0;
    XfTransform_ptr current_matrix();
    void current_matrix(XfTransform_ptr _p);
    void push_matrix();
    void pop_matrix();
    void premultiply(XfTransform_ptr t);
    void clip() = 0;
    void clip_rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1);
    void push_clipping() = 0;
    void pop_clipping() = 0;
    Boolean is_visible(XfRegion_ptr r) = 0;
    XfRegion_ptr visible() = 0;
    void comment(XfCharString_ptr s);
    void page_number(XfCharString_ptr s);

protected:
    void begin_paint();
    void end_paint();

    SharedFrescoObjectImpl object_;

    TransformImpl* matrix_;
    PainterMatrixStack* transforms_;

    PathImpl path_;

    Brush_ptr brush_;
    Color_ptr color_;
    Font_ptr font_;
    Pattern_ptr pattern_;
};

#endif
