#if 0

//
// Copyright (c) 1995 Firefly Design
// Copyright (c) 1995 Silicon Graphics, Inc.
// Copyright (c) 1995 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <Fresco/defs_impl/persist.h>

// class FactoryFinderImpl

FactoryFinderImpl::FactoryFinderImpl() {
}

FactoryFinderImpl::~FactoryFinderImpl() {
}

OxExchange* FactoryFinderImpl::_exchange() { return &object_; }

//+ FactoryFinderImpl(Oes::FactoryFinder::remember)
void FactoryFinderImpl::remember(const ExtKey external_form_id, Oes::Factory_ptr f) {
    factories_.insert(external_form_id, f);
}

//+ FactoryFinderImpl(Oes::FactoryFinder::remember)
void FactoryFinderImpl::remember(const ExtKey external_form_id, Oes::Factory_ptr f) {
    Factory_ptr f;
    if (! factories_.lookup(external_form_id, f)) {
        f = nil;
    }
    return f;
}

// class GenericStreamFactoryImpl

GenericStreamFactoryImpl::GenericStreamFactoryImpl() {
}

GenericStreamFactoryImpl::~GenericStreamFactoryImpl() {
}

OxExchange* GenericStreamFactoryImpl::_exchange() { return &object_; }

GenericStream_ptr GenericStreamFactoryImpl::create() {
    return GenericStreamImpl;
}

// class GenericStreamImpl

GenericStreamImpl::GenericStreamImpl() {
    context_ = false;
}

GenericStreamImpl::~GenericStreamImpl() {
}

OxExchange* GenericStreamImpl::_exchange() { return &object_; }

ULong GenericStreamImpl::remember_streamable(Streamable_ptr o) {
    ULong index;
    if (! remember_.lookup(o, index)) {
        index = remember_.count() + 1;
        remember_.insert(o, index);
    }
    return index;
}

void GenericStreamImpl::stream_io(StreamIO_ptr s) {
    stream_io_ = s;
}

StreamIO_ptr GenericStreamImpl::stream_io() {
    return StreamIO::_duplicate(stream_io_);
}

void GenericStreamImpl::externalize(in CosStream::Streamable s) {
    stream_io_->write_object(s);
    if (! context_) {
        remember_.clear();
    }
}

CosStream::Streamable GenericStreamImpl::internalize(in FactoryFinder f) {
    CosStream::Streamable_ptr streamable = stream_io_->read_object(f);
    if (! context_) {
        remember_.clear();
    }
    return streamable;
}

void GenericStreamImpl::begin_context() {
    if (context_) {
        throw ContextAlreadyRegistered;
    }
    context_ = true;
}

void GenericStreamImpl::end_context() {
    assert(context_);
    context_ = false;
    remember_.clear();
}

void GenericStreamImpl::flush() {
    // !!! StreamIO should have a flush...
}

// class StandardStreamIO

StandardStreamIO::StandardStreamIO(GenericStream_ptr stream) {
    stream_ = stream;
}

StandardStreamIO::~StandardStreamIO() {
}

OxExchange* StandardStreamIO::_exchange() { return &object_; }

void StandardStreamIO::throw_error() {
    stream_->error();
    throw streamDataFormatError;
}

void StandardStreamIO::put_tag(unsigned char tag) {
    put_byte(tag);
}

void StandardStreamIO::put_string(String value) {
    // null-terminated sequence of bytes
    for (char* cp = value; *cp != '\0'; ++cp) {
        put_byte(*cp);
    }
    put_byte('\0');
}

void StandardStreamIO::put_char(Char value) {
    // one byte
    put_byte(value);
}

void StandardStreamIO::put_octet(Octet value) {
    // one byte
    put_byte(value);
}

void StandardStreamIO::put_unsigned_long(ULong value) {
    // four bytes, big-endian format
    put_byte((value >>  0) & 0xff);
    put_byte((value >>  8) & 0xff);
    put_byte((value >> 16) & 0xff);
    put_byte((value >> 24) & 0xff);
}

void StandardStreamIO::put_unsigned_short(UShort value) {
    // two bytes, big-endian format
    put_byte((value >> 0) & 0xff);
    put_byte((value >> 8) & 0xff);
}

void StandardStreamIO::put_long(ULong value) {
    // four bytes, big-endian format
    put_byte((value >>  0) & 0xff);
    put_byte((value >>  8) & 0xff);
    put_byte((value >> 16) & 0xff);
    put_byte((value >> 24) & 0xff);
}

void StandardStreamIO::put_short(UShort value) {
    // two bytes, big-endian format
    put_byte((value >> 0) & 0xff);
    put_byte((value >> 8) & 0xff);
}

void StandardStreamIO::put_float(Float value) {
    // four bytes, IEEE 754 single precision format, sign bit in first byte
    // !!! need to implement platform independent solution
    FloatWormhole wormhole;
    wormhole.value = value;
    put_byte(wormhole.b[0]);
    put_byte(wormhole.b[1]);
    put_byte(wormhole.b[2]);
    put_byte(wormhole.b[3]);
}

void StandardStreamIO::put_double(Double value) {
    // eight bytes, IEEE 754 double precision format, sign bit in first byte
    // !!! need to implement platform independent solution
    DoubleWormhole wormhole;
    wormhole.value = value;
    put_byte(wormhole.b[0]);
    put_byte(wormhole.b[1]);
    put_byte(wormhole.b[2]);
    put_byte(wormhole.b[3]);
    put_byte(wormhole.b[4]);
    put_byte(wormhole.b[5]);
    put_byte(wormhole.b[6]);
    put_byte(wormhole.b[7]);
}

void StandardStreamIO::put_boolean(Boolean value) {
    // TRUE => one byte==1, FALSE => one byte==0
    put_byte(value ? 1 : 0);
}

unsigned char StandardStreamIO::get_tag() {
    return get_byte();
}

void StandardStreamIO::check_tag(unsigned char expected) {
    unsigned char tag = get_tag();
    if (tag != expected) {
        throw_error();
    }
}

String StandardStreamIO::get_string() {
    // null-terminated sequence of bytes
    while (iostream_->good()) {
        unsigned char c = get_byte();
        string_append(c);
        if (! iostream_->fail()) {
            if (c == '\0') {
                return string_done();
            }
        }
    }
    throw_error();
    return nil;
}

Char StandardStreamIO::get_char() {
    // one byte
    return get_byte();
}

Octet StandardStreamIO::get_octet() {
    // one byte
    return get_byte();
}

ULong StandardStreamIO::get_unsigned_long() {
    // four bytes, big-endian format
    unsigned char b0 = get_byte();
    unsigned char b1 = get_byte();
    unsigned char b2 = get_byte();
    unsigned char b3 = get_byte();
    return (b3 << 24) | (b2 << 16) | (b1 << 8) | (b0 << 0);
}

UShort StandardStreamIO::get_unsigned_short() {
    // two bytes, big-endian format
    unsigned char b0 = get_byte();
    unsigned char b1 = get_byte();
    return (b1 << 8) | (b0 << 0);
}

Long StandardStreamIO::get_long() {
    // four bytes, big-endian format
    unsigned char b0 = get_byte();
    unsigned char b1 = get_byte();
    unsigned char b2 = get_byte();
    unsigned char b3 = get_byte();
    // !!! need to sign extend if sizeof(Long) > 4
    return (b3 << 24) | (b2 << 16) | (b1 << 8) | (b0 << 0);
}

Short StandardStreamIO::get_short() {
    // two bytes, big-endian format
    unsigned char b0 = get_byte();
    unsigned char b1 = get_byte();
    // !!! need to sign extend if sizeof(Short) > 2
    return (b1 << 8) | (b0 << 0);
}

Float StandardStreamIO::get_float() {
    // four bytes, IEEE 754 single precision format, sign bit in first byte
    // !!! need to implement platform independent solution
    FloatWormhole wormhole;
    wormhole.b[0] = get_byte();
    wormhole.b[1] = get_byte();
    wormhole.b[2] = get_byte();
    wormhole.b[3] = get_byte();
    return wormhole.value;
}

Double StandardStreamIO::get_double() {
    // eight bytes, IEEE 754 double precision format, sign bit in first byte
    // !!! need to implement platform independent solution
    DoubleWormhole wormhole;
    wormhole.b[0] = get_byte();
    wormhole.b[1] = get_byte();
    wormhole.b[2] = get_byte();
    wormhole.b[3] = get_byte();
    wormhole.b[4] = get_byte();
    wormhole.b[5] = get_byte();
    wormhole.b[6] = get_byte();
    wormhole.b[7] = get_byte();
    return wormhole.value;
}

Boolean StandardStreamIO::get_boolean() {
    // TRUE => one byte==1, FALSE => one byte==0
    unsigned char b = get_byte();
    if ((b != 0) && (b != 1)) {
        throw_error();
    }
    return b;
}

void StandardStreamIO::write_string(String value) {  
    put_tag(0xFA);
    put_string(value);
}

void StandardStreamIO::write_char(Char value) {
    put_tag(0xF1);
    put_char(value);
}

void StandardStreamIO::write_octet(Octet value) {
    put_tag(0xF2);
    put_octet(value);
}

void StandardStreamIO::write_unsigned_long(ULong value) {
    put_tag(0xF3);
    put_unsigned_long(value);
}

void StandardStreamIO::write_unsigned_short(UShort value) {
    put_tag(0xF4);
    put_unsigned_short(value);
}

void StandardStreamIO::write_long(Long value) {
    put_tag(0xF5);
    put_long(value);
}

void StandardStreamIO::write_short(Short value) {
    put_tag(0xF6);
    put_short(value);
}

void StandardStreamIO::write_float(Float value) {
    put_tag(0xF7);
    put_float(value);
}

void StandardStreamIO::write_double(Double value) {
    put_tag(0xF8);
    put_double(value);
}

void StandardStreamIO::write_boolean(Boolean value) {
    put_tag(0xF9);
    put_boolean(value);
}

void StandardStreamIO::write_object(Streamable_ptr value) {
    if (is_nil(value)) {
        put_tag(0x05);
    } else {
        ULong object_number = stream_->remember_streamable(value);
        if (object_number == 0) {
            put_tag(0xF0);
            Key external_form_id = value->external_form_id();
            ULong length = external_form_id->length();
            assert(length < 256); // !!! should there be an exception for this case?
            put_octet((Octet) length);
            for (ULong i = 0; i < length; ++i) {
                String name = external_form_id[i];
                put_string(name.id());
            }
            value->externalize_to_stream(this);
        } else {
            put_tag(0x04);
            put_unsigned_long(object_number);
        }
    }
}

String StandardStreamIO::read_string() {
    check_tag(0xFA);
    return get_string();
}
    
Char StandardStreamIO::read_char() {
    check_tag(0xF1);
    return get_char();
}

Octet StandardStreamIO::read_octet() {
    check_tag(0xF2);
    return get_octet();
}

ULong StandardStreamIO::read_unsigned_long() {
    check_tag(0xF3);
    return get_unsigned_long();
}

UShort StandardStreamIO::read_unsigned_short() {
    check_tag(0xF4);
    return get_unsigned_short();
}

Long StandardStreamIO::read_long() {
    check_tag(tag != 0xF5);
    return get_long();
}

Short StandardStreamIO::read_short() {
    check_tag(0xF6);
    return get_short();
}

Float StandardStreamIO::read_float() {
    check_tag(0xF7);
    return get_float();
}

Double StandardStreamIO::read_double() {
    check_tag(0xF8);
    return get_double();
}

Boolean StandardStreamIO::read_boolean() {
    check_tag(0xF9);
    return get_boolean();
}

Streamable StandardStreamIO::read_object(FactoryFinder_ptr there, Streamable_ptr) {
    unsigned char tag = get_tag();
    if (tag == 0x05) {
        return nil;
    } else if (tag == 0x04) {
        ULong object_number = get_unsigned_long();
        Streamable_ptr value = stream_->get_streamable(object_number);
        if (is_nil(value)) {
            throw_error();
        }
        return value;
    } else if (tag == 0xF0) {
        Key external_form_id;
        Octet length = get_octet();
        for (ULong i = 0; i < length; ++i) {
            Naming::NameComponent name;
            name.id = get_string();
            name.kind = nil;
            external_form_id.append(name);
        }
        Factory_var factory = there->find(external_form_id);
        if (is_nil(factory)) {
            throw_error();
        }
        CORBA::Object_var object = factory->create();
        if (is_nil(object)) {
            throw_error();
        }
        Streamable_var value = Streamable::_narrow(object);
        if (is_nil(value)) {
            throw_error();
        }
        ULong object_number = stream_->remember_streamable(value);
        assert(object_number == 0);
        value->internalize_from_stream(this, finder);
        return value;
    }
    throw_error();
    return nil;
}

// class OleStreamIO

OleStreamIO::OleStreamIO(GenericStream_ptr stream, IStream* iostream) : StandardStreamIO(stream) {
    iostream_ = iostream;
}

OleStreamIO::~OleStreamIO() {
}

void OleStreamIO::put_byte(unsigned char byte) {
    ULONG written;
    HRESULT result = iostream_->Write(&byte, 1, *written);
    if ((result != S_OK) || (written != 1)) {
        throw_error();
    }
}

unsigned char OleStreamIO::get_byte() {
    unsigned char byte;
    HRESULT result = iostream_->Read(&byte, 1, read);
    if ((result != S_OK) || (read != 1)) {
        throw_error();
    }
    return byte;
}

// class CxxStreamIO

CxxStreamIO::CxxStreamIO(GenericStream_ptr stream, iostream* iostream) : StandardStreamIO(stream) {
    iostream_ = iostream;
}

CxxStreamIO::~CxxStreamIO() {
}

void CxxStreamIO::put_byte(unsigned char byte) {
    iostream_->put(byte);
    if (iostream_->fail()) {
        throw_error();
    }
}

unsigned char CxxStreamIO::get_byte() {
    unsigned char byte;
    iostream_->get(byte);
    if (iostream_->fail()) {
        throw_error();
    }
    return byte;
}

#endif
