#if 0

//
// Copyright (c) 1995 Firefly Design
// Copyright (c) 1995 Silicon Graphics, Inc.
// Copyright (c) 1995 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_persist_h
#define Fresco_persist_h

#include <Fresco/defs/oes.h>

#include <Fresco/defs_impl/fobjects.h>

class export FactoryFinderImpl : public Oes::FactoryFinder {
public:
    FactoryFinderImpl();
    ~FactoryFinderImpl();
    
    OxExchange* _exchange();

    //+ Oes::FactoryFinder::*
    /* FactoryFinder */
    void remember(const ExtKey external_form_id, Oes::Factory_ptr f);
    Oes::Factory_ptr find(const ExtKey external_form_id);
    //+

private:
    SharedFrescoObjectImpl object_;
    Table<Oes::Key_ptr,Oes::Factory_ptr> factories_;
};

class export GenericStreamFactoryImpl : public GenericStreamFactory {
public:
    GenericStreamFactoryImpl();
    ~GenericStreamFactoryImpl();

    OxExchange* _exchange();

    //+ Oes::GenericStreamFactory::*
    /* GenericStreamFactory */
    Oes::GenericStream_ptr create();
    //+
};

class export GenericStreamImpl : public GenericStream {
public:
    GenericStreamImpl();
    ~GenericStreamImpl();

    OxExchange* _exchange();

    // Return the streamables object number if it has already been encountered,
    // otherwise assign an object number and return 0.
    // This function is used during writing and reading.
    ULong remember_streamable(Streamable_ptr o);
    
    //+ Oes::GenericStream::*
    /* Stream */
    void externalize(Oes::Streamable_ptr theObject);
    Oes::Streamable_ptr internalize(Oes::FactoryFinder_ptr there);
    void begin_context();
    void end_context();
    void flush();
    /* GenericStream */
    Oes::StreamIO_ptr stream_io();
    void stream_io(Oes::StreamIO_ptr _p);
    //+

private:
    StreamIO_var stream_io_;
    Boolean context_;
    Table<Streamable_ptr,ULong> remember_;
};

class export StandardStreamIO : public StreamIO {
public:
    ~StandardStreamIO();

    OxExchange* _exchange();

    //+ Oes::StreamIO::*
    /* StreamIO */
    void write_string(const char* value);
    void write_char(Char value);
    void write_octet(Octet value);
    void write_unsigned_long(ULong value);
    void write_unsigned_short(UShort value);
    void write_long(Long value);
    void write_short(Short value);
    void write_float(Float value);
    void write_double(Double value);
    void write_object(Oes::Streamable_ptr value);
    char* read_string();
    Char read_char();
    Octet read_octet();
    ULong read_unsigned_long();
    UShort read_unsigned_short();
    Long read_long();
    Short read_short();
    Float read_float();
    Double read_double();
    Boolean read_boolean();
    Oes::Streamable_ptr read_object(Oes::FactoryFinder_ptr there, Oes::Streamable_ptr value);
    //+

protected:
    StandardStreamIO(GenericStream_ptr stream);

    virtual void put_byte() = 0;
    virtual unsigned char get_byte() = 0;
    
    void throw_error();
    
private:
    union FloatWormhole {
        Float value;
        unsigned char b[4];
    };

    union DoubleWormhole {
        Double value;
        unsigned char b[8];
    };

    void put_tag(unsigned char tag);
    
    void put_string(String value);
    void put_char(Char value);
    void put_octet(Octet value);
    void put_unsigned_long(ULong value);
    void put_unsigned_short(UShort value);
    void put_long(Long value);
    void put_short(Short value);
    void put_float(Float value);
    void put_double(Double value);
    void put_boolean(Boolean value);
    
    
    unsigned char get_tag();
    void check_tag(unsigned char expected);
    
    String get_string();
    Char get_char();
    Octet get_octet();
    ULong get_unsigned_long();
    UShort get_unsigned_short();
    Long get_long();
    Short get_short();
    Float get_float();
    Double get_double();
    Boolean get_boolean();

    GenericStream_ptr stream_;
};

class export OleStreamIO : public StandardStreamIO {
public:
    OleStreamIO(GenericStream_ptr stream, IStream* iostream);
    ~OleStreamIO();

private:
    void put_byte();
    unsigned char get_byte();
    
    IStream* iostream_;
};

class export CxxStreamIO : public StandardStreamIO {
public:
    CxxStreamIO(GenericStream_ptr stream, iostream* iostream);
    ~CxxStreamIO();

private:
    void put_byte();
    unsigned char get_byte();
    
    iostream* iostream_;
};

#endif

#endif
