//
// $XConsortium$
//

//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_impl_polyglyph_h
#define Fresco_defs_impl_polyglyph_h

#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/list.h>

class PolyGlyphOffset;

declarePtrListType(class export,PolyGlyphOffsetList,PolyGlyphOffset)

//- PolyGlyph*
class export PolyGlyph : public GlyphImpl {
    //. PolyGlyph is a default implementation for aggregate glyphs.
    //. PolyGlyph inherits most operations from GlyphImpl, redefining
    //. only those operations relating to aggregates.  PolyGlyph
    //. represents children as a list of PolyGlyphOffset objects
    //. that denote the child and the child's index into the list.
public:
    PolyGlyph();
    virtual ~PolyGlyph();

    void append(XfGlyph_ptr g); //+ XfGlyph::append
    void prepend(XfGlyph_ptr g); //+ XfGlyph::prepend
    XfGlyphOffset_ptr first_child_offset(); //+ XfGlyph::first_child_offset
    XfGlyphOffset_ptr last_child_offset(); //+ XfGlyph::last_child_offset

    PolyGlyphOffsetList children_;

    //- new_offset
    virtual PolyGlyphOffset* new_offset(Long index, Glyph_ptr child);
	//. Return a new offset for this polyglyph with
	//. the given index and child.  By default, this will simply
	//. create a new PolyGlyphOffset object.

    //- children_requests
    Glyph::Requisition* children_requests(Glyph::Requisition* req, Long n);
	//. Return an array of requisitions for the polyglyph's children.
	//. If the given integer is as large as the number of children,
	//. then the given pointer will be used for the result array.
	//. Otherwise, an array will be dynamically-allocated that
	//. the caller should delete.

    void visit_trail(Long, GlyphTraversal_ptr t);

    //- child_allocate
    virtual void child_allocate(Long index, Glyph::AllocationInfo& a);
	//. Determine the allocation for the given child given the
	//. allocation information for the parent.

    //- fixup
    virtual void fixup(Long start, Long delta);
	//. This operation is called when a list item is added
	//. or removed, meaning that the indices in all offsets
	//. must be updated.

    //- change
    virtual void change(Long index);
	//. Note that the request of the indexed child
	//. may have changed.

    //- modified
    virtual void modified();
	//. Note that the child list has been modified.  A polyglyph
	//. will normally cache its requisition; if so this operation
	//. should invalidate the cache.
};

Ox_implementation_decl(PolyGlyphOffset,GlyphOffset,SharedFrescoObjectImpl)
public:
    PolyGlyphOffset(PolyGlyph* parent, Long index, Glyph_ptr child);
    ~PolyGlyphOffset();

    //+ XfGlyphOffset::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyphOffset */
    XfGlyph_ptr parent();
    XfGlyph_ptr child();
    XfGlyphOffset_ptr next_child();
    XfGlyphOffset_ptr prev_child();
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void insert(XfGlyph_ptr g);
    void replace(XfGlyph_ptr g);
    void remove();
    void notify();
    void traverse(XfGlyphTraversal_ptr t);
    void visit_trail(XfGlyphTraversal_ptr t);
    void child_allocate(XfGlyph::AllocationInfo& a);
    //+

    PolyGlyph* parent_;
    Long index_;
    Glyph_ptr child_;
    Tag remove_tag_;

    GlyphOffset_ptr offset(Long index);
};

#endif
