/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_defs_impl_rasters_h
#define Fresco_defs_impl_rasters_h

#include <Fresco/defs/drawing.h>

#include <Fresco/defs_impl/fobjects.h>

class export RasterImageImpl : public XfRaster {
public:
    struct Pixel {
        Color::Intensity r, g, b;
	Float a;
    };

    RasterImageImpl(const XfRaster::Info& info, Pixel* image);
    ~RasterImageImpl();

    OxExchange* _exchange();

    //+ XfRaster::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfRaster */
    Boolean equal(XfRaster_ptr r);
    ULong hash();
    void raster_info(XfRaster::Info& i);
    void set_color(XfRaster::Index row, XfRaster::Index col, XfColor_ptr c);
    void set_color_blend(XfRaster::Index row, XfRaster::Index col, XfColor_ptr c, Float blend);
    void set_mask(XfRaster::Index row, XfRaster::Index col, Boolean on);
    XfColor_ptr get_color(XfRaster::Index row, XfRaster::Index col);
    void get_color_blend(XfRaster::Index row, XfRaster::Index col, XfColor_out c, Float& blend);
    Boolean get_mask(XfRaster::Index row, XfRaster::Index col);
    XfCoord scale();
    void scale(XfCoord _p);
    //+

private:
    SharedFrescoObjectImpl object_;
    Coord scale_;
    Raster::Info info_;
    Pixel* image_;
};

class export RasterBitmapImpl : public Raster {
public:
    RasterBitmapImpl(const Raster::Info& info, unsigned char* image);
    RasterBitmapImpl(
        const DrawingKit::Data8& d,
        Raster::Index rows, Raster::Index columns,
        Raster::Index origin_row, Raster::Index origin_column
    );
    ~RasterBitmapImpl();

    OxExchange* _exchange();

    //+ XfRaster::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfRaster */
    Boolean equal(XfRaster_ptr r);
    ULong hash();
    void raster_info(XfRaster::Info& i);
    void set_color(XfRaster::Index row, XfRaster::Index col, XfColor_ptr c);
    void set_color_blend(XfRaster::Index row, XfRaster::Index col, XfColor_ptr c, Float blend);
    void set_mask(XfRaster::Index row, XfRaster::Index col, Boolean on);
    XfColor_ptr get_color(XfRaster::Index row, XfRaster::Index col);
    void get_color_blend(XfRaster::Index row, XfRaster::Index col, XfColor_out c, Float& blend);
    Boolean get_mask(XfRaster::Index row, XfRaster::Index col);
    XfCoord scale();
    void scale(XfCoord _p);
    //+

private:
    SharedFrescoObjectImpl object_;
    Coord scale_;
    Raster::Info info_;
    unsigned char* image_;
    long bytes_per_row_;
};

#endif
