//
// $XConsortium$
//

//
// Copyright (c) 1992-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/file.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/styles.h>
#include <Fresco/defs_impl/table.h>
#include <Fresco/defs_impl/threads.h>
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

implementPtrList(StringList,CharString)

//
// Create an attribute.
//

StyleValueImpl::StyleValueImpl(CharString_ptr name, PathName* path) {
    name_ = new CharStringImpl(name);
    path_ = path;
    uninitialized_ = true;
    value_ = nil;
    any_value_ = nil;
    object_.lock_ = ThreadLib::thread_kit_ptr()->mutex_ptr();
}

//
// Dispose of an attribute's state.
//

StyleValueImpl::~StyleValueImpl() {
    CORBA::release(name_);
    delete_path(path_);
    CORBA::release(value_);
    CORBA::release(any_value_);
}

OxExchange* StyleValueImpl::_exchange() { return &object_; }

//+ StyleValueImpl(XfFrescoObject::=object_.)
XfTag StyleValueImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void StyleValueImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void StyleValueImpl::disconnect() {
    object_.disconnect();
}
void StyleValueImpl::notify_observers() {
    object_.notify_observers();
}
void StyleValueImpl::update() {
    object_.update();
}
//+

//+ StyleValueImpl(XfStyleValue::binding_name)
XfCharString_ptr StyleValueImpl::binding_name() {
    return CharString::_duplicate(name_);
}

//+ StyleValueImpl(XfStyleValue::uninitialized)
Boolean StyleValueImpl::uninitialized() {
    return uninitialized_;
}

//+ StyleValueImpl(XfStyleValue::priority=p)
void StyleValueImpl::priority(Long p) {
    priority_ = p;
}

//+ StyleValueImpl(XfStyleValue::priority?)
Long StyleValueImpl::priority() {
    Long p = priority_;
    return p;
}

//+ StyleValueImpl(XfStyleValue::is_on)
Boolean StyleValueImpl::is_on() {
    Boolean b;
    return read_boolean(b) && b;
}

//+ StyleValueImpl(XfStyleValue::read_boolean)
Boolean StyleValueImpl::read_boolean(Boolean& b) {
    CharString_var s;
    if (read_string(s.out())) {
	CharString_var s1 = FrescoLib::string_ref("on");
	CharString_var s2 = FrescoLib::string_ref("true");
	b = s->case_insensitive_equal(s1) || s->case_insensitive_equal(s2);
	return true;
    }
    return false;
}

//+ StyleValueImpl(XfStyleValue::write_boolean)
void StyleValueImpl::write_boolean(Boolean b) {
    write_string(FrescoLib::tmp_string_ref(b ? "on" : "off"));
}

//+ StyleValueImpl(XfStyleValue::read_coord)
Boolean StyleValueImpl::read_coord(XfCoord& c) {
    Boolean b = false;
    CharString_var s;
    if (read_string(s.out())) {
	CharStringBuffer buf(s);
	Coord pts = Coord(1.0);
	char* p = buf.string();
	const char* end = p + buf.length();
	if (p < end && (*p == '-' || *p == '+')) {
	    ++p;
	}
	Boolean dot = false;
	for (; p < end; p++) {
	    if (!dot && *p == '.') {
		dot = true;
	    } else if (!isspace(*p) && !isdigit(*p)) {
		if (strcmp(p, "mm") == 0) {
		    pts = Coord(72.0 / 25.4);
		} else if (strcmp(p, "cm") == 0) {
		    pts = Coord(72.0 / 2.54);
		} else if (strcmp(p, "in") == 0) {
		    pts = Coord(72.0);
		} else if (strcmp(p, "pt") != 0) {
		    return false;
		}
		*p = '\0';
		break;
	    }
	}
	float f = (float)strtod(buf.string(), &p);
	if (p != buf.string()) {
	    c = f * pts;
	    b = true;
	}
    }
    return b;
}

//
// Writing coords, longs, and doubles is not currently implemented.
//

//+ StyleValueImpl(XfStyleValue::write_coord)
void StyleValueImpl::write_coord(XfCoord) { }

//+ StyleValueImpl(XfStyleValue::read_integer)
Boolean StyleValueImpl::read_integer(Long& i) {
    Boolean b = false;
    CharString_var s;
    if (read_string(s.out())) {
	CharStringBuffer buf(s);
	char* ptr;
	Long n = strtol(buf.string(), &ptr, 0);
	if (ptr != buf.string()) {
	    i = n;
	    b = true;
	}
    }
    return b;
}

//+ StyleValueImpl(XfStyleValue::write_integer)
void StyleValueImpl::write_integer(Long) { }

//+ StyleValueImpl(XfStyleValue::read_real)
Boolean StyleValueImpl::read_real(Double& d) {
    Boolean b = false;
    CharString_var s;
    if (read_string(s.out())) {
	CharStringBuffer buf(s);
	char* ptr;
	double r = strtod(buf.string(), &ptr);
	if (ptr != buf.string()) {
	    d = r;
	    b = true;
	}
    }
    return b;
}

//+ StyleValueImpl(XfStyleValue::write_real)
void StyleValueImpl::write_real(Double) { }

//+ StyleValueImpl(XfStyleValue::read_string)
Boolean StyleValueImpl::read_string(XfCharString_out s) {
    Boolean b = false;
    if (!uninitialized_ && value_ != nil) {
	s = CharString::_duplicate(value_);
	b = true;
    }
    return b;
}

//+ StyleValueImpl(XfStyleValue::write_string)
void StyleValueImpl::write_string(XfCharString_ptr s) {
    uninitialized_ = false;
    CORBA::release(value_);
    value_ = parse_value(s);
    modified();
}

//+ StyleValueImpl(XfStyleValue::read_object)
Boolean StyleValueImpl::read_object(CORBA::Object_out s) {
    Boolean b = false;
    if (any_value_ != nil) {
	s = CORBA::Object::_duplicate(any_value_);
	b = true;
    }
    return b;
}

//+ StyleValueImpl(XfStyleValue::write_object)
void StyleValueImpl::write_object(CORBA::Object_ptr s) {
    uninitialized_ = false;
    CORBA::release(any_value_);
    any_value_ = CORBA::Object::_duplicate(s);
    modified();
}

//+ StyleValueImpl(XfStyleValue::lock)
void StyleValueImpl::lock() {
    object_.lock_->acquire();
}

//+ StyleValueImpl(XfStyleValue::unlock)
void StyleValueImpl::unlock() {
    object_.lock_->release();
}

//
// Scan an attribute value, replace \<char> as follows:
//
//	\<newline>	nothing
//	\n		newline
//	\<backslash>	"\"
//

CharString_ptr StyleValueImpl::parse_value(CharString_ptr v) {
    CharString_ptr r;
    CharStringBuffer buf(v);
    if (CharStringImpl::index(buf.string(), buf.length(), '\\') == nil) {
	r = new CharStringImpl(v);
    } else {
	char fixed_dst[200];
	const char* src = buf.string();
	Long n = buf.length();
	const char* src_end = src + n;
	char* dst_start = n < sizeof(fixed_dst) ? fixed_dst : new char[n + 1];
	char* dst = dst_start;
	for (; src < src_end; src++) {
	    if (*src == '\\') {
		++src;
		switch (*src) {
		case '\n':
		    // eliminate
		    break;
		case 'n':
		    *dst++ = '\n';
		    break;
		case '\\':
		    *dst++ = *src;
		    break;
		default:
		    *dst++ = '\\';
		    *dst++ = *src;
		    break;
		}
	    } else {
		*dst++ = *src;
	    }
	}
	*dst = '\0';
	r = new CharStringImpl(dst_start, dst - dst_start);
	if (dst_start != fixed_dst) {
	    delete dst_start;
	}
    }
    return r;
}

void StyleValueImpl::delete_path(PathName* list) {
    if (list != nil) {
	for (ListItr(StringList) i(*list); i.more(); i.next()) {
	    CORBA::release(i.cur());
	}
	delete list;
    }
}

//
// Notify observers assuming that we already hold our lock.
//

void StyleValueImpl::modified() {
    object_.SharedFrescoObjectImpl::notify_observers();
}

// class SharedStyleImpl

declareList(StyleList,SharedStyleImpl::Info)
implementList(StyleList,SharedStyleImpl::Info)

declarePtrList(StyleValueImplList,StyleValueImpl)
implementPtrList(StyleValueImplList,StyleValueImpl)

struct StyleValueTableEntry {
    StyleValueImplList** entries;
    long avail;
    long used;
};

declareTable(StyleValueTable,CharString_ptr,StyleValueTableEntry*)
implementTable(StyleValueTable,CharString_ptr,StyleValueTableEntry*)

SharedStyleImpl::SharedStyleImpl(CORBA::Object_ptr obj) {
    object_ = obj;
    style_ = nil;
    lock_ = ThreadLib::thread_kit_ptr()->mutex_ptr();
    name_ = nil;
    aliases_ = nil;
    parent_ = nil;
    links_ = nil;
    unlink_ = nil;
    table_ = nil;
    children_ = nil;
    modified_ = true;
}

SharedStyleImpl::~SharedStyleImpl() {
    CORBA::release(name_);
    if (is_not_nil(parent_)) {
	parent_->unlink_child(unlink_);
    }
    StyleValueTable* t = table_;
    if (t != nil) {
	for (TableIterator(StyleValueTable) i(*t); i.more(); i.next()) {
	    StyleValueTableEntry* e = i.cur_value();
	    CORBA::release(i.cur_key());
	    for (Long j = 0; j < e->used; j++) {
		StyleValueImplList* vl = e->entries[j];
		if (vl != nil) {
		    for (
			ListItr(StyleValueImplList) a(*vl); a.more(); a.next()
		    ) {
			CORBA::release(a.cur());
		    }
		    delete vl;
		}
	    }
	    delete e->entries;
	    delete e;
	}
	delete t;
    }
    StyleValueImpl::delete_path(aliases_);
    if (children_ != nil) {
	//
	// Set links_ to nil to avoid having unlink_parent call
	// unlink_child on this and run though the child list.
	//
	links_ = nil;
	for (ListItr(StyleList) i(*children_); i.more(); i.next()) {
	    SharedStyleImpl::Info& info = i.cur_ref();
	    info.child->unlink_parent();
	}
	delete children_;
    }
}

//
// Because of multiple inheritance, the objref given to CORBA::release
// may not be the object we need to free when the refcount goes to zero.
// So, we intercept the call and change the parameter to the objref
// that we have stored.
//

void SharedStyleImpl::release(CORBA::Object_ptr) {
    LockedFrescoObjectImpl::release(object_);
}

class StyleCloner : public StyleVisitorImpl {
public:
    StyleCloner(SharedStyleImpl* s);
    virtual ~StyleCloner();

    //+ XfStyleVisitor::=
    Boolean visit_alias(XfCharString_ptr n);
    Boolean visit_attribute(XfStyleValue_ptr a);
    Boolean visit_style(XfStyleContext_ptr s);
    //+
private:
    SharedStyleImpl* style_impl_;
};

StyleCloner::StyleCloner(SharedStyleImpl* s) { style_impl_ = s; }
StyleCloner::~StyleCloner() { }

//+ StyleCloner(XfStyleVisitor::visit_alias)
Boolean StyleCloner::visit_alias(XfCharString_ptr n) {
    style_impl_->alias(n);
    return true;
}

//+ StyleCloner(XfStyleVisitor::visit_attribute)
Boolean StyleCloner::visit_attribute(XfStyleValue_ptr a) {
    StyleValue_var n = style_impl_->bind(CharString_var(a->binding_name()));
    CharString_var s;
    if (a->read_string(s.out())) {
	n->write_string(s);
	n->priority(a->priority());
    }
    return true;
}

//+ StyleCloner(XfStyleVisitor::visit_style)
Boolean StyleCloner::visit_style(XfStyleContext_ptr s) {
    StyleContext_var ns = style_impl_->new_style();
    ns->merge(s);
    ns->link_parent(style_impl_->style_);
    return true;
}

//+ SharedStyleImpl(XfStyleContext::new_style)
XfStyleContext_ptr SharedStyleImpl::new_style() {
    return new StyleImpl;
}

//+ SharedStyleImpl(XfStyleContext::parent_style)
XfStyleContext_ptr SharedStyleImpl::parent_style() {
    return StyleContext::_duplicate(parent_);
}

//
// Link a style to its parent.  We don't want children to control
// the lifetime of the parent, so we don't keep a "counted" reference
// to the parent.
//

//+ SharedStyleImpl(XfStyleContext::link_parent)
void SharedStyleImpl::link_parent(XfStyleContext_ptr parent) {
    if (is_not_nil(parent_)) {
	parent_->unlink_child(unlink_);
    }
    if (is_not_nil(parent)) {
	unlink_ = parent->link_child(style_);
    }
    parent_ = parent;
}

//+ SharedStyleImpl(XfStyleContext::unlink_parent)
void SharedStyleImpl::unlink_parent() {
    if (is_not_nil(parent_)) {
	parent_->unlink_child(unlink_);
	parent_ = nil;
	CORBA::release(style_);
    }
}

//+ SharedStyleImpl(XfStyleContext::link_child)
XfTag SharedStyleImpl::link_child(XfStyleContext_ptr child) {
    if (children_ == nil) {
	children_ = new StyleList(5);
    }
    ++links_;
    SharedStyleImpl::Info info;
    info.child = StyleContext::_duplicate(child);
    info.tag = links_;
    children_->append(info);
    return links_;
}

//+ SharedStyleImpl(XfStyleContext::unlink_child)
void SharedStyleImpl::unlink_child(XfTag link_tag) {
    if (children_ != nil && links_ != nil) {
	for (ListUpdater(StyleList) i(*children_); i.more(); i.next()) {
	    SharedStyleImpl::Info& info = i.cur_ref();
	    if (info.tag == link_tag) {
		i.remove_cur();
		break;
	    }
	}
    }
}

//+ SharedStyleImpl(XfStyleContext::merge)
void SharedStyleImpl::merge(XfStyleContext_ptr s) {
    if (name_ == nil) {
	CharString_var name = s->context_name();
	if (is_not_nil(name)) {
	    name_ = new CharStringImpl(name);
	}
    }
    StyleVisitor_var v = new StyleCloner(this);
    s->visit_aliases(v);
    s->visit_attributes(v);
    s->visit_styles(v);
}

//+ SharedStyleImpl(XfStyleContext::context_name=s)
void SharedStyleImpl::context_name(XfCharString_ptr s) {
    name_ = CharString::_duplicate(s);
}

//+ SharedStyleImpl(XfStyleContext::context_name?)
XfCharString_ptr SharedStyleImpl::context_name() {
    return CharString::_duplicate(name_);
}

//+ SharedStyleImpl(XfStyleContext::alias)
void SharedStyleImpl::alias(XfCharString_ptr s) {
    if (aliases_ == nil) {
	aliases_ = new PathName(5);
    }
    aliases_->prepend(CharString::_duplicate(s));
}

//
// Bind a name to an attribute.  If the name is already defined,
// return the attribute.  Otherwise, create a new attribute.
//

//+ SharedStyleImpl(XfStyleContext::bind)
XfStyleValue_ptr SharedStyleImpl::bind(XfCharString_ptr n) {
    PathName* path = parse_name(n);
    if (path == nil) {
	// irrelevant attribute: A*B where A doesn't match
	return nil;
    }

    if (table_ == nil) {
	table_ = new StyleValueTable(50);
    }

    StyleValueTableEntry* e = find_tail_entry(path, true);
    Long count = path->count();
    if (e->avail <= count) {
	Long new_avail = count + 5;
	StyleValueImplList** new_list = new StyleValueImplList*[new_avail];
	for (Long i = 0; i < e->avail; i++) {
	    new_list[i] = e->entries[i];
	}
	for (i = e->avail; i < new_avail; i++) {
	    new_list[i] = nil;
	}
	delete e->entries;
	e->entries = new_list;
	e->avail = new_avail;
    }
    if (e->entries[count] == nil) {
	e->entries[count] = new StyleValueImplList;
    }
    e->used = Math::max(e->used, long(count + 1));
    StyleValueImplList* list = e->entries[count];
    for (ListItr(StyleValueImplList) i(*list); i.more(); i.next()) {
	StyleValueImpl* a = i.cur();
	if (same_path(a->path(), path)) {
	    StyleValueImpl::delete_path(path);
	    return StyleValue::_duplicate(a);
	}
    }
    StyleValueImpl* a = new StyleValueImpl(n, path);
    list->append(a);
    return StyleValue::_duplicate(a);
}

//
// Parse a name of the form *A*B*C into the list of names A, B, C.
// Strip the first name (e.g., A) if it matches the style's name
// or an alias.
//

PathName* SharedStyleImpl::parse_name(CharString_ptr s) {
    CharStringBuffer buf(s);
    Boolean leading_star = false;
    char* p = buf.string();
    char* end = p + buf.length();
    if (p < end && *p == '*') {
	leading_star = true;
	++p;
    }
    match_priority_ = 0;
    PathName* list = new PathName;
    Boolean first = true;
    for (;;) {
	Long i = find_separator(p, end);
	if (i == -1) {
	    list->append(new CharStringImpl(p, end - p));
	    break;
	}
	CharString_ptr name = new CharStringImpl(p, i);
	p += (i + 1);
	if (first) {
	    first = false;
	    Long q = match(name);
	    if (q != 0) {
		match_priority_ = (q == 1) ? 2 : 1;
		CORBA::release(name);
		continue;
	    } else if (!leading_star) {
		CORBA::release(name);
		StyleValueImpl::delete_path(list);
		list = nil;
		break;
	    }
	}
	list->append(name);
    }
    return list;
}

//
// Return the index of the next separator ("*" or ".") in the string.
// If no separator is present, return -1.
//

Long SharedStyleImpl::find_separator(const char* start, const char* end) {
    for (const char* p = start; p < end; p++) {
	if (*p == '*' || *p == '.') {
	    return p - start;
	}
    }
    return -1;
}

//
// Check to see if a given name matches the style's name
// or any of its aliases.
//
// Return value:
//     0 - no match
//     1 - name match
//     2 and up - index of alias match plus 2
//

//+ SharedStyleImpl(XfStyleContext::match)
Long SharedStyleImpl::match(XfCharString_ptr n) {
    Long match = 0;
    if (is_not_nil(name_) && name_->equal(n)) {
	match = 1;
    } else if (aliases_ != nil) {
	Long possible_match = 2;
	for (ListItr(StringList) i(*aliases_); i.more(); i.next()) {
	    if (n->equal(i.cur())) {
		match = possible_match;
		break;
	    }
	    ++possible_match;
	}
    }
    return match;
}

//
// Compare to lists of strings.
//

Boolean SharedStyleImpl::same_path(PathName* p1, PathName* p2) {
    if (p1->count() != p2->count()) {
	return false;
    }
    ListItr(StringList) i1(*p1);
    ListItr(StringList) i2(*p2);
    for (; i1.more(); i1.next(), i2.next()) {
	CharString_ptr c1 = i1.cur();
	CharString_ptr c2 = i2.cur();
	if (!c1->equal(c2)) {
	    return false;
	}
    }
    return true;
}

void SharedStyleImpl::load_file(const char* filename, Long priority) {
    InputFile* f = InputFile::open(filename);
    if (f == nil) {
	return;
    }
    const char* start;
    Long len = f->read(start);
    if (len > 0) {
	load_list(start, len, priority);
    }
    f->close();
    delete f;
}

void SharedStyleImpl::load_list(const char* str, Long length, Long priority) {
    const char* p = str;
    const char* q = p + length;
    const char* start = p;
    for (; p < q; p++) {
	if (*p == '\n') {
	    if (p > start && *(p-1) != '\\') {
		load_property(start, p - start, priority);
		start = p + 1;
	    }
	}
    }
}

void SharedStyleImpl::load_property(
    const char* prop, Long length, Long priority
) {
    const char* p = prop;
    Long n = length;
    strip(p, n);
    if (n == 0 || *p == '!') {
	return;
    }
    const char* colon = CharStringImpl::index(p, n, ':');
    if (colon == nil) {
	missing_colon(p, n);
    } else {
	const char* name = p;
	Long name_length = colon - p;
	const char* value = colon + 1;
	Long value_length = n - name_length - 1;
	strip(name, name_length);
	strip(value, value_length);
	if (name_length == 0) {
	    bad_property_name(p, length);
	} else if (value_length == 0) {
	    bad_property_value(p, length);
	} else {
	    StyleValue_var a = bind(
		CharString_var(new CharStringImpl(name, name_length))
	    );
	    if (is_not_nil(a)) {
		Long p = priority + match_priority_;
		if (a->uninitialized() || p >= a->priority()) {
		    a->write_string(
			CharString_var(new CharStringImpl(value, value_length))
		    );
		    a->priority(p);
		}
	    }
	}
    }
}

void SharedStyleImpl::strip(const char*& s, Long& length) {
    if (length == 0)
 	return;
    const char* p = s;
    for (; isspace(*p); p++);
    const char* q = s + length - 1;
    for (; q >= s && isspace(*q); q--);
    s = p;
    length = q - p + 1;
}

//
// Errors are nops for now.
//

void SharedStyleImpl::missing_colon(const char*, Long) { }
void SharedStyleImpl::bad_property_name(const char*, Long) { }
void SharedStyleImpl::bad_property_value(const char*, Long) { }

//
// Test if an attribute is defined and readable as a boolean set to true.
//

//+ SharedStyleImpl(XfStyleContext::is_on)
Boolean SharedStyleImpl::is_on(XfCharString_ptr n) {
    StyleValue_var a = resolve(n);
    Boolean b;
    return is_not_nil(a) && a->read_boolean(b) && b;
}

//
// Find the attribute bound to a given name, if any.
//

//+ SharedStyleImpl(XfStyleContext::resolve)
XfStyleValue_ptr SharedStyleImpl::resolve(XfCharString_ptr n) {
    StyleValueTableEntry* e = find_entry(n, false);
    if (e != nil) {
	StyleValueImplList* list = e->entries[0];
	if (list != nil && list->count() != 0) {
	    return StyleValue::_duplicate(list->item(0));
	}
    }

    StyleContext_var s = parent_style();
    for (; is_not_nil(s); s = s->parent_style()) {
	StyleValue_ptr a = s->resolve_wildcard(n, style_);
	if (is_not_nil(a)) {
	    return a;
	}
    }
    return nil;
}

//
// Try to match the given name against wildcard entries.
// The start style's name and aliases are possible wildcard prefixes.
//

//+ SharedStyleImpl(XfStyleContext::resolve_wildcard)
XfStyleValue_ptr SharedStyleImpl::resolve_wildcard(XfCharString_ptr n, XfStyleContext_ptr start) {
    StyleValueTableEntry* e = find_entry(n, false);
    if (e != nil) {
	if (e->used > 0) {
	    StyleValue_ptr a = wildcard_match(e, start);
	    if (is_not_nil(a)) {
		return a;
	    }
	}
	StyleValueImplList* list = e->entries[0];
	if (list != nil) {
	    return StyleValue::_duplicate(list->item(0));
	}
    }
    return nil;
}

//
// Short-hand for find an entry using the tail of a given path.
// Since the tail shouldn't be stored with the path, it is removed
// after used for the lookup.
//

StyleValueTableEntry* SharedStyleImpl::find_tail_entry(
    PathName* path, Boolean force
) {
    Long t = path->count() - 1;
    CharString_ptr str = path->item(t);
    StyleValueTableEntry* e = find_entry(str, force);
    CORBA::release(str);
    path->remove(t);
    return e;
}

//
// Lookup a given name in the attribute table.
// Since there can be multiple wildcard entries with the same name,
// we return the table entry if the name is found or nil if not.
//
// If the insert parameter is true, then create a new entry if the name
// is not defined.
//

StyleValueTableEntry* SharedStyleImpl::find_entry(
    CharString_ptr name, Boolean insert
) {
    StyleValueTableEntry* e = nil;
    if (table_ == nil && insert) {
	table_ = new StyleValueTable(50);
    }
    if (table_ != nil && !table_->find(e, name) && insert) {
	e = new StyleValueTableEntry;
	e->entries = new StyleValueImplList*[3];
	e->avail = 3;
	e->used = 0;
	for (Long i = 0; i < e->avail; i++) {
	    e->entries[i] = nil;
	}
	table_->insert(CharString::_duplicate(name), e);
    }
    return e;
}

//
// Check if the given table entry contains a match for the
// list of styles starting from the given one.
//

class StyleMatcher : public StyleVisitorImpl {
public:
    StyleMatcher();
    virtual ~StyleMatcher();

    void setup(
	SharedStyleImpl*, StyleValueTableEntry*,
	StyleContext_ptr start, StyleContext_ptr cur
    );

    Boolean visit_alias(XfCharString_ptr n); //+ XfStyleVisitor::visit_alias

    SharedStyleImpl* style_;
    StyleValueTableEntry* entry_;
    StyleContext_ptr start_;
    StyleContext_ptr cur_;
    StyleValue_ptr match_;
};

StyleMatcher::StyleMatcher() {
    style_ = nil;
    start_ = nil;
    cur_ = nil;
    match_ = nil;
}

StyleMatcher::~StyleMatcher() { }

void StyleMatcher::setup(
    SharedStyleImpl* style, StyleValueTableEntry* e,
    StyleContext_ptr start, StyleContext_ptr cur
) {
    style_ = style;
    entry_ = e;
    start_ = start;
    cur_ = cur;
    match_ = nil;
}

//+ StyleMatcher(XfStyleVisitor::visit_alias)
Boolean StyleMatcher::visit_alias(XfCharString_ptr n) {
    StyleValue_ptr a = style_->wildcard_match_name(n, entry_, cur_);
    if (is_not_nil(a)) {
	match_ = a;
	return false;
    }
    return true;
}

StyleValue_ptr SharedStyleImpl::wildcard_match(
    StyleValueTableEntry* e, StyleContext_ptr start
) {
    StyleValue_ptr found = nil;
    StyleMatcher matcher;
    StyleContext_var s = StyleContext::_duplicate(start);
    while (is_not_nil(s)) {
	CharString_var name = s->context_name();
	if (is_not_nil(name)) {
	    StyleValue_ptr a = wildcard_match_name(name, e, s);
	    if (is_not_nil(a)) {
		found = a;
		break;
	    }
	}
	matcher.setup(this, e, start, s);
	s->visit_aliases(&matcher);
	if (is_not_nil(matcher.match_)) {
	    found = matcher.match_;
	    break;
	}
	s = s->parent_style();
    }
    return found;
}

StyleValue_ptr SharedStyleImpl::wildcard_match_name(
    CharString_ptr name, StyleValueTableEntry* e, StyleContext_ptr cur
) {
    for (Long i = e->used - 1; i >= 1; i--) {
	StyleValueImplList* list = e->entries[i];
	if (list != nil) {
	    Long best_match = 0;
	    StyleValue_ptr matched = nil;
	    for (ListItr(StyleValueImplList) a(*list); a.more(); a.next()) {
		StyleValueImpl* attr = a.cur();
		PathName* path = attr->path();
		if (name->equal(path->item(i - 1))) {
		    if (i == 1) {
			return StyleValue::_duplicate(attr);
		    } else {
			Long new_match = finish_wildcard_match(
			    cur, path, i - 2
			);
			if (new_match > best_match) {
			    matched = attr;
			    best_match = new_match;
			}
		    }
		}
	    }
	    if (matched != nil) {
		return StyleValue::_duplicate(matched);
	    }
	}
    }
    return nil;
}

Long SharedStyleImpl::finish_wildcard_match(
    StyleContext_ptr cur, PathName* path, Long p_index
) {
    Long matched = 0;
    StyleContext_var s = StyleContext::_duplicate(cur);
    Long p_cur = p_index;
    while (p_cur >= 0 && is_not_nil(s)) {
	Long m = s->match(path->item(p_cur));
	if (m != 0) {
	    --p_cur;
	    matched += m;
	}
	s = s->parent_style();
    }
    return matched;
}

//+ SharedStyleImpl(XfStyleContext::unbind)
void SharedStyleImpl::unbind(XfCharString_ptr n) {
    if (table_ == nil) {
	return;
    }
    PathName* path = parse_name(n);
    if (path == nil) {
	return;
    }
    StyleValueTableEntry* e = find_tail_entry(path, false);
    Long p = path->count();
    if (e != nil && e->used > p) {
	StyleValueImplList* vl = e->entries[p];
	if (vl != nil) {
	    for (ListUpdater(StyleValueImplList) i(*vl); i.more(); i.next()) {
		StyleValueImpl* a = i.cur();
		if (same_path(a->path(), path)) {
		    CORBA::release(a);
		    i.remove_cur();
		    break;
		}
	    }
	    if (vl->count() == 0) {
		delete vl;
		e->entries[p] = nil;
	    }
	}
    }
    StyleValueImpl::delete_path(path);
}

//+ SharedStyleImpl(XfStyleContext::visit_aliases)
void SharedStyleImpl::visit_aliases(XfStyleVisitor_ptr v) {
    if (aliases_ != nil) {
	Long n = aliases_->count();
	for (Long i = n - 1; i >= 0; i--) {
	    if (!v->visit_alias(aliases_->item(i))) {
		break;
	    }
	}
    }
}

//+ SharedStyleImpl(XfStyleContext::visit_attributes)
void SharedStyleImpl::visit_attributes(XfStyleVisitor_ptr v) {
    if (table_ != nil) {
	for (TableIterator(StyleValueTable) i(*table_); i.more(); i.next()) {
	    StyleValueTableEntry* e = i.cur_value();
	    for (Long j = 0; j < e->used; j++) {
		StyleValueImplList* vl = e->entries[j];
		if (vl != nil) {
		    for (
			ListItr(StyleValueImplList) a(*vl); a.more(); a.next()
		    ) {
			if (!v->visit_attribute(a.cur())) {
			    return;
			}
		    }
		}
	    }
	}
    }
}

//+ SharedStyleImpl(XfStyleContext::visit_styles)
void SharedStyleImpl::visit_styles(XfStyleVisitor_ptr v) {
    if (children_ != nil) {
	for (ListItr(StyleList) i(*children_); i.more(); i.next()) {
	    if (!v->visit_style(i.cur_ref().child)) {
		break;
	    }
	}
    }
}

//+ SharedStyleImpl(XfStyleContext::lock)
void SharedStyleImpl::lock() {
    lock_->acquire();
}

//+ SharedStyleImpl(XfStyleContext::unlock)
void SharedStyleImpl::unlock() {
    lock_->release();
}

// class StyleImpl

StyleImpl::StyleImpl() : exch_(this) {
    exch_.style_ = this;
}

StyleImpl::StyleImpl(CORBA::Object_ptr obj) : exch_(obj) {
    exch_.style_ = this;
}

StyleImpl::~StyleImpl() { }

Ox_implementation_body(StyleImpl)

//+ StyleImpl(XfFrescoObject::=exch_.)
XfTag StyleImpl::attach(XfFrescoObject_ptr observer) {
    return exch_.attach(observer);
}
void StyleImpl::detach(XfTag attach_tag) {
    exch_.detach(attach_tag);
}
void StyleImpl::disconnect() {
    exch_.disconnect();
}
void StyleImpl::notify_observers() {
    exch_.notify_observers();
}
void StyleImpl::update() {
    exch_.update();
}
//+

//+ StyleImpl(XfStyleContext::=exch_.)
XfStyleContext_ptr StyleImpl::new_style() {
    return exch_.new_style();
}
XfStyleContext_ptr StyleImpl::parent_style() {
    return exch_.parent_style();
}
void StyleImpl::link_parent(XfStyleContext_ptr parent) {
    exch_.link_parent(parent);
}
void StyleImpl::unlink_parent() {
    exch_.unlink_parent();
}
XfTag StyleImpl::link_child(XfStyleContext_ptr child) {
    return exch_.link_child(child);
}
void StyleImpl::unlink_child(XfTag link_tag) {
    exch_.unlink_child(link_tag);
}
void StyleImpl::merge(XfStyleContext_ptr s) {
    exch_.merge(s);
}
XfCharString_ptr StyleImpl::context_name() {
    return exch_.context_name();
}
void StyleImpl::context_name(XfCharString_ptr _p) {
    exch_.context_name(_p);
}
void StyleImpl::alias(XfCharString_ptr s) {
    exch_.alias(s);
}
Boolean StyleImpl::is_on(XfCharString_ptr n) {
    return exch_.is_on(n);
}
XfStyleValue_ptr StyleImpl::bind(XfCharString_ptr n) {
    return exch_.bind(n);
}
void StyleImpl::unbind(XfCharString_ptr n) {
    exch_.unbind(n);
}
XfStyleValue_ptr StyleImpl::resolve(XfCharString_ptr n) {
    return exch_.resolve(n);
}
XfStyleValue_ptr StyleImpl::resolve_wildcard(XfCharString_ptr n, XfStyleContext_ptr start) {
    return exch_.resolve_wildcard(n, start);
}
Long StyleImpl::match(XfCharString_ptr n) {
    return exch_.match(n);
}
void StyleImpl::visit_aliases(XfStyleVisitor_ptr v) {
    exch_.visit_aliases(v);
}
void StyleImpl::visit_attributes(XfStyleVisitor_ptr v) {
    exch_.visit_attributes(v);
}
void StyleImpl::visit_styles(XfStyleVisitor_ptr v) {
    exch_.visit_styles(v);
}
void StyleImpl::lock() {
    exch_.lock();
}
void StyleImpl::unlock() {
    exch_.unlock();
}
//+

//
// Print out the name of a style (just for debugging purposes).
//

class StyleAliasPrinter : public StyleVisitorImpl {
public:
    StyleAliasPrinter();
    ~StyleAliasPrinter();

    Boolean visit_alias(CharString_ptr);
};

void StyleImpl::print_style_name(StyleContext_ptr s) {
    CharString_var name = s->context_name();
    if (is_nil(name)) {
	printf("(noname)");
    } else {
	CharStringBuffer buf(name);
	printf("%.*s", buf.length(), buf.string());
    }
    s->visit_aliases(StyleVisitor_var(new StyleAliasPrinter));
}

StyleAliasPrinter::StyleAliasPrinter() { }
StyleAliasPrinter::~StyleAliasPrinter() { }

Boolean StyleAliasPrinter::visit_alias(CharString_ptr s) {
    CharStringBuffer buf(s);
    printf(",%.*s", buf.length(), buf.string());
    return true;
}

// class StyleVisitorImpl

StyleVisitorImpl::StyleVisitorImpl() { }
StyleVisitorImpl::~StyleVisitorImpl() { }

OxExchange* StyleVisitorImpl::_exchange() { return &object_; }

//+ StyleVisitorImpl(XfFrescoObject::=object_.)
XfTag StyleVisitorImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void StyleVisitorImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void StyleVisitorImpl::disconnect() {
    object_.disconnect();
}
void StyleVisitorImpl::notify_observers() {
    object_.notify_observers();
}
void StyleVisitorImpl::update() {
    object_.update();
}
//+

Boolean StyleVisitorImpl::visit_alias(CharString_ptr) { return false; }
Boolean StyleVisitorImpl::visit_attribute(StyleValue_ptr) { return false; }
Boolean StyleVisitorImpl::visit_style(StyleContext_ptr) { return false; }
