//
// $XConsortium$
//

//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_impl_style_h
#define Fresco_defs_impl_style_h

#include <Fresco/defs/style.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/thread.h>

class StyleValueImplList;
class StyleValueTable;
struct StyleValueTableEntry;
class StyleList;

declarePtrListType(class export,StringList,CharString)

typedef StringList PathName;

class export StyleValueImpl : public StyleValue {
public:
    StyleValueImpl(CharString_ptr name, PathName* path);
    ~StyleValueImpl();

    OxExchange* _exchange();

    //+ XfStyleValue::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfStyleValue */
    XfCharString_ptr binding_name();
    Boolean uninitialized();
    Long priority();
    void priority(Long _p);
    Boolean is_on();
    Boolean read_boolean(Boolean& b);
    void write_boolean(Boolean b);
    Boolean read_coord(XfCoord& c);
    void write_coord(XfCoord c);
    Boolean read_integer(Long& i);
    void write_integer(Long i);
    Boolean read_real(Double& d);
    void write_real(Double d);
    Boolean read_string(XfCharString_out s);
    void write_string(XfCharString_ptr s);
    Boolean read_object(CORBA::Object_out s);
    void write_object(CORBA::Object_ptr s);
    void lock();
    void unlock();
    //+

    PathName* path() { return path_; }
    CharString_ptr parse_value(CharString_ptr);

    static void delete_path(PathName*);
protected:
    LockedFrescoObjectImpl object_;
    CharString_ptr name_;
    PathName* path_;
    Boolean uninitialized_;
    CharString_ptr value_;
    CORBA::Object_ptr any_value_;
    long priority_;

    void modified();
};

//
// SharedStyleImpl provides an implementation of the StyleContext operations
// suitable for using with StyleContext implementations (e.g., Viewer).
//

class export SharedStyleImpl : public LockedFrescoObjectImpl {
public:
    SharedStyleImpl(CORBA::Object_ptr);
    ~SharedStyleImpl();

    // release will delete objref rather than itself
    void release(CORBA::Object_ptr);

    //+ XfStyleContext::=
    XfStyleContext_ptr new_style();
    XfStyleContext_ptr parent_style();
    void link_parent(XfStyleContext_ptr parent);
    void unlink_parent();
    XfTag link_child(XfStyleContext_ptr child);
    void unlink_child(XfTag link_tag);
    void merge(XfStyleContext_ptr s);
    XfCharString_ptr context_name();
    void context_name(XfCharString_ptr _p);
    void alias(XfCharString_ptr s);
    Boolean is_on(XfCharString_ptr n);
    XfStyleValue_ptr bind(XfCharString_ptr n);
    void unbind(XfCharString_ptr n);
    XfStyleValue_ptr resolve(XfCharString_ptr n);
    XfStyleValue_ptr resolve_wildcard(XfCharString_ptr n, XfStyleContext_ptr start);
    Long match(XfCharString_ptr n);
    void visit_aliases(XfStyleVisitor_ptr v);
    void visit_attributes(XfStyleVisitor_ptr v);
    void visit_styles(XfStyleVisitor_ptr v);
    void lock();
    void unlock();
    //+

    struct Info {
	StyleContext_ptr child;
	Tag tag;
    };

    CORBA::Object_ptr object_;
    StyleContext_ptr style_;

    void load_file(const char* filename, Long priority);
    void load_list(const char* str, Long length, Long priority);
    void load_property(const char* property, Long length, Long priority);

    StyleValue_ptr wildcard_match(
	StyleValueTableEntry* e, StyleContext_ptr start
    );
    StyleValue_ptr wildcard_match_name(
	CharString_ptr name, StyleValueTableEntry*, StyleContext_ptr cur
    );
protected:
    CharString_ptr name_;
    PathName* aliases_;
    StyleContext_ptr parent_;
    Tag links_;
    Tag unlink_;
    StyleValueTable* table_;
    StyleList* children_;
    Boolean modified_;
    long match_priority_;

    PathName* parse_name(CharString_ptr name);
    Long find_separator(const char* start, const char* end);
    Boolean same_path(PathName* list1, PathName* list2);

    void strip(const char*& prop, Long& length);
    const char* index(const char* s, Long length, char c);
    void missing_colon(const char* s, Long length);
    void bad_property_name(const char* s, Long length);
    void bad_property_value(const char* s, Long length);

    StyleValueTableEntry* find_tail_entry(PathName*, Boolean force);
    StyleValueTableEntry* find_entry(CharString_ptr, Boolean force);
    Long finish_wildcard_match(StyleContext_ptr, PathName*, Long p_index);
};

//
// StyleImpl is an implementation of StyleContext that can be
// constructed directly.
//

Ox_implementation_decl(StyleImpl,StyleContext,SharedStyleImpl)
public:
    StyleImpl();
    StyleImpl(CORBA::Object_ptr);
    ~StyleImpl();

    //+ XfStyleContext::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfStyleContext */
    XfStyleContext_ptr new_style();
    XfStyleContext_ptr parent_style();
    void link_parent(XfStyleContext_ptr parent);
    void unlink_parent();
    XfTag link_child(XfStyleContext_ptr child);
    void unlink_child(XfTag link_tag);
    void merge(XfStyleContext_ptr s);
    XfCharString_ptr context_name();
    void context_name(XfCharString_ptr _p);
    void alias(XfCharString_ptr s);
    Boolean is_on(XfCharString_ptr n);
    XfStyleValue_ptr bind(XfCharString_ptr n);
    void unbind(XfCharString_ptr n);
    XfStyleValue_ptr resolve(XfCharString_ptr n);
    XfStyleValue_ptr resolve_wildcard(XfCharString_ptr n, XfStyleContext_ptr start);
    Long match(XfCharString_ptr n);
    void visit_aliases(XfStyleVisitor_ptr v);
    void visit_attributes(XfStyleVisitor_ptr v);
    void visit_styles(XfStyleVisitor_ptr v);
    void lock();
    void unlock();
    //+

    static void print_style_name(StyleContext_ptr);
};

//
// Default StyleVisitor implementation cancels iteration immediately.
//
// Other StyleVisitor implementations can derive from StyleVisitorImpl
// and just redefine the desired iteration method or methods.
//

class export StyleVisitorImpl : public StyleVisitor {
public:
    StyleVisitorImpl();
    ~StyleVisitorImpl();

    OxExchange* _exchange();

    //+ XfStyleVisitor::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfStyleVisitor */
    Boolean visit_alias(XfCharString_ptr n);
    Boolean visit_attribute(XfStyleValue_ptr a);
    Boolean visit_style(XfStyleContext_ptr s);
    //+
protected:
    SharedFrescoObjectImpl object_;
};

#endif
