//
// $XConsortium$
//

//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_impl_traversal_h
#define Fresco_defs_impl_traversal_h

#include <Fresco/defs/display.h>
#include <Fresco/defs/glyph.h>
#include <Fresco/defs_impl/fobjects.h>

class GTStack;

class export GlyphTraversalImpl : public GlyphTraversal {
public:
    GlyphTraversalImpl(GlyphTraversal::Operation, Window_ptr, Damage_ptr);
    GlyphTraversalImpl(const GlyphTraversalImpl&);
    ~GlyphTraversalImpl();

    OxExchange* _exchange();

    //+ XfGlyphTraversal::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyphTraversal */
    XfGlyphTraversal::Operation op();
    XfGlyphTraversal::Operation swap_op(XfGlyphTraversal::Operation op);
    Boolean is_pick_op();
    void begin_viewer(XfViewer_ptr v);
    void end_viewer();
    XfEventFilter_ptr event_filter();
    void traverse_child(XfGlyphOffset_ptr o, XfRegion_ptr allocation);
    void visit();
    XfGlyphTraversal_ptr trail();
    XfGlyph_ptr current_glyph();
    XfGlyphOffset_ptr current_offset();
    XfViewer_ptr current_viewer();
    Boolean forward();
    Boolean backward();
    XfDisplay_ptr current_display();
    XfScreen_ptr current_screen();
    XfWindow_ptr current_window();
    XfPainter_ptr current_painter();
    void current_painter(XfPainter_ptr _p);
    XfRegion_ptr allocation();
    Boolean bounds(XfVertex& lower, XfVertex& upper, XfVertex& origin);
    Boolean allocation_is_visible();
    XfTransform_ptr current_transform();
    XfDamage_ptr damaged();
    void hit();
    Long hit_info();
    void hit_info(Long _p);
    XfGlyphTraversal_ptr picked();
    void clear();
    //+

    struct Info {
	Viewer_ptr viewer;
	Glyph_ptr glyph;
	GlyphOffset_ptr offset;
	Region_ptr allocation;
	Transform_ptr transformation;
    };

    void push(
	Viewer_ptr v, Glyph_ptr g, GlyphOffset_ptr o,
	Region_ptr allocation, Transform_ptr transform
    );
    void pop();
protected:
    SharedFrescoObjectImpl object_;
    GlyphTraversal::Operation op_;
    GTStack* stack_;
    long index_;
    Painter_ptr painter_;
    Window_ptr window_;
    Damage_ptr damage_;
    GlyphTraversalImpl* picked_;
    long hit_info_;

    GlyphTraversalImpl::Info* top();
    GlyphTraversalImpl::Info* cur();
};

class export PickTraversal : public GlyphTraversalImpl {
public:
    PickTraversal(GlyphTraversal::Operation, Window_ptr, Damage_ptr);
    PickTraversal(const PickTraversal&);
    ~PickTraversal();

    void begin_viewer(XfViewer_ptr v); //+ XfGlyphTraversal::begin_viewer
    void traverse_child(XfGlyphOffset_ptr o, XfRegion_ptr allocation); //+ XfGlyphTraversal::traverse_child
    void hit(); //+ XfGlyphTraversal::hit

    void truncate_stack();
};

#endif
