//
// $XConsortium$
//

//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

//
// Common viewer implementations
//

#include <Fresco/defs/display.h>
#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/memory.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/defs_impl/viewers.h>

// class ViewerOffset

ViewerOffset::ViewerOffset(ViewerImpl* g) { parent_ = g; }
ViewerOffset::~ViewerOffset() { }

OxExchange* ViewerOffset::_exchange() { return parent_->_exchange(); }

//+ ViewerOffset(XfFrescoObject::*parent_->)
/* XfFrescoObject */
XfTag ViewerOffset::attach(XfFrescoObject_ptr observer) {
    return parent_->attach(observer);
}
void ViewerOffset::detach(XfTag attach_tag) {
    parent_->detach(attach_tag);
}
void ViewerOffset::disconnect() {
    parent_->disconnect();
}
void ViewerOffset::notify_observers() {
    parent_->notify_observers();
}
void ViewerOffset::update() {
    parent_->update();
}
//+

//+ ViewerOffset(XfGlyphOffset::parent)
XfGlyph_ptr ViewerOffset::parent() {
    return Glyph::_duplicate(parent_);
}

//+ ViewerOffset(XfGlyphOffset::child)
XfGlyph_ptr ViewerOffset::child() {
    return Glyph::_duplicate(child_);
}

//+ ViewerOffset(XfGlyphOffset::next_child)
XfGlyphOffset_ptr ViewerOffset::next_child() {
    return nil;
}

//+ ViewerOffset(XfGlyphOffset::prev_child)
XfGlyphOffset_ptr ViewerOffset::prev_child() {
    return nil;
}

//+ ViewerOffset(XfGlyphOffset::allocations)
void ViewerOffset::allocations(XfGlyph::AllocationInfoSeq& a) {
    ULong start = a._length;
    parent_->allocations(a);
    for (ULong i = start; i < a._length; i++) {
	parent_->child_allocate(a._buffer[i]);
    }
}

//+ ViewerOffset(XfGlyphOffset::child_allocate)
void ViewerOffset::child_allocate(XfGlyph::AllocationInfo& a) {
    parent_->child_allocate(a);
}

//
// The aggregate operations (insert, replace, and remove) and nops for now.
//

//+ ViewerOffset(XfGlyphOffset::insert)
void ViewerOffset::insert(XfGlyph_ptr) { }

//+ ViewerOffset(XfGlyphOffset::replace)
void ViewerOffset::replace(XfGlyph_ptr) { }

//+ ViewerOffset(XfGlyphOffset::remove)
void ViewerOffset::remove() { }

//+ ViewerOffset(XfGlyphOffset::notify)
void ViewerOffset::notify() {
    parent_->need_resize();
}

//+ ViewerOffset(XfGlyphOffset::traverse)
void ViewerOffset::traverse(XfGlyphTraversal_ptr t) {
    child_->traverse(t);
}

//+ ViewerOffset(XfGlyphOffset::visit_trail)
void ViewerOffset::visit_trail(XfGlyphTraversal_ptr t) {
    parent_->visit_trail(t);
}

ViewerImpl::ViewerImpl(Boolean tr) : style_(this), offset_(this) {
    offset_.child_ = nil;
    parent_ = nil;
    next_ = nil;
    prev_ = nil;
    first_ = nil;
    last_ = nil;
    entered_ = false;
    pressed_ = false;
    transparent_ = tr;
    recorded_time_ = false;
    display_ = nil;
    filter_ = 0;
    style_.alias(FrescoLib::tmp_string_ref("Viewer"));
}

ViewerImpl::~ViewerImpl() {
    CORBA::release(offset_.child_);
    Viewer_ptr v, nextv;
    for (v = first_; is_not_nil(v); v = nextv) {
	nextv = v->next_viewer();
	v->remove_viewer();
    }
}

OxExchange* ViewerImpl::_exchange() { return &style_.exch_; }

//+ ViewerImpl(XfFrescoObject::=style_.exch_.)
XfTag ViewerImpl::attach(XfFrescoObject_ptr observer) {
    return style_.exch_.attach(observer);
}
void ViewerImpl::detach(XfTag attach_tag) {
    style_.exch_.detach(attach_tag);
}
void ViewerImpl::disconnect() {
    style_.exch_.disconnect();
}
void ViewerImpl::notify_observers() {
    style_.exch_.notify_observers();
}
void ViewerImpl::update() {
    style_.exch_.update();
}
//+

//+ ViewerImpl(XfGlyph::style=s)
void ViewerImpl::style(XfStyleContext_ptr s) { style_.link_parent(s); }

//+ ViewerImpl(XfGlyph::style?)
XfStyleContext_ptr ViewerImpl::style() {
    return StyleContext::_duplicate(&style_);
}

//+ ViewerImpl(XfGlyph::transformation)
XfTransform_ptr ViewerImpl::transformation() { return nil; }

//+ ViewerImpl(XfGlyph::request)
void ViewerImpl::request(XfGlyph::Requisition& r) {
    offset_.child_->request(r);
}

//
// If the viewer is not transparent then we quickly compute
// the extension assuming that the contents do not draw
// outside the origin.
//

//+ ViewerImpl(XfGlyph::extension)
void ViewerImpl::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    if (transparent_) {
	offset_.child_->extension(a, r);
    } else {
	GlyphImpl::default_extension(a, r);
    }
}

//+ ViewerImpl(XfGlyph::shape)
void ViewerImpl::shape(XfRegion_ptr r) { offset_.child_->shape(r); }

//+ ViewerImpl(XfGlyph::allocations)
void ViewerImpl::allocations(XfGlyph::AllocationInfoSeq& a) {
    for (ListItr(GlyphOffsetList) i(glyph_parents_); i.more(); i.next()) {
	i.cur()->allocations(a);
    }
}

//+ ViewerImpl(XfGlyph::traverse)
void ViewerImpl::traverse(XfGlyphTraversal_ptr t) {
    Boolean visit_child, hit;
    switch (t->op()) {
    case GlyphTraversal::pick_top:
    case GlyphTraversal::pick_any:
    case GlyphTraversal::pick_all:
	visit_child = t->allocation_is_visible();
	hit = !transparent_;
	break;
    default:
	visit_child = true;
	hit = false;
	break;
    }
    if (visit_child) {
	t->begin_viewer(Viewer_ptr(this));
	t->traverse_child(&offset_, Region_var(t->allocation()));
	if (hit && is_nil(GlyphTraversal_var(t->picked()))) {
	    t->hit();
	}
	t->end_viewer();
    }
}

//+ ViewerImpl(XfGlyph::draw)
void ViewerImpl::draw(XfGlyphTraversal_ptr) { }

//+ ViewerImpl(XfGlyph::pick)
void ViewerImpl::pick(XfGlyphTraversal_ptr) { }

//+ ViewerImpl(XfGlyph::body=g)
void ViewerImpl::body(XfGlyph_ptr g) {
    if (is_not_nil(offset_.child_)) {
	offset_.child_->remove_parent(offset_.remove_tag_);
    }
    CORBA::release(offset_.child_);
    offset_.child_ = Glyph::_duplicate(g);
    offset_.remove_tag_ = g->add_parent(&offset_);
}

//+ ViewerImpl(XfGlyph::body?)
XfGlyph_ptr ViewerImpl::body() {
    return Glyph::_duplicate(offset_.child_);
}

//+ ViewerImpl(XfGlyph::append)
void ViewerImpl::append(XfGlyph_ptr g) {
    offset_.child_->append(g);
}

//+ ViewerImpl(XfGlyph::prepend)
void ViewerImpl::prepend(XfGlyph_ptr g) {
    offset_.child_->prepend(g);
}

//+ ViewerImpl(XfGlyph::add_parent)
XfTag ViewerImpl::add_parent(XfGlyphOffset_ptr parent_offset) {
    long n = glyph_parents_.count();
    glyph_parents_.append(parent_offset);
    return n;
}

//+ ViewerImpl(XfGlyph::remove_parent)
void ViewerImpl::remove_parent(XfTag add_tag) {
    long n = long(add_tag);
    if (n >= 0 && n < glyph_parents_.count()) {
	glyph_parents_.remove(n);
    }
}

//+ ViewerImpl(XfGlyph::first_child_offset)
XfGlyphOffset_ptr ViewerImpl::first_child_offset() {
    return offset_.child_->first_child_offset();
}

//+ ViewerImpl(XfGlyph::last_child_offset)
XfGlyphOffset_ptr ViewerImpl::last_child_offset() {
    return offset_.child_->last_child_offset();
}

//+ ViewerImpl(XfGlyph::parent_offsets)
void ViewerImpl::parent_offsets(XfGlyph::OffsetSeq& parents) {
    GlyphImpl::get_parent_offsets(parents, glyph_parents_);
}

//+ ViewerImpl(XfGlyph::need_redraw)
void ViewerImpl::need_redraw() {
    Glyph::AllocationInfoSeq alist;
    allocations(alist);
    RegionImpl r;
    for (ULong i = 0; i < alist._length; i++) {
	Glyph::AllocationInfo& a = alist._buffer[i];
	if (is_not_nil(a.damaged)) {
	    extension(a, &r);
	    a.damaged->extend(&r);
	}
    }
}

//+ ViewerImpl(XfGlyph::need_redraw_region)
void ViewerImpl::need_redraw_region(XfRegion_ptr r) {
    Glyph::AllocationInfoSeq alist;
    allocations(alist);
    RegionImpl dr;
    for (ULong i = 0; i < alist._length; i++) {
	Glyph::AllocationInfo& a = alist._buffer[i];
	if (is_not_nil(a.damaged)) {
	    dr.copy(r);
	    dr.apply_transform(a.transformation);
	    a.damaged->extend(&dr);
	}
    }
}

//+ ViewerImpl(XfGlyph::need_resize)
void ViewerImpl::need_resize() {
    for (ListItr(GlyphOffsetList) i(glyph_parents_); i.more(); i.next()) {
	i.cur()->notify();
    }
}

//+ ViewerImpl(XfGlyph::restore_trail) !t
Boolean ViewerImpl::restore_trail(XfGlyphTraversal_ptr) {
    return false;
}

void ViewerImpl::visit_trail(GlyphTraversal_ptr) { }
void ViewerImpl::child_allocate(Glyph::AllocationInfo&) { }

//+ ViewerImpl(XfGlyph::clone_glyph)
XfGlyph_ptr ViewerImpl::clone_glyph() {
    return nil;
}

//+ ViewerImpl(XfViewer::parent_viewer)
XfViewer_ptr ViewerImpl::parent_viewer() {
    return Viewer::_duplicate(parent_);
}

//+ ViewerImpl(XfViewer::next_viewer)
XfViewer_ptr ViewerImpl::next_viewer() { return Viewer::_duplicate(next_); }

//+ ViewerImpl(XfViewer::prev_viewer)
XfViewer_ptr ViewerImpl::prev_viewer() { return Viewer::_duplicate(prev_); }

//+ ViewerImpl(XfViewer::first_viewer)
XfViewer_ptr ViewerImpl::first_viewer() { return Viewer::_duplicate(first_); }

//+ ViewerImpl(XfViewer::last_viewer)
XfViewer_ptr ViewerImpl::last_viewer() { return Viewer::_duplicate(last_); }

//+ ViewerImpl(XfViewer::append_viewer)
void ViewerImpl::append_viewer(XfViewer_ptr v) {
    Viewer_ptr nv = Viewer::_duplicate(v);
    nv->set_viewer_links(this, last_, nil);
    last_ = nv;
    if (is_nil(first_)) {
	first_ = last_;
    }
}

//+ ViewerImpl(XfViewer::prepend_viewer)
void ViewerImpl::prepend_viewer(XfViewer_ptr v) {
    Viewer_ptr nv = Viewer::_duplicate(v);
    nv->set_viewer_links(this, nil, first_);
    first_ = nv;
    if (is_nil(last_)) {
	last_ = nv;
    }
}

//+ ViewerImpl(XfViewer::set_viewer_links)
void ViewerImpl::set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next) {
    if (is_not_nil(parent)) {
	parent_ = parent;
	style_.link_parent(parent->style());
	prev_ = prev;
	if (is_not_nil(prev)) {
	    prev->set_viewer_links(nil, nil, this);
	}
	next_ = next;
	if (is_not_nil(next)) {
	    next->set_viewer_links(nil, this, nil);
	}
    } else {
	if (is_not_nil(prev)) {
	    prev_ = prev;
	}
	if (is_not_nil(next)) {
	    next_ = next;
	}
    }
}

//+ ViewerImpl(XfViewer::insert_viewer)
void ViewerImpl::insert_viewer(XfViewer_ptr v) {
    Viewer_ptr nv = Viewer::_duplicate(v);
    nv->set_viewer_links(parent_, prev_, this);
    if (prev_ == nil) {
	parent_->set_first_viewer(nv);
    }
    prev_ = nv;
}

//+ ViewerImpl(XfViewer::replace_viewer)
void ViewerImpl::replace_viewer(XfViewer_ptr v) {
    v->set_viewer_links(parent_, prev_, next_);
    if (is_nil(prev_)) {
	parent_->set_first_viewer(v);
    }
    if (is_nil(next_)) {
	parent_->set_last_viewer(v);
    }
    prev_ = nil;
    next_ = nil;
    CORBA::release(Viewer_ptr(this));
}

//+ ViewerImpl(XfViewer::remove_viewer)
void ViewerImpl::remove_viewer() {
    if (is_nil(prev_)) {
	parent_->set_first_viewer(next_);
    } else if (is_nil(next_)) {
	parent_->set_last_viewer(prev_);
    }
    parent_ = nil;
    prev_ = nil;
    next_ = nil;
    CORBA::release(Viewer_ptr(this));
}

//+ ViewerImpl(XfViewer::set_first_viewer)
void ViewerImpl::set_first_viewer(XfViewer_ptr v) {
    first_ = v;
    if (is_nil(v)) {
	last_ = nil;
    }
}

//+ ViewerImpl(XfViewer::set_last_viewer)
void ViewerImpl::set_last_viewer(XfViewer_ptr v) {
    last_ = v;
    if (is_nil(v)) {
	first_ = nil;
    }
}

//+ ViewerImpl(XfViewer::request_focus)
XfFocus_ptr ViewerImpl::request_focus(XfViewer_ptr requestor, Boolean temporary) {
    return (
	is_nil(parent_) ? nil : parent_->request_focus(requestor, temporary)
    );
}

//+ ViewerImpl(XfViewer::receive_focus) !f, !primary
Boolean ViewerImpl::receive_focus(XfFocus_ptr, Boolean) {
    return nil;
}

//+ ViewerImpl(XfViewer::lose_focus) !f, !primary, !temporary
void ViewerImpl::lose_focus(XfFocus_ptr, Boolean, Boolean) { }

//+ ViewerImpl(XfViewer::first_focus)
Boolean ViewerImpl::first_focus() {
    return false;
}

//+ ViewerImpl(XfViewer::last_focus)
Boolean ViewerImpl::last_focus() {
    return false;
}

//+ ViewerImpl(XfViewer::next_focus)
Boolean ViewerImpl::next_focus() {
    return false;
}

//+ ViewerImpl(XfViewer::prev_focus)
Boolean ViewerImpl::prev_focus() {
    return false;
}

//+ ViewerImpl(XfViewer::handle)
Boolean ViewerImpl::handle(XfGlyphTraversal_ptr t, XfEvent_ptr e) {
    Boolean b = false;
    switch (e->type()) {
    case Event::motion:
    case Event::enter:
    case Event::leave:
    case Event::down:
    case Event::up:
	b = handle_position_event(t, e);
	break;
    case Event::other:
	b = other(t, e);
	break;
    default:
	b = focus_event(t, e);
	break;
    }
    return b;
}

//+ ViewerImpl(XfViewer::close)
void ViewerImpl::close() { }

Boolean ViewerImpl::handle_position_event(GlyphTraversal_ptr t, Event_ptr e) {
    Boolean b = false;
    switch (e->type()) {
    case Event::motion:
    case Event::enter:
    case Event::leave:
	b = pressed_ ? drag(t, e) : move(t, e);
	break;
    case Event::down:
	if (!pressed_ && press(t, e)) {
	    b = true;
	    pressed_ = true;
	    button_ = e->pointer_button();
	    grab(t);
	}
	break;
    case Event::up:
	if (pressed_ && e->pointer_button() == button_) {
	    pressed_ = false;
	    ungrab();
	    unsigned long time = e->time();
	    if (recorded_time_ && e->double_click(click_time_)) {
		b = double_click(t, e);
		recorded_time_ = false;
	    } else {
		b = release(t, e);
		click_time_ = time;
		recorded_time_ = true;
	    }
	}
	break;
    default:
	b = other(t, e);
	break;
    }
    return b;
}

Boolean ViewerImpl::focus_event(GlyphTraversal_ptr t, Event_ptr e) {
    Boolean b = false;
    switch (e->type()) {
    case Event::key_press:
	b = key_press(t, e);
	break;
    case Event::key_release:
	b = key_release(t, e);
	break;
    default:
	break;
    }
    return b;
}

Boolean ViewerImpl::move(GlyphTraversal_ptr, Event_ptr) { return false; }
Boolean ViewerImpl::press(GlyphTraversal_ptr, Event_ptr) { return false; }
Boolean ViewerImpl::drag(GlyphTraversal_ptr, Event_ptr) { return false; }
Boolean ViewerImpl::release(GlyphTraversal_ptr, Event_ptr) { return false; }
Boolean ViewerImpl::double_click(GlyphTraversal_ptr t, Event_ptr e) {
    return release(t, e);
}

Boolean ViewerImpl::key_press(GlyphTraversal_ptr, Event_ptr) {
    return false;
}

Boolean ViewerImpl::key_release(GlyphTraversal_ptr, Event_ptr) {
    return false;
}

Boolean ViewerImpl::other(GlyphTraversal_ptr, Event_ptr) { return false; }

Boolean ViewerImpl::inside(GlyphTraversal_ptr t) {
    Boolean b = t->allocation_is_visible();
    if (b && transparent_) {
	GlyphTraversal::Operation op = t->swap_op(GlyphTraversal::pick_any);
	t->traverse_child(&offset_, Region_var(t->allocation()));
	t->swap_op(op);
	b = is_not_nil(GlyphTraversal_var(t->picked()));
    }
    return b;
}

void ViewerImpl::grab(GlyphTraversal_ptr t) {
    if (is_nil(display_)) {
	display_ = t->current_display();
	filter_ = display_->add_filter(EventFilter_var(t->event_filter()));
    }
}

void ViewerImpl::ungrab() {
    if (!entered_ && !pressed_ && is_not_nil(display_)) {
	display_->remove_filter(filter_);
	CORBA::release(display_);
	display_ = nil;
    }
}

// class ActiveViewer

ActiveViewer::ActiveViewer(Boolean tr) : ViewerImpl(tr) { }
ActiveViewer::~ActiveViewer() { }

//
// Detect enter/leave into a viewer.  This is a little tricky
// because it needs to work in conjunction with the grabbing
// done by button press and release.  We assume that the
// grab and ungrab operations take care of ensuring a single grab
// is active, but we need to watch for re-entering a viewer
// with the button pressed.  When the button is pressed and
// we got the event while not inside the viewer, we must check
// to see if the new event is now inside the viewer.
//

Boolean ActiveViewer::move(GlyphTraversal_ptr t, Event_ptr e) {
    if (!entered_) {
	if (!pressed_ || inside(t)) {
	    entered_ = true;
	    grab(t);
	    enter();
	}
    } else if (e->type() == Event::leave || !inside(t)) {
	entered_ = false;
	leave();
	ungrab();
    }
    return true;
}

Boolean ActiveViewer::drag(GlyphTraversal_ptr t, Event_ptr e) {
    return move(t, e);
}

void ActiveViewer::enter() { }
void ActiveViewer::leave() { }

FocusImpl::FocusImpl() { }
FocusImpl::~FocusImpl() { }

OxExchange* FocusImpl::_exchange() { return &object_; }

//+ FocusImpl(XfFrescoObject::=object_.)
XfTag FocusImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void FocusImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void FocusImpl::disconnect() {
    object_.disconnect();
}
void FocusImpl::notify_observers() {
    object_.notify_observers();
}
void FocusImpl::update() {
    object_.update();
}
//+

//+ FocusImpl(XfFocus::map_keystroke)
void FocusImpl::map_keystroke(XfEvent::KeySym, XfAction_ptr) { }

//+ FocusImpl(XfFocus::map_keychord)
void FocusImpl::map_keychord(const XfEvent::KeyChord&, XfAction_ptr) { }

//+ FocusImpl(XfFocus::menu_bar)
XfMenuItem_ptr FocusImpl::menu_bar() { return nil; }

// class MainViewer

//
// We assume the window holds a reference to a main viewer and
// will be destroyed about the same time, so a main viewer doesn't
// hold a reference back to the window.
//

MainViewer::MainViewer() : ViewerImpl(true) {
    window_ = nil;
    requested_ = false;
    focus_ = new FocusImpl;
    damage_ = nil;
}

MainViewer::~MainViewer() {
    CORBA::release(focus_);
}

void MainViewer::allocate(Coord x, Coord y, Coord z) {
    if (! requested_) {
        request(requisition_);
	requested_ = true;
    }
    allocation_->defined_ = true;
    allocation_->lower_.x = Coord(0);
    allocation_->lower_.y = Coord(0);
    allocation_->lower_.z = Coord(0);
    allocation_->upper_.x = x;
    allocation_->upper_.y = y;
    allocation_->upper_.z = z;
    allocation_->xalign_ = requisition_.x.align;
    allocation_->yalign_ = requisition_.y.align;
    allocation_->zalign_ = requisition_.z.align;
}

//+ MainViewer(XfGlyph::allocations)
void MainViewer::allocations(XfGlyph::AllocationInfoSeq& a) {
    if (a.length() >= a.maximum()) {
	ULong n = a.maximum() == 0 ? 10 : a.maximum() + a.maximum();
	Glyph::AllocationInfo* buf = new Glyph::AllocationInfo[n];
	for (ULong ai = 0; ai < a.length(); ai++) {
	    buf[ai] = a[ai];
	}
	ULong len = a.length();
	a.load(n, len, buf, true);
    }
    if (allocation_->defined_) {
	Glyph::AllocationInfo& i = a[a.length()];
	i.allocation = new RegionImpl;
	i.allocation->copy(allocation_);
	i.transformation = new TransformImpl;
	i.damaged = Damage::_duplicate(damage_);
	++a._length;
    }
}

//+ MainViewer(XfGlyph::traverse)
void MainViewer::traverse(XfGlyphTraversal_ptr t) {
    ViewerImpl::traverse(t);
    damage_ = t->damaged();
    CORBA::release(damage_);
}

//+ MainViewer(XfGlyph::need_redraw)
void MainViewer::need_redraw() {
    if (is_not_nil(damage_)) {
	damage_->incur();
    }
}

//+ MainViewer(XfGlyph::need_resize)
void MainViewer::need_resize() {
    if (allocation_->defined_) {
        request(requisition_);

        allocation_->xalign_ = requisition_.x.align;
        allocation_->yalign_ = requisition_.y.align;
        allocation_->zalign_ = requisition_.z.align;
    }

// !!! Should resize the window if it is not within the bodys request. -denis
//    need_redraw();
}

//+ MainViewer(XfViewer::handle)
Boolean MainViewer::handle(XfGlyphTraversal_ptr t, XfEvent_ptr e) {
    if (e->type() == Event::focus_in) {
        if (!is_nil(requestor_)) {
            if (! requestor_->receive_focus(focus_, true)) {
                requestor_ = nil;
            }
        }
        return true;
    }

    if (e->type() == Event::focus_out) {
        if (!is_nil(requestor_)) {
            requestor_->lose_focus(focus_, true, true);
        }
        return true;
    }

    Viewer_var v = requestor_;
    while (is_not_nil(v) && !is_eq(v, this)) {
	if (v->handle(t, e)) {
	    return true;
	}
	v = v->parent_viewer();
    }
    return ViewerImpl::handle(t, e);
}

//+ MainViewer(XfViewer::request_focus)
XfFocus_ptr MainViewer::request_focus(XfViewer_ptr requestor, Boolean temporary) {
    if (!is_nil(requestor_)) {
	requestor_->lose_focus(focus_, true, temporary);
	Viewer_var v = requestor_->parent_viewer();
	for (; !is_nil(v); v = v->parent_viewer()) {
	    v->lose_focus(focus_, false, temporary);
	}
    }

    requestor_ = nil;

    if (!is_nil(requestor)) {
	if (requestor->receive_focus(focus_, true)) {
// !!! should lose_focus(false) if main viewer hasn't had a focus_in? -denis
            requestor_ = Viewer::_duplicate(requestor);
	    Viewer_var v = requestor_->parent_viewer();
	    for (; !is_nil(v); v = v->parent_viewer()) {
	        v->receive_focus(focus_, false);
	    }
        }
    }

    return is_nil(requestor_) ? nil : Focus::_duplicate(focus_);
}

//
// Normal main viewer close action is to exit the display run loop.
//

void MainViewer::close() {
    if (is_not_nil(window_)) {
	Screen_var s = window_->screen_ptr();
	Display_var d = s->display_ptr();
	d->run(false);
    }
}

// class BgViewer

BgViewer::BgViewer() : ViewerImpl(true) {
    background_ = nil;
}

BgViewer::~BgViewer() {
    CORBA::release(background_);
}

void BgViewer::traverse(GlyphTraversal_ptr t) {
    if (t->op() == GlyphTraversal::draw && t->allocation_is_visible()) {
	draw(t);
    }
    ViewerImpl::traverse(t);
}

void BgViewer::draw(GlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Vertex lower, upper, origin;
    t->bounds(lower, upper, origin);
    if (is_nil(background_)) {
	DrawingKit_var d = Display_var(t->current_display())->drawing_kit();
	background_ = d->background(StyleContext_var(style()));
    }
    p->current_color(background_);
    p->fill_rect(lower.x, lower.y, upper.x, upper.y);
}
