//
// $XConsortium$
//

//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_impl_viewers_h
#define Fresco_defs_impl_viewers_h

#include <Fresco/defs/viewer.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/styles.h>

class RegionImpl;
class ViewerImpl;

class export ViewerOffset : public GlyphOffset {
public:
    ViewerOffset(ViewerImpl*);
    ~ViewerOffset();

    OxExchange* _exchange();

    //+ XfGlyphOffset::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyphOffset */
    XfGlyph_ptr parent();
    XfGlyph_ptr child();
    XfGlyphOffset_ptr next_child();
    XfGlyphOffset_ptr prev_child();
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void insert(XfGlyph_ptr g);
    void replace(XfGlyph_ptr g);
    void remove();
    void notify();
    void traverse(XfGlyphTraversal_ptr t);
    void visit_trail(XfGlyphTraversal_ptr t);
    void child_allocate(XfGlyph::AllocationInfo& a);
    //+

    ViewerImpl* parent_;
    Glyph_ptr child_;
    Tag remove_tag_;
};

//- ViewerImpl*
class export ViewerImpl : _Ix_public Viewer {
    //. ViewerImpl is an implementation of the viewer interface.
    //. Because every viewer has one and only one distinct style,
    //. ViewerImpl uses a special ViewerStyle object to provide
    //. the style directly.  This approach means there is only
    //. one object for a viewer and its style.
    //.
    //. ViewerImpl also adds some implementation-specific virtual functions
    //. for input handling.  ViewerImpl implements Viewer::handle
    //. by inspecting the event type and calling the appropriate
    //. virtual function.
public:
    ViewerImpl(Boolean transparent = false);
    ~ViewerImpl();

    OxExchange* _exchange();

    //+ XfViewer::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyph */
    XfGlyph_ptr clone_glyph();
    XfStyleContext_ptr style();
    void style(XfStyleContext_ptr _p);
    XfTransform_ptr transformation();
    void request(XfGlyph::Requisition& r);
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r);
    void shape(XfRegion_ptr r);
    void traverse(XfGlyphTraversal_ptr t);
    void draw(XfGlyphTraversal_ptr t);
    void pick(XfGlyphTraversal_ptr t);
    XfGlyph_ptr body();
    void body(XfGlyph_ptr _p);
    void append(XfGlyph_ptr g);
    void prepend(XfGlyph_ptr g);
    XfTag add_parent(XfGlyphOffset_ptr parent_offset);
    void remove_parent(XfTag add_tag);
    XfGlyphOffset_ptr first_child_offset();
    XfGlyphOffset_ptr last_child_offset();
    void parent_offsets(XfGlyph::OffsetSeq& parents);
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void need_redraw();
    void need_redraw_region(XfRegion_ptr r);
    void need_resize();
    Boolean restore_trail(XfGlyphTraversal_ptr t);
    /* XfViewer */
    XfViewer_ptr parent_viewer();
    XfViewer_ptr next_viewer();
    XfViewer_ptr prev_viewer();
    XfViewer_ptr first_viewer();
    XfViewer_ptr last_viewer();
    void append_viewer(XfViewer_ptr v);
    void prepend_viewer(XfViewer_ptr v);
    void insert_viewer(XfViewer_ptr v);
    void replace_viewer(XfViewer_ptr v);
    void remove_viewer();
    void set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next);
    void set_first_viewer(XfViewer_ptr v);
    void set_last_viewer(XfViewer_ptr v);
    XfFocus_ptr request_focus(XfViewer_ptr requestor, Boolean temporary);
    Boolean receive_focus(XfFocus_ptr f, Boolean primary);
    void lose_focus(XfFocus_ptr f, Boolean primary, Boolean temporary);
    Boolean first_focus();
    Boolean last_focus();
    Boolean next_focus();
    Boolean prev_focus();
    Boolean handle(XfGlyphTraversal_ptr t, XfEvent_ptr e);
    void close();
    //+

    //- handle_position_event
    virtual Boolean handle_position_event(GlyphTraversal_ptr t, Event_ptr e);
	//. This operation looks at the type of event and calls
	//. move, press, drag, or whatever the appropriate function is.
	//. The chosen function depends on some state as well
	//. as the event type.  If a pointer button has been pressed,
	//. (and is still down), then a motion event will call drag,
	//. otherwise it will call move.  A button-down event
	//. will be ignored if a button is already down, otherwise
	//. it will call press and grab.  A button-up event
	//. will call ungrab and either release or double_click,
	//. depending on whether the timestamp of the up event
	//. is within a preset threshold.

    //- focus_event
    virtual Boolean focus_event(GlyphTraversal_ptr t, Event_ptr e);
	//. Handle an event from a focusable device, such as a keyboard.
	//. This operation looks at the type of event and calls
	//. key_press, key_release, or whatever the appropriate function is.

    //- move
    virtual Boolean move(GlyphTraversal_ptr t, Event_ptr e);
	//. Respond to a pointer motion event.  The default
	//. is to do nothing and return false.

    //- press
    virtual Boolean press(GlyphTraversal_ptr t, Event_ptr e);
	//. Respond to a button press event.  The default
	//. is to do nothing and return false.

    //- drag
    virtual Boolean drag(GlyphTraversal_ptr t, Event_ptr e);
	//. Respond to a pointer motion event while a button
	//. is pressed.  The default is to do nothing and return false.

    //- release
    virtual Boolean release(GlyphTraversal_ptr t, Event_ptr e);
	//. Respond to a button up event.  The default
	//. is to do nothing and return false.

    //- double_click
    virtual Boolean double_click(GlyphTraversal_ptr t, Event_ptr e);
	//. Respond to a button up event that rapidly follows
	//. a previous up event.  The default is to call release and
	//. return the value from release.

    //- key_press
    virtual Boolean key_press(GlyphTraversal_ptr t, Event_ptr e);
	//. Respond to a key being pressed.  The default is to do
	//. nothing and return false.

    //- key_release
    virtual Boolean key_release(GlyphTraversal_ptr t, Event_ptr e);
	//. Respond to a key being released.  The default is to do
	//. nothing and return false.

    //- other
    virtual Boolean other(GlyphTraversal_ptr t, Event_ptr e);
	//. Respond to an implementation-dependent event.  The default
	//. is to do nothing and return false.

    //- inside
    virtual Boolean inside(GlyphTraversal_ptr t);
	//. Test whether the given traversal state has clipping set
	//. that intersects with this viewer.  This operation is useful
	//. for testing if a positional event is inside a viewer
	//. when the positional event is the result of a grab
	//. (other positional events will necessarily be inside the viewer,
	//. otherwise it wouldn't be given the event to handle).

    //- grab
    virtual void grab(GlyphTraversal_ptr t);
	//. Set up a filter on the display (see Display::add_filter)
	//. so that further positional events will be passed directly
	//. to this viewer's handle_position_event operation.

    //- ungrab
    virtual void ungrab();
	//. Remove the filter for this viewer.  If the same viewer
	//. calls several grabs, then only one filter will be created
	//. but it will not be removed until a corresponding number
	//. of ungrab calls are made.

    virtual void visit_trail(GlyphTraversal_ptr t);
    virtual void child_allocate(Glyph::AllocationInfo&);
protected:
    StyleImpl style_;
    GlyphOffsetList glyph_parents_;
    ViewerOffset offset_;
    Viewer_ptr parent_;
    Viewer_ptr next_;
    Viewer_ptr prev_;
    Viewer_ptr first_;
    Viewer_ptr last_;

    Boolean entered_ : 1;
    Boolean pressed_ : 1;
    Boolean transparent_ : 1;
    Event::ButtonIndex button_;
    Boolean recorded_time_;
    unsigned long click_time_;
    Display_ptr display_;
    Tag filter_;
};

//- ActiveViewer*
class export ActiveViewer : public ViewerImpl {
    //. The ActiveViewer class checks motion events by implementing
    //. ViewerImpl::move and ViewerImpl::drag to notice when
    //. the pointer enters or leaves the viewer.
public:
    ActiveViewer(Boolean transparent = false);
    ~ActiveViewer();

    Boolean move(GlyphTraversal_ptr, Event_ptr);
    Boolean drag(GlyphTraversal_ptr, Event_ptr);

    //- enter
    virtual void enter();
	//. Notice that the pointer is now inside the viewer.  The default
	//. is to do nothing.

    //- leave
    virtual void leave();
	//. Notice that the pointer is now outside the viewer.  The default
	//. is to do nothing.
};

class export FocusImpl : public Focus {
public:
    FocusImpl();
    ~FocusImpl();

    OxExchange* _exchange();

    //+ XfFocus::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfFocus */
    void map_keystroke(XfEvent::KeySym k, XfAction_ptr a);
    void map_keychord(const XfEvent::KeyChord& k, XfAction_ptr a);
    XfMenuItem_ptr menu_bar();
    //+
protected:
    SharedFrescoObjectImpl object_;
};

class export MainViewer : public ViewerImpl {
public:
    MainViewer();
    ~MainViewer();

    void allocate(Coord x, Coord y, Coord z);

    void allocations(XfGlyph::AllocationInfoSeq& a); //+ XfGlyph::allocations
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    void need_redraw(); //+ XfGlyph::need_redraw
    void need_resize(); //+ XfGlyph::need_resize
    Boolean handle(XfGlyphTraversal_ptr t, XfEvent_ptr e); //+ XfViewer::handle
    XfFocus_ptr request_focus(XfViewer_ptr requestor, Boolean temporary); //+ XfViewer::request_focus
    void close(); //+ XfViewer::close

    Window_ptr window_;
    RegionImpl* allocation_;

protected:
    Boolean requested_;
    Glyph::Requisition requisition_;
    FocusImpl* focus_;
    Damage_ptr damage_;
    Viewer_var requestor_;
};

//- BgViewer*
class export BgViewer : public ViewerImpl {
    //. A background viewer just fills its allocation with its background
    //. color (from its style).
public:
    BgViewer();
    ~BgViewer();

    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
protected:
    Color_ptr background_;
};

#endif
