/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_defs_impls_windows_h
#define Fresco_defs_impls_windows_h

#include <Fresco/defs/display.h>
#include <Fresco/defs/viewer.h>

#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/styles.h>
#include <Fresco/defs_impl/thread.h>

class GlyphTraversalImpl;
class MainViewer;
class RegionImpl;
class WindowBase;

class export DamageImpl : public Damage {
public:
    DamageImpl(WindowBase*);

    OxExchange* _exchange();

    //+ XfDamage::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfDamage */
    void incur();
    void extend(XfRegion_ptr r);
    XfRegion_ptr current();
    void begin_update();
    void end_update();
    XfWindow_ptr window_ptr();
    //+

private:
    SharedFrescoObjectImpl object_;
    WindowBase* window_;
};

class export WindowBase : public Window {
public:
    WindowBase(Display_ptr display, Viewer_ptr viewer);
    ~WindowBase();

    // !!! There is no way to unconfigure a window, maybe configured and
    // !!! unconfigure should be interface members. -denis
    virtual Boolean configured();
    // The x and y values are only filled in if the window has been configured.
    virtual void get_client_geometry(
        PixelCoord& x, PixelCoord& y, PixelCoord& w, PixelCoord& h
    );
    
public:
    OxExchange* _exchange();

    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfWindow */
    Boolean double_buffered();
    void double_buffered(Boolean _p);
    XfCursor_ptr default_cursor();
    void default_cursor(XfCursor_ptr _p);
    XfColor_ptr cursor_foreground();
    void cursor_foreground(XfColor_ptr _p);
    XfColor_ptr cursor_background();
    void cursor_background(XfColor_ptr _p);
    XfCharString_ptr geometry();
    void geometry(XfCharString_ptr _p);
    XfWindow_ptr icon();
    void icon(XfWindow_ptr _p);
    Boolean iconic();
    void iconic(Boolean _p);
    XfCharString_ptr title();
    void title(XfCharString_ptr _p);
    Long xor_pixel();
    void xor_pixel(Long _p);
    XfScreen_ptr screen_ptr() = 0;
    XfStyleContext_ptr style();
    XfViewer_ptr main_viewer();
    void configure(const XfWindow::Placement& p) = 0;
    void get_configuration(Boolean position, XfWindow::Placement& p);
    void configure_notify(XfCoord width, XfCoord height);
    void move_notify(XfCoord left, XfCoord bottom);
    void map() = 0;
    void unmap() = 0;
    Boolean is_mapped() = 0;
    void map_notify();
    void unmap_notify();
    void iconify() = 0;
    void deiconify() = 0;
    void raise() = 0;
    void lower() = 0;
    void redraw(XfCoord left, XfCoord bottom, XfCoord width, XfCoord height);
    void repair() = 0;
    Boolean hits(XfEvent_ptr e);
    void handle_event(XfEvent_ptr e);
    void grab_pointer(XfCursor_ptr c) = 0;
    void ungrab_pointer() = 0;

    //+ XfDamage::=
    void incur();
    void extend(XfRegion_ptr r);
    XfRegion_ptr current();
    void begin_update();
    void end_update();
    XfWindow_ptr window_ptr();
    //+

protected:
    void set_attribute(const char* name, CharString_ptr value);
    CharString_ptr get_attribute(const char* name);

    void render();

    SharedFrescoObjectImpl object_;
    StyleImpl* style_;
    MainViewer* viewer_;
    Mutex_var lock_;

    DamageImpl* wdamage_;
    Mutex_var damage_lock_;
    RegionImpl* damage_;

    GlyphTraversalImpl* pick_;

    Boolean configured_;
    PixelCoord left_;
    PixelCoord top_;
    PixelCoord width_;
    PixelCoord height_;
    Coord x_align_;
    Coord y_align_;

    Cursor_ptr cursor_;
    Color_ptr cursor_foreground_;
    Color_ptr cursor_background_;

    Window_ptr icon_;

    Boolean double_buffer_;
    Long update_nesting_;
    GlyphTraversalImpl* draw_;
};

#endif
