/*
 * event.cxx
 */

#include <stdio.h>

#include <Fresco/defs/fresco.h>
#include <Fresco/figures/figures.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/widgets.h>

#include <Fresco/defs_impl/viewers.h>

static WidgetKit_var widgets_;
static LayoutKit_var layouts_;
static FigureKit_var figures_;
static FigureStyle_var figstyle_;

class EventView {
public:
    Glyph_ptr body();
    void show(Event* event);

    void label(const char* string);
    Glyph_ptr enum_row(const char* string, unsigned count, char* values[]);
    void set(Glyph_ptr deck, int value);
    Telltale_ptr boolean_row(const char* string);
    void set(Telltale_ptr telltale, Boolean value);
    FieldSubject_ptr long_row(const char* string);
    void set(FieldSubject_ptr field_subject, Long value);
    FieldSubject_ptr unsigned_long_row(const char* string);
    void set(FieldSubject_ptr field_subject, ULong value);
    FieldSubject_ptr float_row(const char* string);
    void set(FieldSubject_ptr field_subject, float value);

    Grid_var grid_;

    Glyph_var type_;
    FieldSubject_var time_;
    Telltale_var positional_;
    FieldSubject_var pointer_x_;
    FieldSubject_var pointer_y_;
    FieldSubject_var pointer_button_;
    Telltale_var button_1_is_down_;
    Telltale_var button_2_is_down_;
    Telltale_var button_3_is_down_;
    Telltale_var control_modifier_is_down_;
    Telltale_var shift_modifier_is_down_;
    Telltale_var caps_lock_modifier_is_down_;
    Telltale_var meta_modifier_is_down_;
    Telltale_var alt_modifier_is_down_;
    FieldSubject_var key_;
    FieldSubject_var character_;
};

Glyph_ptr EventView::body() {
    Grid::Index index;
    index.row = 16;
    index.col = 2;
    grid_ = layouts_->fixed_grid(index);

    static char* type_names[] = {
        "motion", "enter", "leave", "down", "up",
        "key press", "key release", "focus in", "focus out", "other"
    };
    type_ = enum_row("type", 10, type_names);
    time_ = unsigned_long_row("time");
    positional_ = boolean_row("positional");
    pointer_x_ = float_row("pointer x");
    pointer_y_ = float_row("pointer y");
    pointer_button_ = long_row("pointer button");
    button_1_is_down_ = boolean_row("button 1 is down");
    button_2_is_down_ = boolean_row("button 2 is down");
    button_3_is_down_ = boolean_row("button 3 is down");
    control_modifier_is_down_ = boolean_row("control modifier is down");
    shift_modifier_is_down_ = boolean_row("shift modifier is down");
    caps_lock_modifier_is_down_ = boolean_row("caps lock modifier is down");
    meta_modifier_is_down_ = boolean_row("meta modifier is down");
    alt_modifier_is_down_ = boolean_row("alt modifier is down");
    key_ = unsigned_long_row("key");
    character_ = unsigned_long_row("character");
    
    return layouts_->back(
	Glyph_var(widgets_->outset_frame(
	    Glyph_var(layouts_->margin(grid_, Coord(20)))
        )),
	Glyph_var(widgets_->filler())
    ); 
}

void EventView::show(Event* event) {
    set(type_, event->type());
    set(time_, event->time());
    set(positional_, event->positional());
    set(pointer_x_, event->pointer_x());
    set(pointer_y_, event->pointer_y());
    set(pointer_button_, event->pointer_button());
    set(button_1_is_down_, event->button_is_down(1));
    set(button_2_is_down_, event->button_is_down(2));
    set(button_3_is_down_, event->button_is_down(3));
    set(control_modifier_is_down_, event->modifier_is_down(Event::control));
    set(shift_modifier_is_down_, event->modifier_is_down(Event::shift));
    set(caps_lock_modifier_is_down_, event->modifier_is_down(Event::capslock));
    set(meta_modifier_is_down_, event->modifier_is_down(Event::meta));
    set(alt_modifier_is_down_, event->modifier_is_down(Event::alt));
    set(key_, event->key());
    set(character_, event->character());
}

void EventView::label(const char* string) {
    CharString_var char_string = FrescoLib::string_ref(string);
    grid_->append(Glyph_var(figures_->label(figstyle_, char_string)));
}

Glyph_ptr EventView::enum_row(const char* string, unsigned count, char* values[]) {
    label(string);
    Glyph_ptr deck = layouts_->deck();
    for (unsigned i = 0; i < count; ++i) {
        CharString_var char_string = FrescoLib::string_ref(values[i]);
        deck->append(Glyph_var(figures_->label(figstyle_, char_string)));
    }
    grid_->append(deck);
    return deck;
}

void EventView::set(Glyph_ptr deck, int value) {
}

Telltale_ptr EventView::boolean_row(const char* string) {
    label(string);
    Telltale_ptr telltale = widgets_->telltale_group();
    grid_->append(Glyph_var(widgets_->checkmark(telltale)));
    return telltale;
}

void EventView::set(Telltale_ptr telltale, Boolean value) {
    if (value) {
        telltale->set(Telltale::chosen);
    } else {
        telltale->clear(Telltale::chosen);
    }
}

FieldSubject_ptr EventView::long_row(const char* string) {
    return float_row(string);
}

void EventView::set(FieldSubject_ptr field_subject, Long value) {
    char buffer[12];
    sprintf(buffer, "%d", value);
    field_subject->replace_all(CharString_var(FrescoLib::string_ref(buffer)));
}

FieldSubject_ptr EventView::unsigned_long_row(const char* string) {
    return float_row(string);
}

void EventView::set(FieldSubject_ptr field_subject, ULong value) {
    char buffer[12];
    sprintf(buffer, "%ul", value);
    field_subject->replace_all(CharString_var(FrescoLib::string_ref(buffer)));
}

FieldSubject_ptr EventView::float_row(const char* string) {
    label(string);
    FieldSubject_ptr field_subject = widgets_->field_subject();
    grid_->append(Glyph_var(widgets_->field_editor(field_subject, 5, 5)));
    return field_subject;
}

void EventView::set(FieldSubject_ptr field_subject, float value) {
    char buffer[12];
    sprintf(buffer, "%f", value);
    field_subject->replace_all(CharString_var(FrescoLib::string_ref(buffer)));
}

class Input : public ViewerImpl {
public:
    Input(EventView& event_view);

    Boolean receive_focus(XfFocus_ptr focus, Boolean primary);
    Boolean handle(GlyphTraversal_ptr glyph_traversal, Event_ptr event);

private:
    EventView& event_view_;
};

Input::Input(EventView& event_view) : ViewerImpl(false), event_view_(event_view) {
    body(Glyph_var(widgets_->filler()));
}

Boolean Input::receive_focus(XfFocus_ptr, Boolean) {
    return true;
}

Boolean Input::handle(GlyphTraversal_ptr, Event_ptr event) {
    event_view_.show(event);
    return true;
}

int main(int argc, char** argv) {
    FrescoContext* fc = FrescoLib::init_default_context(
	"FTutorial", argc, argv
    );

    Display_var display = fc->display_ptr();
    if (is_nil(display)) {
        return 1;
    }
    Screen_var screen = display->default_screen();

    widgets_ = Fresco_resolve_object(WidgetKit, "widgets_impl");
    layouts_ = Fresco_resolve_object(LayoutKit, "layouts_impl");
    figures_ = Fresco_resolve_object(FigureKit, "figures_impl");
    figstyle_ = figures_->default_style();
    
    EventView event_view;
    Window_var application = screen->application(
        Viewer_var(widgets_->viewer_group(event_view.body()))
    );
    application->map();

    Viewer_var input = new Input(event_view);
    Window_var top_level = screen->top_level(input, application);
    Viewer_var parent = input->parent_viewer();
    parent->request_focus(input, false);
    top_level->map();

    display->run(true);

    CORBA::release(fc);
    return 0;
}
