/*
 * transform.cxx
 */

#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/figures/figures.h>
#include <Fresco/layouts/layouts.h>

#include <Fresco/defs_impl/drawings.h>
#include <Fresco/defs_impl/rasters.h>

static const unsigned int checkmark_width = 32;
static const unsigned int checkmark_height = 12;
static unsigned char checkmark_bits[] = {
    0x00, 0x00, 0x3e, 0x00,
    0x00, 0xc0, 0x0f, 0x00,
    0x04, 0xf0, 0x03, 0x00,
    0x1e, 0xf8, 0x00, 0x00,
    0x3f, 0x3e, 0x00, 0x00,
    0xbf, 0x1f, 0x00, 0x00,
    0xf8, 0x07, 0x00, 0x00,
    0xf0, 0x03, 0x00, 0x00,
    0xf0, 0x01, 0x00, 0x00,
    0xe0, 0x00, 0x00, 0x00,
    0x60, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00
};
static Raster::Info checkmark_info = {
    checkmark_height, checkmark_width, 0, 0
};

static RasterImageImpl::Pixel image_data[] = {
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},

    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {1.0f, 0.0f, 0.0f, 1.0f},
    {1.0f, 0.0f, 0.0f, 1.0f},
    {1.0f, 0.0f, 0.0f, 0.0f},
    {1.0f, 0.0f, 0.0f, 0.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},

    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {1.0f, 0.0f, 0.0f, 1.0f},
    {1.0f, 0.0f, 0.0f, 1.0f},
    {1.0f, 0.0f, 0.0f, 0.0f},
    {1.0f, 0.0f, 0.0f, 0.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},

    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 1.0f}
};
static Raster::Info image_info = {
    4, 8, 0, 0
};

int main(int argc, char** argv) {
    FrescoContext* fresco_context = FrescoLib::init_default_context(
	"FTutorial", argc, argv
    );
    DrawingKit_var drawing = fresco_context->default_drawing_kit();
    FigureKit_var figures = Fresco_resolve_object(FigureKit, "figures_impl");
    FigureStyle_var figure_style = figures->default_style();
    LayoutKit_var layouts = Fresco_resolve_object(LayoutKit, "layouts_impl");

    // Create a path figure.
    FigureKit::Vertices vertices(4, 4);
    vertices[0].x = Coord(0);
    vertices[0].y = Coord(0);
    vertices[0].z = Coord(0);
    vertices[1].x = Coord(0);
    vertices[1].y = Coord(17);
    vertices[1].z = Coord(0);
    vertices[2].x = Coord(17);
    vertices[2].y = Coord(17);
    vertices[2].z = Coord(17);
    vertices[3].x = Coord(7);
    vertices[3].y = Coord(0);
    vertices[3].z = Coord(0);
    Glyph_var figure = figures->polygon(
        FigureKit::fill, figure_style, vertices
    );
    Glyph_var figure_group = figures->group();
    figure_group->append(figure);   

    // Create a stencil figure.    
    Raster_var stencil_raster = new RasterBitmapImpl(checkmark_info, checkmark_bits);
    Glyph_var stencil = figures->stencil(figure_style, stencil_raster);
    Glyph_var stencil_group = figures->group();
    stencil_group->append(stencil);   

    // Create a image figure.
    Raster_var image_raster = new RasterImageImpl(image_info, image_data);
    Glyph_var image = figures->image(image_raster);
    Glyph_var image_group = figures->group();
    image_group->append(image);   
    
    // Create a label figure.
    Glyph_var label = figures->label(
        figure_style, FrescoLib::tmp_string_ref("hi")
    );
    Glyph_var label_group = figures->group();
    label_group->append(label);   

    const float angle = 45.0f;
    Transform_var(figure_group->transformation())->rotate(angle, Z_axis);
    Transform_var(stencil_group->transformation())->rotate(angle, Z_axis);
    Transform_var(image_group->transformation())->rotate(angle, Z_axis);
    Transform_var(label_group->transformation())->rotate(angle, Z_axis);

    // Put all the figures in to a horizontal layout.
    Grid::Index size;
    size.col = 4;
    size.row = 2;
    Glyph_var box = layouts->fixed_grid(size);
    box->append(Glyph_var(layouts->align(stencil_group, 0.5f, 0.5f)));
    box->append(Glyph_var(layouts->align(image_group, 0.5f, 0.5f)));
    box->append(Glyph_var(layouts->align(label_group, 0.5f, 0.5f)));
    box->append(Glyph_var(layouts->align(figure_group, 0.5f, 0.5f)));
    box->append(Glyph_var(layouts->align(stencil, 0.5f, 0.5f)));
    box->append(Glyph_var(layouts->align(image, 0.5f, 0.5f)));
    box->append(Glyph_var(layouts->align(label, 0.5f, 0.5f)));
    box->append(Glyph_var(layouts->align(figure, 0.5f, 0.5f)));

    FrescoLib::run(nil, box);

    CORBA::release(fresco_context);
    return 0;
}
