/*
 * tutorial.cxx
 */

#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/glyphs.h>
#include "ftutorial.h"

/* class FrescoObjectImpl */

OxExchange* FrescoObjectImpl::_exchange() { return &object_; }

//+ FrescoObjectImpl(XfFrescoObject::=object_.)
Tag FrescoObjectImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void FrescoObjectImpl::detach(Tag attach_tag) {
    object_.detach(attach_tag);
}
void FrescoObjectImpl::disconnect() {
    object_.disconnect();
}
void FrescoObjectImpl::notify_observers() {
    object_.notify_observers();
}
void FrescoObjectImpl::update() {
    object_.update();
}
//+

/* class Requestor */

Requestor::Requestor(Glyph_ptr g, const Glyph::Requisition& r) {
    body(g);
    req_ = r;
}

Requestor::~Requestor() { }

Requestor::Requestor(Glyph_ptr g,
    Coord xspan, Coord yspan,
    Alignment xalign, Alignment yalign
) {
    body(g);
    GlyphImpl::default_requisition(req_);
    Glyph::Requirement* rx = GlyphImpl::requirement(req_, X_axis);
    Glyph::Requirement* ry = GlyphImpl::requirement(req_, Y_axis);
    rx->align = xalign;
    ry->align = yalign;
    rx->natural = xspan;
    ry->natural = yspan; 
}

//+ Requestor(Glyph::request)
void Requestor::request(Glyph::Requisition& r) {
    r = req_;
}

/* class ProxyObserver */

ProxyObserver::ProxyObserver(FrescoObject_ptr subj, Action_ptr action) {
    subj_ = FrescoObject::_duplicate(subj);
    action_ = Action::_duplicate(action);
    tag_ = subj_->attach(this);
}

ProxyObserver::~ProxyObserver() {
    subj_->detach(tag_);
}


//+ ProxyObserver(XfFrescoObject::update)
void ProxyObserver::update() {
    action_->execute();
}

/* WinMain for Win32 */
#ifdef WIN32
#include <wtypes.h>

extern int main(int, char**);

int WINAPI WinMain(
    HINSTANCE  hInstance,       // handle of current instance
    HINSTANCE  hPrevInstance,   // handle of previous instance
    LPSTR  lpszCmdLine, // address of command line
    int  nCmdShow       // show state of window
) {
    int argc = 1;
    char* argv[] = {"Tutorial", 0};
    return main(argc, argv);
}
#endif

