/*
 * ftutorial.h
 */

#ifndef _tutorial_h
#define _tutorial_h

#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/action.h>

/* class FrescoObjectImpl */

class FrescoObjectImpl : public FrescoObject {
public:
    OxExchange* _exchange();
    
    //+ FrescoObject::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    //+
private:
    SharedFrescoObjectImpl object_;
};

//- Requestor*
class Requestor : public MonoGlyph {
    //. A Requestor always returns a constant requisition. It is
    //. useful as a wrapper for glyphs whose requisition changes
    //. but who wish to be laid out as though their requisition
    //. were constant.
public:
    Requestor(Glyph_ptr g,
	 Coord xspan = 1, Coord yspan = 1,
         Alignment xalign = .5, Alignment yalign = .5
    );
    Requestor(Glyph_ptr g, const Glyph::Requisition& r);
    ~Requestor();
    void request(Glyph::Requisition& r); //+ Glyph::request

protected:
    Glyph::Requisition req_;
};



/* ProxyObserver */
    
/* Since a given observer cannot attach itself to more than one object,
 * a general mechanism is needed to define a "proxy" object that
 * will attach itself on behalf of the observer. When update() is
 * called on the proxy, it makes a callback to the given observer.
 */

class ProxyObserver : public FrescoObjectImpl {
public:
    ProxyObserver(FrescoObject_ptr subj, Action_ptr action);
    ~ProxyObserver();
    void update(); //+ XfFrescoObject::update
protected:
    FrescoObject_var subj_;
    Action_var action_;		
    Tag tag_;
};

#endif

