/*
 * rgb.cxx
 */

#include <Fresco/defs/fresco.h>
#include <Fresco/figures/figures.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/widgets.h>
#include "floatview.h"
#include "rgbview.h"

static Glyph_ptr label(FigureKit_ptr figures, const char* s) {
    return figures->label(
        figures->default_style(), FrescoLib::tmp_string_ref(s)
    );
}

static Glyph_ptr value_controller(
    Viewer_ptr vg, Adjustment_ptr a, FieldSubject_ptr fs,
    const char* name, Coord scrollbar_size
) {
    FigureKit_var figures = Fresco_resolve_object(FigureKit, "figures_impl");
    WidgetKit_var widgets = Fresco_resolve_object(WidgetKit, "widgets_impl");
    LayoutKit_var layouts = Fresco_resolve_object(LayoutKit, "layouts_impl");
    long nchars = 5;
    Viewer_var fe = widgets->field_editor(fs, nchars, nchars);
    vg->append_viewer(fe);
    Viewer_var scrollbar = widgets->scroll_bar(Y_axis, a);
    Glyph_var sbar = layouts->vfixed(scrollbar, scrollbar_size);
    Glyph_var lab = label(figures, name);
    Glyph_var inset_fe = widgets->inset_frame(fe);
    Glyph_ptr vbox = layouts->vbox_align_elements(Alignment(0.5));
    vbox->append(sbar);
    vbox->append(Glyph_var(layouts->vglue(Coord(4.0), Coord(0), Coord(0))));
    vbox->append(inset_fe);
    vbox->append(lab);
    return vbox;
}
    
int main(int argc, char** argv) {
    FrescoContext* fc = FrescoLib::init_default_context(
	"FTutorial", argc, argv
    );
    FigureKit_var figures = Fresco_resolve_object(FigureKit, "figures_impl");
    WidgetKit_var widgets = Fresco_resolve_object(WidgetKit, "widgets_impl");
    LayoutKit_var layouts = Fresco_resolve_object(LayoutKit, "layouts_impl");
    FigureStyle_var figstyle = figures->new_style(
	FigureStyle_var(figures->default_style())
    );
    Coord size = Coord(100);
    Glyph_var rect = figures->rectangle(
	FigureKit::fill, figstyle, Coord(0),Coord(0), size,size
    );
    FrescoObject_var figview = new ColoredFigView(rect, figstyle);
    Adjustment_var r = widgets->bounded_float(0.0f, 100.0f, 50.0f);
    Adjustment_var g = widgets->bounded_float(0.0f, 100.0f, 50.0f);
    Adjustment_var b = widgets->bounded_float(0.0f, 100.0f, 50.0f);
    FrescoObject_var rgbview = new RgbView(r,g,b, figstyle);

    FieldSubject_var rs = widgets->field_subject();
    FieldSubject_var gs = widgets->field_subject();
    FieldSubject_var bs = widgets->field_subject();
    FrescoObject_var rfv = new FloatView(rs, r);
    FrescoObject_var gfv = new FloatView(gs, g);
    FrescoObject_var bfv = new FloatView(bs, b);

    Glyph_var color_area = widgets->inset_frame(rect);
    Viewer_var vg = widgets->viewer_group(nil);
    Glyph_var hbox = layouts->hbox_align_elements(Alignment(0.5));
    hbox->append(color_area);
    hbox->append(Glyph_var(layouts->hglue(Coord(15), Coord(30), Coord(0))));
    hbox->append(Glyph_var(value_controller(vg, r, rs, "red", size-8)));
    hbox->append(Glyph_var(value_controller(vg, g, gs, "green", size-8)));
    hbox->append(Glyph_var(value_controller(vg, b, bs, "blue",  size-8)));
    Glyph_var glyph = layouts->back(
	Glyph_var(widgets->outset_frame(
	    Glyph_var(layouts->margin(hbox, Coord(10.0)))
	)),
	Glyph_var(widgets->filler())
    );
    FrescoLib::run(vg, glyph);  
    CORBA::release(fc);
    return 0;
}

