/*
 * rgbview.cxx
 */

#include <Fresco/defs/fresco.h>
#include <Fresco/widgets/widgets.h>
#include <Fresco/figures/figures.h>
#include "ftutorial.h"
#include "rgbview.h"

/* class RgbView */

declareActionCallback(RgbView);
implementActionCallback(RgbView);

RgbView::RgbView(
    Adjustment_ptr r, Adjustment_ptr g, Adjustment_ptr b,
    FigureStyle_ptr fs
) {
    r_ = Adjustment::_duplicate(r);
    g_ = Adjustment::_duplicate(g);
    b_ = Adjustment::_duplicate(b);
    figstyle_ = FigureStyle::_duplicate(fs);
    rcallback_ = new ProxyObserver(
	r_, new ActionCallback(RgbView)(this, &RgbView::update_red)
    );
    gcallback_ = new ProxyObserver(
	g_, new ActionCallback(RgbView)(this, &RgbView::update_green)
    );
    bcallback_ = new ProxyObserver(
	b_, new ActionCallback(RgbView)(this, &RgbView::update_blue)
    );
    drawkit_ = FrescoLib::current_context()->default_drawing_kit();
    update_red();
    update_green();
    update_blue();
}

RgbView::~RgbView() { }

void RgbView::update_red() {
    Color::Intensity r, g, b, val;
    Color_var(figstyle_->foreground())->rgb(r,g,b);
    val = value(r_);
    Color_var fg = drawkit_->color_rgb(val, g, b);
    figstyle_->foreground(fg);
    figstyle_->notify_observers();
}

void RgbView::update_green() {
    Color::Intensity r, g, b, val;
    Color_var(figstyle_->foreground())->rgb(r,g,b);
    val = value(g_);
    Color_var fg = drawkit_->color_rgb(r, val, b);
    figstyle_->foreground(fg);
    figstyle_->notify_observers();
}    

void RgbView::update_blue() {
    Color::Intensity r, g, b, val;
    Color_var(figstyle_->foreground())->rgb(r,g,b);
    val = value(b_);
    Color_var fg = drawkit_->color_rgb(r, g, val);
    figstyle_->foreground(fg);
    figstyle_->notify_observers();
}

Color::Intensity RgbView::value(Adjustment_ptr a) {
    Adjustment::Settings s;
    a->get_settings(s);
    return s.cur_lower / 100;
}


/* class ColoredFigView */

ColoredFigView::ColoredFigView(Glyph_ptr fig, FigureStyle_ptr fs) {
    fig_ = Glyph::_duplicate(fig);
    figstyle_ = FigureStyle::_duplicate(fs);
    tag_ = figstyle_->attach(this);
}

ColoredFigView::~ColoredFigView() {
    figstyle_->detach(tag_);
}

void ColoredFigView::update() {
    fig_->need_redraw();
}
