/*
 * simpleview.cxx
 */

#include <Fresco/defs/fresco.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/widgets.h>
#include <stdio.h>
#include "ftutorial.h"

class SimpleView : public FrescoObjectImpl {
public:
    SimpleView(Adjustment_ptr adj) {
	adj_ = Adjustment::_duplicate(adj);
	tag_ = adj_->attach(this);
    }
    ~SimpleView() {
	adj_->detach(tag_);
    }
    void update() {
	Adjustment::Settings s;
	adj_->get_settings(s);
	printf("%3.2f\n", s.cur_lower);
    }    
protected:
    Adjustment_var adj_;
    Tag tag_;
};

int main(int argc, char** argv) {
    FrescoContext* fc = FrescoLib::init_default_context(
	"FTutorial", argc, argv
    );
    WidgetKit_var widgets = Fresco_resolve_object(WidgetKit, "widgets_impl");
    LayoutKit_var layouts = Fresco_resolve_object(LayoutKit, "layouts_impl");

    Adjustment_var adj = widgets->bounded_float(0.0f, 50.0f, 0.0f);
    FrescoObject_var simple = new SimpleView(adj);
    Viewer_var scrollbar = widgets->scroll_bar(X_axis, adj);
    FrescoLib::run(scrollbar, Glyph_var(layouts->hfixed(scrollbar, Coord(140))));
    CORBA::release(fc);
    return 0;
}
 
