//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

//
// FigureKit -- structured graphics
//

#include "figimpls.h"
#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fcontext.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/dload.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/polyglyph.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/styles.h>
#include <Fresco/defs_impl/threads.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/Ox/exchange.h>
#include <string.h>

implementList(VertexList,Vertex)

class FigureStyleImpl;

Ox_implementation_decl(FigureKitImpl,FigureKit,SharedFrescoObjectImpl)
public:
    FigureKitImpl();
    ~FigureKitImpl();

    //+ FigureKit::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* FigureKit */
    FigureStyle_ptr default_style();
    FigureStyle_ptr new_style(XfStyleContext_ptr parent);
    XfGlyph_ptr figure_root(XfGlyph_ptr child);
    XfGlyph_ptr image(XfRaster_ptr r);
    XfGlyph_ptr stencil(FigureStyle_ptr s, XfRaster_ptr r);
    XfGlyph_ptr label(FigureStyle_ptr s, XfCharString_ptr str);
    XfGlyph_ptr point(FigureStyle_ptr s, XfCoord x, XfCoord y);
    XfGlyph_ptr line(FigureStyle_ptr s, XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1);
    XfGlyph_ptr rectangle(FigureKit::Mode m, FigureStyle_ptr s, XfCoord left, XfCoord bottom, XfCoord right, XfCoord top);
    XfGlyph_ptr circle(FigureKit::Mode m, FigureStyle_ptr s, XfCoord x, XfCoord y, XfCoord r);
    XfGlyph_ptr ellipse(FigureKit::Mode m, FigureStyle_ptr s, XfCoord x, XfCoord y, XfCoord r1, XfCoord r2);
    XfGlyph_ptr open_bspline(FigureKit::Mode m, FigureStyle_ptr s, const FigureKit::Vertices& v);
    XfGlyph_ptr closed_bspline(FigureKit::Mode m, FigureStyle_ptr s, const FigureKit::Vertices& v);
    XfGlyph_ptr multiline(FigureKit::Mode m, FigureStyle_ptr s, const FigureKit::Vertices& v);
    XfGlyph_ptr polygon(FigureKit::Mode m, FigureStyle_ptr s, const FigureKit::Vertices& v);
    XfGlyph_ptr fitter(XfGlyph_ptr g);
    XfGlyph_ptr group();
    Figure_ptr general_figure(FigureKit::Mode m, FigureStyle_ptr s, Boolean closed, Boolean curved);
    //+
protected:
    FigureStyle_var style_;
};

Ox_implementation_decl(FigureStyleImpl,FigureStyle,SharedStyleImpl)
public:
    FigureStyleImpl();
    ~FigureStyleImpl();

    //+ FigureStyle::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfStyleContext */
    XfStyleContext_ptr new_style();
    XfStyleContext_ptr parent_style();
    void link_parent(XfStyleContext_ptr parent);
    void unlink_parent();
    XfTag link_child(XfStyleContext_ptr child);
    void unlink_child(XfTag link_tag);
    void merge(XfStyleContext_ptr s);
    XfCharString_ptr context_name();
    void context_name(XfCharString_ptr _p);
    void alias(XfCharString_ptr s);
    Boolean is_on(XfCharString_ptr n);
    XfStyleValue_ptr bind(XfCharString_ptr n);
    void unbind(XfCharString_ptr n);
    XfStyleValue_ptr resolve(XfCharString_ptr n);
    XfStyleValue_ptr resolve_wildcard(XfCharString_ptr n, XfStyleContext_ptr start);
    Long match(XfCharString_ptr n);
    void visit_aliases(XfStyleVisitor_ptr v);
    void visit_attributes(XfStyleVisitor_ptr v);
    void visit_styles(XfStyleVisitor_ptr v);
    void lock();
    void unlock();
    /* FigureStyle */
    XfColor_ptr background();
    void background(XfColor_ptr _p);
    XfBrush_ptr brush_attr();
    void brush_attr(XfBrush_ptr _p);
    XfFont_ptr font_attr();
    void font_attr(XfFont_ptr _p);
    XfColor_ptr foreground();
    void foreground(XfColor_ptr _p);
    //+

    DrawingKit_var drawing_kit() {
	return FrescoLib::current_context()->default_drawing_kit();
    }
private:
    Color_var background_;
    Brush_var brush_;
    Font_var font_;
    Color_var foreground_;
};

dload(FigureKitImpl)

FigureKitImpl::FigureKitImpl() {
    FrescoContext* fc = FrescoLib::current_context();
    DrawingKit_var d = fc->default_drawing_kit();
    style_ = new_style(StyleContext_var(d->style()));
}

FigureKitImpl::~FigureKitImpl() { }

Ox_implementation_body(FigureKitImpl)

//+ FigureKitImpl(XfFrescoObject::=exch_.)
XfTag FigureKitImpl::attach(XfFrescoObject_ptr observer) {
    return exch_.attach(observer);
}
void FigureKitImpl::detach(XfTag attach_tag) {
    exch_.detach(attach_tag);
}
void FigureKitImpl::disconnect() {
    exch_.disconnect();
}
void FigureKitImpl::notify_observers() {
    exch_.notify_observers();
}
void FigureKitImpl::update() {
    exch_.update();
}
//+

//+ FigureKitImpl(FigureKit::default_style)
FigureStyle_ptr FigureKitImpl::default_style() {
    return FigureStyle::_duplicate(style_);
}

//+ FigureKitImpl(FigureKit::new_style)
FigureStyle_ptr FigureKitImpl::new_style(XfStyleContext_ptr parent) {
    FigureStyleImpl* fs = new FigureStyleImpl;
    fs->link_parent(parent);
    return fs;
}

//+ FigureKitImpl(FigureKit::figure_root)
XfGlyph_ptr FigureKitImpl::figure_root(XfGlyph_ptr child) {
    Glyph_ptr g = new TransformAllocator(
        Alignment(0.5), Alignment(0.5), Alignment(0.5), Alignment(0.5), Alignment(0.5), Alignment(0.5)
    );
    g->body(child);
    return g;
}

//+ FigureKitImpl(FigureKit::image)
XfGlyph_ptr FigureKitImpl::image(XfRaster_ptr r) {
    return new FigureImage(r);
}
 
//+ FigureKitImpl(FigureKit::stencil)
XfGlyph_ptr FigureKitImpl::stencil(FigureStyle_ptr s, XfRaster_ptr r) {
    return new FigureStencil(s, r);
}

//+ FigureKitImpl(FigureKit::label)
XfGlyph_ptr FigureKitImpl::label(FigureStyle_ptr s, XfCharString_ptr str) {
    return new FigureLabel(s, str);
}

//+ FigureKitImpl(FigureKit::point)
XfGlyph_ptr FigureKitImpl::point(FigureStyle_ptr s, XfCoord x, XfCoord y) {
    return line(s, x, y, x, y);
}

//+ FigureKitImpl(FigureKit::line)
XfGlyph_ptr FigureKitImpl::line(FigureStyle_ptr s, XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1) {
    Figure* f = new FigureImpl(FigureKit::stroke, s, false, false, 2);
    f->add_point(x0, y0);
    f->add_point(x1, y1);
    return f;
}

//+ FigureKitImpl(FigureKit::rectangle)
XfGlyph_ptr FigureKitImpl::rectangle(FigureKit::Mode m, FigureStyle_ptr s, XfCoord left, XfCoord bottom, XfCoord right, XfCoord top) {
    Figure* f = new FigureImpl(m, s, true, false, 4);
    f->add_point(left, bottom);
    f->add_point(left, top);
    f->add_point(right, top);
    f->add_point(right, bottom);
    return f;
}

static const float magic = 0.5522847498307934f; // 4/3 * (sqrt(2) - 1)

//+ FigureKitImpl(FigureKit::circle)
XfGlyph_ptr FigureKitImpl::circle(FigureKit::Mode m, FigureStyle_ptr s, XfCoord x, XfCoord y, XfCoord r) {
    float r0 = magic * r;

    Figure* f = new FigureImpl(m, s, true, true, 25);
    f->add_point(x + r, y);
    f->add_curve(x, y - r, x + r, y - r0, x + r0, y - r);
    f->add_curve(x - r, y, x - r0, y - r, x - r, y - r0);
    f->add_curve(x, y + r, x - r, y + r0, x - r0, y + r);
    f->add_curve(x + r, y, x + r0, y + r, x + r, y + r0);
    return f;
}

static const float p0 = 1.00000000f;
static const float p1 = 0.89657547f;   // cos 30 * sqrt(1 + tan 15 * tan 15)
static const float p2 = 0.70710678f;   // cos 45
static const float p3 = 0.51763809f;   // cos 60 * sqrt(1 + tan 15 * tan 15)
static const float p4 = 0.26794919f;   // tan 15

//+ FigureKitImpl(FigureKit::ellipse)
XfGlyph_ptr FigureKitImpl::ellipse(FigureKit::Mode m, FigureStyle_ptr s, XfCoord x, XfCoord y, XfCoord r1, XfCoord r2) {
    float px0 = p0 * r1, py0 = p0 * r2;
    float px1 = p1 * r1, py1 = p1 * r2;
    float px2 = p2 * r1, py2 = p2 * r2;
    float px3 = p3 * r1, py3 = p3 * r2;
    float px4 = p4 * r1, py4 = p4 * r2;

    Figure* f = new FigureImpl(m, s, true, true, 25);
    f->add_point(x + r1, y);
    f->add_curve(x + px2, y + py2, x + px0, y + py4, x + px1, y + py3);
    f->add_curve(x, y + r2, x + px3, y + py1, x + px4, y + py0);
    f->add_curve(x - px2, y + py2, x - px4, y + py0, x - px3, y + py1);
    f->add_curve(x - r1, y, x - px1, y + py3, x - px0, y + py4);
    f->add_curve(x - px2, y - py2, x - px0, y - py4, x - px1, y - py3);
    f->add_curve(x, y - r2, x - px3, y - py1, x - px4, y - py0);
    f->add_curve(x + px2, y - py2, x + px4, y - py0, x + px3, y - py1);
    f->add_curve(x + r1, y, x + px1, y - py3, x + px0, y - py4);
    return f;
}

//+ FigureKitImpl(FigureKit::open_bspline)
XfGlyph_ptr FigureKitImpl::open_bspline(FigureKit::Mode m, FigureStyle_ptr s, const FigureKit::Vertices& v) {
    long n = v._length;
    Vertex* vv = v._buffer;
    Figure* f = new FigureImpl(m, s, false, true, (n + 2) * 3 + 1);
    f->Bspline_move_to(vv[0].x, vv[0].y, vv[0].x, vv[0].y, vv[0].x, vv[0].y);
    f->Bspline_curve_to(vv[0].x, vv[0].y, vv[0].x, vv[0].y, vv[1].x, vv[1].y);
    for (long i = 1; i < n - 1; ++i) {
        f->Bspline_curve_to(
	    vv[i].x, vv[i].y, vv[i-1].x, vv[i-1].y, vv[i+1].x, vv[i+1].y
	);
    }
    f->Bspline_curve_to(
	vv[n-1].x, vv[n-1].y, vv[n-2].x, vv[n-2].y, vv[n-1].x, vv[n-1].y
    );
    f->Bspline_curve_to(
	vv[n-1].x, vv[n-1].y, vv[n-1].x, vv[n-1].y, vv[n-1].x, vv[n-1].y
    );
    return f;
}

//+ FigureKitImpl(FigureKit::closed_bspline)
XfGlyph_ptr FigureKitImpl::closed_bspline(FigureKit::Mode m, FigureStyle_ptr s, const FigureKit::Vertices& v) {
    long n = v._length;
    Vertex* vv = v._buffer;
    Figure* f = new FigureImpl(m, s, true, true, n * 3 + 1);
    f->Bspline_move_to(
	vv[0].x, vv[0].y, vv[n-1].x, vv[n-1].y, vv[1].x, vv[1].y
    );
    for (long i = 1; i < n - 1; ++i) {
        f->Bspline_curve_to(
	    vv[i].x, vv[i].y, vv[i-1].x, vv[i-1].y, vv[i+1].x, vv[i+1].y
	);
    }
    f->Bspline_curve_to(
	vv[n-1].x, vv[n-1].y, vv[n-2].x, vv[n-2].y, vv[0].x, vv[0].y
    );
    f->Bspline_curve_to(
	vv[0].x, vv[0].y, vv[n-1].x, vv[n-1].y, vv[1].x, vv[1].y
    );
    return f;
}

//+ FigureKitImpl(FigureKit::multiline)
XfGlyph_ptr FigureKitImpl::multiline(FigureKit::Mode m, FigureStyle_ptr s, const FigureKit::Vertices& v) {
    long n = v._length;
    Vertex* vv = v._buffer;
    Figure* f = new FigureImpl(m, s, false, false, n);
    f->add_point(vv[0].x, vv[0].y);
    for (long i = 1; i < n; ++i) {
        f->add_point(vv[i].x, vv[i].y);
    }
    return f;
}

//+ FigureKitImpl(FigureKit::polygon)
XfGlyph_ptr FigureKitImpl::polygon(FigureKit::Mode m, FigureStyle_ptr s, const FigureKit::Vertices& v) {
    long n = v._length;
    Vertex* vv = v._buffer;
    Figure* f = new FigureImpl(m, s, true, false, n);
    f->add_point(vv[0].x, vv[0].y);
    for (long i = 1; i < n; ++i) {
        f->add_point(vv[i].x, vv[i].y);
    }
    return f;
}

//+ FigureKitImpl(FigureKit::fitter)
XfGlyph_ptr FigureKitImpl::fitter(XfGlyph_ptr g) {
    /* unimplemented */
    return g;
}

//+ FigureKitImpl(FigureKit::group)
XfGlyph_ptr FigureKitImpl::group() {
    return new PolyFigure;
}

//+ FigureKitImpl(FigureKit::general_figure)
Figure_ptr FigureKitImpl::general_figure(FigureKit::Mode m, FigureStyle_ptr s, Boolean closed, Boolean curved) {
    return new FigureImpl(m, s, closed, curved);
}

/* class FigureImpl */

FigureImpl::FigureImpl(
    FigureKit::Mode mode, FigureStyle_ptr style,
    Boolean closed, Boolean curved, Long coords
) {
    mode_ = mode;
    tx_ = new TransformImpl;
    style_ = FigureStyle::_duplicate(style);
    closed_ = closed;
    curved_ = curved;
    v_ = new VertexList(coords);
}

FigureImpl::FigureImpl(FigureImpl* f) {
    mode_ = f->mode_;
    tx_ = new TransformImpl;
    tx_->copy(f->tx_);
    style_ = FigureStyle::_duplicate(f->style_);
    closed_ = f->closed_;
    curved_ = f->curved_;
    v_ = new VertexList(f->v_->count());
    for (Long i = 0; i < f->v_->count(); i++) {
	v_->append(f->v_->item_ref(i));
    }
    vmin_ = f->vmin_;
    vmax_ = f->vmax_;
}

FigureImpl::~FigureImpl() {
    CORBA::release(tx_);
    CORBA::release(style_);
    delete v_;
}

OxExchange* FigureImpl::_exchange() { return GlyphImpl::_exchange(); }

//+ FigureImpl(XfFrescoObject::=GlyphImpl::)
XfTag FigureImpl::attach(XfFrescoObject_ptr observer) {
    return GlyphImpl::attach(observer);
}
void FigureImpl::detach(XfTag attach_tag) {
    GlyphImpl::detach(attach_tag);
}
void FigureImpl::disconnect() {
    GlyphImpl::disconnect();
}
void FigureImpl::notify_observers() {
    GlyphImpl::notify_observers();
}
void FigureImpl::update() {
    GlyphImpl::update();
}
//+

XfGlyph_ptr FigureImpl::clone_glyph() {
    return GlyphImpl::clone_glyph();
}
XfStyleContext_ptr FigureImpl::style() {
    return GlyphImpl::style();
}
void FigureImpl::style(XfStyleContext_ptr _p) {
    GlyphImpl::style(_p);
}
void FigureImpl::shape(XfRegion_ptr r) {
    GlyphImpl::shape(r);
}
void FigureImpl::traverse(XfGlyphTraversal_ptr t) {
    GlyphImpl::traverse(t);
}
XfGlyph_ptr FigureImpl::body() {
    return GlyphImpl::body();
}
void FigureImpl::body(XfGlyph_ptr _p) {
    GlyphImpl::body(_p);
}
void FigureImpl::append(XfGlyph_ptr g) {
    GlyphImpl::append(g);
}
void FigureImpl::prepend(XfGlyph_ptr g) {
    GlyphImpl::prepend(g);
}
XfTag FigureImpl::add_parent(XfGlyphOffset_ptr parent_offset) {
    return GlyphImpl::add_parent(parent_offset);
}
void FigureImpl::remove_parent(XfTag add_tag) {
    GlyphImpl::remove_parent(add_tag);
}
XfGlyphOffset_ptr FigureImpl::first_child_offset() {
    return GlyphImpl::first_child_offset();
}
XfGlyphOffset_ptr FigureImpl::last_child_offset() {
    return GlyphImpl::last_child_offset();
}
void FigureImpl::parent_offsets(XfGlyph::OffsetSeq& parents) {
    GlyphImpl::parent_offsets(parents);
}
void FigureImpl::allocations(XfGlyph::AllocationInfoSeq& a) {
    GlyphImpl::allocations(a);
}
void FigureImpl::need_redraw() {
    GlyphImpl::need_redraw();
}
void FigureImpl::need_redraw_region(XfRegion_ptr r) {
    GlyphImpl::need_redraw_region(r);
}
void FigureImpl::need_resize() {
    GlyphImpl::need_resize();
}
Boolean FigureImpl::restore_trail(XfGlyphTraversal_ptr t) {
    return GlyphImpl::restore_trail(t);
}

//+ FigureImpl(Figure::add_point)
void FigureImpl::add_point(XfCoord x, XfCoord y) {
    if (v_->count() == 0) {
	vmin_.x = x;
	vmax_.x = x;
	vmin_.y = y;
	vmax_.y = y;
	vmin_.z = Coord(0);
	vmax_.z = Coord(0);
    } else {
        vmin_.x = Math::min(vmin_.x, x);
        vmax_.x = Math::max(vmax_.x, x);
        vmin_.y = Math::min(vmin_.y, y);
        vmax_.y = Math::max(vmax_.y, y);
    }
    Vertex v;
    v.x = x;
    v.y = y;
    v_->append(v);
}

//+ FigureImpl(Figure::add_curve)
void FigureImpl::add_curve(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2) {
    add_point(x1, y1);
    add_point(x2, y2);
    add_point(x, y);
}

//+ FigureImpl(Figure::Bspline_move_to)
void FigureImpl::Bspline_move_to(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2) {
    Coord p1x = (x + x + x1) / 3;
    Coord p1y = (y + y + y1) / 3;
    Coord p2x = (x + x + x2) / 3;
    Coord p2y = (y + y + y2) / 3;
    add_point((p1x + p2x) / 2, (p1y + p2y) / 2);
}

//+ FigureImpl(Figure::Bspline_curve_to)
void FigureImpl::Bspline_curve_to(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2) {
    Coord p1x = (x + x + x1) / 3;
    Coord p1y = (y + y + y1) / 3;
    Coord p2x = (x + x + x2) / 3;
    Coord p2y = (y + y + y2) / 3;
    Coord p3x = (x1 + x1 + x) / 3;
    Coord p3y = (y1 + y1 + y) / 3;
    add_curve((p1x + p2x) / 2, (p1y + p2y) / 2, p3x, p3y, p1x, p1y);
}

//+ FigureImpl(XfGlyph::transformation)
XfTransform_ptr FigureImpl::transformation() {
    return Transform::_duplicate(tx_);
}

//+ FigureImpl(XfGlyph::request)
void FigureImpl::request(XfGlyph::Requisition& r) {
    if (v_->count() > 0) {
	RegionImpl region;
	region.defined_ = true;
	region.lower_ = vmin_;
	region.upper_ = vmax_;
	region.xalign_ = region.yalign_ = region.zalign_ = Coord(0);
	region.apply_transform(tx_);
	Coord x_lead = -region.lower_.x, x_trail = region.upper_.x;
	Coord y_lead = -region.lower_.y, y_trail = region.upper_.y;
	Coord z_lead = -region.lower_.z, z_trail = region.upper_.z;
	GlyphImpl::require_lead_trail(
	    r.x, x_lead, x_lead, x_lead, x_trail, x_trail, x_trail
	);
	GlyphImpl::require_lead_trail(
	    r.y, y_lead, y_lead, y_lead, y_trail, y_trail, y_trail
	);
	GlyphImpl::require_lead_trail(
	    r.z, z_lead, z_lead, z_lead, z_trail, z_trail, z_trail
	);
    } else {
	r.x.defined = false;
	r.y.defined = false;
	r.z.defined = false;
    }
}

//+ FigureImpl(XfGlyph::extension)
void FigureImpl::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    if (v_->count() > 0) {
	RegionImpl region;
	region.defined_ = true;
	region.lower_ = vmin_;
	region.upper_ = vmax_;
	region.xalign_ = region.yalign_ = region.zalign_ = Coord(0);
	TransformImpl t;
        if (is_not_nil(a.transformation)) {
	    t.copy(a.transformation);
        }
	t.premultiply(tx_);
	region.apply_transform(&t);
	if (mode_ == FigureKit::stroke || mode_ == FigureKit::fill_stroke) {
	    Coord w = Coord(1);
	    Brush_var b = style_->brush_attr();
	    if (is_not_nil(b)) {
		Coord bw = b->width();
		if (!Math::equal(bw, float(0), float(1e-2))) {
		    w = bw;
		}
	    }
	    region.lower_.x -= w; region.upper_.x += w;
	    region.lower_.y -= w; region.upper_.y += w;
	    region.lower_.z -= w; region.upper_.z += w;
	}
	r->merge_union(&region);
    }
}

//+ FigureImpl(XfGlyph::draw)
void FigureImpl::draw(XfGlyphTraversal_ptr t) {
    if (v_->count() > 0) {
	Painter_var p = t->current_painter();
	p->push_matrix();
	p->premultiply(tx_);

	/*
	 * Should do bounding box culling here.  The bounding box is
	 * described in object coordinates by vmin_, vmax_.
	 * Then we should just do p->is_visible(bbox) for the test.
	 */

        p->begin_path();
        p->move_to(v_->item_ref(0).x, v_->item_ref(0).y);
        if (curved_) {
            for (long i = 1; i < v_->count(); i += 3) {
		long i1 = i + 1;
		long i2 = i1 + 1;
                p->curve_to(
                    v_->item_ref(i2).x, v_->item_ref(i2).y,
		    v_->item_ref(i).x, v_->item_ref(i).y,
		    v_->item_ref(i1).x, v_->item_ref(i1).y
                );
            }
        } else {
            for (Long i = 1; i < v_->count(); ++i) {
                p->line_to(v_->item_ref(i).x, v_->item_ref(i).y);
            }
        }
        if (closed_) {
            p->close_path();
        }
	if (mode_ == FigureKit::stroke || mode_ == FigureKit::fill_stroke) {
	    p->current_brush(Brush_var(style_->brush_attr()));
	}
	if (mode_ == FigureKit::fill || mode_ == FigureKit::stroke) {
	    p->current_color(Color_var(style_->foreground()));
	}
	switch (mode_) {
	case FigureKit::fill:
	    p->fill();
	    break;
	case FigureKit::stroke:
	    p->stroke();
	    break;
	case FigureKit::fill_stroke:
	    p->current_color(Color_var(style_->background()));
	    p->fill();
	    p->current_color(Color_var(style_->foreground()));
	    p->stroke();
	    break;
	}
	p->pop_matrix();
    }
}

/*
 * Picking just does a bounding box test for now.
 */

//+ FigureImpl(XfGlyph::pick)
void FigureImpl::pick(XfGlyphTraversal_ptr t) {
    if (v_->count() > 0) {
	Painter_var p = t->current_painter();
	p->push_matrix();
	p->premultiply(tx_);
	RegionImpl bbox;
	bbox.defined_ = true;
	bbox.lower_ = vmin_;
	bbox.upper_ = vmax_;
	bbox.xalign_ = Alignment(0);
	bbox.yalign_ = Alignment(0);
	bbox.zalign_ = Alignment(0);
	if (p->is_visible(&bbox)) {
	    t->hit();
	}
	p->pop_matrix();
    }
}

/*
 * Reset the figure's list of vertices
 */

//+ FigureImpl(Figure::reset)
void FigureImpl::reset() {
    v_->remove_all();
}

//+ FigureImpl(Figure::copy_figure)
Figure_ptr FigureImpl::copy_figure() {
    return new FigureImpl(this);
}

/* class PolyFigure */

PolyFigure::PolyFigure() {
    tx_ = new TransformImpl;
    bbox_ = new RegionImpl;
}

PolyFigure::PolyFigure(PolyFigure* pf) {
    tx_ = new TransformImpl;
    tx_->copy(pf->tx_);
    bbox_ = new RegionImpl;
    bbox_->defined_ = pf->bbox_->defined_;
    if (bbox_->defined_) {
        bbox_->copy(pf->bbox_);
    }
}

PolyFigure::~PolyFigure() {
    CORBA::release(tx_);
    CORBA::release(bbox_);
}

void PolyFigure::update_bbox() {
    if (!bbox_->defined_) {
	Long n = children_.count();
	if (n > 0) {
	    Glyph::AllocationInfo a;
	    RegionImpl region;
	    for (Long i = 0; i < n; i++) {
		children_.item(i)->child_->extension(a, &region);
		bbox_->merge_union(&region);
	    }
	}
    }
}

void PolyFigure::child_allocate(Long, Glyph::AllocationInfo& a) {
    a.allocation = nil;
    a.transformation->premultiply(tx_);
}

//+ PolyFigure(XfGlyph::request)
void PolyFigure::request(XfGlyph::Requisition& r) {
    if (children_.count() != 0) {
	Glyph::AllocationInfo a;
	RegionImpl region;
	extension(a, &region);
	Coord x_lead = -region.lower_.x, x_trail = region.upper_.x;
	Coord y_lead = -region.lower_.y, y_trail = region.upper_.y;
	Coord z_lead = -region.lower_.z, z_trail = region.upper_.z;
	GlyphImpl::require_lead_trail(
	    r.x, x_lead, x_lead, x_lead, x_trail, x_trail, x_trail
	);
	GlyphImpl::require_lead_trail(
	    r.y, y_lead, y_lead, y_lead, y_trail, y_trail, y_trail
	);
	GlyphImpl::require_lead_trail(
	    r.z, z_lead, z_lead, z_lead, z_trail, z_trail, z_trail
	);
    } else {
	r.x.defined = false;
	r.y.defined = false;
	r.z.defined = false;
    }
}

// If given transform is nil, PolyFigure::extension considers this a flag
// meaning "ok to calculate an imprecise extension" and thus bounding box
// can be used.
    
//+ PolyFigure(XfGlyph::extension)
void PolyFigure::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    Glyph::AllocationInfo ga;
    ga.transformation = new TransformImpl;
    if (is_not_nil(a.transformation)) {
	Glyph::AllocationInfo ga;
	ga.transformation = new TransformImpl;
	ga.transformation->copy(a.transformation);
	ga.transformation->premultiply(tx_);
	ga.allocation = a.allocation;
	ga.damaged = a.damaged;
	for (Long i = 0; i < children_.count(); i++) {
	    children_.item(i)->child_->extension(ga, r);
	}
    } else {
	RegionImpl region;
	update_bbox();
	if (bbox_->defined_) {
	    region.copy(bbox_);
	    region.apply_transform(tx_);
	    r->merge_union(&region);
	}
    }
}

//+ PolyFigure(XfGlyph::traverse)
void PolyFigure::traverse(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    if (is_not_nil(p)) {
	RegionImpl region;
	update_bbox();
	if (bbox_->defined_) {
	    region.copy(bbox_);
	    region.apply_transform(tx_);
	    if (!p->is_visible(&region)) {
		return;
	    }
	}
    }
    Boolean tx = is_not_nil(p) && !tx_->identity();
    if (tx) {
	p->push_matrix();
	p->premultiply(tx_);
    }
    GlyphTraversal::Operation op = t->op();
    if (op == GlyphTraversal::pick_top || op == GlyphTraversal::pick_any) {
	for (Long i = children_.count() - 1; i >= 0; i--) {
	    t->traverse_child(children_.item(i), nil);
	    if (is_not_nil(GlyphTraversal_var(t->picked()))) {
		break;
	    }
	}
    } else {
	for (ListItr(PolyGlyphOffsetList) i(children_); i.more(); i.next()) {
	    t->traverse_child(i.cur(), nil);
	}
    }
    if (tx) {
	p->pop_matrix();
    }
}

//+ PolyFigure(XfGlyph::transformation)
XfTransform_ptr PolyFigure::transformation() {
    return Transform::_duplicate(tx_);
}

//+ PolyFigure(XfGlyph::need_redraw)
void PolyFigure::need_redraw() {
    GlyphImpl::need_redraw();
	// Calling GlyphImpl::need_redraw does not allow us to take
	// advantage of bbox_ for damage. However, to do damage with
	// bbox_, we would need to grow the transformed bbox_ to compensate
	// for the brush size of leaves. (In truth, we should do the same for
	// the cull test as well since currently culling will prevent redraw
	// when the outer part of an outer leaf's brush is damaged.)
}

//+ PolyFigure(XfGlyph::need_resize)
void PolyFigure::need_resize() {
    bbox_->defined_ = false;
    PolyGlyph::need_resize();
}

void PolyFigure::visit_trail(Long, GlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Transform_var(p->current_matrix())->premultiply(tx_);
}

/* class FigureRaster */

FigureRaster::FigureRaster(Raster_ptr raster) {
    raster_ = Raster::_duplicate(raster);
}
 
FigureRaster::~FigureRaster () {
}
 
//+ FigureRaster(XfGlyph::request)
void FigureRaster::request(XfGlyph::Requisition& r) {
    Raster::Info info;
    raster_->raster_info(info);
    Coord scale = raster_->scale();
    if (Math::equal(scale, Coord(0), Coord(1e-2))) {
        // !!! this only works for a 72 dpi screen -denis
        scale = Coord(1);
    }
    Coord x_lead = info.origin_x * scale;
    Coord x_trail = (info.columns - info.origin_x) * scale;
    Coord y_lead = info.origin_y * scale;
    Coord y_trail = (info.rows - info.origin_y) * scale;
    GlyphImpl::require_lead_trail(
        r.x, x_lead, x_lead, x_lead, x_trail, x_trail, x_trail
    );
    GlyphImpl::require_lead_trail(
        r.y, y_lead, y_lead, y_lead, y_trail, y_trail, y_trail
    );
}
 
//+ FigureRaster(XfGlyph::extension)
void FigureRaster::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    Raster::Info info;
    raster_->raster_info(info);
    Coord scale = raster_->scale();
    if (Math::equal(scale, Coord(0), Coord(1e-2))) {
        // !!! this only works for a 72 dpi screen -denis
        scale = Coord(1);
    }
    RegionImpl region;
    region.defined_ = true;
    region.lower_.x = -info.origin_x * scale;
    region.lower_.y = -info.origin_y * scale;
    region.lower_.z = Coord(0);
    region.upper_.x = (info.columns - info.origin_x) * scale;
    region.upper_.y = (info.rows - info.origin_y) * scale;
    region.upper_.z = Coord(0);
    r->copy(&region);
}

//+ FigureRaster(XfGlyph::pick) !t
void FigureRaster::pick(XfGlyphTraversal_ptr) {
}

/* class FigureImage */
 
FigureImage::FigureImage(Raster_ptr raster) : FigureRaster(raster) {
}
 
FigureImage::~FigureImage () {
}
 
//+ FigureImage(XfGlyph::draw)
void FigureImage::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    p->image(raster_, Coord(0), Coord(0));
}
 
/* class FigureStencil */
 
FigureStencil::FigureStencil(
    FigureStyle_ptr s, Raster_ptr raster
) : FigureRaster(raster) {
    style_ = FigureStyle::_duplicate(s);
}
 
FigureStencil::~FigureStencil () {
    CORBA::release(style_);
}
 
//+ FigureStencil(XfGlyph::draw)
void FigureStencil::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    p->current_color(style_->foreground());
    p->stencil(raster_, Coord(0), Coord(0));
}
 
/* class FigureLabel */

FigureLabel::FigureLabel(FigureStyle_ptr s, CharString_ptr str) {
    style_ = FigureStyle::_duplicate(s);
    text_ = CharString::_duplicate(str);
    Font_var(style_->font_attr())->string_info(text_, info_);
}

FigureLabel::FigureLabel(FigureLabel* fl) {
    style_ = FigureStyle::_duplicate(fl->style_);
    text_ = CharString::_duplicate(fl->text_);
    info_ = fl->info_;
}

FigureLabel::~FigureLabel() {
    CORBA::release(style_);
}

//+ FigureLabel(XfGlyph::request)
void FigureLabel::request(XfGlyph::Requisition& r) {
    Coord h = info_.font_ascent + info_.font_descent;
    Alignment a;
    if (Math::equal(h, Coord(0), Coord(1e-2))) {
	a = Alignment(0);
    } else {
	a = info_.font_descent / h;
    }
    GlyphImpl::require(r.x, info_.width, Coord(0), Coord(0), Alignment(0));
    GlyphImpl::require(r.y, h, Coord(0), Coord(0), a);
}

//+ FigureLabel(XfGlyph::extension)
void FigureLabel::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    RegionImpl region;
    region.defined_ = true;
    region.lower_.x = Coord(0);
    region.lower_.y = -info_.font_descent;
    region.lower_.z = Coord(0);
    region.upper_.x = info_.width;
    region.upper_.y = info_.font_ascent;
    region.upper_.z = Coord(0);
    r->copy(&region);
}

//+ FigureLabel(XfGlyph::draw)
void FigureLabel::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Region_var a = t->allocation();
    p->current_color(Color_var(style_->foreground()));
    Font_var f = style_->font_attr();
    p->current_font(f);
    Vertex v;
    if (is_not_nil(a)) {
	a->origin(v);
    } else {
	v.x = Coord(0);
	v.y = info_.font_descent;
	v.z = Coord(0);
    }
    Font::Info info;
    CharStringBuffer buf(text_);
    const char* cur = buf.string();
    const char* end = &cur[buf.length()];
    for (; cur < end; cur++) {
	f->char_info(*cur, info);
	p->character(*cur, info.width, v.x, v.y);
	v.x += info.width;
    }
}

//+ FigureLabel(XfGlyph::pick)
void FigureLabel::pick(XfGlyphTraversal_ptr) { }

/* class FigureStyleImpl */

FigureStyleImpl::FigureStyleImpl() : exch_(this) {
    exch_.style_ = this;
}

FigureStyleImpl::~FigureStyleImpl() { }

Ox_implementation_body(FigureStyleImpl)

//+ FigureStyleImpl(XfFrescoObject::=exch_.)
XfTag FigureStyleImpl::attach(XfFrescoObject_ptr observer) {
    return exch_.attach(observer);
}
void FigureStyleImpl::detach(XfTag attach_tag) {
    exch_.detach(attach_tag);
}
void FigureStyleImpl::disconnect() {
    exch_.disconnect();
}
void FigureStyleImpl::notify_observers() {
    exch_.notify_observers();
}
void FigureStyleImpl::update() {
    exch_.update();
}
//+

//+ FigureStyleImpl(XfStyleContext::=exch_.)
XfStyleContext_ptr FigureStyleImpl::new_style() {
    return exch_.new_style();
}
XfStyleContext_ptr FigureStyleImpl::parent_style() {
    return exch_.parent_style();
}
void FigureStyleImpl::link_parent(XfStyleContext_ptr parent) {
    exch_.link_parent(parent);
}
void FigureStyleImpl::unlink_parent() {
    exch_.unlink_parent();
}
XfTag FigureStyleImpl::link_child(XfStyleContext_ptr child) {
    return exch_.link_child(child);
}
void FigureStyleImpl::unlink_child(XfTag link_tag) {
    exch_.unlink_child(link_tag);
}
void FigureStyleImpl::merge(XfStyleContext_ptr s) {
    exch_.merge(s);
}
XfCharString_ptr FigureStyleImpl::context_name() {
    return exch_.context_name();
}
void FigureStyleImpl::context_name(XfCharString_ptr _p) {
    exch_.context_name(_p);
}
void FigureStyleImpl::alias(XfCharString_ptr s) {
    exch_.alias(s);
}
Boolean FigureStyleImpl::is_on(XfCharString_ptr n) {
    return exch_.is_on(n);
}
XfStyleValue_ptr FigureStyleImpl::bind(XfCharString_ptr n) {
    return exch_.bind(n);
}
void FigureStyleImpl::unbind(XfCharString_ptr n) {
    exch_.unbind(n);
}
XfStyleValue_ptr FigureStyleImpl::resolve(XfCharString_ptr n) {
    return exch_.resolve(n);
}
XfStyleValue_ptr FigureStyleImpl::resolve_wildcard(XfCharString_ptr n, XfStyleContext_ptr start) {
    return exch_.resolve_wildcard(n, start);
}
Long FigureStyleImpl::match(XfCharString_ptr n) {
    return exch_.match(n);
}
void FigureStyleImpl::visit_aliases(XfStyleVisitor_ptr v) {
    exch_.visit_aliases(v);
}
void FigureStyleImpl::visit_attributes(XfStyleVisitor_ptr v) {
    exch_.visit_attributes(v);
}
void FigureStyleImpl::visit_styles(XfStyleVisitor_ptr v) {
    exch_.visit_styles(v);
}
void FigureStyleImpl::lock() {
    exch_.lock();
}
void FigureStyleImpl::unlock() {
    exch_.unlock();
}
//+

//+ FigureStyleImpl(FigureStyle::brush_attr=b)
void FigureStyleImpl::brush_attr(XfBrush_ptr b) {
    brush_ = Brush::_duplicate(b);
}

//+ FigureStyleImpl(FigureStyle::brush_attr?)
XfBrush_ptr FigureStyleImpl::brush_attr() {
    return Brush::_duplicate(brush_);
}

//+ FigureStyleImpl(FigureStyle::font_attr=f)
void FigureStyleImpl::font_attr(XfFont_ptr f) {
    font_ = Font::_duplicate(f);
}

//+ FigureStyleImpl(FigureStyle::font_attr?)
XfFont_ptr FigureStyleImpl::font_attr() {
    if (is_nil(font_)) {
	DrawingKit_var d = drawing_kit();
	font_ = d->default_font(this);
	if (is_nil(font_)) {
	    font_ = d->find_font(FrescoLib::tmp_string_ref("fixed"));
	}
    }
    return Font::_duplicate(font_);
}

//+ FigureStyleImpl(FigureStyle::foreground=c)
void FigureStyleImpl::foreground(XfColor_ptr c) {
    foreground_ = Color::_duplicate(c);
}

//+ FigureStyleImpl(FigureStyle::foreground?)
XfColor_ptr FigureStyleImpl::foreground() {
    if (is_nil(foreground_)) {
	DrawingKit_var d = drawing_kit();
	foreground_ = d->foreground(this);
    }
    return Color::_duplicate(foreground_);
}

//+ FigureStyleImpl(FigureStyle::background=c)
void FigureStyleImpl::background(XfColor_ptr c) {
    background_ = Color::_duplicate(c);
}

//+ FigureStyleImpl(FigureStyle::background?)
XfColor_ptr FigureStyleImpl::background() {
    if (is_nil(background_)) {
	DrawingKit_var d = drawing_kit();
	background_ = d->background(this);
    }
    return Color::_duplicate(background_);
}
