//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//
 
#ifndef Fresco_figures_figures_h
#define Fresco_figures_figures_h
 
#include <Fresco/defs/drawing.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/polyglyph.h>
#include <Fresco/figures/figures.h>

declareList(VertexList,Vertex)

class TransformImpl;

class FigureImpl : _Ix_public Figure, _Ix_public GlyphImpl {
public:
    FigureImpl(
	FigureKit::Mode, FigureStyle_ptr,
        Boolean closed, Boolean curved, Long coords = 0
    );
    FigureImpl(FigureImpl*);
    ~FigureImpl();

    OxExchange* _exchange();

    _Ix_ClassInterfaceOps(Figure)

    //+ Figure::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyph */
    XfGlyph_ptr clone_glyph();
    XfStyleContext_ptr style();
    void style(XfStyleContext_ptr _p);
    XfTransform_ptr transformation();
    void request(XfGlyph::Requisition& r);
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r);
    void shape(XfRegion_ptr r);
    void traverse(XfGlyphTraversal_ptr t);
    void draw(XfGlyphTraversal_ptr t);
    void pick(XfGlyphTraversal_ptr t);
    XfGlyph_ptr body();
    void body(XfGlyph_ptr _p);
    void append(XfGlyph_ptr g);
    void prepend(XfGlyph_ptr g);
    XfTag add_parent(XfGlyphOffset_ptr parent_offset);
    void remove_parent(XfTag add_tag);
    XfGlyphOffset_ptr first_child_offset();
    XfGlyphOffset_ptr last_child_offset();
    void parent_offsets(XfGlyph::OffsetSeq& parents);
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void need_redraw();
    void need_redraw_region(XfRegion_ptr r);
    void need_resize();
    Boolean restore_trail(XfGlyphTraversal_ptr t);
    /* Figure */
    void add_point(XfCoord x, XfCoord y);
    void add_curve(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2);
    void Bspline_move_to(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2);
    void Bspline_curve_to(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2);
    Figure_ptr copy_figure();
    void reset();
    //+

public:
    TransformImpl* tx_;
    FigureStyle_ptr style_;
    VertexList* v_;
    Vertex vmin_;
    Vertex vmax_;
    FigureKit::Mode mode_;
    Boolean closed_ : 1;
    Boolean curved_ : 1;
};

class PolyFigure : public PolyGlyph {
public:
    PolyFigure();
    PolyFigure(PolyFigure*);
    ~PolyFigure();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    XfTransform_ptr transformation(); //+ XfGlyph::transformation
    void need_redraw(); //+ XfGlyph::need_redraw
    void need_resize(); //+ XfGlyph::need_resize
    void visit_trail(Long, GlyphTraversal_ptr t);
    void child_allocate(Long, Glyph::AllocationInfo& a);
 
    static void corners(
	Transform_ptr t, Coord& left, Coord& bottom, Coord& right, Coord& top
    );
protected:
    TransformImpl* tx_;
    RegionImpl* bbox_;

    void update_bbox();
};

class FigureRaster : public GlyphImpl {
public:
    FigureRaster(Raster_ptr);
    ~FigureRaster();
 
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    void pick(XfGlyphTraversal_ptr t); //+ XfGlyph::pick
public:
    Raster_var raster_;
};
 
class FigureImage : public FigureRaster {
public:
    FigureImage(Raster_ptr);
    ~FigureImage();
 
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
};
 
class FigureStencil : public FigureRaster {
public:
    FigureStencil(FigureStyle_ptr, Raster_ptr);
    ~FigureStencil();
 
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
public:
    FigureStyle_ptr style_;
};

class FigureLabel : public GlyphImpl {
public:
    FigureLabel(FigureStyle_ptr, CharString_ptr);
    FigureLabel(FigureLabel*);
    ~FigureLabel();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
    void pick(XfGlyphTraversal_ptr t); //+ XfGlyph::pick
public:
    FigureStyle_ptr style_;
    CharString_var text_;
    Font::Info info_;
};

#endif
