//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_layouts_impl_box_h
#define Fresco_layouts_impl_box_h

#include <Fresco/defs_impl/polyglyph.h>

//
// Box -- PolyGlyph with layout manager
//
// The box "consumes" the given layout manager; that is, it will be
// responsible for destroying the layout manager when the box
// is deallocated.
//

class LayoutManager;

class Box : public PolyGlyph {
public:
    Box(LayoutManager*);
    virtual ~Box();

    void request(Glyph::Requisition& r); //+ XfGlyph::request
    void extension(const Glyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    void need_resize(); //+ XfGlyph::need_resize

    RegionImpl** children_allocations(Region_ptr a);
    void traverse_with_allocation(GlyphTraversal_ptr t, Region_ptr a);
    void traverse_without_allocation(GlyphTraversal_ptr t);
    void child_allocate(Long index, Glyph::AllocationInfo& a);
    void modified();
private:
    static const int static_size_;
    LayoutManager* layout_;
    Boolean requested_;
    Glyph::Requisition requisition_;
};

// BoxAlignElements -- Box that aligns children
 
class BoxAlignElements : public Box {
public:
    BoxAlignElements(LayoutManager*, Axis, Alignment);
    virtual ~BoxAlignElements();

    void append(XfGlyph_ptr g); //+ XfGlyph::append
    void prepend(XfGlyph_ptr g); //+ XfGlyph::prepend
private:
    Axis axis_;
    Alignment alignment_;
};

#endif
