//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include "deck.h"
#include "layout_mgr.h"

// class DeckOffset

DeckOffset::DeckOffset(
    PolyGlyph* parent, Long index, Glyph_ptr child
) : PolyGlyphOffset(parent, index, child) { }

DeckOffset::~DeckOffset() { }

//+ DeckOffset(XfGlyphOffset::allocations)
void DeckOffset::allocations(XfGlyph::AllocationInfoSeq& a) {
    if (index_ == parent_->children_.count() - 1) {
	PolyGlyphOffset::allocations(a);
    }
}

// class DeckImpl

DeckImpl::DeckImpl() { requested_ = false; }
DeckImpl::~DeckImpl() { }

//+ DeckImpl(XfGlyph::request)
void DeckImpl::request(XfGlyph::Requisition& r) {
    if (!requested_) {
	GlyphImpl::init_requisition(requisition_);
	Long n = children_.count();
	if (n > 0) {
	    Glyph::Requisition req[10];
	    Glyph::Requisition* r = children_requests(req, 10);
	    LayoutAlign x(X_axis);
	    x.request(n, r, requisition_);
	    LayoutAlign y(Y_axis);
	    y.request(n, r, requisition_);
	    if (r != req) {
		delete [] r;
	    }
	}
	requested_ = true;
    }
    r = requisition_;
}

//+ DeckImpl(XfGlyph::extension)
void DeckImpl::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    Long n = children_.count();
    if (n > 0) {
	children_.item(n - 1)->child_->extension(a, r);
    }
}

//+ DeckImpl(XfGlyph::traverse)
void DeckImpl::traverse(XfGlyphTraversal_ptr t) {
    Long n = children_.count();
    if (n > 0) {
	t->traverse_child(children_.item(n - 1), Region_var(t->allocation()));
    }
}

PolyGlyphOffset* DeckImpl::new_offset(Long index, Glyph_ptr child) {
    return new DeckOffset(this, index, child);
}

void DeckImpl::modified() {
    requested_ = false;
    need_redraw();
}
