//
// Copyright (c) 1994-95 Firefly Design
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

#ifndef Fresco_layouts_impl_grid_h
#define Fresco_layouts_impl_grid_h

#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/layouts/layouts.h>

// struct Span

struct Span {
    Coord lower_;
    Coord upper_;
    Alignment align_;
};

// class GridOffset

class GridImpl;

class GridOffset : public GlyphOffset {
public:
    GridOffset(GridImpl* parent, Glyph_ptr child, Grid::Index index);
    ~GridOffset();

    Glyph::Requirement requirement(Axis axis);
    GridOffset*& next(Axis axis) { return next_[int(axis)]; }

    OxExchange* _exchange();
    
    //+ XfGlyphOffset::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyphOffset */
    XfGlyph_ptr parent();
    XfGlyph_ptr child();
    XfGlyphOffset_ptr next_child();
    XfGlyphOffset_ptr prev_child();
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void insert(XfGlyph_ptr g);
    void replace(XfGlyph_ptr g);
    void remove();
    void notify();
    void traverse(XfGlyphTraversal_ptr t);
    void visit_trail(XfGlyphTraversal_ptr t);
    void child_allocate(XfGlyph::AllocationInfo& a);
    //+

    Grid::Index index_;

private:
    SharedFrescoObjectImpl object_;
    GridImpl* parent_;
    Glyph_var child_;
    Tag remove_tag_;
    GridOffset* next_[2];
};

// class GridDimension

class GridDimension {
public:
    GridDimension();
    ~GridDimension();

    void initialize(Long count);

    Long count_;
    GridOffset** offsets_;
    Glyph::Requirement* requirements_;
};

// class GridImpl

class GridImpl : public Grid {
public:
    GridImpl(const Index& upper);
    ~GridImpl();

    OxExchange* _exchange();
 
    //+ Grid::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyph */
    XfGlyph_ptr clone_glyph();
    XfStyleContext_ptr style();
    void style(XfStyleContext_ptr _p);
    XfTransform_ptr transformation();
    void request(XfGlyph::Requisition& r);
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r);
    void shape(XfRegion_ptr r);
    void traverse(XfGlyphTraversal_ptr t);
    void draw(XfGlyphTraversal_ptr t);
    void pick(XfGlyphTraversal_ptr t);
    XfGlyph_ptr body();
    void body(XfGlyph_ptr _p);
    void append(XfGlyph_ptr g);
    void prepend(XfGlyph_ptr g);
    XfTag add_parent(XfGlyphOffset_ptr parent_offset);
    void remove_parent(XfTag add_tag);
    XfGlyphOffset_ptr first_child_offset();
    XfGlyphOffset_ptr last_child_offset();
    void parent_offsets(XfGlyph::OffsetSeq& parents);
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void need_redraw();
    void need_redraw_region(XfRegion_ptr r);
    void need_resize();
    Boolean restore_trail(XfGlyphTraversal_ptr t);
    /* Grid */
    void replace(XfGlyph_ptr g, const Grid::Index& i);
    Grid::Index find(XfGlyphTraversal_ptr t);
    void allocate_cell(XfRegion_ptr given, const Grid::Index& i, XfRegion_ptr a);
    void request_range(XfGlyph::Requisition& r, const Grid::Range& a);
    void traverse_range(XfGlyphTraversal_ptr t, const Grid::Range& a);
    Grid::Index find_range(XfGlyphTraversal_ptr t, const Grid::Range& a);
    Grid::Index upper();
    //+

private:
    void cache_request();
    void partial_request(
        Axis axis, Long lower, Long upper, XfGlyph::Requirement& r
    );
    void full_request(Axis axis, Axis direction);
    Span* full_allocate(Axis axis, Region_ptr given);
    void traverse_with_allocation(
        GlyphTraversal_ptr t, Region_ptr given, const Grid::Range& range
    );
    void traverse_without_allocation(
        GlyphTraversal_ptr t, const Grid::Range& range
    );

    SharedFrescoObjectImpl object_;
    GlyphOffsetList parents_;

    GridDimension dimensions_[2];

    Grid::Index cursor_;

    Boolean requested_;
    Requisition requisition_;
};

// class SubGridOffset

class SubGridImpl;

class SubGridOffset : public GlyphOffset {
public:
    SubGridOffset(SubGridImpl* parent, Grid_ptr child, const Grid::Range& range);
    ~SubGridOffset();

    OxExchange* _exchange();

    void request(Glyph::Requisition& r);

    //+ XfGlyphOffset::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyphOffset */
    XfGlyph_ptr parent();
    XfGlyph_ptr child();
    XfGlyphOffset_ptr next_child();
    XfGlyphOffset_ptr prev_child();
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void insert(XfGlyph_ptr g);
    void replace(XfGlyph_ptr g);
    void remove();
    void notify();
    void traverse(XfGlyphTraversal_ptr t);
    void visit_trail(XfGlyphTraversal_ptr t);
    void child_allocate(XfGlyph::AllocationInfo& a);
    //+

protected:
    SharedFrescoObjectImpl object_;
    SubGridImpl* parent_;
    Grid_var child_;
    Tag remove_tag_;
    Grid::Range range_;
};

// class SubGridImpl

class SubGridImpl : public GlyphImpl {
public:
    SubGridImpl(Grid_ptr grid, const Grid::Range& range);

    OxExchange* _exchange();

    void visit_trail(GlyphTraversal_ptr t);
    void child_allocate(Glyph::AllocationInfo&);

    //+ XfGlyph::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyph */
    XfGlyph_ptr clone_glyph();
    XfStyleContext_ptr style();
    void style(XfStyleContext_ptr _p);
    XfTransform_ptr transformation();
    void request(XfGlyph::Requisition& r);
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r);
    void shape(XfRegion_ptr r);
    void traverse(XfGlyphTraversal_ptr t);
    void draw(XfGlyphTraversal_ptr t);
    void pick(XfGlyphTraversal_ptr t);
    XfGlyph_ptr body();
    void body(XfGlyph_ptr _p);
    void append(XfGlyph_ptr g);
    void prepend(XfGlyph_ptr g);
    XfTag add_parent(XfGlyphOffset_ptr parent_offset);
    void remove_parent(XfTag add_tag);
    XfGlyphOffset_ptr first_child_offset();
    XfGlyphOffset_ptr last_child_offset();
    void parent_offsets(XfGlyph::OffsetSeq& parents);
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void need_redraw();
    void need_redraw_region(XfRegion_ptr r);
    void need_resize();
    Boolean restore_trail(XfGlyphTraversal_ptr t);
    //+

private:
    SubGridOffset offset_;
};

#endif
