//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include "box.h"
#include "deck.h"
#include "grid.h"
#include "layout_mgr.h"
#include "place.h"
#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/dload.h>
#include <Fresco/layouts/layouts.h>

Fresco_file_id(Fresco_layouts_impl_kit_cxx)

class LayoutKitImpl : public LayoutKit {
public:
    LayoutKitImpl();
    ~LayoutKitImpl();

    OxExchange* _exchange();

    //+ LayoutKit::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* LayoutKit */
    XfCoord fil();
    void fil(XfCoord _p);
    Grid_ptr fixed_grid(const Grid::Index& upper);
    XfGlyph_ptr fixed_range(Grid_ptr g, const Grid::Range& r);
    XfGlyph_ptr hbox();
    XfGlyph_ptr vbox();
    XfGlyph_ptr hbox_first_aligned();
    XfGlyph_ptr vbox_first_aligned();
    XfGlyph_ptr hbox_align_elements(XfAlignment align);
    XfGlyph_ptr vbox_align_elements(XfAlignment align);
    XfGlyph_ptr overlay();
    XfGlyph_ptr deck();
    XfGlyph_ptr back(XfGlyph_ptr g, XfGlyph_ptr under);
    XfGlyph_ptr front(XfGlyph_ptr g, XfGlyph_ptr over);
    XfGlyph_ptr between(XfGlyph_ptr g, XfGlyph_ptr under, XfGlyph_ptr over);
    XfGlyph_ptr glue(XfAxis a, XfCoord natural, XfCoord stretch, XfCoord shrink, XfAlignment align);
    XfGlyph_ptr glue_requisition(const XfGlyph::Requisition& r);
    XfGlyph_ptr hfil();
    XfGlyph_ptr hglue_fil(XfCoord natural);
    XfGlyph_ptr hglue(XfCoord natural, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr hglue_aligned(XfCoord natural, XfCoord stretch, XfCoord shrink, XfAlignment a);
    XfGlyph_ptr hspace(XfCoord natural);
    XfGlyph_ptr vfil();
    XfGlyph_ptr vglue_fil(XfCoord natural);
    XfGlyph_ptr vglue(XfCoord natural, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr vglue_aligned(XfCoord natural, XfCoord stretch, XfCoord shrink, XfAlignment a);
    XfGlyph_ptr vspace(XfCoord natural);
    XfGlyph_ptr shape_of(XfGlyph_ptr g);
    XfGlyph_ptr shape_of_xy(XfGlyph_ptr gx, XfGlyph_ptr gy);
    XfGlyph_ptr shape_of_xyz(XfGlyph_ptr gx, XfGlyph_ptr gy, XfGlyph_ptr gz);
    XfGlyph_ptr strut(XfFont_ptr f, XfCoord natural, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr hstrut(XfCoord right_bearing, XfCoord left_bearing, XfCoord natural, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr vstrut(XfCoord ascent, XfCoord descent, XfCoord natural, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr spaces(Long count, XfCoord each, XfFont_ptr f, XfColor_ptr c);
    XfGlyph_ptr align(XfGlyph_ptr g, XfAlignment x, XfAlignment y);
    XfGlyph_ptr align_axis(XfGlyph_ptr g, XfAxis a, XfAlignment align);
    XfGlyph_ptr halign(XfGlyph_ptr g, XfAlignment x);
    XfGlyph_ptr valign(XfGlyph_ptr g, XfAlignment y);
    XfGlyph_ptr fixed(XfGlyph_ptr g, XfCoord x, XfCoord y);
    XfGlyph_ptr fixed_axis(XfGlyph_ptr g, XfAxis a, XfCoord size);
    XfGlyph_ptr hfixed(XfGlyph_ptr g, XfCoord x);
    XfGlyph_ptr vfixed(XfGlyph_ptr g, XfCoord y);
    XfGlyph_ptr flexible(XfGlyph_ptr g, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr flexible_fil(XfGlyph_ptr g);
    XfGlyph_ptr flexible_axis(XfGlyph_ptr g, XfAxis a, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr hflexible(XfGlyph_ptr g, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr vflexible(XfGlyph_ptr g, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr natural(XfGlyph_ptr g, XfCoord x, XfCoord y);
    XfGlyph_ptr natural_axis(XfGlyph_ptr g, XfAxis a, XfCoord size);
    XfGlyph_ptr hnatural(XfGlyph_ptr g, XfCoord x);
    XfGlyph_ptr vnatural(XfGlyph_ptr g, XfCoord y);
    XfGlyph_ptr margin(XfGlyph_ptr g, XfCoord all);
    XfGlyph_ptr margin_lrbt(XfGlyph_ptr g, XfCoord lmargin, XfCoord rmargin, XfCoord bmargin, XfCoord tmargin);
    XfGlyph_ptr margin_flexible(XfGlyph_ptr g, XfCoord margin, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr margin_lrbt_flexible(XfGlyph_ptr g, XfCoord lmargin, XfCoord lstretch, XfCoord lshrink, XfCoord rmargin, XfCoord rstretch, XfCoord rshrink, XfCoord bmargin, XfCoord bstretch, XfCoord bshrink, XfCoord tmargin, XfCoord tstretch, XfCoord tshrink);
    XfGlyph_ptr hmargin(XfGlyph_ptr g, XfCoord both);
    XfGlyph_ptr hmargin_lr(XfGlyph_ptr g, XfCoord lmargin, XfCoord rmargin);
    XfGlyph_ptr hmargin_lr_flexible(XfGlyph_ptr g, XfCoord lmargin, XfCoord lstretch, XfCoord lshrink, XfCoord rmargin, XfCoord rstretch, XfCoord rshrink);
    XfGlyph_ptr vmargin(XfGlyph_ptr g, XfCoord both);
    XfGlyph_ptr vmargin_bt(XfGlyph_ptr g, XfCoord bmargin, XfCoord tmargin);
    XfGlyph_ptr vmargin_bt_flexible(XfGlyph_ptr g, XfCoord bmargin, XfCoord bstretch, XfCoord bshrink, XfCoord tmargin, XfCoord tstretch, XfCoord tshrink);
    XfGlyph_ptr lmargin(XfGlyph_ptr g, XfCoord natural);
    XfGlyph_ptr lmargin_flexible(XfGlyph_ptr g, XfCoord natural, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr rmargin(XfGlyph_ptr g, XfCoord natural);
    XfGlyph_ptr rmargin_flexible(XfGlyph_ptr g, XfCoord natural, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr bmargin(XfGlyph_ptr g, XfCoord natural);
    XfGlyph_ptr bmargin_flexible(XfGlyph_ptr g, XfCoord natural, XfCoord stretch, XfCoord shrink);
    XfGlyph_ptr tmargin(XfGlyph_ptr g, XfCoord natural);
    XfGlyph_ptr tmargin_flexible(XfGlyph_ptr g, XfCoord natural, XfCoord stretch, XfCoord shrink);
    //+
protected:
    Coord fil_;
    SharedFrescoObjectImpl object_;

    void init_types();
};

//
// Glue - filler glyph
//

class Glue : public GlyphImpl {
public:
    Glue(Axis, Coord natural, Coord stretch, Coord shrink, Alignment a);
    Glue(const Glyph::Requisition&);
    virtual ~Glue();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
private:
    Glyph::Requisition requisition_;
};

//
// ShapeOf -- request the same size(s) as another glyph
//

class ShapeOf : public GlyphImpl {
public:
    ShapeOf(Glyph_ptr x, Glyph_ptr y, Glyph_ptr z);
    virtual ~ShapeOf();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
private:
    Glyph_ptr x_;
    Glyph_ptr y_;
    Glyph_ptr z_;
};

//
// Space -- draw space with font metrics
//

class Space : public GlyphImpl {
public:
    Space(long count, Coord each, Font_ptr, Color_ptr);
    virtual ~Space();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
    void pick(XfGlyphTraversal_ptr t); //+ XfGlyph::pick
private:
    long count_;
    Coord each_;
    Font_ptr font_;
    Color_ptr color_;
    Coord width_;
    Coord height_;
    float alignment_;
};

//
// Strut -- prop open space aligned by a font
//

class Strut : public GlyphImpl {
public:
    Strut(Font_ptr, Coord natural = 0, Coord stretch = 0, Coord shrink = 0);
    virtual ~Strut();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
private:
    Font_ptr font_;
    Coord natural_;
    Coord stretch_;
    Coord shrink_;
    Coord height_;
    float alignment_;
};

//
// HStrut -- prop horizontal space
//

class HStrut : public GlyphImpl {
public:
    HStrut(
        Coord right_bearing, Coord left_bearing = 0,
        Coord natural = 0, Coord stretch = 0, Coord shrink = 0
    );
    virtual ~HStrut();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
private:
    Coord left_bearing_;
    Coord right_bearing_;
    Coord natural_;
    Coord stretch_;
    Coord shrink_;
};

//
// VStrut -- prop vertical space
//

class VStrut : public GlyphImpl {
public:
    VStrut(
        Coord ascent, Coord descent = 0,
        Coord natural = 0, Coord stretch = 0, Coord shrink = 0
    );
    virtual ~VStrut();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
private:
    Coord ascent_;
    Coord descent_;
    Coord natural_;
    Coord stretch_;
    Coord shrink_;
};

// class LayoutKitImpl

dload(LayoutKitImpl)

LayoutKitImpl::LayoutKitImpl() {
    fil_ = infinite_coord;
}

LayoutKitImpl::~LayoutKitImpl() { }

OxExchange* LayoutKitImpl::_exchange() { return &object_; }

//+ LayoutKitImpl(XfFrescoObject::=object_.)
XfTag LayoutKitImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void LayoutKitImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void LayoutKitImpl::disconnect() {
    object_.disconnect();
}
void LayoutKitImpl::notify_observers() {
    object_.notify_observers();
}
void LayoutKitImpl::update() {
    object_.update();
}
//+

//+ LayoutKitImpl(LayoutKit::fil=c)
void LayoutKitImpl::fil(XfCoord c) { fil_ = c; }

//+ LayoutKitImpl(LayoutKit::fil?)
XfCoord LayoutKitImpl::fil() { return fil_; }

//+ LayoutKitImpl(LayoutKit::fixed_grid)
Grid_ptr LayoutKitImpl::fixed_grid(const Grid::Index& upper) {
    return new GridImpl(upper);
}

//+ LayoutKitImpl(LayoutKit::fixed_range)
XfGlyph_ptr LayoutKitImpl::fixed_range(Grid_ptr g, const Grid::Range& r) {
    return new SubGridImpl(g, r);
}

//+ LayoutKitImpl(LayoutKit::hbox)
XfGlyph_ptr LayoutKitImpl::hbox() {
    return new Box(
	new LayoutSuperpose(
	    new LayoutTile(X_axis), new LayoutAlign(Y_axis)
	)
    );
}

//+ LayoutKitImpl(LayoutKit::vbox)
XfGlyph_ptr LayoutKitImpl::vbox() {
    return new Box(
	new LayoutSuperpose(
	    new LayoutTileReversed(Y_axis), new LayoutAlign(X_axis)
	)
    );
}

//+ LayoutKitImpl(LayoutKit::hbox_first_aligned)
XfGlyph_ptr LayoutKitImpl::hbox_first_aligned() {
    return new Box(
	new LayoutSuperpose(
	    new LayoutTileFirstAligned(X_axis), new LayoutAlign(Y_axis)
	)
    );
}

//+ LayoutKitImpl(LayoutKit::vbox_first_aligned)
XfGlyph_ptr LayoutKitImpl::vbox_first_aligned() {
    return new Box(
	new LayoutSuperpose(
	    new LayoutTileReversedFirstAligned(Y_axis),
	    new LayoutAlign(X_axis)
	)
    );
}

//+ LayoutKitImpl(LayoutKit::hbox_align_elements)
XfGlyph_ptr LayoutKitImpl::hbox_align_elements(XfAlignment align) {
    return new BoxAlignElements(
        new LayoutSuperpose(
            new LayoutTile(X_axis), new LayoutAlign(Y_axis)
        ),
        Y_axis,
        align
    );
}
 
//+ LayoutKitImpl(LayoutKit::vbox_align_elements)
XfGlyph_ptr LayoutKitImpl::vbox_align_elements(XfAlignment align) {
    return new BoxAlignElements(
        new LayoutSuperpose(
            new LayoutTileReversed(Y_axis), new LayoutAlign(X_axis)
        ),
        X_axis,
        align
    );
}

//+ LayoutKitImpl(LayoutKit::overlay)
XfGlyph_ptr LayoutKitImpl::overlay() {
    return new Box(
	new LayoutSuperpose(
	    new LayoutAlign(X_axis), new LayoutAlign(Y_axis)
	)
    );
}

//+ LayoutKitImpl(LayoutKit::deck)
XfGlyph_ptr LayoutKitImpl::deck() {
    return new DeckImpl;
}

//+ LayoutKitImpl(LayoutKit::back)
XfGlyph_ptr LayoutKitImpl::back(XfGlyph_ptr g, XfGlyph_ptr under) {
    return new LayoutLayer(g, under, nil);
}

//+ LayoutKitImpl(LayoutKit::front)
XfGlyph_ptr LayoutKitImpl::front(XfGlyph_ptr g, XfGlyph_ptr over) {
    return new LayoutLayer(g, nil, over);
}

//+ LayoutKitImpl(LayoutKit::between)
XfGlyph_ptr LayoutKitImpl::between(XfGlyph_ptr g, XfGlyph_ptr under, XfGlyph_ptr over) {
    return new LayoutLayer(g, under, over);
}

//+ LayoutKitImpl(LayoutKit::glue)
XfGlyph_ptr LayoutKitImpl::glue(XfAxis a, XfCoord natural, XfCoord stretch, XfCoord shrink, XfAlignment align) {
    return new Glue(a, natural, stretch, shrink, align);
}

//+ LayoutKitImpl(LayoutKit::glue_requisition)
XfGlyph_ptr LayoutKitImpl::glue_requisition(const XfGlyph::Requisition& r) {
    return new Glue(r);
}

//+ LayoutKitImpl(LayoutKit::hfil)
XfGlyph_ptr LayoutKitImpl::hfil() {
    return new Glue(X_axis, Coord(0), fil_, Coord(0), Alignment(0));
}

//+ LayoutKitImpl(LayoutKit::hglue_fil)
XfGlyph_ptr LayoutKitImpl::hglue_fil(XfCoord natural) {
    return new Glue(X_axis, natural, fil_, Coord(0), Alignment(0));
}

//+ LayoutKitImpl(LayoutKit::hglue)
XfGlyph_ptr LayoutKitImpl::hglue(XfCoord natural, XfCoord stretch, XfCoord shrink) {
    return new Glue(X_axis, natural, stretch, shrink, Alignment(0));
}

//+ LayoutKitImpl(LayoutKit::hglue_aligned)
XfGlyph_ptr LayoutKitImpl::hglue_aligned(XfCoord natural, XfCoord stretch, XfCoord shrink, XfAlignment a) {
    return new Glue(X_axis, natural, stretch, shrink, a);
}

//+ LayoutKitImpl(LayoutKit::hspace)
XfGlyph_ptr LayoutKitImpl::hspace(XfCoord natural) {
    return new Glue(X_axis, natural, Coord(0), Coord(0), Alignment(0));
}

//+ LayoutKitImpl(LayoutKit::vfil)
XfGlyph_ptr LayoutKitImpl::vfil() {
    return new Glue(Y_axis, Coord(0), fil_, Coord(0), Alignment(0));
}

//+ LayoutKitImpl(LayoutKit::vglue_fil)
XfGlyph_ptr LayoutKitImpl::vglue_fil(XfCoord natural) {
    return new Glue(Y_axis, natural, fil_, Coord(0), Alignment(0));
}

//+ LayoutKitImpl(LayoutKit::vglue)
XfGlyph_ptr LayoutKitImpl::vglue(XfCoord natural, XfCoord stretch, XfCoord shrink) {
    return new Glue(Y_axis, natural, stretch, shrink, Alignment(0));
}

//+ LayoutKitImpl(LayoutKit::vglue_aligned)
XfGlyph_ptr LayoutKitImpl::vglue_aligned(XfCoord natural, XfCoord stretch, XfCoord shrink, XfAlignment a) {
    return new Glue(Y_axis, natural, stretch, shrink, a);
}

//+ LayoutKitImpl(LayoutKit::vspace)
XfGlyph_ptr LayoutKitImpl::vspace(XfCoord natural) {
    return new Glue(Y_axis, natural, Coord(0), Coord(0), Alignment(0));
}

//+ LayoutKitImpl(LayoutKit::shape_of)
XfGlyph_ptr LayoutKitImpl::shape_of(XfGlyph_ptr g) {
    return new ShapeOf(g, nil, nil);
}

//+ LayoutKitImpl(LayoutKit::shape_of_xy)
XfGlyph_ptr LayoutKitImpl::shape_of_xy(XfGlyph_ptr gx, XfGlyph_ptr gy) {
    return new ShapeOf(gx, gy, nil);
}

//+ LayoutKitImpl(LayoutKit::shape_of_xyz)
XfGlyph_ptr LayoutKitImpl::shape_of_xyz(XfGlyph_ptr gx, XfGlyph_ptr gy, XfGlyph_ptr gz) {
    return new ShapeOf(gx, gy, gz);
}

//+ LayoutKitImpl(LayoutKit::strut)
XfGlyph_ptr LayoutKitImpl::strut(XfFont_ptr f, XfCoord natural, XfCoord stretch, XfCoord shrink) {
    return new Strut(f, natural, stretch, shrink);
}

//+ LayoutKitImpl(LayoutKit::hstrut)
XfGlyph_ptr LayoutKitImpl::hstrut(XfCoord right_bearing, XfCoord left_bearing, XfCoord natural, XfCoord stretch, XfCoord shrink) {
    return new HStrut(right_bearing, left_bearing, natural, stretch, shrink);
}

//+ LayoutKitImpl(LayoutKit::vstrut)
XfGlyph_ptr LayoutKitImpl::vstrut(XfCoord ascent, XfCoord descent, XfCoord natural, XfCoord stretch, XfCoord shrink) {
    return new VStrut(ascent, descent, natural, stretch, shrink);
}

//+ LayoutKitImpl(LayoutKit::spaces)
XfGlyph_ptr LayoutKitImpl::spaces(Long count, XfCoord each, XfFont_ptr f, XfColor_ptr c) {
    return new Space(count, each, f, c);
}

//+ LayoutKitImpl(LayoutKit::align)
XfGlyph_ptr LayoutKitImpl::align(XfGlyph_ptr g, XfAlignment x, XfAlignment y) {
    return new Placement(
        g,
        new LayoutSuperpose(
            new LayoutCenter(X_axis, x), new LayoutCenter(Y_axis, y)
        )
    );
}
 
//+ LayoutKitImpl(LayoutKit::align_axis)
XfGlyph_ptr LayoutKitImpl::align_axis(XfGlyph_ptr g, XfAxis a, XfAlignment align) {
    return new Placement(g, new LayoutCenter(a, align));
}
 
//+ LayoutKitImpl(LayoutKit::halign)
XfGlyph_ptr LayoutKitImpl::halign(XfGlyph_ptr g, XfAlignment x) {
    return align_axis(g, X_axis, x);
}
 
//+ LayoutKitImpl(LayoutKit::valign)
XfGlyph_ptr LayoutKitImpl::valign(XfGlyph_ptr g, XfAlignment y) {
    return align_axis(g, Y_axis, y);
}

//+ LayoutKitImpl(LayoutKit::fixed)
XfGlyph_ptr LayoutKitImpl::fixed(XfGlyph_ptr g, XfCoord x, XfCoord y) {
    return new Placement(
	g,
	new LayoutSuperpose(
	    new LayoutFixed(X_axis, x), new LayoutFixed(Y_axis, y)
	)
    );
}

//+ LayoutKitImpl(LayoutKit::fixed_axis)
XfGlyph_ptr LayoutKitImpl::fixed_axis(XfGlyph_ptr g, XfAxis a, XfCoord size) {
    return new Placement(g, new LayoutFixed(a, size));
}

//+ LayoutKitImpl(LayoutKit::hfixed)
XfGlyph_ptr LayoutKitImpl::hfixed(XfGlyph_ptr g, XfCoord x) {
    return fixed_axis(g, X_axis, x);
}

//+ LayoutKitImpl(LayoutKit::vfixed)
XfGlyph_ptr LayoutKitImpl::vfixed(XfGlyph_ptr g, XfCoord y) {
    return fixed_axis(g, Y_axis, y);
}

//+ LayoutKitImpl(LayoutKit::flexible)
XfGlyph_ptr LayoutKitImpl::flexible(XfGlyph_ptr g, XfCoord stretch, XfCoord shrink) {
    return new Placement(
	g,
	new LayoutSuperpose(
	    new LayoutVariable(X_axis, stretch, shrink),
	    new LayoutVariable(Y_axis, stretch, shrink)
	)
    );
}

//+ LayoutKitImpl(LayoutKit::flexible_fil)
XfGlyph_ptr LayoutKitImpl::flexible_fil(XfGlyph_ptr g) {
    return flexible(g, fil_, fil_);
}

//+ LayoutKitImpl(LayoutKit::flexible_axis)
XfGlyph_ptr LayoutKitImpl::flexible_axis(XfGlyph_ptr g, XfAxis a, XfCoord stretch, XfCoord shrink) {
    return new Placement(g, new LayoutVariable(a, stretch, shrink));
}

//+ LayoutKitImpl(LayoutKit::hflexible)
XfGlyph_ptr LayoutKitImpl::hflexible(XfGlyph_ptr g, XfCoord stretch, XfCoord shrink) {
    return flexible_axis(g, X_axis, stretch, shrink);
}

//+ LayoutKitImpl(LayoutKit::vflexible)
XfGlyph_ptr LayoutKitImpl::vflexible(XfGlyph_ptr g, XfCoord stretch, XfCoord shrink) {
    return flexible_axis(g, Y_axis, stretch, shrink);
}

//+ LayoutKitImpl(LayoutKit::natural)
XfGlyph_ptr LayoutKitImpl::natural(XfGlyph_ptr g, XfCoord x, XfCoord y) {
    return new Placement(
	g,
	new LayoutSuperpose(
	    new LayoutNatural(X_axis, x),
	    new LayoutNatural(Y_axis, y)
	)
    );
}

//+ LayoutKitImpl(LayoutKit::natural_axis)
XfGlyph_ptr LayoutKitImpl::natural_axis(XfGlyph_ptr g, XfAxis a, XfCoord size) {
    return new Placement(g, new LayoutNatural(a, size));
}

//+ LayoutKitImpl(LayoutKit::hnatural)
XfGlyph_ptr LayoutKitImpl::hnatural(XfGlyph_ptr g, XfCoord x) {
    return natural_axis(g, X_axis, x);
}

//+ LayoutKitImpl(LayoutKit::vnatural)
XfGlyph_ptr LayoutKitImpl::vnatural(XfGlyph_ptr g, XfCoord y) {
    return natural_axis(g, Y_axis, y);
}

//+ LayoutKitImpl(LayoutKit::margin)
XfGlyph_ptr LayoutKitImpl::margin(XfGlyph_ptr g, XfCoord all) {
    return new Placement(g, new LayoutMargin(all));
}

//+ LayoutKitImpl(LayoutKit::margin_flexible)
XfGlyph_ptr LayoutKitImpl::margin_flexible(XfGlyph_ptr g, XfCoord margin, XfCoord stretch, XfCoord shrink) {
    return new Placement(
	g,
	new LayoutMargin(
	    margin, stretch, shrink, margin, stretch, shrink,
	    margin, stretch, shrink, margin, stretch, shrink
	)
    );
}

//+ LayoutKitImpl(LayoutKit::margin_lrbt)
XfGlyph_ptr LayoutKitImpl::margin_lrbt(XfGlyph_ptr g, XfCoord lmargin, XfCoord rmargin, XfCoord bmargin, XfCoord tmargin) {
    return new Placement(
	g, new LayoutMargin(lmargin, rmargin, bmargin, tmargin)
    );
}

//+ LayoutKitImpl(LayoutKit::margin_lrbt_flexible)
XfGlyph_ptr LayoutKitImpl::margin_lrbt_flexible(XfGlyph_ptr g, XfCoord lmargin, XfCoord lstretch, XfCoord lshrink, XfCoord rmargin, XfCoord rstretch, XfCoord rshrink, XfCoord bmargin, XfCoord bstretch, XfCoord bshrink, XfCoord tmargin, XfCoord tstretch, XfCoord tshrink) {
    return new Placement(
	g,
	new LayoutMargin(
	    lmargin, lstretch, lshrink, rmargin, rstretch, rshrink,
	    bmargin, bstretch, bshrink, tmargin, tstretch, tshrink
	)
    );
}

//+ LayoutKitImpl(LayoutKit::hmargin)
XfGlyph_ptr LayoutKitImpl::hmargin(XfGlyph_ptr g, XfCoord both) {
    return margin_lrbt(g, both, both, Coord(0), Coord(0));
}

//+ LayoutKitImpl(LayoutKit::hmargin_lr)
XfGlyph_ptr LayoutKitImpl::hmargin_lr(XfGlyph_ptr g, XfCoord lmargin, XfCoord rmargin) {
    return margin_lrbt(g, lmargin, rmargin, Coord(0), Coord(0));
}

//+ LayoutKitImpl(LayoutKit::hmargin_lr_flexible)
XfGlyph_ptr LayoutKitImpl::hmargin_lr_flexible(XfGlyph_ptr g, XfCoord lmargin, XfCoord lstretch, XfCoord lshrink, XfCoord rmargin, XfCoord rstretch, XfCoord rshrink) {
    return margin_lrbt_flexible(
	g,
	lmargin, lstretch, lshrink, rmargin, rstretch, rshrink,
	Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0)
    );
}

//+ LayoutKitImpl(LayoutKit::vmargin)
XfGlyph_ptr LayoutKitImpl::vmargin(XfGlyph_ptr g, XfCoord both) {
    return margin_lrbt(g, Coord(0), Coord(0), both, both);
}

//+ LayoutKitImpl(LayoutKit::vmargin_bt)
XfGlyph_ptr LayoutKitImpl::vmargin_bt(XfGlyph_ptr g, XfCoord bmargin, XfCoord tmargin) {
    return margin_lrbt(g, Coord(0), Coord(0), bmargin, tmargin);
}

//+ LayoutKitImpl(LayoutKit::vmargin_bt_flexible)
XfGlyph_ptr LayoutKitImpl::vmargin_bt_flexible(XfGlyph_ptr g, XfCoord bmargin, XfCoord bstretch, XfCoord bshrink, XfCoord tmargin, XfCoord tstretch, XfCoord tshrink) {
    return margin_lrbt_flexible(
	g,
	Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0),
	bmargin, bstretch, bshrink, tmargin, tstretch, tshrink
    );
}

//+ LayoutKitImpl(LayoutKit::lmargin)
XfGlyph_ptr LayoutKitImpl::lmargin(XfGlyph_ptr g, XfCoord natural) {
    return margin_lrbt(g, natural, Coord(0), Coord(0), Coord(0));
}

//+ LayoutKitImpl(LayoutKit::lmargin_flexible)
XfGlyph_ptr LayoutKitImpl::lmargin_flexible(XfGlyph_ptr g, XfCoord natural, XfCoord stretch, XfCoord shrink) {
    return margin_lrbt_flexible(
	g, natural, stretch, shrink, Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0)
    );
}

//+ LayoutKitImpl(LayoutKit::rmargin)
XfGlyph_ptr LayoutKitImpl::rmargin(XfGlyph_ptr g, XfCoord natural) {
    return margin_lrbt(g, Coord(0), natural, Coord(0), Coord(0));
}

//+ LayoutKitImpl(LayoutKit::rmargin_flexible)
XfGlyph_ptr LayoutKitImpl::rmargin_flexible(XfGlyph_ptr g, XfCoord natural, XfCoord stretch, XfCoord shrink) {
    return margin_lrbt_flexible(
	g, Coord(0), Coord(0), Coord(0), natural, stretch, shrink, Coord(0), Coord(0),
	Coord(0), Coord(0), Coord(0), Coord(0)
    );
}

//+ LayoutKitImpl(LayoutKit::bmargin)
XfGlyph_ptr LayoutKitImpl::bmargin(XfGlyph_ptr g, XfCoord natural) {
    return margin_lrbt(g, Coord(0), Coord(0), natural, Coord(0));
}

//+ LayoutKitImpl(LayoutKit::bmargin_flexible)
XfGlyph_ptr LayoutKitImpl::bmargin_flexible(XfGlyph_ptr g, XfCoord natural, XfCoord stretch, XfCoord shrink) {
    return margin_lrbt_flexible(
	g, Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), natural, stretch,
	shrink, Coord(0), Coord(0), Coord(0)
    );
}

//+ LayoutKitImpl(LayoutKit::tmargin)
XfGlyph_ptr LayoutKitImpl::tmargin(XfGlyph_ptr g, XfCoord natural) {
    return margin_lrbt(g, Coord(0), Coord(0), Coord(0), natural);
}

//+ LayoutKitImpl(LayoutKit::tmargin_flexible)
XfGlyph_ptr LayoutKitImpl::tmargin_flexible(XfGlyph_ptr g, XfCoord natural, XfCoord stretch, XfCoord shrink) {
    return margin_lrbt_flexible(
	g, Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0), Coord(0),
	Coord(0), natural, stretch, shrink
    );
}

// class Glue

Glue::Glue(
    Axis a, Coord natural, Coord stretch, Coord shrink, Alignment align
) {
    GlyphImpl::init_requisition(requisition_);
    Glyph::Requirement* r = GlyphImpl::requirement(requisition_, a);
    if (r != nil) {
	GlyphImpl::require(*r, natural, stretch, shrink, align);
    }
}

Glue::Glue(const Glyph::Requisition& r) {
    requisition_ = r;
}

Glue::~Glue() { }

//+ Glue(XfGlyph::request)
void Glue::request(XfGlyph::Requisition& r) {
    r = requisition_;
}

// class Space

Space::Space(long count, float each, Font_ptr f, Color_ptr c) {
    count_ = count;
    each_ = each;
    font_ = Font::_duplicate(f);
    color_ = Color::_duplicate(c);
    if (is_not_nil(font_)) {
	Font::Info i;
	font_->font_info(i);
	Coord ascent = i.ascent;
	Coord descent = i.descent;
	font_->char_info(' ', i);
	width_ = i.width * each_ * count_;
	height_ = ascent + descent;
	alignment_ = (height_ == 0) ? 0 : descent / height_;
    } else {
	width_ = Coord(0);
	height_ = Coord(0);
	alignment_ = Alignment(0);
    }
}

Space::~Space() {
    CORBA::release(font_);
    CORBA::release(color_);
}

//+ Space(XfGlyph::request)
void Space::request(XfGlyph::Requisition& r) {
    require(r.x, width_, width_ * Coord(4), width_ / Coord(3), Alignment(0));
    require(r.y, height_, Coord(0), Coord(0), alignment_);
}

//+ Space(XfGlyph::draw)
void Space::draw(XfGlyphTraversal_ptr t) {
    if (count_ > 0) {
	Vertex v0, v1, v;
	t->bounds(v0, v1, v);
	Coord each = (v1.x - v0.x) / count_;
	Painter_var p = t->current_painter();
	p->current_font(font_);
	p->current_color(color_);
        for (long i = 0; i < count_; ++i) {
            p->character(' ', each, v.x, v.y);
            v.x += each;
        }
    }
}

//+ Space(XfGlyph::pick)
void Space::pick(XfGlyphTraversal_ptr) { }

// class Strut

Strut::Strut(Font_ptr font, Coord natural, Coord stretch, Coord shrink) {
    font_ = Font::_duplicate(font);
    if (is_not_nil(font_)) {
	Font::Info b;
	font_->font_info(b);
	height_ = b.ascent + b.descent;
	alignment_ = (height_ == 0) ? 0 : b.descent / height_;
    }
    natural_ = natural;
    stretch_ = stretch;
    shrink_ = shrink;
}

Strut::~Strut() {
    CORBA::release(font_);
}

//+ Strut(XfGlyph::request)
void Strut::request(XfGlyph::Requisition& r) {
    require(r.x, natural_, stretch_, shrink_, Alignment(0));
    require(r.y, height_, Coord(0), Coord(0), alignment_);
}

// class HStrut

HStrut::HStrut(
    Coord right_bearing, Coord left_bearing,
    Coord natural, Coord stretch, Coord shrink
) {
    left_bearing_ = left_bearing;
    right_bearing_ = right_bearing;
    natural_ = natural;
    stretch_ = stretch;
    shrink_ = shrink;
}

HStrut::~HStrut() { }

//+ HStrut(XfGlyph::request)
void HStrut::request(XfGlyph::Requisition& r) {
    Coord width = left_bearing_ + right_bearing_;
    require(r.x, width, Coord(0), Coord(0), (width == 0) ? 0 : left_bearing_ / width);
    require(r.y, natural_, stretch_, shrink_, Alignment(0));
}

// class VStrut

VStrut::VStrut(
    Coord ascent, Coord descent, Coord natural, Coord stretch, Coord shrink
) {
    ascent_ = ascent;
    descent_ = descent;
    natural_ = natural;
    stretch_ = stretch;
    shrink_ = shrink;
}

VStrut::~VStrut() { }

//+ VStrut(XfGlyph::request)
void VStrut::request(XfGlyph::Requisition& r) {
    Coord height = ascent_ + descent_;
    require(r.x, natural_, stretch_, shrink_, Alignment(0));
    require(r.y, height, Coord(0), Coord(0), (height == 0) ? 0 : descent_ / height);
}

// class ShapeOf

ShapeOf::ShapeOf(Glyph_ptr x, Glyph_ptr y, Glyph_ptr z) {
    x_ = Glyph::_duplicate(x);
    y_ = Glyph::_duplicate(y);
    z_ = Glyph::_duplicate(z);
}

ShapeOf::~ShapeOf() {
    CORBA::release(x_);
    CORBA::release(y_);
    CORBA::release(z_);
}

//+ ShapeOf(XfGlyph::request)
void ShapeOf::request(XfGlyph::Requisition& r) {
    if (is_nil(y_) && is_nil(z_)) {
	x_->request(r);
    } else {
	Glyph::Requisition req;
	if (is_not_nil(x_)) {
	    x_->request(req);
	    r.x = req.x;
	}
	if (is_not_nil(y_)) {
	    y_->request(req);
	    r.y = req.y;
	}
	if (is_not_nil(z_)) {
	    z_->request(req);
	    r.z = req.z;
	}
    }
}
