//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include "place.h"
#include "layout_mgr.h"
#include <Fresco/defs/drawing.h>
#include <Fresco/defs_impl/transform.h>

// class Placement

Placement::Placement(Glyph_ptr g, LayoutManager* layout) {
    body(g);
    layout_ = layout;
    result_ = new RegionImpl;
}

Placement::~Placement() {
    delete layout_;
    CORBA::release(result_);
}

//+ Placement(XfGlyph::request)
void Placement::request(XfGlyph::Requisition& r) {
    MonoGlyph::request(r);
    layout_->request(0, nil, r);
}

void Placement::child_allocate(Glyph::AllocationInfo& a) {
    Region_ptr given = a.allocation;
    result_->copy(given);
    Glyph::Requisition r;
    GlyphImpl::init_requisition(r);
    MonoGlyph::request(r);
    layout_->allocate(1, &r, given, &result_);

    TransformImpl tx;
    normal_transform(result_, &tx);
    if (is_nil(a.transformation)) {
	a.transformation = new TransformImpl;
    }
    a.transformation->premultiply(&tx);
    given->copy(result_);
}

//+ Placement(XfGlyph::traverse)
void Placement::traverse(XfGlyphTraversal_ptr t) {
    Region_var given = t->allocation();
    if (is_not_nil(given)) {
	result_->copy(given);
	Glyph::Requisition r;
	GlyphImpl::init_requisition(r);
	MonoGlyph::request(r);
	layout_->allocate(1, &r, given, &result_);
	TransformImpl tx;
	normal_transform(result_, &tx);
	Painter_var p = t->current_painter();
	p->push_matrix();
	p->premultiply(&tx);
	t->traverse_child(&offset_, result_);
	p->pop_matrix();
    } else {
	MonoGlyph::traverse(t);
    }
}

void Placement::normal_origin(RegionImpl* r, Vertex& o) {
    r->origin(o);
    r->lower_.x -= o.x; r->upper_.x -= o.x;
    r->lower_.y -= o.y; r->upper_.y -= o.y;
    r->lower_.z -= o.z; r->upper_.z -= o.z;
}

void Placement::normal_transform(RegionImpl* r, TransformImpl* tx) {
    Vertex o;
    normal_origin(r, o);
    tx->translate(o);
}

// class LayoutLayer

LayoutLayer::LayoutLayer(
    Glyph_ptr between, Glyph_ptr under, Glyph_ptr over
) {
    body(between);
    under_ = Glyph::_duplicate(under);
    over_ = Glyph::_duplicate(over);
}

LayoutLayer::~LayoutLayer() {
    CORBA::release(under_);
    CORBA::release(over_);
}

//+ LayoutLayer(XfGlyph::traverse)
void LayoutLayer::traverse(XfGlyphTraversal_ptr t) {
    if (is_not_nil(under_)) {
	under_->traverse(t);
    }
    MonoGlyph::traverse(t);
    if (is_not_nil(over_)) {
	over_->traverse(t);
    }
}
