//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_layouts_impl_place_h
#define Fresco_layouts_impl_place_h

#include <Fresco/defs_impl/glyphs.h>

class LayoutManager;
class RegionImpl;

//
// Placement -- glyph that places its body with a layout manager.
//
// The placement "consumes" the given layout; that is, it will
// destroy the layout manager when the placement is deallocated.
//

class Placement : public MonoGlyph {
public:
    Placement(Glyph_ptr, LayoutManager*);
    virtual ~Placement();

    void request(Glyph::Requisition& r); //+ XfGlyph::request
    void traverse(GlyphTraversal_ptr t); //+ XfGlyph::traverse

    void child_allocate(Glyph::AllocationInfo& a);
    static void normal_origin(RegionImpl*, Vertex&);
    static void normal_transform(RegionImpl*, TransformImpl*);
private:
    LayoutManager* layout_;
    RegionImpl* result_;
};

//
// LayoutLayer -- place glyph between two other glyphs
//

class LayoutLayer : public MonoGlyph {
public:
    LayoutLayer(Glyph_ptr between, Glyph_ptr under, Glyph_ptr over);
    virtual ~LayoutLayer();

    void traverse(GlyphTraversal_ptr t); //+ XfGlyph::traverse
private:
    Glyph_ptr under_;
    Glyph_ptr over_;
};

#endif
