//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-93 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating tyo the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/names.h>
#include "applics.h"

// class InsetImpl

InsetImpl::InsetImpl(
    FieldSubject_ptr fsubj, FieldEditorData_ptr fe_data,
    ExclusiveEditMgr_ptr fe_mgr, CharString_ptr viewer_class
) {
    fsubj_ = FieldSubject::_duplicate(fsubj);
    fe_data_ = FieldEditorData::_duplicate(fe_data);
    fe_mgr_ = ExclusiveEditMgr::_duplicate(fe_mgr);
    fsubj_tag_ = fe_mgr_->attach(fsubj_);
    viewer_class_ = CharString::_duplicate(viewer_class);
}

InsetImpl::~InsetImpl() {
    fe_mgr_->detach_subject(fsubj_tag_);
}

OxExchange* InsetImpl::_exchange() { return &object_; }

//+ InsetImpl(XfFrescoObject::=object_.)
XfTag InsetImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void InsetImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void InsetImpl::disconnect() {
    object_.disconnect();
}
void InsetImpl::notify_observers() {
    object_.notify_observers();
}
void InsetImpl::update() {
    object_.update();
}
//+

//+ InsetImpl(Inset::edit_manager)
EditMgr_ptr InsetImpl::edit_manager() {
    return ExclusiveEditMgr::_duplicate(fe_mgr_);
}

//+ InsetImpl(Inset::subject)
XfFrescoObject_ptr InsetImpl::subject() {
    return FieldSubject::_duplicate(fsubj_);
}

//+ InsetImpl(Inset::viewer_data)
ViewerData_ptr InsetImpl::viewer_data() {
    return FieldEditorData::_duplicate(fe_data_);
}

//+ InsetImpl(Inset::viewer_class)
XfCharString_ptr InsetImpl::viewer_class() {
    return CharString::_duplicate(viewer_class_);
}


// class RegistryImpl

RegistryImpl::RegistryImpl() { }
RegistryImpl::~RegistryImpl() { }

OxExchange* RegistryImpl::_exchange() { return &object_; }

//+ RegistryImpl(XfFrescoObject::=object_.)
XfTag RegistryImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void RegistryImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void RegistryImpl::disconnect() {
    object_.disconnect();
}
void RegistryImpl::notify_observers() {
    object_.notify_observers();
}
void RegistryImpl::update() {
    object_.update();
}
//+

//+ RegistryImpl(Registry::bind)
void RegistryImpl::bind(XfCharString_ptr viewer_class, ViewerCreator_ptr vc) {  // $$$ hack
    if (is_nil(viewer_class_hack1_)) {
	viewer_class_hack1_ = CharString::_duplicate(viewer_class);
	viewer_creator_hack1_ = ViewerCreator::_duplicate(vc);
    } else if (is_nil(viewer_class_hack2_)) {
	viewer_class_hack2_ = CharString::_duplicate(viewer_class);
	viewer_creator_hack2_ = ViewerCreator::_duplicate(vc);
    } else {
	abort();
    }
}

//+ RegistryImpl(Registry::viewer_ptr)
XfViewer_ptr RegistryImpl::viewer_ptr(Inset_ptr i) {   // $$$ hack
    CharString_var type = i->viewer_class();
    if (type->equal(viewer_class_hack1_)) {
	return viewer_creator_hack1_->create_viewer(i);
    } else if (type->equal(viewer_class_hack2_)) {
	return viewer_creator_hack2_->create_viewer(i);
    } else {
	abort();
	return nil;
    }
}

// class EditMgrImpl

declarePtrList(CommandList,Command)
implementPtrList(CommandList,Command)

EditMgrImpl::EditMgrImpl() {
    commands_ = new CommandList();
    current_ = -1;
}

EditMgrImpl::~EditMgrImpl() {
    Long count = commands_->count();
    for (Long i=0; i < count; i++) {
	CORBA::release(commands_->item(i));
    }
    delete commands_;
}

OxExchange* EditMgrImpl::_exchange() { return &object_; }

//+ EditMgrImpl(XfFrescoObject::=object_.)
XfTag EditMgrImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void EditMgrImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void EditMgrImpl::disconnect() {
    object_.disconnect();
}
void EditMgrImpl::notify_observers() {
    object_.notify_observers();
}
void EditMgrImpl::update() {
    object_.update();
}
//+

//+ EditMgrImpl(EditMgr::log)
void EditMgrImpl::log(Command_ptr cmd) {   
    commands_->append(Command::_duplicate(cmd));
    ++current_;
}

//+ EditMgrImpl(EditMgr::unlog)
void EditMgrImpl::unlog() {   
    Long n = commands_->count() - 1;
    if (n >= 0) {
	CORBA::release(commands_->item(n));
	commands_->remove(n);
	--current_;
    }
    // else raise exception
}

//+ EditMgrImpl(EditMgr::forward)
Boolean EditMgrImpl::forward() {
    if (current_ < commands_->count()) {
	++current_;
	return true;
    }
    return false;
}

//+ EditMgrImpl(EditMgr::backward)
Boolean EditMgrImpl::backward() {
    if (current_ > 0) {
	--current_;
	return true;
    }
    return false;
}

//+ EditMgrImpl(EditMgr::current_command)
Command_ptr EditMgrImpl::current_command() {
    return Command::_duplicate(commands_->item(current_));
}

//+ EditMgrImpl(EditMgr::empty)
void EditMgrImpl::empty() {
    commands_->remove_all();
    current_ = -1;
}

//+ EditMgrImpl(EditMgr::trim)
void EditMgrImpl::trim() {
    Long count = commands_->count();
    for (Long i = current_+1; i<count; i++) {
	commands_->remove(i);
    }
}

// class ExclusiveEditMgrImpl

declarePtrList(SubjectList,FrescoObject)
implementPtrList(SubjectList,FrescoObject)

#define IllegalSubject 0xffffffff

ExclusiveEditMgrImpl::ExclusiveEditMgrImpl() {
    subjects_ = new SubjectList();
    subject_ = IllegalSubject;
}

ExclusiveEditMgrImpl::~ExclusiveEditMgrImpl() {
    Long count = subjects_->count();
    for (Long i=0; i < count; i++) {
	CORBA::release(subjects_->item(i));
    }
    delete subjects_;
}

OxExchange* ExclusiveEditMgrImpl::_exchange() {
    return EditMgrImpl::_exchange();
}


//+ ExclusiveEditMgrImpl(EditMgr::*EditMgrImpl::)
/* XfFrescoObject */
XfTag ExclusiveEditMgrImpl::attach(XfFrescoObject_ptr observer) {
    return EditMgrImpl::attach(observer);
}
void ExclusiveEditMgrImpl::detach(XfTag attach_tag) {
    EditMgrImpl::detach(attach_tag);
}
void ExclusiveEditMgrImpl::disconnect() {
    EditMgrImpl::disconnect();
}
void ExclusiveEditMgrImpl::notify_observers() {
    EditMgrImpl::notify_observers();
}
void ExclusiveEditMgrImpl::update() {
    EditMgrImpl::update();
}
/* EditMgr */
void ExclusiveEditMgrImpl::log(Command_ptr cmd) {
    EditMgrImpl::log(cmd);
}
void ExclusiveEditMgrImpl::unlog() {
    EditMgrImpl::unlog();
}
Boolean ExclusiveEditMgrImpl::forward() {
    return EditMgrImpl::forward();
}
Boolean ExclusiveEditMgrImpl::backward() {
    return EditMgrImpl::backward();
}
Command_ptr ExclusiveEditMgrImpl::current_command() {
    return EditMgrImpl::current_command();
}
void ExclusiveEditMgrImpl::empty() {
    EditMgrImpl::empty();
}
void ExclusiveEditMgrImpl::trim() {
    EditMgrImpl::trim();
}
//+

//+ ExclusiveEditMgrImpl(ExclusiveEditMgr::attach_subject)
XfTag ExclusiveEditMgrImpl::attach_subject(XfFrescoObject_ptr subject) {
    subjects_->append(FrescoObject::_duplicate(subject));
    return subjects_->count()-1;
}

//+ ExclusiveEditMgrImpl(ExclusiveEditMgr::detach_subject)
void ExclusiveEditMgrImpl::detach_subject(XfTag t) {
    subjects_->remove(t);
    subjects_->insert(t, nil);  // Keep external tags intact. 
}

//+ ExclusiveEditMgrImpl(ExclusiveEditMgr::subject)
XfFrescoObject_ptr ExclusiveEditMgrImpl::subject(XfTag t) {
    return FrescoObject::_duplicate(subjects_->item(t));
}

//+ ExclusiveEditMgrImpl(ExclusiveEditMgr::begin_edit)
Boolean ExclusiveEditMgrImpl::begin_edit(XfTag subject) {
    if (subject_ == -1) {
	subject_ = subject;
	start_ = commands_->count();
	return true;
    } else {
	return false;
    }	
}

//+ ExclusiveEditMgrImpl(ExclusiveEditMgr::end_edit)
void ExclusiveEditMgrImpl::end_edit() {   
    if (subject_ == IllegalSubject) {
	printf("Attempting to end EditMgr transaction when "
	       "no transaction is current\n");
	abort();
    }
    Tag subj_count = subjects_->count();
    Long cmd_count = commands_->count();
    if (start_ < cmd_count) {
	for (Tag i=0; i < subj_count; i++) {
	    if (i != subject_) {
		FrescoObject_ptr subj = subjects_->item(i);
		if (is_not_nil(subj)) {
		    for (Long j = start_; j< cmd_count; j++) {
			Command_ptr cmd = commands_->item(j);
			execute(cmd, subj);
		    }
		}
	    }
	}
    }
    subject_ = IllegalSubject;
    notify_observers();
}

//+ ExclusiveEditMgrImpl(ExclusiveEditMgr::abort_edit)
void ExclusiveEditMgrImpl::abort_edit() {
    notify_observers();
    abort();  // not implemented
}


// class PersistentNamingContext

PersistentNamingContext::PersistentNamingContext() {
    widgets_ = Fresco_resolve_object(WidgetKit, "widgets_impl");
}

PersistentNamingContext::~PersistentNamingContext() { } 

OxExchange* PersistentNamingContext::_exchange() { return &object_; }

CORBA::Object_var PersistentNamingContext::tmp_ = nil;

//+ PersistentNamingContext(Naming::NamingContext::resolve)
CORBA::Object_ptr PersistentNamingContext::resolve(const Naming::Name& n) {
    const char* name = NamingLib::to_string(n);
    if (strcmp(name, "MyString") == 0) {
	// Temp code: create a new subject, viewerdata, inset, etc.
	if (is_nil(tmp_)) {
	    ExclusiveEditMgr_var fe_mgr = widgets_->field_edit_manager();
	    FieldSubject_var fsubj = widgets_->field_subject();
	    Long count = fsubj->count();
	    CharString_var str = FrescoLib::string_ref("hello");
	    fsubj->replace_all(str);

	    FieldEditorData_var fe_data = widgets_->field_editor_data(20, 0);
	    tmp_ = widgets_->field_inset(fsubj, fe_data, fe_mgr);
	}
    }
    return CORBA::Object::_duplicate(tmp_);
}

//+ PersistentNamingContext(XfFrescoObject::=object_.)
XfTag PersistentNamingContext::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void PersistentNamingContext::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void PersistentNamingContext::disconnect() {
    object_.disconnect();
}
void PersistentNamingContext::notify_observers() {
    object_.notify_observers();
}
void PersistentNamingContext::update() {
    object_.update();
}
//+

//+ PersistentNamingContext(Naming::NamingContext::bind) !n !obj
void PersistentNamingContext::bind(const Naming::Name&, CORBA::Object_ptr) {
    // unimplemented
}

//+ PersistentNamingContext(Naming::NamingContext::bind_context) !n !nc
void PersistentNamingContext::bind_context(const Naming::Name&, Naming::NamingContext_ptr) {
    // unimplemented
}

//+ PersistentNamingContext(Naming::NamingContext::rebind) !n !obj
void PersistentNamingContext::rebind(const Naming::Name&, CORBA::Object_ptr) {
    // unimplemented
}

//+ PersistentNamingContext(Naming::NamingContext::unbind) !n
void PersistentNamingContext::unbind(const Naming::Name&) {
    // unimplemented
}

//+ PersistentNamingContext(Naming::NamingContext::new_context) !n
Naming::NamingContext_ptr PersistentNamingContext::new_context() {
    // unimplemented
    return nil;
}

//+ PersistentNamingContext(Naming::NamingContext::bind_new_context) !n
Naming::NamingContext_ptr PersistentNamingContext::bind_new_context(const Naming::Name&) {
    // unimplemented
    return nil;
}

//+ PersistentNamingContext(Naming::NamingContext::list_names) !start,how_many,bl
Naming::BindingList* PersistentNamingContext::list_names(ULong, ULong) {
    // unimplemented
    return nil;
}

//+ PersistentNamingContext(Naming::NamingContext::list_all)
Naming::BindingList* PersistentNamingContext::list_all() {
    // unimplemented
    return nil;
}

//+ PersistentNamingContext(Naming::NamingContext::destroy)
void PersistentNamingContext::destroy() {
    // unimplemented
}
