/*
 * Copyright (c) 1993-94 Silicon Graphics, Inc.
 * Copyright (c) 1993-94 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_applic_applic_h
#define Fresco_applic_applic_h

#include <Fresco/widgets/applic.h>
#include <Fresco/widgets/widgets.h>
#include <Fresco/defs_impl/naming.h>     // $$$ goes elsewhere


class InsetImpl : public Inset {
public:
    InsetImpl(
	FieldSubject_ptr fsubj, FieldEditorData_ptr fe_data,
	ExclusiveEditMgr_ptr fe_mgr, CharString_ptr viewer_class
    );
    ~InsetImpl();
    
    OxExchange* _exchange();
    
    //+ Inset::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* Inset */
    EditMgr_ptr edit_manager();
    XfFrescoObject_ptr subject();
    ViewerData_ptr viewer_data();
    XfCharString_ptr viewer_class();
    //+
    
private:
    SharedFrescoObjectImpl object_;
    FieldSubject_var fsubj_;
    FieldEditorData_var fe_data_;
    ExclusiveEditMgr_var fe_mgr_;
    CharString_var viewer_class_;
    Tag fsubj_tag_;
};

class RegistryImpl : public Registry {
public:
    RegistryImpl();
    ~RegistryImpl();
    
    OxExchange* _exchange();
    
    //+ Registry::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* Registry */
    void bind(XfCharString_ptr viewer_class, ViewerCreator_ptr vc);
    XfViewer_ptr viewer_ptr(Inset_ptr i);
    //+
    
private:
    SharedFrescoObjectImpl object_;
    CharString_var viewer_class_hack1_;
    CharString_var viewer_class_hack2_;
    ViewerCreator_var viewer_creator_hack1_;
    ViewerCreator_var viewer_creator_hack2_;
};


class CommandList;
class EditMgrImpl : _Ix_public EditMgr {
public:
    EditMgrImpl();
    ~EditMgrImpl();
    
    OxExchange* _exchange();

    //+ EditMgr::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* EditMgr */
    void log(Command_ptr cmd);
    void unlog();
    Boolean forward();
    Boolean backward();
    Command_ptr current_command();
    void empty();
    void trim();
    //+
      
protected:
    CommandList* commands_;
private:
    SharedFrescoObjectImpl object_;
    Long current_;
};

class SubjectList;

class ExclusiveEditMgrImpl : public ExclusiveEditMgr, public EditMgrImpl {
public:
    ExclusiveEditMgrImpl();
    ~ExclusiveEditMgrImpl();

    OxExchange* _exchange();
    
    _Ix_ClassInterfaceOps(ExclusiveEditMgr)
    
    //+ ExclusiveEditMgr::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* EditMgr */
    void log(Command_ptr cmd);
    void unlog();
    Boolean forward();
    Boolean backward();
    Command_ptr current_command();
    void empty();
    void trim();
    /* ExclusiveEditMgr */
    XfTag attach_subject(XfFrescoObject_ptr subject);
    void detach_subject(XfTag t);
    XfFrescoObject_ptr subject(XfTag t);
    Boolean begin_edit(XfTag subject);
    void abort_edit();
    void end_edit();
    //+
    
protected:
    virtual void execute(Command_ptr command, FrescoObject_ptr subject) = 0;
	// Execute commands on given subject.

private:
    SubjectList* subjects_;
//---     TagList* observers_;
    Tag subject_;
    Long start_;
};


// PersistentNamingContext ($$$ Goes somewhere else)

class PersistentNamingContext : public Naming::NamingContext {
public:
    PersistentNamingContext();
    ~PersistentNamingContext();

    OxExchange* _exchange();

    //+ XfFrescoObject::=
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    //+

    //+ Naming::NamingContext::*
    /* NamingContext */
    void bind(const Naming::Name& n, CORBA::Object_ptr obj);
    void bind_context(const Naming::Name& n, Naming::NamingContext_ptr nc);
    void rebind(const Naming::Name& n, CORBA::Object_ptr obj);
    void unbind(const Naming::Name& n);
    CORBA::Object_ptr resolve(const Naming::Name& n);
    Naming::NamingContext_ptr new_context();
    Naming::NamingContext_ptr bind_new_context(const Naming::Name& n);
    Naming::BindingList* list_names(ULong start, ULong how_many);
    Naming::BindingList* list_all();
    void destroy();
    //+
private:
    SharedFrescoObjectImpl object_;
    WidgetKit_var widgets_;
    static CORBA::Object_var tmp_;
};


#endif
