//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-93 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <Fresco/widgets/widgets.h>
#include "bfloat.h"

BoundedFloat::BoundedFloat(Coord lower, Coord upper, Coord value) {
    lower_ = lower;
    span_ = upper - lower;
    scroll_incr_ = span_ * Coord(0.04);
    page_incr_ = span_ * Coord(0.2);
    value_ = value;
}

BoundedFloat::~BoundedFloat() { }

OxExchange* BoundedFloat::_exchange() { return &object_; }

//+ BoundedFloat(XfFrescoObject::=object_.)
Tag BoundedFloat::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void BoundedFloat::detach(Tag attach_tag) {
    object_.detach(attach_tag);
}
void BoundedFloat::disconnect() {
    object_.disconnect();
}
void BoundedFloat::notify_observers() {
    object_.notify_observers();
}
void BoundedFloat::update() {
    object_.update();
}
//+

//+ BoundedFloat(Adjustment::small_scroll=s)
void BoundedFloat::small_scroll(Coord s) { scroll_incr_ = s; }

//+ BoundedFloat(Adjustment::small_scroll?)
Coord BoundedFloat::small_scroll() { return scroll_incr_; }

//+ BoundedFloat(Adjustment::large_scroll=p)
void BoundedFloat::large_scroll(Coord p) { page_incr_ = p; }

//+ BoundedFloat(Adjustment::large_scroll?)
Coord BoundedFloat::large_scroll() { return page_incr_; }

//+ BoundedFloat(Adjustment::get_settings)
void BoundedFloat::get_settings(Adjustment::Settings& s) {
    s.lower = lower_;
    s.upper = lower_ + span_;
    s.length = span_;
    s.cur_lower = value_;
    s.cur_upper = value_;
    s.cur_length = Coord(0);
}

//+ BoundedFloat(Adjustment::begin)
void BoundedFloat::begin() { get_settings(saved_); }

//+ BoundedFloat(Adjustment::commit)
void BoundedFloat::commit() { }

//+ BoundedFloat(Adjustment::cancel)
void BoundedFloat::cancel() {
    lower_ = saved_.lower;
    span_ = saved_.length;
    value_ = saved_.cur_lower;
}

//+ BoundedFloat(Adjustment::scroll_forward)
void BoundedFloat::scroll_forward() {
    scroll_to(value_ + scroll_incr_);
}

//+ BoundedFloat(Adjustment::scroll_backward)
void BoundedFloat::scroll_backward() {
    scroll_to(value_ - scroll_incr_);
}

//+ BoundedFloat(Adjustment::page_forward)
void BoundedFloat::page_forward() {
    scroll_to(value_ + page_incr_);
}

//+ BoundedFloat(Adjustment::page_backward)
void BoundedFloat::page_backward() {
    scroll_to(value_ - page_incr_);
}

//+ BoundedFloat(Adjustment::scroll_to)
void BoundedFloat::scroll_to(Coord lower) {
    constrain(lower);
    if (lower != value_) {
	value_ = lower;
	notify_observers();
    }
}

//+ BoundedFloat(Adjustment::scroll_by)
void BoundedFloat::scroll_by(Coord delta) {
    scroll_to(value_ + delta);
}

//+ BoundedFloat(Adjustment::scale_to)
void BoundedFloat::scale_to(Coord) { }

//+ BoundedFloat(Adjustment::constrain)
void BoundedFloat::constrain(Coord& c) {
    Coord a = lower_, b = a + span_;
    if (c < a) {
	c = a;
    } else if (c > b){
	c = b;
    }
}
