/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_widgets_impl_box_h
#define Fresco_widgets_impl_box_h

#include "menu_kit.h"
#include <Fresco/defs_impl/viewers.h>
#include <Fresco/widgets/menus.h>
#include <Fresco/widgets/widgets.h>

// class MKBoxInfo

class MKBoxInfo : public FrescoObject {
public:
    MKBoxInfo(MenuItem_ptr item, Glyph_ptr glyph, Telltale_ptr telltale);
    ~MKBoxInfo();

    OxExchange* _exchange();

    //+ XfFrescoObject::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    //+

    MenuItem_var item_;
    Glyph_var glyph_;
    Telltale_var telltale_;
    PopupMenu_var menu_;
    Tag tag_;

protected:
    SharedFrescoObjectImpl object_;
};

declarePtrList(MKBoxInfoList,MKBoxInfo)

// class MKBox

class MKBox : public ViewerImpl {
public:
    ~MKBox();

    Boolean handle(XfGlyphTraversal_ptr t, XfEvent_ptr e); //+ XfViewer::handle

    void menu_manager(MenuManager_ptr);
    const MKBoxInfoList& children();
    void select(const MenuItem::Path& path, Boolean flap);
    void selection(MenuItem::Path& path);

protected:
    MKBox(MenuKitImpl::Info*);

    void grab(GlyphTraversal_ptr t);
    void ungrab();
    void pop_selection();
    void select_item(MenuItem_ptr item);
    void open_selection(const MenuItem::Path& path, Boolean flap);
    void close_selection();
    void unselect_item();
    void selection_allocation(RegionImpl& a);

    virtual PopupMenu_ptr create_popup_menu(MenuItem_ptr item) = 0;

    SharedFrescoObjectImpl object_;
    GlyphOffsetList parents_;
    MenuKitImpl::Info* info_;
    MKBoxInfoList children_;
    Boolean requested_;
    Requisition requisition_;
    MKBoxInfo* selection_;
    MenuManager_ptr manager_;
    Tag grab_tag_;
    Display_ptr grab_display_;
};

class MKHBox : public MKBox {
public:
    MKHBox(MenuKitImpl::Info*, MenuItem_ptr menu);
 
    /* MKBox */
    PopupMenu_ptr create_popup_menu(MenuItem_ptr item);
};
 
class MKVBox : public MKBox {
public:
    MKVBox(MenuKitImpl::Info*, MenuItem_ptr menu);
 
    /* MKBox */
    PopupMenu_ptr create_popup_menu(MenuItem_ptr item);
};
 
#endif
