//
// Copyright (c) 1994-95 Firefly Design
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

#ifndef Fresco_widgets_browser_h
#define Fresco_widgets_browser_h

#include <Fresco/widgets/widgets.h>
#include "viewport.h"

class WKGridBrowser : public GridBrowser, public WKViewport {
public:
    WKGridBrowser(Grid_ptr, XfColor_ptr);
    ~WKGridBrowser();

    OxExchange* _exchange();

    //+ GridBrowser::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyph */
    XfGlyph_ptr clone_glyph();
    XfStyleContext_ptr style();
    void style(XfStyleContext_ptr _p);
    XfTransform_ptr transformation();
    void request(XfGlyph::Requisition& r);
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r);
    void shape(XfRegion_ptr r);
    void traverse(XfGlyphTraversal_ptr t);
    void draw(XfGlyphTraversal_ptr t);
    void pick(XfGlyphTraversal_ptr t);
    XfGlyph_ptr body();
    void body(XfGlyph_ptr _p);
    void append(XfGlyph_ptr g);
    void prepend(XfGlyph_ptr g);
    XfTag add_parent(XfGlyphOffset_ptr parent_offset);
    void remove_parent(XfTag add_tag);
    XfGlyphOffset_ptr first_child_offset();
    XfGlyphOffset_ptr last_child_offset();
    void parent_offsets(XfGlyph::OffsetSeq& parents);
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void need_redraw();
    void need_redraw_region(XfRegion_ptr r);
    void need_resize();
    Boolean restore_trail(XfGlyphTraversal_ptr t);
    /* XfViewer */
    XfViewer_ptr parent_viewer();
    XfViewer_ptr next_viewer();
    XfViewer_ptr prev_viewer();
    XfViewer_ptr first_viewer();
    XfViewer_ptr last_viewer();
    void append_viewer(XfViewer_ptr v);
    void prepend_viewer(XfViewer_ptr v);
    void insert_viewer(XfViewer_ptr v);
    void replace_viewer(XfViewer_ptr v);
    void remove_viewer();
    void set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next);
    void set_first_viewer(XfViewer_ptr v);
    void set_last_viewer(XfViewer_ptr v);
    XfFocus_ptr request_focus(XfViewer_ptr requestor, Boolean temporary);
    Boolean receive_focus(XfFocus_ptr f, Boolean primary);
    void lose_focus(Boolean temporary);
    Boolean first_focus();
    Boolean last_focus();
    Boolean next_focus();
    Boolean prev_focus();
    Boolean handle(XfGlyphTraversal_ptr t, XfEvent_ptr e);
    void close();
    /* Viewport */
    Adjustment_ptr scroll_adjustment(XfAxis a);
    /* GridBrowser */
    void select(const Grid::IndexSeq& s);
    Grid::IndexSeq* selection();
    XfAction_ptr select_action();
    void select_action(XfAction_ptr _p);
    //+

protected:
    void damage(Region_ptr a, Region_ptr b, const Grid::Index& index);
    void allocation(
        Region_ptr a, Region_ptr b, const Grid::Index& index, Region_ptr r
    );
    void select_move(XfGlyphTraversal_ptr, XfEvent_ptr, Long cols, Long rows);
    void select_point(XfGlyphTraversal_ptr, XfEvent_ptr);


    Grid_var grid_;
    XfColor_var color_;
    Grid::IndexSeq selection_;
    Action_var select_action_;
};

#endif
