//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-93 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <Fresco/defs/display.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/action.h>
#include <Fresco/widgets/widgets.h>
#include "buttons.h"

ButtonImpl::ButtonImpl(Telltale_ptr t, Action_ptr a) {
    state_ = nil;
    action_ = nil;
    tag_ = 0;
    state(t);
    click_action(a);
    style_.alias(FrescoLib::tmp_string_ref("Button"));
}

ButtonImpl::~ButtonImpl() {
    detach_state();
    CORBA::release(action_);
}

OxExchange* ButtonImpl::_exchange() { return ActiveViewer::_exchange(); }

//+ ButtonImpl(XfFrescoObject::attach)
XfTag ButtonImpl::attach(XfFrescoObject_ptr observer) {
    return ActiveViewer::attach(observer);
}

//+ ButtonImpl(XfFrescoObject::detach)
void ButtonImpl::detach(XfTag attach_tag) { ActiveViewer::detach(attach_tag); }

//+ ButtonImpl(XfFrescoObject::disconnect)
void ButtonImpl::disconnect() {
    state_ = nil;
}

//+ ButtonImpl(XfFrescoObject::notify_observers)
void ButtonImpl::notify_observers() { ActiveViewer::notify_observers(); }

//+ ButtonImpl(XfFrescoObject::update)
void ButtonImpl::update() { need_redraw(); }

//+ ButtonImpl(XfGlyph::=ActiveViewer::)
XfGlyph_ptr ButtonImpl::clone_glyph() {
    return ActiveViewer::clone_glyph();
}
XfStyleContext_ptr ButtonImpl::style() {
    return ActiveViewer::style();
}
void ButtonImpl::style(XfStyleContext_ptr _p) {
    ActiveViewer::style(_p);
}
XfTransform_ptr ButtonImpl::transformation() {
    return ActiveViewer::transformation();
}
void ButtonImpl::request(XfGlyph::Requisition& r) {
    ActiveViewer::request(r);
}
void ButtonImpl::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    ActiveViewer::extension(a, r);
}
void ButtonImpl::shape(XfRegion_ptr r) {
    ActiveViewer::shape(r);
}
void ButtonImpl::traverse(XfGlyphTraversal_ptr t) {
    ActiveViewer::traverse(t);
}
void ButtonImpl::draw(XfGlyphTraversal_ptr t) {
    ActiveViewer::draw(t);
}
void ButtonImpl::pick(XfGlyphTraversal_ptr t) {
    ActiveViewer::pick(t);
}
XfGlyph_ptr ButtonImpl::body() {
    return ActiveViewer::body();
}
void ButtonImpl::body(XfGlyph_ptr _p) {
    ActiveViewer::body(_p);
}
void ButtonImpl::append(XfGlyph_ptr g) {
    ActiveViewer::append(g);
}
void ButtonImpl::prepend(XfGlyph_ptr g) {
    ActiveViewer::prepend(g);
}
XfTag ButtonImpl::add_parent(XfGlyphOffset_ptr parent_offset) {
    return ActiveViewer::add_parent(parent_offset);
}
void ButtonImpl::remove_parent(XfTag add_tag) {
    ActiveViewer::remove_parent(add_tag);
}
XfGlyphOffset_ptr ButtonImpl::first_child_offset() {
    return ActiveViewer::first_child_offset();
}
XfGlyphOffset_ptr ButtonImpl::last_child_offset() {
    return ActiveViewer::last_child_offset();
}
void ButtonImpl::parent_offsets(XfGlyph::OffsetSeq& parents) {
    ActiveViewer::parent_offsets(parents);
}
void ButtonImpl::allocations(XfGlyph::AllocationInfoSeq& a) {
    ActiveViewer::allocations(a);
}
void ButtonImpl::need_redraw() {
    ActiveViewer::need_redraw();
}
void ButtonImpl::need_redraw_region(XfRegion_ptr r) {
    ActiveViewer::need_redraw_region(r);
}
void ButtonImpl::need_resize() {
    ActiveViewer::need_resize();
}
Boolean ButtonImpl::restore_trail(XfGlyphTraversal_ptr t) {
    return ActiveViewer::restore_trail(t);
}
//+

//+ ButtonImpl(XfViewer::=ActiveViewer::)
XfViewer_ptr ButtonImpl::parent_viewer() {
    return ActiveViewer::parent_viewer();
}
XfViewer_ptr ButtonImpl::next_viewer() {
    return ActiveViewer::next_viewer();
}
XfViewer_ptr ButtonImpl::prev_viewer() {
    return ActiveViewer::prev_viewer();
}
XfViewer_ptr ButtonImpl::first_viewer() {
    return ActiveViewer::first_viewer();
}
XfViewer_ptr ButtonImpl::last_viewer() {
    return ActiveViewer::last_viewer();
}
void ButtonImpl::append_viewer(XfViewer_ptr v) {
    ActiveViewer::append_viewer(v);
}
void ButtonImpl::prepend_viewer(XfViewer_ptr v) {
    ActiveViewer::prepend_viewer(v);
}
void ButtonImpl::insert_viewer(XfViewer_ptr v) {
    ActiveViewer::insert_viewer(v);
}
void ButtonImpl::replace_viewer(XfViewer_ptr v) {
    ActiveViewer::replace_viewer(v);
}
void ButtonImpl::remove_viewer() {
    ActiveViewer::remove_viewer();
}
void ButtonImpl::set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next) {
    ActiveViewer::set_viewer_links(parent, prev, next);
}
void ButtonImpl::set_first_viewer(XfViewer_ptr v) {
    ActiveViewer::set_first_viewer(v);
}
void ButtonImpl::set_last_viewer(XfViewer_ptr v) {
    ActiveViewer::set_last_viewer(v);
}
XfFocus_ptr ButtonImpl::request_focus(XfViewer_ptr requestor, Boolean temporary) {
    return ActiveViewer::request_focus(requestor, temporary);
}
Boolean ButtonImpl::receive_focus(XfFocus_ptr f, Boolean primary) {
    return ActiveViewer::receive_focus(f, primary);
}
void ButtonImpl::lose_focus(XfFocus_ptr f, Boolean primary, Boolean temporary) {
    ActiveViewer::lose_focus(f, primary, temporary);
}
Boolean ButtonImpl::first_focus() {
    return ActiveViewer::first_focus();
}
Boolean ButtonImpl::last_focus() {
    return ActiveViewer::last_focus();
}
Boolean ButtonImpl::next_focus() {
    return ActiveViewer::next_focus();
}
Boolean ButtonImpl::prev_focus() {
    return ActiveViewer::prev_focus();
}
Boolean ButtonImpl::handle(XfGlyphTraversal_ptr t, XfEvent_ptr e) {
    return ActiveViewer::handle(t, e);
}
void ButtonImpl::close() {
    ActiveViewer::close();
}
//+

//+ ButtonImpl(Button::state=t)
void ButtonImpl::state(XfTelltale_ptr t) {
    detach_state();
    state_ = Telltale::_duplicate(t);
    tag_ = state_->attach(Button_ptr(this));
}

//+ ButtonImpl(Button::state?)
XfTelltale_ptr ButtonImpl::state() {
    return Telltale::_duplicate(state_);
}

//+ ButtonImpl(Button::click_action=a)
void ButtonImpl::click_action(XfAction_ptr a) {
    action_ = Action::_duplicate(a);
}

//+ ButtonImpl(Button::click_action?)
XfAction_ptr ButtonImpl::click_action() {
    return Action::_duplicate(action_);
}

void ButtonImpl::enter() {
    if (state_->test(Telltale::enabled)) {
	state_->set(Telltale::visible);
    }
}

void ButtonImpl::leave() {
    if (state_->test(Telltale::enabled)) {
	state_->clear(Telltale::visible);
    }
}

Boolean ButtonImpl::press(GlyphTraversal_ptr, Event_ptr) {
    if (state_->test(Telltale::enabled)) {
	state_->set(Telltale::active);
    }
    return true;
}

Boolean ButtonImpl::release(GlyphTraversal_ptr t, Event_ptr) {
    Telltale_ptr s = state_;
    if (s->test(Telltale::enabled)) {
	s->clear(Telltale::active);
	if (inside(t)) {
	    Boolean chosen = s->test(Telltale::chosen);
	    if (s->test(Telltale::toggle)) {
		if (chosen) {
		    s->clear(Telltale::chosen);
		} else {
		    s->set(Telltale::chosen);
		}
		click();
	    } else {
		if (s->test(Telltale::choosable)) {
		    s->set(Telltale::chosen);
		}
		if (!chosen) {
		    click();
		}
	    }
	}
    }
    return true;
}

void ButtonImpl::click() {
    if (is_not_nil(action_)) {
	state_->set(Telltale::running);
	action_->execute();
	state_->clear(Telltale::running);
    }
}

void ButtonImpl::detach_state() {
    if (is_not_nil(state_)) {
	state_->detach(tag_);
	CORBA::release(state_);
    }
}
