/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#include "dialog_kit.h"
#include <Fresco/defs/display.h>
#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs/fcontext.h>
#include <Fresco/defs/viewer.h>
#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/names.h>
#include <Fresco/defs_impl/viewers.h>
#include <Fresco/figures/figures.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/dialogs.h>
#include <Fresco/widgets/widgets.h>
#include <Fresco/widgets/applic.h>
#include <stdio.h>
#include <string.h>

// class DKMessageDialog

DKMessageDialog::DKMessageDialog(
    DialogKitImpl::Info* info, CharString_ptr t, Glyph_ptr g
) {
    info_ = info;
    title_ = CharString::_duplicate(t);
    message_glyph_ = Glyph::_duplicate(g);
    flavor_glyph_ = nil;
    for (int i = 0; i < none; ++i) {
        choices_[i].choice = none;
        choices_[i].glyph = nil;
        choices_[i].action = nil;
    }
    choice_ = none;
    default_button_ = nil;
    viewer_ = nil;
    window_ = nil;
}

DKMessageDialog::~DKMessageDialog() {
    CORBA::release(title_);
    CORBA::release(message_glyph_);
    CORBA::release(flavor_glyph_);
    for (int i = 0; i < none; ++i) {
        CORBA::release(choices_[i].glyph);
        CORBA::release(choices_[i].action);
    }
    CORBA::release(default_button_);
    CORBA::release(viewer_);
    CORBA::release(window_);
}

OxExchange* DKMessageDialog::_exchange() { return &object_; }

//+ DKMessageDialog(XfFrescoObject::=object_.)
XfTag DKMessageDialog::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void DKMessageDialog::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void DKMessageDialog::disconnect() {
    object_.disconnect();
}
void DKMessageDialog::notify_observers() {
    object_.notify_observers();
}
void DKMessageDialog::update() {
    object_.update();
}
//+

void DKMessageDialog::show(Window_ptr w) {
    if (is_nil(window_)) {
        build();

        Screen_ptr screen = w->screen_ptr();
        window_ = screen->top_level(viewer_, w);
        window_->title(title_);
    }
    // !!! want to be able to change transient_for of TransientWindow -denis

    position(w);
    window_->map();
}

void DKMessageDialog::hide() {
    window_->unmap();
}

void DKMessageDialog::done() {
    done_ = true;
    done_default_ = false;
    choice_ = none;
}

void DKMessageDialog::post_modal(Window_ptr w) {
    if (is_nil(window_)) {
        build();

        Screen_ptr screen = w->screen_ptr();
        window_ = screen->transient(viewer_, w);
        window_->title(title_);
    }
    // !!! want to be able to change transient_for of TransientWindow -denis

    Display_ptr display = window_->screen_ptr()->display_ptr();
    Tag filter_tag = display->add_filter(new DKFilter(this));
    position(w);
    window_->map();

    done_ = false;
    while (! done_) {
        display->run_one_event();
    }

    if (done_default_) {
        activate_default();
    }
    window_->unmap();
    display->remove_filter(filter_tag);
}

void DKMessageDialog::show_flavor(MessageDialog::Flavor, Glyph_ptr g) {
  flavor_glyph_ = Glyph::_duplicate(g);
}

void DKMessageDialog::show_choice(
    MessageDialog::Choice c, Glyph_ptr g, Action_ptr a
) {
    for (int i = 0; (i < none) && (choices_[i].choice != none) &&
        (choices_[i].choice != c); ++i) ;
    ChoiceInfo& info = choices_[i];
    info.choice = c;
    CORBA::release(info.glyph);
    info.glyph = Glyph::_duplicate(g);
    CORBA::release(info.action);
    info.action = Action::_duplicate(a);
}

void DKMessageDialog::hide_choice(MessageDialog::Choice c) {
    for (int i = 0; (i < none) && (choices_[i].choice != none) &&
        (choices_[i].choice != c); ++i) ;
    {
        ChoiceInfo& info = choices_[i];
        info.choice = c;
        CORBA::release(info.glyph);
        CORBA::release(info.action);
    }

    for (; i < (none - 1); ++i) {
        choices_[i] = choices_[i + 1];
    }
    {
        ChoiceInfo& info = choices_[i];
        info.choice = none;
        info.glyph = nil;
        info.action = nil;
    }
}

void DKMessageDialog::choice_action(MessageDialog::Choice c, Action_ptr a) {
    for (int i = 0; (i < none) && (choices_[i].choice != none) &&
        (choices_[i].choice != c); ++i) ;
    ChoiceInfo& info = choices_[i];
    info.choice = c;
    CORBA::release(info.action);
    info.action = Action::_duplicate(a);
}

void DKMessageDialog::set_choice(MessageDialog::Choice c) {
    done_ = true;
    done_default_ = false;
    choice_ = c;
}

MessageDialog::Choice DKMessageDialog::get_choice() {
    return choice_;
}

MessageDialog::Choice DKMessageDialog::post_modal_choice(Window_ptr w) {
    post_modal(w);
    return get_choice();
}

Window_ptr DKMessageDialog::window() {
    return window_;
}

void DKMessageDialog::done_default() {
    done_ = true;
    done_default_ = true;
    choice_ = choices_[0].choice;
}

Glyph_ptr DKMessageDialog::custom_glyph() {
    return nil;
}

void DKMessageDialog::build() {
    LayoutKit_ptr layout = info_->layout;
    WidgetKit_ptr widget = info_->widget;

    Glyph_ptr info_box = layout->hbox_align_elements(Alignment(0.5));
    info_box->append(Glyph_var(layout->hfil()));
    if (is_not_nil(flavor_glyph_)) {
        info_box->append(Glyph_var(flavor_glyph_));
        info_box->append(Glyph_var(layout->hspace(info_->flavor_space)));
    }
    info_box->append(Glyph_var(message_glyph_));
    info_box->append(Glyph_var(layout->hfil()));

    Glyph_ptr info_area = layout->vbox_align_elements(Alignment(0.5));
    info_area->append(Glyph_var(info_box));
    Glyph_ptr custom = custom_glyph();
    if (is_not_nil(custom)) {
        info_area->append(Glyph_var(custom));
    }

    Glyph_ptr button_box = layout->hbox_align_elements(Alignment(0.5));
    button_box->append(Glyph_var(layout->hfil()));
    for (int i = 0; (i < none) && (choices_[i].choice != none); ++i) {
        ChoiceInfo& info = choices_[i];
        if (i == 0) {
            default_button_ = widget->default_button(info.glyph, info.action);
            button_box->append(default_button_);
        } else {
            button_box->append(Glyph_var(layout->hspace(info_->button_space)));
            button_box->append(
                Glyph_var(widget->push_button(info.glyph, info.action))
            );
        }
    }
    button_box->append(Glyph_var(layout->hfil()));

    Glyph_ptr top_box = layout->vbox_align_elements(Alignment(0.5));
    top_box->append(
        Glyph_var(layout->margin(Glyph_var(info_area), info_->message_area_margin))
    );
    top_box->append(Glyph_var(widget->vseparator()));
    top_box->append(
        Glyph_var(layout->margin(Glyph_var(button_box), info_->button_area_margin))
    );

    viewer_ = new ViewerImpl(true);
    viewer_->body(
        Glyph_var(layout->back(Glyph_var(top_box), Glyph_var(widget->filler())))
    );
}

void DKMessageDialog::position(Window_ptr show_for) {
    Window::Placement p;
    show_for->get_configuration(true, p);
    p.x += -(p.align_x * p.width) + (p.width / 2);
    p.y += -(p.align_y * p.height) + (p.height / 2);
    p.align_x = p.align_y = Alignment(0.5);
    Glyph::Requisition r;
    window_->main_viewer()->request(r);
    p.width = r.x.natural;
    p.height = r.y.natural;
    window_->configure(p);
}

void DKMessageDialog::activate_default() {
    Display_ptr display = window_->screen_ptr()->display_ptr();

    Telltale_ptr telltale = default_button_->state();
    Action_ptr action = default_button_->click_action();

    // set the button state
    telltale->set(Telltale::visible);
    telltale->set(Telltale::active);
    telltale->set(Telltale::chosen);
    telltale->set(Telltale::running);

    // wait for the user to see it
    display->repair();
    display->flush_and_wait();
    FrescoLib::delay(info_->delay);

    // execute the action
    action->execute();

    // clear the button state
/*
!!! this commented out section is a hack to avoid deadlocking the current
implementation of display.  It deadlocks when the display damage list has
the last reference to a window. -denis
    telltale->clear(Telltale::running);
    telltale->clear(Telltale::chosen);
    telltale->clear(Telltale::active);
    telltale->clear(Telltale::visible);
*/
    CORBA::release(telltale);
    CORBA::release(action);
}

// class DKPromptDialog

DKPromptDialog::DKPromptDialog(
    DialogKitImpl::Info* info, CharString_ptr title, Glyph_ptr message
) : DKMessageDialog(info, title, message) {
    WidgetKit_ptr widget = info->widget;
    FieldSubject_var fsubj = widget->field_subject();
    field_editor_ = widget->field_editor(
	fsubj, info->field_width, info->field_maximum
    );
}

DKPromptDialog::~DKPromptDialog() {
    CORBA::release(field_editor_);
}

OxExchange* DKPromptDialog::_exchange() {
    return DKMessageDialog::_exchange();
}

//+ DKPromptDialog(XfFrescoObject::=DKMessageDialog::)
XfTag DKPromptDialog::attach(XfFrescoObject_ptr observer) {
    return DKMessageDialog::attach(observer);
}
void DKPromptDialog::detach(XfTag attach_tag) {
    DKMessageDialog::detach(attach_tag);
}
void DKPromptDialog::disconnect() {
    DKMessageDialog::disconnect();
}
void DKPromptDialog::notify_observers() {
    DKMessageDialog::notify_observers();
}
void DKPromptDialog::update() {
    DKMessageDialog::update();
}
//+

//+ DKPromptDialog(Dialog::=DKMessageDialog::)
void DKPromptDialog::show(XfWindow_ptr show_for) {
    DKMessageDialog::show(show_for);
}
void DKPromptDialog::hide() {
    DKMessageDialog::hide();
}
void DKPromptDialog::done() {
    DKMessageDialog::done();
}
void DKPromptDialog::post_modal(XfWindow_ptr show_for) {
    DKMessageDialog::post_modal(show_for);
}
//+

//+ DKPromptDialog(MessageDialog::=DKMessageDialog::)
void DKPromptDialog::show_flavor(MessageDialog::Flavor f, XfGlyph_ptr g) {
    DKMessageDialog::show_flavor(f, g);
}
void DKPromptDialog::show_choice(MessageDialog::Choice c, XfGlyph_ptr g, XfAction_ptr a) {
    DKMessageDialog::show_choice(c, g, a);
}
void DKPromptDialog::hide_choice(MessageDialog::Choice c) {
    DKMessageDialog::hide_choice(c);
}
void DKPromptDialog::choice_action(MessageDialog::Choice c, XfAction_ptr a) {
    DKMessageDialog::choice_action(c, a);
}
void DKPromptDialog::set_choice(MessageDialog::Choice c) {
    DKMessageDialog::set_choice(c);
}
MessageDialog::Choice DKPromptDialog::get_choice() {
    return DKMessageDialog::get_choice();
}
MessageDialog::Choice DKPromptDialog::post_modal_choice(XfWindow_ptr show_for) {
    return DKMessageDialog::post_modal_choice(show_for);
}
//+

FieldEditor_ptr DKPromptDialog::field_editor() {
    return FieldEditor::_duplicate(field_editor_);
}

Glyph_ptr DKPromptDialog::custom_glyph() {
    WidgetKit_ptr widget = info_->widget;

    return widget->inset_frame(field_editor_);
}

// class DKFilter

DKFilter::DKFilter(DKMessageDialog* d) {
    dialog_ = d;
}

OxExchange* DKFilter::_exchange() { return &object_; }

//+ DKFilter(XfFrescoObject::=object_.)
XfTag DKFilter::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void DKFilter::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void DKFilter::disconnect() {
    object_.disconnect();
}
void DKFilter::notify_observers() {
    object_.notify_observers();
}
void DKFilter::update() {
    object_.update();
}
//+

Boolean DKFilter::handle(Event_ptr e) {
    Event::TypeId type = e->type();
    if (
        (type == Event::motion) ||
        (type == Event::down) || (type == Event::up) ||
        (type == Event::key_press) || (type == Event::key_release) ||
        (type == Event::enter) || (type == Event::leave)
    ) {
        Window_ptr w = dialog_->window();
        // !!! should this be a style attribute for international use? -denis
        if ((type == Event::key_press) && (e->character() == '\r')) {
            dialog_->done_default();
        } else if (e->hits(Viewer_var(w->main_viewer()))) {
            w->handle_event(e);
        } else {
            if (type == Event::down) {
                w->screen_ptr()->display_ptr()->ring_bell(1.0f);
            }
        }
    }
    return true;
}

// class DKAction

DKAction::DKAction(MessageDialog_ptr d, MessageDialog::Choice c) {
    dialog_ = d;
    choice_ = c;
}

void DKAction::execute() {
    dialog_->set_choice(choice_);
}

// class DialogKitImpl

#define xm_error_width 20
#define xm_error_height 20
static unsigned char xm_error_bits[] = {
   0x00, 0x00, 0x00, 0xc0, 0x0f, 0x00, 0xf0, 0x3a, 0x00, 0x58, 0x55, 0x00,
   0x2c, 0xa0, 0x00, 0x56, 0x40, 0x01, 0xaa, 0x80, 0x02, 0x46, 0x81, 0x01,
   0x8a, 0x82, 0x02, 0x06, 0x85, 0x01, 0x0a, 0x8a, 0x02, 0x06, 0x94, 0x01,
   0x0a, 0xe8, 0x02, 0x14, 0x50, 0x01, 0x28, 0xb0, 0x00, 0xd0, 0x5f, 0x00,
   0xa0, 0x2a, 0x00, 0x40, 0x15, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

#define xm_information_width 11
#define xm_information_height 24
static unsigned char xm_information_bits[] = {
   0x00, 0x00, 0x78, 0x00, 0x54, 0x00, 0x2c, 0x00, 0x54, 0x00, 0x28, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x7e, 0x00, 0x2a, 0x00, 0x5c, 0x00, 0x28, 0x00,
   0x58, 0x00, 0x28, 0x00, 0x58, 0x00, 0x28, 0x00, 0x58, 0x00, 0x28, 0x00,
   0x58, 0x00, 0xae, 0x01, 0x56, 0x01, 0xaa, 0x00, 0x00, 0x00, 0x00, 0x00};


#define xm_question_width 22
#define xm_question_height 22
static unsigned char xm_question_bits[] = {
   0xf0, 0x3f, 0x00, 0x58, 0x55, 0x00, 0xac, 0xaa, 0x00, 0xd6, 0x5f, 0x01,
   0xea, 0xbf, 0x02, 0xf6, 0x7f, 0x01, 0xea, 0xba, 0x02, 0xf6, 0x7d, 0x05,
   0xea, 0xba, 0x0a, 0x56, 0x7d, 0x15, 0xaa, 0xbe, 0x1e, 0x56, 0x5f, 0x01,
   0xac, 0xaf, 0x02, 0x58, 0x57, 0x01, 0xb0, 0xaf, 0x00, 0x60, 0x55, 0x01,
   0xa0, 0xaa, 0x00, 0x60, 0x17, 0x00, 0xa0, 0x2f, 0x00, 0x60, 0x17, 0x00,
   0xb0, 0x2a, 0x00, 0x50, 0x55, 0x00};


#define xm_warning_width 9
#define xm_warning_height 22
static unsigned char xm_warning_bits[] = {
   0x00, 0x00, 0x18, 0x00, 0x2c, 0x00, 0x56, 0x00, 0x2a, 0x00, 0x56, 0x00,
   0x2a, 0x00, 0x56, 0x00, 0x2c, 0x00, 0x14, 0x00, 0x2c, 0x00, 0x14, 0x00,
   0x2c, 0x00, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3c, 0x00, 0x14, 0x00,
   0x2c, 0x00, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00};

#define xm_working_width 21
#define xm_working_height 23
static unsigned char xm_working_bits[] = {
   0x00, 0x00, 0x00, 0xfe, 0xff, 0x0f, 0xaa, 0xaa, 0x0a, 0x44, 0x55, 0x06,
   0xcc, 0x2a, 0x02, 0x44, 0x55, 0x06, 0xcc, 0x2a, 0x02, 0x84, 0x15, 0x06,
   0x8c, 0x2a, 0x02, 0x04, 0x15, 0x06, 0x0c, 0x0a, 0x02, 0x04, 0x06, 0x06,
   0x0c, 0x0b, 0x02, 0x84, 0x15, 0x06, 0xcc, 0x2a, 0x02, 0x44, 0x55, 0x06,
   0xcc, 0x2a, 0x02, 0x44, 0x55, 0x06, 0xcc, 0x2a, 0x02, 0x44, 0x55, 0x06,
   0xfe, 0xff, 0x0f, 0x56, 0x55, 0x05, 0x00, 0x00, 0x00};

struct StyleData {
    const char* path;
    const char* value;
};

static StyleData kit_attributes[] = {
    { "*DialogKit*delay", "0.2" },
    { "*DialogKit*flavorSpace", "20.0" },
    { "*DialogKit*buttonSpace", "20.0" },
    { "*DialogKit*messageAreaMargin", "10.0" },
    { "*DialogKit*buttonAreaMargin", "10.0" },
    { "*DialogKit*fieldWidth", "30" },
    { "*DialogKit*fieldMaximum", "0" },
    { nil }
};

DialogKitImpl::DialogKitImpl(WidgetKit* widget) {
    FrescoContext* fc = FrescoLib::current_context();
    info_.drawing = fc->default_drawing_kit();
    info_.layout = LayoutKit::_narrow(
        fc->resolve(*NamingLib::to_name("layouts_impl"))
    );
    info_.figure = FigureKit::_narrow(
        fc->resolve(*NamingLib::to_name("figures_impl"))
    );
    info_.widget = widget;

    StyleContext_var s = fc->style();
    style_ = info_.figure->new_style(s);
    style_->alias(FrescoLib::tmp_string_ref("DialogKit"));
    for (StyleData* d = kit_attributes; d->path != nil; d++) {
        StyleValue_var a = s->bind(FrescoLib::tmp_string_ref(d->path));
        if (a->uninitialized() || a->priority() <= -10) {
            a->write_string(FrescoLib::tmp_string_ref(d->value));
            a->priority(-10);
        }
    }

    load();
}

DialogKitImpl::~DialogKitImpl() {
    unload();
}

OxExchange* DialogKitImpl::_exchange() { return &object_; }

//+ DialogKitImpl(XfFrescoObject::=object_.)
XfTag DialogKitImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void DialogKitImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void DialogKitImpl::disconnect() {
    object_.disconnect();
}
void DialogKitImpl::notify_observers() {
    object_.notify_observers();
}
void DialogKitImpl::update() {
    object_.update();
}
//+

#define flavor(f)\
 dialog->show_flavor(MessageDialog::f, Glyph_var(flavor_glyph(MessageDialog::f)))
#define choice(c)\
 dialog->show_choice(MessageDialog::c, Glyph_var(choice_glyph(MessageDialog::c)),\
 new DKAction(dialog, MessageDialog::c))

MessageDialog_ptr DialogKitImpl::error(CharString_ptr t, Glyph_ptr g) {
    MessageDialog_ptr dialog = new DKMessageDialog(&info_, t, g);

    flavor(error);

    choice(ok);
    choice(cancel);
    choice(help);

    return dialog;
}

MessageDialog_ptr DialogKitImpl::information(CharString_ptr t, Glyph_ptr g) {
    MessageDialog_ptr dialog = new DKMessageDialog(&info_, t, g);

    flavor(information);

    choice(ok);
    choice(help);

    return dialog;
}

MessageDialog_ptr DialogKitImpl::question(CharString_ptr t, Glyph_ptr g) {
    MessageDialog_ptr dialog = new DKMessageDialog(&info_, t, g);

    flavor(question);

    choice(yes);
    choice(no);
    choice(cancel);
    choice(help);

    return dialog;
}

MessageDialog_ptr DialogKitImpl::warning(CharString_ptr t, Glyph_ptr g) {
    MessageDialog_ptr dialog = new DKMessageDialog(&info_, t, g);

    flavor(warning);

    choice(ok);
    choice(cancel);
    choice(help);

    return dialog;
}

MessageDialog_ptr DialogKitImpl::working(CharString_ptr t, Glyph_ptr g) {
    MessageDialog_ptr dialog = new DKMessageDialog(&info_, t, g);

    flavor(working);

    choice(stop);

    return dialog;
}

PromptDialog_ptr DialogKitImpl::prompt(CharString_ptr t, Glyph_ptr g) {
    PromptDialog_ptr dialog = new DKPromptDialog(&info_, t, g);

    choice(ok);
    choice(apply);
    choice(reset);
    choice(cancel);
    choice(help);

    return dialog;
}

Glyph_ptr DialogKitImpl::flavor_glyph(MessageDialog::Flavor f) {
    return Glyph::_duplicate(info_.flavors[(int) f]);
}

Glyph_ptr DialogKitImpl::choice_glyph(MessageDialog::Choice c) {
    return Glyph::_duplicate(info_.choices[(int) c]);
}

void DialogKitImpl::load() {
    load_float("delay", info_.delay);

    load_coord("flavorSpace", info_.flavor_space);
    load_coord("buttonSpace", info_.button_space);
    load_coord("messageAreaMargin", info_.message_area_margin);
    load_coord("buttonAreaMargin", info_.button_area_margin);

    load_Long("fieldWidth", info_.field_width);
    load_Long("fieldMaximum", info_.field_maximum);

    load_stencil(
        "error", info_.flavors[MessageDialog::error],
        xm_error_bits, xm_error_width, xm_error_height
    );
    load_stencil(
        "information", info_.flavors[MessageDialog::information],
        xm_information_bits, xm_information_width, xm_information_height
    );
    load_stencil(
        "question", info_.flavors[MessageDialog::question],
        xm_question_bits, xm_question_width, xm_question_height
    );
    load_stencil(
        "warning", info_.flavors[MessageDialog::warning],
        xm_warning_bits, xm_warning_width, xm_warning_height
    );
    load_stencil(
        "working", info_.flavors[MessageDialog::working],
        xm_working_bits, xm_working_width, xm_working_height
    );

    load_label("OK", info_.choices[MessageDialog::ok]);
    load_label("Yes", info_.choices[MessageDialog::yes]);
    load_label("Retry", info_.choices[MessageDialog::retry]);
    load_label("Apply", info_.choices[MessageDialog::apply]);
    load_label("Reset", info_.choices[MessageDialog::reset]);
    load_label("Close", info_.choices[MessageDialog::close]);
    load_label("Pause", info_.choices[MessageDialog::pause]);
    load_label("No", info_.choices[MessageDialog::no]);
    load_label("Stop", info_.choices[MessageDialog::stop]);
    load_label("Resume", info_.choices[MessageDialog::resume]);
    load_label("Cancel", info_.choices[MessageDialog::cancel]);
    load_label("Help", info_.choices[MessageDialog::help]);
}

void DialogKitImpl::load_coord(const char* name, Coord& c) {
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
	a->read_coord(c);
    }
}

void DialogKitImpl::load_float(const char* name, float& f) {
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
	double d;
	if (a->read_real(d)) {
	    f = float(d);
	}
    }
}

void DialogKitImpl::load_Long(const char* name, Long& l) {
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
	Long v;
	if (a->read_integer(v)) {
	    l = Long(v);
	}
    }
}

static Raster_ptr make_bitmap(
    DrawingKit_ptr d,
    unsigned char* data, unsigned int width, unsigned int height
) {
    Long n = height * ((width + 7) / 8);
    DrawingKit::Data8 seq(n, n, data);
    return d->bitmap_from_data(seq, height, width, 0, 0);
}

void DialogKitImpl::load_stencil(
    const char* name, Glyph_var& g,
    unsigned char* data, unsigned int width, unsigned int height
) {
    DrawingKit_ptr drawing = info_.drawing;
    FigureKit_ptr figure = info_.figure;

    Raster_ptr raster = nil;
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
        CharString_ptr s;
        if (a->read_string(s)) {
            raster = drawing->raster_open(s);
            CORBA::release(s);
            // !!! should warn user if can't get raster -denis
        }
    }
    if (is_nil(raster)) {
        raster = make_bitmap(drawing, data, width, height);
    }
    g = figure->stencil(style_, Raster_var(raster));
}

void DialogKitImpl::load_label(const char* name, Glyph_var& g) {
    FigureKit_ptr figure = info_.figure;

    g = nil;
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
        CharString_ptr s;
        if (a->read_string(s)) {
            g = figure->label(style_, s);
            CORBA::release(s);
        }
    }
    if (is_nil(g)) {
        g = figure->label(style_, FrescoLib::tmp_string_ref(name));
    }
}

void DialogKitImpl::unload() {
}
