/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_widgets_impl_dialog_kit_h
#define Fresco_widgets_impl_dialog_kit_h

#include <Fresco/defs_impl/action.h>
#include <Fresco/figures/figures.h>
#include <Fresco/widgets/dialogs.h>
#include <Fresco/widgets/widgets.h>

class DrawingKit;
class LayoutKit;

// class DialogKitImpl

class DialogKitImpl : public DialogKit {
    //. DialogKitImpl provides an implementation of the DialogKit
    //. interface that creates objects with a Motif-like look and feel.
public:
    DialogKitImpl(WidgetKit*);
    ~DialogKitImpl();

    OxExchange* _exchange();

    //+ DialogKit::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* DialogKit */
    MessageDialog_ptr error(XfCharString_ptr title, XfGlyph_ptr message);
    MessageDialog_ptr information(XfCharString_ptr title, XfGlyph_ptr message);
    MessageDialog_ptr question(XfCharString_ptr title, XfGlyph_ptr message);
    MessageDialog_ptr warning(XfCharString_ptr title, XfGlyph_ptr message);
    MessageDialog_ptr working(XfCharString_ptr title, XfGlyph_ptr message);
    PromptDialog_ptr prompt(XfCharString_ptr title, XfGlyph_ptr message);
    XfGlyph_ptr flavor_glyph(MessageDialog::Flavor f);
    XfGlyph_ptr choice_glyph(MessageDialog::Choice c);
    //+

    struct Info {
        DrawingKit* drawing;
        LayoutKit* layout;
        FigureKit* figure;
        WidgetKit* widget;

        Coord delay;

        Coord flavor_space;
        Coord button_space;
        Coord message_area_margin;
        Coord button_area_margin;

        Long field_width;
        Long field_maximum;

        Glyph_var flavors[5];

        Glyph_var choices[12];
    };

protected:
    SharedFrescoObjectImpl object_;
    FigureStyle_var style_;
    DialogKitImpl::Info info_;

    void load();
    void load_coord(const char*, Coord&);
    void load_float(const char*, float&);
    void load_Long(const char*, Long&);
    void load_stencil(
        const char*, Glyph_var&,
        unsigned char* data, unsigned int width, unsigned int height
    );
    void load_label(const char*, Glyph_var&);
    void unload();
};

// class DKMessageDialog

class DKMessageDialog : _Ix_public MessageDialog {
public:
    DKMessageDialog(DialogKitImpl::Info*, CharString_ptr, Glyph_ptr);
    ~DKMessageDialog();

    OxExchange* _exchange();

    //+ MessageDialog::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* Dialog */
    void show(XfWindow_ptr show_for);
    void hide();
    void done();
    void post_modal(XfWindow_ptr show_for);
    /* MessageDialog */
    void show_flavor(MessageDialog::Flavor f, XfGlyph_ptr g);
    void show_choice(MessageDialog::Choice c, XfGlyph_ptr g, XfAction_ptr a);
    void hide_choice(MessageDialog::Choice c);
    void choice_action(MessageDialog::Choice c, XfAction_ptr a);
    void set_choice(MessageDialog::Choice c);
    MessageDialog::Choice get_choice();
    MessageDialog::Choice post_modal_choice(XfWindow_ptr show_for);
    //+

    Window_ptr window();
    void done_default();

protected:
    struct ChoiceInfo {
        Choice choice;
        Glyph_ptr glyph;
        Action_ptr action;
    };

    virtual Glyph_ptr custom_glyph();

    void build();
    void position(Window_ptr);
    void activate_default();

    SharedFrescoObjectImpl object_;
    DialogKitImpl::Info* info_;
    CharString_ptr title_;
    Glyph_ptr message_glyph_;
    Glyph_ptr flavor_glyph_;
    ChoiceInfo choices_[none];
    Button_ptr default_button_;
    Viewer_ptr viewer_;
    Window_ptr window_;
    Choice choice_;
    Boolean done_:1;
    Boolean done_default_:1;
};

// class DKPromptDialog

class DKPromptDialog : _Ix_public DKMessageDialog, _Ix_public PromptDialog {
public:
    DKPromptDialog(DialogKitImpl::Info*, CharString_ptr, Glyph_ptr);
    ~DKPromptDialog();

    OxExchange* _exchange();

    _Ix_ClassInterfaceOps(PromptDialog)

    //+ PromptDialog::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* Dialog */
    void show(XfWindow_ptr show_for);
    void hide();
    void done();
    void post_modal(XfWindow_ptr show_for);
    /* MessageDialog */
    void show_flavor(MessageDialog::Flavor f, XfGlyph_ptr g);
    void show_choice(MessageDialog::Choice c, XfGlyph_ptr g, XfAction_ptr a);
    void hide_choice(MessageDialog::Choice c);
    void choice_action(MessageDialog::Choice c, XfAction_ptr a);
    void set_choice(MessageDialog::Choice c);
    MessageDialog::Choice get_choice();
    MessageDialog::Choice post_modal_choice(XfWindow_ptr show_for);
    /* PromptDialog */
    FieldEditor_ptr field_editor();
    //+

protected:
    Glyph_ptr custom_glyph();

    FieldEditor_ptr field_editor_;
};

// class DKFilter

class DKFilter : public EventFilter {
public:
    DKFilter(DKMessageDialog*);

    OxExchange* _exchange();

    //+ XfEventFilter::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfEventFilter */
    Boolean handle(XfEvent_ptr e);
    //+

protected:
    SharedFrescoObjectImpl object_;
    DKMessageDialog* dialog_;
};

// class DKAction

class DKAction : public ActionImpl {
public:
    DKAction(MessageDialog_ptr, MessageDialog::Choice);

    void execute();

protected:
    MessageDialog_ptr dialog_;
    MessageDialog::Choice choice_;
};

#endif
