/*
 * Copyright (c) 1993-94 Silicon Graphics, Inc.
 * Copyright (c) 1993-94 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_widgets_field_h
#define Fresco_widgets_field_h

#include <Fresco/defs_impl/viewers.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/thread.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/widgets/widgets.h>
#include <Fresco/widgets/menus.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/applic.h>

#include "applics.h"

class BoundedField;
class CharBox;
class FieldEditorImpl_UpdateCallback;  // $$$ Fix UpdateCallback (no template)

class FieldEditorImpl : public FieldEditor, public ViewerImpl {
public:
    FieldEditorImpl(
	ExclusiveEditMgr_ptr mgr, Tag tag, FieldEditorData_ptr
    );
    ~FieldEditorImpl();

    OxExchange* _exchange();

    _Ix_ClassInterfaceOps(FieldEditor)
    
    //+ FieldEditor::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyph */
    XfGlyph_ptr clone_glyph();
    XfStyleContext_ptr style();
    void style(XfStyleContext_ptr _p);
    XfTransform_ptr transformation();
    void request(XfGlyph::Requisition& r);
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r);
    void shape(XfRegion_ptr r);
    void traverse(XfGlyphTraversal_ptr t);
    void draw(XfGlyphTraversal_ptr t);
    void pick(XfGlyphTraversal_ptr t);
    XfGlyph_ptr body();
    void body(XfGlyph_ptr _p);
    void append(XfGlyph_ptr g);
    void prepend(XfGlyph_ptr g);
    XfTag add_parent(XfGlyphOffset_ptr parent_offset);
    void remove_parent(XfTag add_tag);
    XfGlyphOffset_ptr first_child_offset();
    XfGlyphOffset_ptr last_child_offset();
    void parent_offsets(XfGlyph::OffsetSeq& parents);
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void need_redraw();
    void need_redraw_region(XfRegion_ptr r);
    void need_resize();
    Boolean restore_trail(XfGlyphTraversal_ptr t);
    /* XfViewer */
    XfViewer_ptr parent_viewer();
    XfViewer_ptr next_viewer();
    XfViewer_ptr prev_viewer();
    XfViewer_ptr first_viewer();
    XfViewer_ptr last_viewer();
    void append_viewer(XfViewer_ptr v);
    void prepend_viewer(XfViewer_ptr v);
    void insert_viewer(XfViewer_ptr v);
    void replace_viewer(XfViewer_ptr v);
    void remove_viewer();
    void set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next);
    void set_first_viewer(XfViewer_ptr v);
    void set_last_viewer(XfViewer_ptr v);
    XfFocus_ptr request_focus(XfViewer_ptr requestor, Boolean temporary);
    Boolean receive_focus(XfFocus_ptr f, Boolean primary);
    void lose_focus(XfFocus_ptr f, Boolean primary, Boolean temporary);
    Boolean first_focus();
    Boolean last_focus();
    Boolean next_focus();
    Boolean prev_focus();
    Boolean handle(XfGlyphTraversal_ptr t, XfEvent_ptr e);
    void close();
    /* FieldEditor */
    ExclusiveEditMgr_ptr field_edit_manager();
    FieldSubject_ptr field_subject();
    Adjustment_ptr scroll_adjustment();
    //+
    
    void insert_char(XfCharCode c);
    void delete_prev_char();
    void delete_next_char();
    void move_left();
    void move_right();
    void move_to_end_line();
    void move_to_beg_line();
    void select_all();

    Boolean press(GlyphTraversal_ptr, Event_ptr); 
    Boolean drag(GlyphTraversal_ptr, Event_ptr); 
    Boolean release(GlyphTraversal_ptr, Event_ptr);
    Boolean key_press(GlyphTraversal_ptr, Event_ptr);

private:
    LayoutKit_var layouts_;
    ThreadKit_var threads_;
    MenuKit_var menus_;
    BoundedField* bfield_;
    ExclusiveEditMgr_var editmgr_;
    FieldSubject_var fsubj_;
    Tag ftag_;
    Action_var fsubj_callback_;
    Action_var scroll_callback_;
    long max_chars_;
    long char_count_;
    CharBox* charbox_;
    
    Coord visible_span_;
    GlyphOffset_var point_, mark_;
    Coord lmargin_, rmargin_, bmargin_, tmargin_;
    Coord left_, bottom_, right_, top_;

    Boolean have_focus_;
    Boolean have_access_;
    FieldEditorImpl_UpdateCallback* editmgr_watcher_;  // $$$ Fix 
    Boolean text_change_;
    Brush_var edit_brush_;
    Color_var edit_color_;
    Color_var busy_color_;
    Color_var nofocus_color_;
    Thread_var ticker_;
    Mutex_var lock_;
    ConditionVariable_var tickwait_;
    Boolean ticking_;
    float scroll_sensitivity_; 
    Coord edge_;
    Vertex edge_pt_;
    float delay_;
    Coord delta_;
    
    Font_var font_;
    Color_var foreground_, background_;
private:
    void scroll_update();
    void fsubj_update();
    void change_fsubj();
    GlyphOffset_ptr current_point(GlyphTraversal_ptr t);
    Coord find_pos(GlyphOffset_ptr);
    void local_edges(GlyphOffset_ptr o, Coord& left, Coord &right);
    Boolean find_point_during_drag(GlyphTraversal_ptr t);
    void delete_mark_region();
    Thread_ptr start_ticker();
    void tick();
    Boolean is_equal(GlyphOffset_ptr, GlyphOffset_ptr);
    void beep();
    void scroll_to(Coord);
    void scroll_by(Coord);
    Coord scroll_pos();
    void draw_caret(Painter_ptr);
    void draw_region(Painter_ptr);
    void set_color(Painter_ptr);
    void adjust_for_movement_fwd();
    void adjust_for_movement_bwd();
    void begin_text_change();
    void end_text_change();
    void notify_ok_to_edit();
};

// class FieldEditMgr
class FieldEditMgr : public ExclusiveEditMgrImpl {
public:
    void execute(Command_ptr, FrescoObject_ptr);
};

// class FieldSubjectImpl

declareList(CharList, CharCode)
    
class FieldSubjectImpl : public FieldSubject {
public:
    FieldSubjectImpl();
    ~FieldSubjectImpl();

    OxExchange* _exchange();

    //+ FieldSubject::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* FieldSubject */
    Long count();
    void append_char(XfCharCode c);
    void prepend_char(XfCharCode c);
    void insert_char(Long index, XfCharCode c);
    void insert_chars(Long index, XfCharString_ptr str);
    void replace_char(Long index, XfCharCode c);
    void replace_chars(Long first, Long last, XfCharString_ptr str);
    void replace_all(XfCharString_ptr str);
    void delete_chars(Long first, Long last);
    void delete_all();
    XfCharCode get_char(Long index);
    char* get_chars(Long first, Long last);
    char* get_all();
    void copy(FieldSubject_ptr fs);
    //+


protected:
    CharList chars_;
private:
    void do_delete_chars(Long first, Long last);
    void do_insert_chars(Long index, XfCharString_ptr str);
    SharedFrescoObjectImpl object_;
};

class FieldEditorDataImpl : public FieldEditorData {
public:
    FieldEditorDataImpl(Long init_width, Long max_chars);
    ~FieldEditorDataImpl();
    
    OxExchange* _exchange();
    
    //+ FieldEditorData::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* ViewerData */
    void foo();
    /* FieldEditorData */
    Long init_width_in_chars();
    void init_width_in_chars(Long _p);
    Long max_chars_allowed();
    void max_chars_allowed(Long _p);
    //+
    
    Long init_width_, max_chars_;
    
private:
    SharedFrescoObjectImpl object_;
};

class FieldEditorCreator : public ViewerCreator {
public:
    FieldEditorCreator( );
    ~FieldEditorCreator();
    
    OxExchange* _exchange();
    
    //+ ViewerCreator::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* ViewerCreator */
    XfViewer_ptr create_viewer(Inset_ptr ins);
    //+
    
private:
    SharedFrescoObjectImpl object_;
};

class FieldCommandImpl : public FieldCommand {
public:
    FieldCommandImpl();
    ~FieldCommandImpl();
    
    OxExchange* _exchange();

    //+ FieldCommand::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* Command */
    Boolean reversible();
    /* FieldCommand */
    void execute(FieldSubject_ptr fs);
    void unexecute(FieldSubject_ptr fs);
    //+

private:
    SharedFrescoObjectImpl object_;
};

class PasteCommand : public FieldCommandImpl {
public:
    PasteCommand(Long first, Long last, CharString_ptr);
    ~PasteCommand();
    
    //+ Command::=
    Boolean reversible();
    //+
    
    //+ FieldCommand::=
    void execute(FieldSubject_ptr fs);
    void unexecute(FieldSubject_ptr fs);
    //+
protected:
    Long first_, last_;
    CharString_var str_;
};

class CutCommand : public FieldCommandImpl {

public:
    CutCommand(Long first, Long last);
    ~CutCommand();
    
    //+ Command::=
    Boolean reversible();
    //+
    
    //+ FieldCommand::=
    void execute(FieldSubject_ptr fs);
    void unexecute(FieldSubject_ptr fs);
    //+
protected:
    Long first_, last_;
};


#endif
