/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#include "items.h"
#include "menu_bar.h"
#include "popups.h"
#include <Fresco/defs/fresco.h>
#include <Fresco/figures/figures.h>

// class MKItemBase

MKItemBase::MKItemBase(MenuKitImpl::Info* info) {
    info_ = info;
    parent_ = nil;
}

MKItemBase::~MKItemBase() {
};

OxExchange* MKItemBase::_exchange() { return &object_; }

//+ MKItemBase(XfFrescoObject::=object_.)
XfTag MKItemBase::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void MKItemBase::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void MKItemBase::disconnect() {
    object_.disconnect();
}
void MKItemBase::notify_observers() {
    object_.notify_observers();
}
void MKItemBase::update() {
    object_.update();
}
//+
 
//+ MKItemBase(XfMenuItem::type)
XfMenuItem::TypeId MKItemBase::type() {
    return MenuItem::leaf;
}
 
//+ MKItemBase(XfMenuItem::count)
Long MKItemBase::count() {
    return 0;
}
 
//+ MKItemBase(XfMenuItem::item) !i
XfMenuItem_ptr MKItemBase::item(Long) {
    return nil;
}
 
//+ MKItemBase(XfMenuItem::prepend) !item
void MKItemBase::prepend(XfMenuItem_ptr) {
}

//+ MKItemBase(XfMenuItem::append) !item
void MKItemBase::append(XfMenuItem_ptr) {
}

//+ MKItemBase(XfMenuItem::insert) !i !item
void MKItemBase::insert(Long, XfMenuItem_ptr) {
}

//+ MKItemBase(XfMenuItem::remove) !i
void MKItemBase::remove(Long) {
}

//+ MKItemBase(XfMenuItem::remove_all)
void MKItemBase::remove_all() {
}
 
//+ MKItemBase(XfMenuItem::default_item?)
XfMenuItem_ptr MKItemBase::default_item() {
    return nil;
}
 
//+ MKItemBase(XfMenuItem::default_item=)
void MKItemBase::default_item(XfMenuItem_ptr) {
}

//+ MKItemBase(XfMenuItem::zone_glyph) !z
XfGlyph_ptr MKItemBase::zone_glyph(XfMenuItem::Zone) {
    return nil;
}

//+ MKItemBase(XfMenuItem::parent?)
XfMenuItem_ptr MKItemBase::parent() {
    return parent_;
}
 
//+ MKItemBase(XfMenuItem::parent=parent)
void MKItemBase::parent(XfMenuItem_ptr parent) {
    parent_ = parent;
}
 
//+ MKItemBase(XfMenuItem::mnemonic)
XfEvent::KeySym MKItemBase::mnemonic() {
    return 0;
}
 
//+ MKItemBase(XfMenuItem::accelerator)
XfEvent::KeyChord* MKItemBase::accelerator() {
    return nil;
}
 
//+ MKItemBase(XfMenuItem::enabled)
Boolean MKItemBase::enabled() {
    return false;
}
 
//+ MKItemBase(XfMenuItem::enable) !e
void MKItemBase::enable(Boolean) {
}
 
//+ MKItemBase(XfMenuItem::state)
XfTelltale_ptr MKItemBase::state() {
    return nil;
}

//+ MKItemBase(XfMenuItem::item_action=)
void MKItemBase::item_action(XfAction_ptr) {
}
 
//+ MKItemBase(XfMenuItem::item_action?)
XfAction_ptr MKItemBase::item_action() {
    return nil;
}
 
//+ MKItemBase(XfMenuItem::menu_bar)
XfMenuBar_ptr MKItemBase::menu_bar() {
    return nil;
}
 
//+ MKItemBase(XfMenuItem::pulldown_menu)
XfPopupMenu_ptr MKItemBase::pulldown_menu() {
    return nil;
}
 
//+ MKItemBase(XfMenuItem::pullright_menu)
XfPopupMenu_ptr MKItemBase::pullright_menu() {
    return nil;
}

// class MKItemSeparator

MKItemSeparator::MKItemSeparator(MenuKitImpl::Info* info) : MKItemBase(info) {
}

//+ MKItemSeparator(XfMenuItem::type)
XfMenuItem::TypeId MKItemSeparator::type() {
    return MenuItem::separator;
}
 
// class MKMenuItem

MKMenuItem::MKMenuItem(
    MenuKitImpl::Info* info, Event::KeySym mnemonic
) : MKItemBase(info) {
    mnemonic_ = mnemonic;
    enable_ = true;
}

//+ MKMenuItem(XfMenuItem::parent?)
XfMenuItem_ptr MKMenuItem::parent() {
    return parent_;
}

//+ MKMenuItem(XfMenuItem::parent=parent)
void MKMenuItem::parent(XfMenuItem_ptr parent) {
    parent_ = parent;
}

//+ MKMenuItem(XfMenuItem::mnemonic)
XfEvent::KeySym MKMenuItem::mnemonic() {
    return mnemonic_;
}

//+ MKMenuItem(XfMenuItem::enabled)
Boolean MKMenuItem::enabled() {
    return enable_;
}

//+ MKMenuItem(XfMenuItem::enable)
void MKMenuItem::enable(Boolean e) {
    if (enable_ != e) {
        enable_ = e;
        notify_observers();
    }
}

// class MKItemMenu

implementPtrList(MenuItemList,MenuItem)

MKItemMenu::MKItemMenu(
    MenuKitImpl::Info* info, Glyph_ptr l, Event::KeySym mnemonic
) : MKMenuItem(info, mnemonic) {
    label_ = Glyph::_duplicate(l);
    default_ = nil;
}

MKItemMenu::~MKItemMenu() {
    for (ListItr(MenuItemList) i(items_); i.more(); i.next()) {
        CORBA::release(i.cur());
    }
    CORBA::release(default_);
}

//+ MKItemMenu(XfMenuItem::type)
XfMenuItem::TypeId MKItemMenu::type() {
    return MenuItem::submenu;
}
 
//+ MKItemMenu(XfMenuItem::count)
Long MKItemMenu::count() {
    return items_.count();
}

//+ MKItemMenu(XfMenuItem::item)
XfMenuItem_ptr MKItemMenu::item(Long i) {
    return items_.item(i);
}

//+ MKItemMenu(XfMenuItem::default_item?)
XfMenuItem_ptr MKItemMenu::default_item() {
    return default_;
}

//+ MKItemMenu(XfMenuItem::default_item=item)
void MKItemMenu::default_item(XfMenuItem_ptr item) {
    default_ = MenuItem::_duplicate(item);
}

//+ MKItemMenu(XfMenuItem::prepend)
void MKItemMenu::prepend(XfMenuItem_ptr item) {
    MenuItem_ptr mi = MenuItem::_duplicate(item);
    items_.prepend(mi);
    mi->parent(this);

    notify_observers();
}
 
//+ MKItemMenu(XfMenuItem::append)
void MKItemMenu::append(XfMenuItem_ptr item) {
    MenuItem_ptr mi = MenuItem::_duplicate(item);
    items_.append(mi);
    mi->parent(this);

    notify_observers();
}
 
//+ MKItemMenu(XfMenuItem::insert)
void MKItemMenu::insert(Long i, XfMenuItem_ptr item) {
    MenuItem_ptr mi = MenuItem::_duplicate(item);
    items_.insert(i, mi);
    mi->parent(this);

    notify_observers();
}
 
//+ MKItemMenu(XfMenuItem::remove)
void MKItemMenu::remove(Long i) {
    MenuItem_ptr mi = items_.item(i);
    items_.remove(i);
    mi->parent(nil);
    CORBA::release(mi);

    notify_observers();
}
 
//+ MKItemMenu(XfMenuItem::remove_all)
void MKItemMenu::remove_all() {
    Long count = items_.count();
    for (Long i = 0; i < count; ++i) {
        MenuItem_ptr mi = items_.item(i);
        mi->parent(nil);
        CORBA::release(mi);
    }
    items_.remove_all();

    notify_observers();
}

//+ MKItemMenu(XfMenuItem::zone_glyph)
XfGlyph_ptr MKItemMenu::zone_glyph(XfMenuItem::Zone z) {
    return z == MenuItem::label ? Glyph::_duplicate(label_) : nil;
}

//+ MKItemMenu(XfMenuItem::menu_bar)
XfMenuBar_ptr MKItemMenu::menu_bar() {
    return new MKMenuBar(info_, this);
}

//+ MKItemMenu(XfMenuItem::pulldown_menu)
XfPopupMenu_ptr MKItemMenu::pulldown_menu() {
    return new MKPulldownMenu(info_, this);
}

//+ MKItemMenu(XfMenuItem::pullright_menu)
XfPopupMenu_ptr MKItemMenu::pullright_menu() {
    return new MKPullrightMenu(info_, this);
}

// class MKItemButton

MKItemButton::MKItemButton(
    MenuKitImpl::Info* info,
    Glyph_ptr m, Glyph_ptr l, Glyph_ptr s, Event::KeySym mnemonic,
    Event::KeyChord accelerator, Action_ptr action, Telltale_ptr telltale
) : MKMenuItem(info, mnemonic) {
    mark_ = Glyph::_duplicate(m);
    label_ = Glyph::_duplicate(l);
    shortcut_ = Glyph::_duplicate(s);
    accelerator_ = accelerator;
    action_ = Action::_duplicate(action);
    telltale_ = Telltale::_duplicate(telltale);
}

MKItemButton::~MKItemButton() {
}

//+ MKItemButton(XfMenuItem::type)
XfMenuItem::TypeId MKItemButton::type() {
    return MenuItem::leaf;
}
 
//+ MKItemButton(XfMenuItem::zone_glyph)
XfGlyph_ptr MKItemButton::zone_glyph(XfMenuItem::Zone z) {
    switch (z) {
        case MenuItem::mark: return Glyph::_duplicate(mark_);
        case MenuItem::label: return Glyph::_duplicate(label_);
        case MenuItem::shortcut: return Glyph::_duplicate(shortcut_);
		default: return nil;
    }
}

//+ MKItemButton(XfMenuItem::accelerator)
XfEvent::KeyChord* MKItemButton::accelerator() {
    return new Event::KeyChord(accelerator_);
}

//+ MKItemButton(XfMenuItem::state)
XfTelltale_ptr MKItemButton::state() {
    return Telltale::_duplicate(telltale_);
}

//+ MKItemButton(XfMenuItem::item_action=a)
void MKItemButton::item_action(XfAction_ptr a) {
    action_ = Action::_duplicate(a);
}

//+ MKItemButton(XfMenuItem::item_action?)
XfAction_ptr MKItemButton::item_action() {
    return Action::_duplicate(action_);
}
