/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_widgets_impl_items_h
#define Fresco_widgets_impl_items_h

#include "menu_kit.h"
#include <Fresco/defs_impl/list.h>
#include <Fresco/widgets/menus.h>

declarePtrList(MenuItemList,MenuItem)

// class MKItemBase

class MKItemBase : public MenuItem {
public:
    MKItemBase(MenuKitImpl::Info* info);
    ~MKItemBase();

    OxExchange* _exchange();
 
    //+ XfMenuItem::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfMenuItem */
    XfMenuItem::TypeId type();
    Long count();
    XfMenuItem_ptr item(Long i);
    XfMenuItem_ptr default_item();
    void default_item(XfMenuItem_ptr _p);
    void prepend(XfMenuItem_ptr item);
    void append(XfMenuItem_ptr item);
    void insert(Long i, XfMenuItem_ptr item);
    void remove(Long i);
    void remove_all();
    XfMenuItem_ptr parent();
    void parent(XfMenuItem_ptr _p);
    XfGlyph_ptr zone_glyph(XfMenuItem::Zone z);
    XfEvent::KeySym mnemonic();
    XfEvent::KeyChord* accelerator();
    void enable(Boolean e);
    Boolean enabled();
    XfTelltale_ptr state();
    XfAction_ptr item_action();
    void item_action(XfAction_ptr _p);
    XfMenuBar_ptr menu_bar();
    XfPopupMenu_ptr pulldown_menu();
    XfPopupMenu_ptr pullright_menu();
    //+
 
protected:
    SharedFrescoObjectImpl object_;
    MenuKitImpl::Info* info_;
    MenuItem_ptr parent_;
};
 
// class MKItemSeparator

class MKItemSeparator : public MKItemBase {
public:
    MKItemSeparator(MenuKitImpl::Info* info);

    MenuItem::TypeId type();
};

// class MKMenuItem

class MKMenuItem : public MKItemBase {
public:
    MenuItem_ptr parent();
    void parent(MenuItem_ptr);
    Event::KeySym mnemonic();
    Boolean enabled();
    void enable(Boolean);

protected:
    MKMenuItem(MenuKitImpl::Info* info, Event::KeySym mnemonic = 0);

    Event::KeySym mnemonic_;
    Boolean enable_;
};

// class MKItemMenu

class MKItemMenu : public MKMenuItem {
public:
    MKItemMenu(MenuKitImpl::Info* info, Glyph_ptr label, Event::KeySym mnemonic);
    ~MKItemMenu();

    /* MenuItem */
    MenuItem::TypeId type();
    Long count();
    MenuItem_ptr item(Long i);
    MenuItem_ptr default_item();
    void default_item(MenuItem_ptr item);
    void prepend(MenuItem_ptr item);
    void append(MenuItem_ptr item);
    void insert(Long i, MenuItem_ptr item);
    void remove(Long i);
    void remove_all();
    Glyph_ptr zone_glyph(MenuItem::Zone z);
    MenuBar_ptr menu_bar();
    PopupMenu_ptr pulldown_menu();
    PopupMenu_ptr pullright_menu();

protected:
    Glyph_var label_;
    MenuItemList items_;
    MenuItem_ptr default_;
};

// class MKItemButton

class MKItemButton : public MKMenuItem {
public:
    MKItemButton(
        MenuKitImpl::Info* info, Glyph_ptr mark, Glyph_ptr label, Glyph_ptr shortcut,
        Event::KeySym mnemonic, Event::KeyChord accelerator, Action_ptr action,
        Telltale_ptr telltale = nil
    );
    ~MKItemButton();

    /* MenuItem */
    MenuItem::TypeId type();
    Event::KeyChord* accelerator();
    Glyph_ptr zone_glyph(MenuItem::Zone z);
    Telltale_ptr state();
    void item_action(Action_ptr);
    Action_ptr item_action();

protected:
    Glyph_var mark_;
    Glyph_var label_;
    Glyph_var shortcut_;
    Event::KeyChord accelerator_;
    Action_var action_;
    Telltale_var telltale_;
};

#endif
