//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_widgets_kit_h
#define Fresco_widgets_kit_h

#include <Fresco/defs/drawing.h>
#include <Fresco/defs_impl/dload.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/widgets.h>
#include "bevel.h"

class Slider;

//- WidgetKitImpl*
class WidgetKitImpl : public WidgetKit {
    //. WidgetKitImpl provides an implementation of the WidgetKit
    //. interface that creates objects with a Motif-like look and feel.
    //. The bevelled look is slightly enhanced by using several levels
    //. of shading instead of a single color change.
public:
    WidgetKitImpl();
    ~WidgetKitImpl();

    OxExchange* _exchange();

    //+ WidgetKit::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* WidgetKit */
    XfGlyph_ptr inset_frame(XfGlyph_ptr g);
    XfGlyph_ptr outset_frame(XfGlyph_ptr g);
    XfGlyph_ptr bright_inset_frame(XfGlyph_ptr g);
    XfGlyph_ptr hilite_frame(XfGlyph_ptr g, XfTelltale_ptr t);
    XfGlyph_ptr filler();
    XfGlyph_ptr color_filler(XfColor_ptr c);
    XfGlyph_ptr hseparator();
    XfGlyph_ptr vseparator();
    FieldEditor_ptr field_editor(FieldSubject_ptr subj, Long width, Long maximum);
    FieldEditor_ptr managed_field_editor(ExclusiveEditMgr_ptr mgr, XfTag subj, FieldEditorData_ptr fe_data);
    FieldSubject_ptr field_subject();
    ExclusiveEditMgr_ptr field_edit_manager();
    Inset_ptr field_inset(FieldSubject_ptr subj, FieldEditorData_ptr vdata, ExclusiveEditMgr_ptr mgr);
    FieldEditorData_ptr field_editor_data(Long width, Long maximum);
    ViewerCreator_ptr field_editor_creator();
    Registry_ptr _registry_ptr();
    Naming::NamingContext_ptr _persistent_naming_context();
    XfTelltale_ptr telltale_group();
    Button_ptr push_button(XfGlyph_ptr g, XfAction_ptr a);
    Button_ptr default_button(XfGlyph_ptr g, XfAction_ptr a);
    Button_ptr palette_button(XfGlyph_ptr g, XfAction_ptr a);
    XfGlyph_ptr checkmark(XfTelltale_ptr t);
    Button_ptr check_box(XfGlyph_ptr g, XfAction_ptr a);
    XfGlyph_ptr radio_flag(XfTelltale_ptr t);
    Button_ptr radio_button(XfGlyph_ptr g, XfAction_ptr a, XfTelltale_ptr group);
    Adjustment_ptr bounded_float(Float lower, Float upper, Float initial);
    XfViewer_ptr slider(XfAxis a, Adjustment_ptr adj);
    XfViewer_ptr scroll_bar(XfAxis a, Adjustment_ptr adj);
    XfViewer_ptr panner(Adjustment_ptr x, Adjustment_ptr y);
    Button_ptr zoomer(XfCoord scale, Adjustment_ptr x, Adjustment_ptr y, Adjustment_ptr z);
    Button_ptr up_mover(Adjustment_ptr a);
    Button_ptr down_mover(Adjustment_ptr a);
    Button_ptr left_mover(Adjustment_ptr a);
    Button_ptr right_mover(Adjustment_ptr a);
    XfViewer_ptr viewer_group(XfGlyph_ptr g);
    DialogKit_ptr dialog_kit();
    MenuKit_ptr menu_kit();
    //+

    struct Info {
	Coord thickness;
	float check_size;
	float radio_scale;
	Coord mover_size;
	Coord slider_size;
	Coord min_button_width;
	long num_colors;
	Font_var font;
	Color_var* color;
	Raster_var checkmark;
	Color_var checkmark_color;
	Raster_var shadow1;
	Raster_var shadow2;
    };

    Color_ptr brightness(Color_ptr c, float adjust);

    LayoutKit_ptr layouts() { return layouts_; }
    DrawingKit_ptr drawing() { return drawing_; }
protected:
    SharedFrescoObjectImpl object_;
    StyleContext_var style_;
    WidgetKitImpl::Info info_;
    LayoutKit_ptr layouts_;
    DrawingKit_ptr drawing_;
    DialogKit_ptr dialogs_;
    MenuKit_ptr menus_;

    void load();
    Raster_ptr make_bitmap(
	DrawingKit_ptr, unsigned char*, unsigned int width, unsigned int height
    );
    void load_coord(const char*, Coord&);
    void load_float(const char*, float&);
    void unload();
    void make_thumb(Slider*, Axis);
};

enum WKColorIndex {
    none, black, very_dark_gray, dark_gray, medium_gray,
    light_gray, very_light_gray, white,
    yellow, light_yellow, dark_yellow, medium_yellow
};

class WKFrame : public Beveler {
public:
    WKFrame(
	Glyph_ptr, Telltale_ptr, WidgetKitImpl::Info*, Coord thickness,
	float xalign = 0.5, float yalign = 0.5,
	Boolean hmargin = true, Boolean ymargin = true
    );
    ~WKFrame();

    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
    virtual void draw_frame(GlyphTraversal_ptr t);

    static void fill_rect(
	Painter_ptr p, const Vertex& lower, const Vertex& upper, Color_ptr
    );
    static void shade(
	Painter_ptr p, const Vertex& lower, const Vertex& upper,
	WidgetKitImpl::Info* info, const int* colors, int ncolors, Coord* t
    );
protected:
    Telltale_var state_;
    WidgetKitImpl::Info* info_;
};

class WKGlyph : public GlyphImpl {
public:
    WKGlyph(WidgetKitImpl::Info*);
    ~WKGlyph();
protected:
    WidgetKitImpl::Info* info_;
};

#endif
