/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#include "menu_bar.h"
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/table.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/menus.h>
#include <Fresco/widgets/widgets.h>
#include <X11/keysym.h>
#include <ctype.h>

static Event::KeySym key_sym_to_lower(Event::KeySym k) {
    return (Event::KeySym) tolower((int) k);
}

// class MKObserveItem

struct MKObserveItem {
    MKObserveItem() {}
    MKObserveItem(MenuItem_ptr i, Tag t) {
        item = MenuItem::_duplicate(i);
        tag = t;
    }

    MenuItem_var item;
    Tag tag;
};

declareList(MKObserveList,MKObserveItem)
implementList(MKObserveList,MKObserveItem)

// class MKAcceleratorTable

inline unsigned long key_to_hash(const Event::KeyChord& k) {
    long hash = 0;
    for (unsigned long i = 0; i < k.length(); ++i) {
        hash ^= k[i];
    }
    return hash;
}
 
static Boolean key_equal(
    const Event::KeyChord& k1,
    const Event::KeyChord& k2
) {
    if (k1.length() != k2.length()) {
        return false;
    }
    for (Long i = k1.length() - 1; i >= 0; --i) {
        if (k1[i] != k2[i]) {
            return false;
        }
    }
    return true;
}

declareTable(MKAcceleratorTable,Event::KeyChord,MenuItem_ptr)
implementTable(MKAcceleratorTable,Event::KeyChord,MenuItem_ptr)

// class MKMenuBar
 
implementPtrList(MKWindowInfoList,MKWindowInfo)

MKMenuBar::MKMenuBar(MenuKitImpl::Info* info, MenuItem_ptr item) {
    info_ = info;
    item_ = MenuItem::_duplicate(item);
    window_ = nil;
    active_ = false;
    windows_tag_ = 0;
    accelerators_ = nil;
    observe_list_ = new MKObserveList;
    box_ = nil;
    build();
    item_tag_ = item_->attach(EventFilter_ptr(this));
}

MKMenuBar::~MKMenuBar() {
    item_->detach(item_tag_);
    uncache_accelerators();
    CORBA::release(item_);
    CORBA::release(box_);
    for (ListItr(MKWindowInfoList) i(windows_); i.more(); i.next()) {
        delete i.cur();
    }
}

OxExchange* MKMenuBar::_exchange() { return ViewerImpl::_exchange(); }

XfTag MKMenuBar::attach(XfFrescoObject_ptr observer) {
    return ViewerImpl::attach(observer);
}
void MKMenuBar::detach(XfTag attach_tag) {
    ViewerImpl::detach(attach_tag);
}
void MKMenuBar::disconnect() {
    ViewerImpl::disconnect();
}
void MKMenuBar::notify_observers() {
    ViewerImpl::notify_observers();
}

void MKMenuBar::update() {
    if (active_) {
        close_menus();
    }
    build();
    need_redraw();
    need_resize(); // !!! why doesn't monoglyph do this when its body changes? -denis
    uncache_accelerators();
}

//+ MKMenuBar(XfGlyph::=ViewerImpl::)
XfGlyph_ptr MKMenuBar::clone_glyph() {
    return ViewerImpl::clone_glyph();
}
XfStyleContext_ptr MKMenuBar::style() {
    return ViewerImpl::style();
}
void MKMenuBar::style(XfStyleContext_ptr _p) {
    ViewerImpl::style(_p);
}
XfTransform_ptr MKMenuBar::transformation() {
    return ViewerImpl::transformation();
}
void MKMenuBar::request(XfGlyph::Requisition& r) {
    ViewerImpl::request(r);
}
void MKMenuBar::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    ViewerImpl::extension(a, r);
}
void MKMenuBar::shape(XfRegion_ptr r) {
    ViewerImpl::shape(r);
}
void MKMenuBar::traverse(XfGlyphTraversal_ptr t) {
    ViewerImpl::traverse(t);
}
void MKMenuBar::draw(XfGlyphTraversal_ptr t) {
    ViewerImpl::draw(t);
}
void MKMenuBar::pick(XfGlyphTraversal_ptr t) {
    ViewerImpl::pick(t);
}
XfGlyph_ptr MKMenuBar::body() {
    return ViewerImpl::body();
}
void MKMenuBar::body(XfGlyph_ptr _p) {
    ViewerImpl::body(_p);
}
void MKMenuBar::append(XfGlyph_ptr g) {
    ViewerImpl::append(g);
}
void MKMenuBar::prepend(XfGlyph_ptr g) {
    ViewerImpl::prepend(g);
}
XfTag MKMenuBar::add_parent(XfGlyphOffset_ptr parent_offset) {
    return ViewerImpl::add_parent(parent_offset);
}
void MKMenuBar::remove_parent(XfTag add_tag) {
    ViewerImpl::remove_parent(add_tag);
}
XfGlyphOffset_ptr MKMenuBar::first_child_offset() {
    return ViewerImpl::first_child_offset();
}
XfGlyphOffset_ptr MKMenuBar::last_child_offset() {
    return ViewerImpl::last_child_offset();
}
void MKMenuBar::parent_offsets(XfGlyph::OffsetSeq& parents) {
    ViewerImpl::parent_offsets(parents);
}
void MKMenuBar::allocations(XfGlyph::AllocationInfoSeq& a) {
    ViewerImpl::allocations(a);
}
void MKMenuBar::need_redraw() {
    ViewerImpl::need_redraw();
}
void MKMenuBar::need_redraw_region(XfRegion_ptr r) {
    ViewerImpl::need_redraw_region(r);
}
void MKMenuBar::need_resize() {
    ViewerImpl::need_resize();
}
Boolean MKMenuBar::restore_trail(XfGlyphTraversal_ptr t) {
    return ViewerImpl::restore_trail(t);
}
//+

XfViewer_ptr MKMenuBar::parent_viewer() {
    return ViewerImpl::parent_viewer();
}
XfViewer_ptr MKMenuBar::next_viewer() {
    return ViewerImpl::next_viewer();
}
XfViewer_ptr MKMenuBar::prev_viewer() {
    return ViewerImpl::prev_viewer();
}
XfViewer_ptr MKMenuBar::first_viewer() {
    return ViewerImpl::first_viewer();
}
XfViewer_ptr MKMenuBar::last_viewer() {
    return ViewerImpl::last_viewer();
}
void MKMenuBar::append_viewer(XfViewer_ptr v) {
    ViewerImpl::append_viewer(v);
}
void MKMenuBar::prepend_viewer(XfViewer_ptr v) {
    ViewerImpl::prepend_viewer(v);
}
void MKMenuBar::insert_viewer(XfViewer_ptr v) {
    ViewerImpl::insert_viewer(v);
}
void MKMenuBar::replace_viewer(XfViewer_ptr v) {
    ViewerImpl::replace_viewer(v);
}
void MKMenuBar::remove_viewer() {
    ViewerImpl::remove_viewer();
}
void MKMenuBar::set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next) {
    ViewerImpl::set_viewer_links(parent, prev, next);
}
void MKMenuBar::set_first_viewer(XfViewer_ptr v) {
    ViewerImpl::set_first_viewer(v);
}
void MKMenuBar::set_last_viewer(XfViewer_ptr v) {
    ViewerImpl::set_last_viewer(v);
}
XfFocus_ptr MKMenuBar::request_focus(XfViewer_ptr requestor, Boolean temporary) {
    return ViewerImpl::request_focus(requestor, temporary);
}
Boolean MKMenuBar::receive_focus(XfFocus_ptr f, Boolean primary) {
    return ViewerImpl::receive_focus(f, primary);
}
void MKMenuBar::lose_focus(XfFocus_ptr f, Boolean primary, Boolean temporary) {
    ViewerImpl::lose_focus(f, primary, temporary);
}
Boolean MKMenuBar::first_focus() {
    return ViewerImpl::first_focus();
}
Boolean MKMenuBar::last_focus() {
    return ViewerImpl::last_focus();
}
Boolean MKMenuBar::next_focus() {
    return ViewerImpl::next_focus();
}
Boolean MKMenuBar::prev_focus() {
    return ViewerImpl::prev_focus();
}
void MKMenuBar::close() {
    ViewerImpl::close();
}
 
Boolean MKMenuBar::handle(XfGlyphTraversal_ptr t, XfEvent_ptr e) {
    MenuItem_ptr item = find_accelerator(e);
    if (is_not_nil(item)) {
        Action_var action = item->item_action();
        if (is_not_nil(action)) {
            action->execute();
        }
        return true;
    }

    if (e->type() == Event::key_press) {
        Event::KeySym key = e->key();
        if (
            (key == XK_F10) ||
            ((key == XK_Menu) && e->modifier_is_down(Event::shift))
        ) {
            open_menus(Window_var(t->current_window()));
            return true;
        } else if (e->modifier_is_down(Event::alt)) {
            return toplevel_mnemonic(e->character());
        }
    }

    return false;
}

Boolean MKMenuBar::handle(Event_ptr e) {
    if (e->type() == Event::key_press) {
        Event::KeySym key = e->key();
        if (
            (key == XK_F10) ||
            ((key == XK_Menu) && e->modifier_is_down(Event::shift)) ||
            (key == XK_Cancel) ||
            (key == XK_Escape)
        ) {
            close_menus();
        } else if (key == XK_Down) {
            traverse_down();
        } else if (key == XK_Up) {
            traverse_up();
        } else if (key == XK_Left) {
            traverse_left();
        } else if (key == XK_Right) {
            traverse_right();
        } else if (
            (key == XK_Select) ||
            (key == XK_space) ||
            (key == XK_Return)
        ) {
            activate();
        } else if (e->modifier_is_down(Event::alt)) {
            toplevel_mnemonic(e->character());
        } else {
            traverse_mnemonic(e->character());
        }
    } else {
        if (deliverable(e)) {
            // don't need to do anything more
        } else if ((e->type() == Event::up) || (e->type() == Event::down)) {
            close_menus();
        }
    }

    return true;
}

MenuManager_ptr MKMenuBar::menu_manager() {
    return MenuManager::_duplicate(this);
}

MenuItem_ptr MKMenuBar::root() {
    return MenuItem::_duplicate(item_);
}

Screen_ptr MKMenuBar::filter_screen() {
    return window_->screen_ptr();
}

void MKMenuBar::open_menus(XfWindow_ptr window) {
    if (! window_) {
        window_ = window;
        Screen_var screen = window_->screen_ptr();
        Display_var display = screen->display_ptr();
        filter_tag_ = display->add_filter(this);

        MenuItem_ptr item = box_->children().item(0)->item_;
        select(item, false);
    }
}

void MKMenuBar::close_menus() {
    if (info_->grab) {
        window_->ungrab_pointer();
    }
    active_ = false;

    MenuItem::Path path;
    box_->select(path, false);

    Screen_var screen = window_->screen_ptr();
    Display_var display = screen->display_ptr();
    display->remove_filter(filter_tag_);
    window_ = nil;
}

Tag MKMenuBar::add_window(Window_ptr window) {
    MKWindowInfo* info = new MKWindowInfo;
    info->tag = ++windows_tag_;
    info->window = window;
    windows_.append(info);
    return info->tag;
}

void MKMenuBar::remove_window(Tag tag) {
    for (ListUpdater(MKWindowInfoList) i(windows_); i.more(); i.next()) {
        if (i.cur()->tag == tag) {
            i.remove_cur();
            break;
        }
    }
}

MenuItem_ptr MKMenuBar::selection() {
    MenuItem::Path path;
    box_->selection(path);
    return path.length() > 0 ? MenuItem::_duplicate(path[path.length() - 1]) : nil;
}

void MKMenuBar::select(MenuItem_ptr item, Boolean flap) {
    if (! active_) {
        active_ = true;
        if (info_->grab) {
            window_->grab_pointer(nil);
        }
    }

    MenuItem::Path path;
    find_path(path, item);
    box_->select(path, flap);
}

void MKMenuBar::unselect() {
    MenuItem::Path path;
    box_->select(path, false);
}

void MKMenuBar::activate() {
    MenuItem_ptr item = selection();

    // make sure the user sees the selection
    Screen_var screen = window_->screen_ptr();
    Display_var display = screen->display_ptr();
    display->repair();
    display->flush_and_wait();
    FrescoLib::delay(info_->delay);
 
    close_menus();
    display->flush();

    if (is_not_nil(item)) {
        Action_var action = item->item_action();
        if (is_not_nil(action)) {
            action->execute();
        }
        CORBA::release(item);
    }
}

Boolean MKMenuBar::dive(MenuItem_ptr parent) {
    MenuItem_ptr item = MenuKitImpl::first_child(parent);
    if (is_not_nil(item)) {
        select(item, false);
    }
    return item != nil;
}

void MKMenuBar::circulate(MenuItem_ptr child, Long amount, Boolean open) {
    MenuItem_ptr item = nil;
    MenuItem_ptr parent = child->parent();
    Long count = parent->count();
    for (Long i = 0; i < count; ++i) {
        if (parent->item(i) == child) {
            for (Long j = i + amount; j != i; j += amount) {
                MenuItem_ptr next = parent->item((j + count) % count);
                if (next->type() != MenuItem::separator) {
                    item = next;
                    break;
                }
            }
            if (is_not_nil(item)) {
                if (open) {
                    if (dive(item)) {
                        break;
                    }
                }
                select(item, false);
            }
            break;
        }
    }
}

void MKMenuBar::traverse_down() {
    MenuItem::Path path;
    box_->selection(path);

    if (path.length() > 1) {
        circulate(path[path.length() - 1], +1);
    } else if (path.length() == 1) {
        dive(path[0]);
    } else {
        select(box_->children().item(0)->item_, false);
    }
}

void MKMenuBar::traverse_up() {
    MenuItem::Path path;
    box_->selection(path);

    if (path.length() > 1) {
        circulate(path[path.length() - 1], -1);
    }
}

void MKMenuBar::traverse_left() {
    MenuItem::Path path;
    box_->selection(path);

    if (path.length() == 1) {
        circulate(path[0], -1);
    } else if (path.length() > 2) {
        select(path[path.length() - 2], false);
    } else if (path.length() == 2) {
        circulate(path[0], -1, true);
    }
}

void MKMenuBar::traverse_right() {
    MenuItem::Path path;
    box_->selection(path);

    if (path.length() > 1) {
        if (! dive(path[path.length() - 1])) {
            circulate(path[path.length() - 2], +1, true);
        }
    } else if (path.length() == 1) {
        circulate(path[0], +1);
    }
}

void MKMenuBar::traverse_mnemonic(Event::KeySym key) {
    MenuItem_ptr item = find_mnemonic(key);
    if (is_not_nil(item)) {
        if (item->type() == MenuItem::submenu) {
            while (item) {
                MenuItem_ptr mi = MenuKitImpl::first_child(item);
                if (is_not_nil(mi)) {
                    item = mi;
                } else {
                    break;
                }
            }
            select(item, false);
        } else {
            select(item, false);
            if (item->type() == MenuItem::leaf) {
                activate();
            }
        }
    }
}

Boolean MKMenuBar::toplevel_mnemonic(Event::KeySym key) {
    Event::KeySym k = key_sym_to_lower(key);

    const MKBoxInfoList& children = box_->children();
    for (ListItr(MKBoxInfoList) i(children); i.more(); i.next()) {
        MenuItem_ptr item = i.cur()->item_;
        if (key_sym_to_lower(item->mnemonic()) == k) {
            select(MenuKitImpl::first_item(item), false);
            if (item->type() == MenuItem::leaf) {
                activate();
            }
            return true;
        }
    }

    return false;
}

MenuItem_ptr MKMenuBar::find_mnemonic(Event::KeySym key) {
    Event::KeySym k = key_sym_to_lower(key);

    MenuItem::Path path;
    box_->selection(path);
    for (Long i = path.length() - 2; i >= 0; --i) {
        MenuItem_ptr item = path[i];
        Long count = item->count();
        for (Long j = 0; j < count; ++j) {
            MenuItem_ptr mi = item->item(j);
            if (key_sym_to_lower(mi->mnemonic()) == k) {
                return mi;
            }
        }
    }

    const MKBoxInfoList& children = box_->children();
    for (ListItr(MKBoxInfoList) j(children); j.more(); j.next()) {
        MenuItem_ptr item = j.cur()->item_;
        if (key_sym_to_lower(item->mnemonic()) == k) {
            return item;
        }
    }

    return nil;
}

static void key_chord_append(Event::KeyChord& keys, Event::KeySym keysym) {
    keys[keys.length()] = keysym;
    keys.length(keys.length() + 1);
}

MenuItem_ptr MKMenuBar::find_accelerator(Event_ptr e) {
    if (e->type() != Event::key_press) {
        return nil;
    }

    Event::KeyChord keys(6);
    if (e->modifier_is_down(Event::control)) {
        key_chord_append(keys, XK_Control_L);
    }
    if (e->modifier_is_down(Event::shift)) {
        key_chord_append(keys, XK_Shift_L);
    }
    if (e->modifier_is_down(Event::capslock)) {
        key_chord_append(keys, XK_Caps_Lock);
    }
    if (e->modifier_is_down(Event::meta)) {
        key_chord_append(keys, XK_Meta_L);
    }
    if (e->modifier_is_down(Event::alt)) {
        key_chord_append(keys, XK_Alt_L);
    }
    key_chord_append(keys, e->key());

    if (! accelerators_) {
        accelerators_ = new MKAcceleratorTable(256);
        fill_accelerator_table(item_);
    }
    MenuItem_ptr item = nil;
    accelerators_->find(item, keys);
    return is_not_nil(item) && item->enabled() ? item : nil;
}
    
void MKMenuBar::find_path(MenuItem::Path& path, MenuItem_ptr item) {
    while (item_ != item) {
        MenuKitImpl::path_append(path, item);
        item = item->parent();
    }
    MenuKitImpl::path_reverse(path);
}

Boolean MKMenuBar::deliverable(Event_ptr e) {
    if (e->hits(Viewer_var(window_->main_viewer()))) {
        // !!! assumes there are no glyphs on top of the menu bar -denis
        Glyph::AllocationInfoSeq s;
        allocations(s);
        // !!! assumes the menu bar is in only one point in the glyph DAG -denis
        Glyph::AllocationInfo& i = s[0];
        RegionImpl r;
        r.copy(i.allocation);
        r.apply_transform(i.transformation);
        Vertex v;
        v.x = e->pointer_x();
        v.y = e->pointer_y();
        v.z = Coord(0);
        Boolean handle = r.contains(v);
        if (handle) {
            window_->handle_event(e);
        }
        return handle;
    }

    for (ListUpdater(MKWindowInfoList) i(windows_); i.more(); i.next()) {
        Window_ptr w = i.cur()->window;
        if (e->hits(w->main_viewer())) {
            w->handle_event(e);
            return true;
        }
    }

    return false;
}

void MKMenuBar::build() {
    CORBA::release(box_);

    box_ = new MKHBox(info_, item_);
    box_->menu_manager(this);

    WidgetKit_ptr widget = info_->widget;
    LayoutKit_ptr layout = info_->layout;

    body(
        Glyph_var(widget->outset_frame(
            Glyph_var(layout->margin(box_, info_->menu_margin))
        ))
    );
}

void MKMenuBar::fill_accelerator_table(MenuItem_ptr item) {
    Long count = item->count();
    for (Long i = 0; i < count; ++i) {
        MenuItem_ptr mi = item->item(i);
        switch (mi->type()) {
            case MenuItem::submenu: {
                observe_list_->append(
                    MKObserveItem(mi, mi->attach(EventFilter_ptr(this)))
                );
                fill_accelerator_table(mi);
            } break;
            case MenuItem::leaf: {
                Event::KeyChord* a = mi->accelerator();
                if (a != nil) {
                    accelerators_->insert(*a, mi);
                }
                delete a;
            } break;
            case MenuItem::separator: {
            } break;
        }
    }
}

void MKMenuBar::uncache_accelerators() {
    for (MKObserveList_Iterator i(*observe_list_); i.more(); i.next()) {
        MKObserveItem oi = i.cur();
        oi.item->detach(oi.tag);
    }
    observe_list_->remove_all();
    delete accelerators_;
    accelerators_ = nil;
}
