/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_widgets_impl_menu_bar_h
#define Fresco_widgets_impl_menu_bar_h

#include "box.h"
#include "menu_kit.h"
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/widgets/menus.h>

// class MKMenuBar

struct MKWindowInfo {
    Window_ptr window;
    Tag tag;
};

declarePtrList(MKWindowInfoList,MKWindowInfo)

class MKAcceleratorTable;
class MKObserveList;

class MKMenuBar :
    _Ix_public MenuBar, _Ix_public MenuManager, public ViewerImpl, _Ix_public EventFilter {
public:
    MKMenuBar(MenuKitImpl::Info*, MenuItem_ptr);
    ~MKMenuBar();

    OxExchange* _exchange();

    _Ix_ClassInterfaceOps(MenuBar)
		  
    //+ XfMenuBar::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyph */
    XfGlyph_ptr clone_glyph();
    XfStyleContext_ptr style();
    void style(XfStyleContext_ptr _p);
    XfTransform_ptr transformation();
    void request(XfGlyph::Requisition& r);
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r);
    void shape(XfRegion_ptr r);
    void traverse(XfGlyphTraversal_ptr t);
    void draw(XfGlyphTraversal_ptr t);
    void pick(XfGlyphTraversal_ptr t);
    XfGlyph_ptr body();
    void body(XfGlyph_ptr _p);
    void append(XfGlyph_ptr g);
    void prepend(XfGlyph_ptr g);
    XfTag add_parent(XfGlyphOffset_ptr parent_offset);
    void remove_parent(XfTag add_tag);
    XfGlyphOffset_ptr first_child_offset();
    XfGlyphOffset_ptr last_child_offset();
    void parent_offsets(XfGlyph::OffsetSeq& parents);
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void need_redraw();
    void need_redraw_region(XfRegion_ptr r);
    void need_resize();
    Boolean restore_trail(XfGlyphTraversal_ptr t);
    /* XfViewer */
    XfViewer_ptr parent_viewer();
    XfViewer_ptr next_viewer();
    XfViewer_ptr prev_viewer();
    XfViewer_ptr first_viewer();
    XfViewer_ptr last_viewer();
    void append_viewer(XfViewer_ptr v);
    void prepend_viewer(XfViewer_ptr v);
    void insert_viewer(XfViewer_ptr v);
    void replace_viewer(XfViewer_ptr v);
    void remove_viewer();
    void set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next);
    void set_first_viewer(XfViewer_ptr v);
    void set_last_viewer(XfViewer_ptr v);
    XfFocus_ptr request_focus(XfViewer_ptr requestor, Boolean temporary);
    Boolean receive_focus(XfFocus_ptr f, Boolean primary);
    void lose_focus(XfFocus_ptr f, Boolean primary, Boolean temporary);
    Boolean first_focus();
    Boolean last_focus();
    Boolean next_focus();
    Boolean prev_focus();
    Boolean handle(XfGlyphTraversal_ptr t, XfEvent_ptr e);
    void close();
    /* XfMenuBar */
    XfMenuManager_ptr menu_manager();
    //+

    /* XfEventFilter */
    Boolean handle(XfEvent_ptr e);

    /* XfMenuManager */
    MenuItem_ptr root();
    void add_filter(XfWindow_ptr w);
    XfScreen_ptr filter_screen();
    void open_menus(XfWindow_ptr w);
    void close_menus();
    XfTag add_window(XfWindow_ptr w);
    void remove_window(XfTag t);
    void select(MenuItem_ptr item, Boolean flap);
    MenuItem_ptr selection();
    void unselect();
    void activate();

protected:
    Boolean dive(MenuItem_ptr parent);
    void circulate(MenuItem_ptr child, Long amount, Boolean open = false);
    void traverse_down();
    void traverse_up();
    void traverse_left();
    void traverse_right();
    void traverse_mnemonic(Event::KeySym key);
    Boolean toplevel_mnemonic(Event::KeySym key);
    MenuItem_ptr find_mnemonic(Event::KeySym key);
    MenuItem_ptr find_accelerator(Event_ptr e);
    void find_path(MenuItem::Path& path, MenuItem_ptr item);
    Boolean deliverable(Event_ptr e);
    void build();
    void fill_accelerator_table(MenuItem_ptr item);
    void uncache_accelerators();

    MenuKitImpl::Info* info_;
    MenuItem_ptr item_;
    Tag item_tag_;
    MKHBox* box_;
    Window_ptr window_;
    Boolean active_;
    Tag filter_tag_;
    Tag windows_tag_;
    MKWindowInfoList windows_;
    MKAcceleratorTable* accelerators_;
    MKObserveList* observe_list_;
};

#endif
