/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_widgets_impl_menu_kit_h
#define Fresco_widgets_impl_menu_kit_h

#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/figures/figures.h>
#include <Fresco/widgets/menus.h>
#include <Fresco/widgets/widgets.h>

class DrawingKit;
class LayoutKit;
class FigureKit;
class WidgetKit;

// class MenuKitImpl

class MenuKitImpl : public MenuKit {
    //. MenuKitImpl provides an implementation of the MenuKit
    //. interface that creates objects with a Motif-like look and feel.
public:
    MenuKitImpl(WidgetKit*);
    ~MenuKitImpl();
 
    OxExchange* _exchange();

    //+ MenuKit::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* MenuKit */
    XfMenuItem_ptr menu_item(XfGlyph_ptr label_glyph, XfEvent::KeySym mnemonic);
    XfMenuItem_ptr menu_item_string(XfCharString_ptr label_name, XfCharCode mnemonic);
    XfMenuItem_ptr button_item(XfGlyph_ptr label_glyph, XfGlyph_ptr shortcut_glyph, XfEvent::KeySym mnemonic, const XfEvent::KeyChord& accelerator, XfAction_ptr a);
    XfMenuItem_ptr button_item_string(XfCharString_ptr label_name, XfCharCode mnemonic, XfCharString_ptr accelerator_name, XfAction_ptr a);
    XfMenuItem_ptr dialog_button_item_string(XfCharString_ptr label_name, XfCharCode mnemonic, XfCharString_ptr accelerator_name, XfAction_ptr a);
    XfMenuItem_ptr mark_item(XfGlyph_ptr mark_glyph, XfGlyph_ptr label_glyph, XfGlyph_ptr shortcut_glyph, XfEvent::KeySym mnemonic, const XfEvent::KeyChord& accelerator, XfAction_ptr a, XfTelltale_ptr t);
    XfMenuItem_ptr check_box_item(XfGlyph_ptr label_glyph, XfGlyph_ptr shortcut_glyph, XfEvent::KeySym mnemonic, const XfEvent::KeyChord& accelerator, XfAction_ptr a);
    XfMenuItem_ptr check_box_item_string(XfCharString_ptr label_name, XfCharCode mnemonic, XfCharString_ptr accelerator_name, XfAction_ptr a);
    XfMenuItem_ptr radio_item(XfGlyph_ptr label_glyph, XfGlyph_ptr shortcut_glyph, XfEvent::KeySym mnemonic, const XfEvent::KeyChord& accelerator, XfAction_ptr a);
    XfMenuItem_ptr radio_item_string(XfCharString_ptr label_name, XfCharCode mnemonic, XfCharString_ptr accelerator_name, XfAction_ptr a);
    XfMenuItem_ptr separator_item();
    //+
 
    static void path_append(MenuItem::Path& path, MenuItem_ptr item);
    static void path_reverse(MenuItem::Path& path);
    static MenuItem_ptr first_child(MenuItem_ptr item);
    static MenuItem_ptr first_item(MenuItem_ptr item);

    struct Info {
        DrawingKit* drawing;
        LayoutKit* layout;
        FigureKit* figure;
        WidgetKit* widget;

        Coord menu_margin;
        float delay;
        Boolean grab;
        Glyph_ptr menu_shortcut;
        Glyph_ptr menu_separator;
        Glyph_ptr menu_glue;
    };
 
protected:
    Glyph_ptr label(
        CharString_ptr name, CharCode mnemonic, Boolean dialog = false
    );
    Event::KeyChord accelerator(CharString_ptr string);
    Glyph_ptr shortcut(const XfEvent::KeyChord& accelerator);

    void load();
    void load_coord(const char*, Coord&);
    void load_float(const char*, float&);
    void load_boolean(const char*, Boolean&);
    void load_stencil(
        const char* name, Glyph_ptr& g,
        unsigned char* data, unsigned int width, unsigned int height
    );
    void unload();

    SharedFrescoObjectImpl object_;
    FigureStyle_var style_;
    MenuKitImpl::Info info_;
};

#endif
