/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#ifndef Fresco_widgets_impl_popups_h
#define Fresco_widgets_impl_popups_h
 
#include "box.h"
#include "menu_kit.h"
#include <Fresco/widgets/menus.h>

// class MKPopupMenu

class MKPopupMenu : public PopupMenu {
public:
    ~MKPopupMenu();

    OxExchange* _exchange();

    //+ XfPopupMenu::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfPopupMenu */
    void menu_manager(XfMenuManager_ptr m);
    void show(XfRegion_ptr a);
    void select(const XfMenuItem::Path& p, Boolean flap);
    void selection(XfMenuItem::Path& p);
    void hide();
    //+

protected:
    MKPopupMenu(MenuKitImpl::Info* info, MenuItem_ptr item);

    void build(Screen_ptr s);
    Window::Placement position(Region_ptr r);

    virtual void location(
        const Vertex& lower, const Vertex& upper, Coord& x, Coord& y
    ) = 0;

    SharedFrescoObjectImpl object_;
    MenuKitImpl::Info* info_;
    MenuItem_ptr item_;
    Tag item_tag_;
    MKBox* box_;
    MenuManager_ptr manager_;
    Window_ptr window_;
    Tag window_tag_;
};

// class MKPulldownMenu

class MKPulldownMenu : public MKPopupMenu {
public:
    MKPulldownMenu(MenuKitImpl::Info* info, MenuItem_ptr item);
    
    void location(const Vertex& lower, const Vertex& upper, Coord& x, Coord& y);
};

// class MKPullrightMenu

class MKPullrightMenu : public MKPopupMenu {
public:
    MKPullrightMenu(MenuKitImpl::Info* info, MenuItem_ptr item);
    
    void location(const Vertex& lower, const Vertex& upper, Coord& x, Coord& y);
};

#endif
