//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_widgets_slider_h
#define Fresco_widgets_slider_h

#include <Fresco/defs_impl/polyglyph.h>
#include <Fresco/defs_impl/viewers.h>
#include <Fresco/defs_impl/thread.h>
#include <Fresco/layouts/layouts.h>

class Slider;
class Stepper;
class WidgetKitImpl;

class SliderViewer : public ActiveViewer {
public:
    SliderViewer(Slider*, Glyph_ptr bdy);
    ~SliderViewer();

    /* ActiveViewer */
    Boolean move(GlyphTraversal_ptr, Event_ptr);
    Boolean press(GlyphTraversal_ptr, Event_ptr);
    Boolean drag(GlyphTraversal_ptr, Event_ptr);
    Boolean release(GlyphTraversal_ptr, Event_ptr);
protected:
    Slider* slider_;
    Boolean dragging_ : 1;
    Boolean aborted_ : 1;
    Coord xoffset_;
    Coord yoffset_;
    Stepper* stepper_;
};

class Slider : public PolyGlyph {
public:
    Slider(WidgetKitImpl*, StyleContext_ptr);
    ~Slider();

    /* Slider */
    void normal_thumb(Glyph_ptr);
    void visible_thumb(Glyph_ptr);
    void old_thumb(Glyph_ptr);

    Stepper* forward_stepper();
    Stepper* backward_stepper();

    /* PolyGlyph */
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
    void child_allocation(Long, Glyph::AllocationInfo&);
    void update(); //+ XfFrescoObject::update

    /* For SliderViewer */
    void move_thumb(GlyphTraversal_ptr t, Event_ptr);
    Long hit_thumb(GlyphTraversal_ptr t, Event_ptr);
    void show_old_thumb();
    void release_thumb();
    void redraw_thumb();

    virtual Region_ptr thumb_allocation(Region_ptr);
    virtual void move_to(Coord x, Coord y);
    virtual void begin_adjustment();
    virtual void commit_adjustment();
protected:
    WidgetKitImpl* kit_;
    GlyphOffset_ptr normal_thumb_;
    GlyphOffset_ptr visible_thumb_;
    GlyphOffset_ptr old_thumb_;
    Coord min_thumb_size_;
    Boolean showing_visible_;
    Boolean showing_old_thumb_;
    Stepper* forward_;
    Stepper* backward_;
    const long old_thumb_index;
    const long normal_thumb_index;
    const long visible_thumb_index;

    void forward_stepper(Stepper*);
    void backward_stepper(Stepper*);
    void allot_thumb_major_axis(
	Coord slider_size, Coord slider_lower, Adjustment::Settings&,
	float& scale, Coord& lower, Coord& upper, float& align
    );
    void allot_thumb_minor_axis(
	Coord lower, Coord upper,
	Coord& new_lower, Coord& new_upper, float& align
    );
    Boolean changed(Adjustment_ptr, Adjustment::Settings&);

    virtual Region_ptr old_thumb_allocation(Region_ptr);
    virtual Boolean adjustments_changed();
};

class XSlider : public Slider {
public:
    XSlider(WidgetKitImpl*, StyleContext_ptr, Adjustment_ptr);
    ~XSlider();

    void move_to(Coord x, Coord y);
    Region_ptr old_thumb_allocation(Region_ptr);
    Region_ptr thumb_allocation(Region_ptr);
    void disconnect();
    void begin_adjustment();
    void commit_adjustment();
protected:
    Adjustment_ptr adjustment_;
    Tag tag_;
    float xscale_;
    Adjustment::Settings settings_;
    Adjustment::Settings old_settings_;

    Region_ptr allocate_thumb(Region_ptr, Adjustment::Settings&);
    Boolean adjustments_changed();
};

class YSlider : public Slider {
public:
    YSlider(WidgetKitImpl*, StyleContext_ptr, Adjustment_ptr);
    ~YSlider();

    void move_to(Coord x, Coord y);
    Region_ptr old_thumb_allocation(Region_ptr);
    Region_ptr thumb_allocation(Region_ptr);
    void disconnect();
    void begin_adjustment();
    void commit_adjustment();
protected:
    Adjustment_ptr adjustment_;
    Tag tag_;
    float yscale_;
    Adjustment::Settings settings_;
    Adjustment::Settings old_settings_;

    Region_ptr allocate_thumb(Region_ptr, Adjustment::Settings&);
    Boolean adjustments_changed();
};

class XYSlider : public Slider {
public:
    XYSlider(
	WidgetKitImpl*, StyleContext_ptr, Adjustment_ptr x, Adjustment_ptr y
    );
    ~XYSlider();

    void move_to(Coord x, Coord y);
    Region_ptr old_thumb_allocation(Region_ptr);
    Region_ptr thumb_allocation(Region_ptr);
    void disconnect();
    void begin_adjustment();
    void commit_adjustment();
protected:
    Adjustment_ptr x_adjustment_;
    Adjustment_ptr y_adjustment_;
    Tag x_tag_;
    Tag y_tag_;
    float xscale_;
    float yscale_;
    Adjustment::Settings x_settings_;
    Adjustment::Settings y_settings_;
    Adjustment::Settings x_old_settings_;
    Adjustment::Settings y_old_settings_;

    Region_ptr allocate_thumb(
	Region_ptr, Adjustment::Settings& x, Adjustment::Settings& y
    );
    Boolean adjustments_changed();
};

#endif
