//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_widgets_stepper_h
#define Fresco_widgets_stepper_h

#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/thread.h>
#include "buttons.h"

//- Stepper*
class Stepper : public ButtonImpl {
    //. The stepper class implements a button with autorepeat.
public:
    Stepper(StyleContext_ptr, Telltale_ptr, Action_ptr = nil);
    ~Stepper();

    Boolean press(GlyphTraversal_ptr, Event_ptr);
    Boolean release(GlyphTraversal_ptr, Event_ptr);

    void start_stepping();
    void stop_stepping();
protected:
    long start_delay_;
    long next_delay_;
    Mutex_var lock_;
    Thread_var ticker_;

    virtual void adjust();
    void tick();
};

#define declareAdjustStepper(StepperSubclassName) \
class StepperSubclassName : public Stepper { \
public: \
    StepperSubclassName(StyleContext_ptr, Telltale_ptr, Adjustment_ptr); \
    ~StepperSubclassName(); \
\
    void adjust(); \
protected: \
    Adjustment_var adjustment_; \
};

declareAdjustStepper(ForwardScroller)
declareAdjustStepper(BackwardScroller)
declareAdjustStepper(ForwardPager)
declareAdjustStepper(BackwardPager)

#define declareArrowGlyph(ArrowSubclassName) \
class ArrowSubclassName : public GlyphImpl { \
public: \
    ArrowSubclassName(Color_ptr); \
    ~ArrowSubclassName(); \
\
    void draw(GlyphTraversal_ptr); \
protected: \
    Color_var color_; \
};

declareArrowGlyph(UpArrow)
declareArrowGlyph(DownArrow)
declareArrowGlyph(LeftArrow)
declareArrowGlyph(RightArrow)

#endif
