//
// Copyright (c) 1994-95 Firefly Design
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

#ifndef Fresco_widgets_viewport_h
#define Fresco_widgets_viewport_h

#include <Fresco/widgets/widgets.h>
#include <Fresco/defs_impl/viewers.h>

class WKViewport;

class WKViewportAdjustment : public Adjustment {
public:
    WKViewportAdjustment(WKViewport*, XfAxis);
    ~WKViewportAdjustment();

    OxExchange* _exchange();

    //+ Adjustment::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* Adjustment */
    XfCoord small_scroll();
    void small_scroll(XfCoord _p);
    XfCoord large_scroll();
    void large_scroll(XfCoord _p);
    void get_settings(Adjustment::Settings& s);
    void begin();
    void commit();
    void cancel();
    void scroll_forward();
    void scroll_backward();
    void page_forward();
    void page_backward();
    void scroll_to(XfCoord lower);
    void scroll_by(XfCoord delta);
    void scale_to(XfCoord length);
    void constrain(XfCoord& c);
    //+

protected:
    Coord page_amount();

    SharedFrescoObjectImpl object_;
    WKViewport* viewport_;
    XfAxis axis_;
    Coord scroll_incr_;
    Coord page_incr_;
    Adjustment::Settings saved_;
};

class TransformImpl;

class WKViewport : public Viewport, public ViewerImpl {
public:
    WKViewport(XfGlyph_ptr);
    ~WKViewport();

    OxExchange* _exchange();

    //+ Viewport::*
    /* XfFrescoObject */
    XfTag attach(XfFrescoObject_ptr observer);
    void detach(XfTag attach_tag);
    void disconnect();
    void notify_observers();
    void update();
    /* XfGlyph */
    XfGlyph_ptr clone_glyph();
    XfStyleContext_ptr style();
    void style(XfStyleContext_ptr _p);
    XfTransform_ptr transformation();
    void request(XfGlyph::Requisition& r);
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r);
    void shape(XfRegion_ptr r);
    void traverse(XfGlyphTraversal_ptr t);
    void draw(XfGlyphTraversal_ptr t);
    void pick(XfGlyphTraversal_ptr t);
    XfGlyph_ptr body();
    void body(XfGlyph_ptr _p);
    void append(XfGlyph_ptr g);
    void prepend(XfGlyph_ptr g);
    XfTag add_parent(XfGlyphOffset_ptr parent_offset);
    void remove_parent(XfTag add_tag);
    XfGlyphOffset_ptr first_child_offset();
    XfGlyphOffset_ptr last_child_offset();
    void parent_offsets(XfGlyph::OffsetSeq& parents);
    void allocations(XfGlyph::AllocationInfoSeq& a);
    void need_redraw();
    void need_redraw_region(XfRegion_ptr r);
    void need_resize();
    Boolean restore_trail(XfGlyphTraversal_ptr t);
    /* XfViewer */
    XfViewer_ptr parent_viewer();
    XfViewer_ptr next_viewer();
    XfViewer_ptr prev_viewer();
    XfViewer_ptr first_viewer();
    XfViewer_ptr last_viewer();
    void append_viewer(XfViewer_ptr v);
    void prepend_viewer(XfViewer_ptr v);
    void insert_viewer(XfViewer_ptr v);
    void replace_viewer(XfViewer_ptr v);
    void remove_viewer();
    void set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next);
    void set_first_viewer(XfViewer_ptr v);
    void set_last_viewer(XfViewer_ptr v);
    XfFocus_ptr request_focus(XfViewer_ptr requestor, Boolean temporary);
    Boolean receive_focus(XfFocus_ptr f, Boolean primary);
    void lose_focus(Boolean temporary);
    Boolean first_focus();
    Boolean last_focus();
    Boolean next_focus();
    Boolean prev_focus();
    Boolean handle(XfGlyphTraversal_ptr t, XfEvent_ptr e);
    void close();
    /* Viewport */
    Adjustment_ptr scroll_adjustment(XfAxis a);
    //+

    void scroll_to(XfAxis a, Coord lower);
    XfCoord lower(XfAxis a);
    XfCoord length(XfAxis a);
    XfCoord cur_lower(XfAxis a);
    XfCoord cur_length(XfAxis a);

    struct Flags {
        Boolean clipping:1;
        Boolean transforming:1;
    };

protected:
    void cache_requisition();
    void check_allocation(Region_ptr a);
    Region_ptr body_allocation(Region_ptr a);
    WKViewport::Flags push_environment(XfPainter_ptr p, XfRegion_ptr a);
    void pop_environment(XfPainter_ptr p, WKViewport::Flags);

    Coord adjustment(Axis a);

    Coord cur_offset_[2];
    Coord cur_length_[2];
    TransformImpl* tx_;
    WKViewportAdjustment* x_adjustment_;
    WKViewportAdjustment* y_adjustment_;
    Boolean requested_;
    Glyph::Requisition requisition_;
};

#endif
